// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#ifndef CAMERA_H
#define CAMERA_H

#include "vector.hpp"

#define EPSILON			0.001		// epsilon value

// application window size
#define SCREENWIDTH 800		
#define SCREENHEIGHT 600


#define NEARCLIP		0.05		// default camera near clip
#define FARCLIP			500.0		// default far clip

//-----------------------------------------------------------------------------
// Camera class
//=============================================================================
class Camera {
//=============================================================================
public:
	int	id;
	// default values
	Vector	defeyep,deflookp,defupdir;
	float	deffov;

	Vector	eyep;				// virtual camera position in 3D space
	Vector	lookp;				// focus point of camera
	Vector	updir;				// direction pointing upward

	float	viewdist;			// distance from eyepoint to focus point	
	float	fov, hfov, vfov;	// half of the field of view, horizontal and vertical, in degrees.

	float	nearClip, farClip;	// near and far clipping plane distance
	long	hres, vres;			// horizontal and vertical resolution
	float	pixh, pixv;			// pixel size
		
	Vector	X, Y, Z;			// eye coordinate system (right-hand-orientation): X=right, Y=down, Z=viewing direction
  	Vector  R;

	inline Camera();
	inline Camera(int i,Vector ep,Vector lp,Vector up,float f);

	inline int getID() {return id;}

	inline void init();
	inline void completeCamera();
	inline void moveUpDown(float step);
	inline void strafe(float step);
	inline void dolly(float step);
	inline void zoomIn();
	inline void zoomOut();
	inline void orbit(float deltaX, float deltaY);
    inline void reset();
};

//-----------------------------------------------------------------------------
// default camera setup - when no camera specified in file
//-----------------------------------------------------------------------------
Camera::Camera()  {
//-----------------------------------------------------------------------------
	id = 0;
 	defeyep.Set(2.0, 2.0, 0.0);
	deflookp.Set(0.0, 0.0, 0.0);
	defupdir.Set(0, 1.0, 0.0);
	deffov = 22.5;
	reset();
}

//-----------------------------------------------------------------------------
// camera setup
//-----------------------------------------------------------------------------
Camera::Camera(int i,Vector ep,Vector lp,Vector up,float f) {
//-----------------------------------------------------------------------------
	id = i;
	defeyep = ep;	deflookp = lp;	defupdir = up;	deffov = f;
	reset();
}

//-----------------------------------------------------------------------------
// initialize the camera
//-----------------------------------------------------------------------------
void Camera::init()  {
//-----------------------------------------------------------------------------
	R = eyep - lookp;

	nearClip	= NEARCLIP; //EPSILON;
	farClip		= FARCLIP;

	hres		= SCREENWIDTH;
	vres		= SCREENHEIGHT;

	completeCamera();
}

//-----------------------------------------------------------------------------
// reset the camera
//-----------------------------------------------------------------------------
void Camera::reset()  {
//-----------------------------------------------------------------------------
	eyep = defeyep;
	lookp = deflookp;
	updir = defupdir;
	fov = deffov;
 	
	init();
}

//-----------------------------------------------------------------------------
// complete camera
//-----------------------------------------------------------------------------
void Camera::completeCamera() {
//-----------------------------------------------------------------------------
	// set up Z
    Z = lookp - eyep;
	// distance from virtual camera position to focus point
	viewdist = Z.Norm();
	if (viewdist < EPSILON) {
	//	::MessageBox(NULL, "Camera", "eyepoint and look-point coincide", MB_OK);
		return;
	}
	Z *= 1.0 / viewdist;

	// set up X   Camera->X is a direction pointing to the right in the window 
	X = Z % updir;
	float lengthX = X.Norm();
	if (lengthX < EPSILON) {
		return;
	}
	X *= 1.0 / lengthX;

	// set up Y
	Y = Z % X;
	Y.Normalize();

	// compute horizontal and vertical field of view angle from the specified one
	// if the vertical resolution is smaller, it is the specified fov = 45, the other is greater than 45
	if (hres < vres) {
		hfov = fov; 
		vfov = atan(tan(fov * M_PI/180.0) * (float)vres/(float)hres) * 180.0/M_PI;
	} else {
		vfov = fov; 
		hfov = atan(tan(fov * M_PI/180.0) * (float)hres/(float)vres) * 180.0/M_PI;
	}
	
	float tanFovH = tan(hfov * M_PI / 180.0);
	float tanFovV = tan(vfov * M_PI / 180.0);
	
	pixh = 2.0 * tanFovH / (float)(hres);
	pixv = 2.0 * tanFovV / (float)(vres);

	R = lookp - eyep;

}

//-----------------------------------------------------------------------------
// strafe the camera
//-----------------------------------------------------------------------------
void Camera::strafe(float step) {
//-----------------------------------------------------------------------------
	eyep	+= X * step;
	lookp	+= X * step;
	completeCamera();
}

//-----------------------------------------------------------------------------
// move the camera UP/DOWN
//-----------------------------------------------------------------------------
void Camera::moveUpDown(float step) {
//-----------------------------------------------------------------------------
	eyep	+= Y * step;
	lookp	+= Y * step;
	completeCamera();
}

//-----------------------------------------------------------------------------
// dolly the camera
//-----------------------------------------------------------------------------
void Camera::dolly(float step) {
//-----------------------------------------------------------------------------
	eyep	+= Z * step;
	R = eyep - lookp;
	completeCamera();
}

//-----------------------------------------------------------------------------
// zoom in the camera
//-----------------------------------------------------------------------------
void Camera::zoomIn() {
//-----------------------------------------------------------------------------
   if(fov>1) {
    fov-=0.1;
    completeCamera();
   }
}

//-----------------------------------------------------------------------------
// zoom out the camera
//-----------------------------------------------------------------------------
void Camera::zoomOut() {
//-----------------------------------------------------------------------------
   if(fov<70 ) {
    fov+=0.1;
    completeCamera();
   }
}

//-----------------------------------------------------------------------------
// orbit the camera
//-----------------------------------------------------------------------------
void Camera::orbit(float deltaX, float deltaY) { 
//-----------------------------------------------------------------------------
	float phi = 0, theta = 0;
    float LocX = eyep.x - lookp.x;
    float LocY = eyep.y - lookp.y;
    float LocZ = eyep.z - lookp.z;
   
    float dx = 5*deltaX/(float)hres;
    float dy = -3*deltaY/(float)vres;
    
    // calculates theta and phi angles for coord transform
	theta = acos(LocY/R.Norm());
    
     if(LocZ>=0 && LocX>=0) 
      phi = asin(LocZ/(R.Norm()*sin(theta)));
     else if(LocZ>=0 && LocX<=0) 
      phi = M_PI-asin(LocZ/(R.Norm()*sin(theta)));
     else if(LocZ<=0 && LocX<=0) 
      phi = M_PI+asin(fabs(LocZ)/(R.Norm()*sin(theta)));    
     else if(LocZ<=0 && LocX>=0) 
      phi = 2*M_PI-asin(fabs(LocZ)/(R.Norm()*sin(theta))); 
  
    phi	+= dx;
    theta += dy;

    if(theta<0.2) theta = 0.2;
    if(theta>3.0) theta = 3.0;
	
	eyep.setZ(lookp.z + R.Norm()*sin(phi)*sin(theta));
    eyep.setX(lookp.x + R.Norm()*cos(phi)*sin(theta));
    eyep.setY(lookp.y + R.Norm()*cos(theta));
    completeCamera();
}

#endif
