// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#ifndef GPUSPH_H
#define GPUSPH_H

#include "../RenderTexture/RenderTexture.h"
#include "particlesystem.hpp"

#include <Cg/cgGL.h>
#include <malloc.h>
#include <math.h>

//-----------------------------------------------------------------------------
// gpu-accelerated sph simulation class
//=============================================================================
class GpuSPH {
//=============================================================================
private:
	
	// num of neighbours, particle number for textures
	int nLength;
	int pNum;
	
	int itNum;						// number of iterations
	
	float dt;					// timestep	
	float maindensity;				// main density
	float gas_const;				// gas constant
	float mass;					// mass
	float h;					// smoothing radius
	float visc;					// viscosity
	
	// texture datas
	float* neighbourMapData;		
	float* posAttribMapData;		
	float* velAttribMapData;	
	float* densityAndPressureMapData;
	
	// position readback data
	float* downData;
	
	// texture maps
	RenderTexture *neighbourMap[15];
	RenderTexture *distanceMap[15];	
	RenderTexture *posAttribMap;
	RenderTexture *velAttribMap;
	RenderTexture *densityAndPressureMap;
	RenderTexture *forceMap;
	RenderTexture *collisionMap;
	
	// cg profile
	CGprofile   theProfile;	
	
	// cg programs
	CGprogram   densityFragProgram;
	CGprogram   distanceFragProgram;
	CGprogram   forceFragProgram;
	CGprogram   velFragProgram;
	CGprogram   posFragProgram;
	CGprogram   vertexProgram;	
	
	CGparameter textureParam;
	
	// uniform parameters
    CGparameter frameParam, dtParam, heightParam, lenParam, attrlenParam, rankParam;
    CGparameter hParam, massParam, gasconstParam, maindensParam, itNumParam;
    
    // uniform coefficient parameters
    CGparameter poly6_coef, grad_spiky_coef, v_lap_visc_coef;		
	
	// coefficient calculation
	inline float poly6coef() { return 315/(64*PI*(float)pow(h,9));}
	inline float grad_spiky() { return -45/(PI*h*h*h*h*h*h); }
	inline float lap_visc() { return 45/(PI*h*h*h*h*h*h); }
	inline float v_lap_visc() { return visc * lap_visc(); }
	
	// simulation 
	void constructNeighbourMap(ParticleSystem* ps,int num);
	void constructPosAttribMap(ParticleSystem* ps);
	void constructVelAttribMap(ParticleSystem* ps);
	void constructDensityAndPressureMap(ParticleSystem* ps);

	void computeDistance(int num);	
	void computeDensityAndPressure(int num);
	void computeForceAndCollision(int num);
	void computeVelocity(int num);
	void computePosition();

	void downloadDensityPressure(ParticleSystem* ps);	
	void downloadVelocity(ParticleSystem* ps, GridCell*** gridcells);
	void downloadPosition(ParticleSystem* ps);
	
public:
	GpuSPH() { nLength = 1; pNum = 1; neighbourMapData = NULL; posAttribMapData = NULL; velAttribMapData = NULL;}	
	~GpuSPH() { free(downData);}
	
	void init(ParticleSystem* ps);
	void simulate(ParticleSystem* ps, Container** container, int framenum);
};

#endif /*GPUSPH_H*/
