// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#include <stdio.h>
#include <vector>

#include "vector.hpp"
#include "camera.hpp"
#include "container.hpp"
#include "particlesystem.hpp"

//-----------------------------------------------------------------------------
// reader to the scene
//=============================================================================
class ObjectReader {
//=============================================================================
private:
	FILE*	objectfile;
	bool	end;
	int		id;
public:
	ObjectReader();
	ObjectReader(const char* file);

	Camera* nextCamera();
	Container* nextContainer();
	ParticleSystem* nextPartSystem();
};

//------------------------------------------------------------------------------
// open the file
//------------------------------------------------------------------------------
ObjectReader::ObjectReader(const char* file) {
//------------------------------------------------------------------------------
	// if file is NULL, then no objects in scene
	objectfile = NULL;
	if (file != NULL) {
		objectfile = fopen(file,"r");
	}	
	end = false;
	id = -1;
}

//------------------------------------------------------------------------------
// read camera
//------------------------------------------------------------------------------
Camera* ObjectReader::nextCamera() {
//------------------------------------------------------------------------------
	// no more camera
	if (end) {end = false; id = -1; return NULL;}
	// read object
	if (objectfile) {
		float a,b,c;
		Vector eyep,lookp,updir;
		float fov=22.5f;
		char order [20];

		for(;;) {
			fscanf(objectfile,"%s",order);

			// eyep
			if (strcmp(order,"eyep:") == 0) {
				fscanf(objectfile,"%f,%f,%f\n",&a,&b,&c);
				eyep.Set(a,b,c);
			}
			// lookp
			else if (strcmp(order,"lookp:") == 0) {
				fscanf(objectfile,"%f,%f,%f\n",&a,&b,&c);
				lookp.Set(a,b,c);
			}
			// updir
			else if (strcmp(order,"updir:") == 0) {
				fscanf(objectfile,"%f,%f,%f\n",&a,&b,&c);
				updir.Set(a,b,c);
			}
			// fov
			else if (strcmp(order,"fov:") == 0) {
				fscanf(objectfile,"%f\n",&fov);
			}

			// #next
			else if (strcmp(order,"#next") == 0) {
				id++;
				return new Camera(id,eyep,lookp,updir,fov);
			}
			// #end
			else if (strcmp(order,"#end") == 0) {
				end = true;
				id++;
				return new Camera(id,eyep,lookp,updir,fov);
			}
		}
	}
	else return NULL;
}

//------------------------------------------------------------------------------
// read particlesystem
//------------------------------------------------------------------------------
ParticleSystem* ObjectReader::nextPartSystem() {
//------------------------------------------------------------------------------
	// no more container
	if (end) {end = false; id = -1; return NULL;}
	// read object
	if (objectfile) {
		float density,visc,k;
		char type[10] = "";
		char order [20];

		for(;;) {
			fscanf(objectfile,"%s",order);

			// density of the stuff
			if (strcmp(order,"density:") == 0) {
				fscanf(objectfile,"%f\n",&density);
			}
			// viscosity of the stuff
			else if (strcmp(order,"viscosity:") == 0) {
				fscanf(objectfile,"%f\n",&visc);
			}
			// gas constant of the stuff
			else if (strcmp(order,"gasconstant:") == 0) {
				fscanf(objectfile,"%f\n",&k);
			}
			// particlesystem type
			else if (strcmp(order,"type:") == 0) {
				fscanf(objectfile,"%s\n",type);
			}

			// #next
			else if (strcmp(order,"#next") == 0) {
				id++;
				return new ParticleSystem(id,density,visc,k);
			}
			// #end
			else if (strcmp(order,"#end") == 0) {
				end = true;
				id++;
				return new ParticleSystem(id,density,visc,k);
			}
		}
	}
	else return NULL;
}

//------------------------------------------------------------------------------
// read container
//------------------------------------------------------------------------------
Container* ObjectReader::nextContainer() {
//------------------------------------------------------------------------------
	// no more container
	if (end) {
		end = false;
		id = -1;
		if (objectfile) fclose(objectfile);
		return NULL;
	}
	// read object
	if (objectfile != NULL) {
		float a,b,c;
		Vector position;
		float width=1.0f,height=1.0f,depth=1.0f;
		bool open = false;
		bool hole = false;
		int partsysid=-1;
		float partsysheight=1.0f;
		int partnum = 0;
		std::vector<Container*> subcontainers;
		char order [20];

		for(;;) {
			fscanf(objectfile,"%s",order);

			// position
			if (strcmp(order,"position:") == 0) {
				fscanf(objectfile,"%f,%f,%f\n",&a,&b,&c);
				position.Set(a,b,c);
			}
			// width
			else if (strcmp(order,"width:") == 0) {
				fscanf(objectfile,"%f\n",&width);
			}
			// height
			else if (strcmp(order,"height:") == 0) {
				fscanf(objectfile,"%f\n",&height);
			}
			// depth
			else if (strcmp(order,"depth:") == 0) {
				fscanf(objectfile,"%f\n",&depth);
			}
			// particlesystem index
			else if (strcmp(order,"partsysid:") == 0) {
				fscanf(objectfile,"%d\n",&partsysid);
			}
			// number of particles
			else if (strcmp(order,"partnum:") == 0) {
				fscanf(objectfile,"%d\n",&partnum);
			}
			// particlesystem height - in percents
			else if (strcmp(order,"partsysheight:") == 0) {
				fscanf(objectfile,"%f\n",&partsysheight);
			}

			// #subcontainer
			else if (strcmp(order,"#subcontainer") == 0) {
				subcontainers.push_back(nextContainer());
			}
			// #endsub
			else if (strcmp(order,"#endsub") == 0) {
				id++;
				return new Container(id,position,width,height,depth,open,hole,partsysid,partnum,partsysheight,subcontainers);
			}
			// #next
			else if (strcmp(order,"#next") == 0) {
				id++;
				return new Container(id,position,width,height,depth,open,hole,partsysid,partnum,partsysheight,subcontainers);
			}
			// #end
			else if (strcmp(order,"#end") == 0) {
				end = true;
				id++;
				return new Container(id,position,width,height,depth,open,hole,partsysid,partnum,partsysheight,subcontainers);
			}
		}
	}
	else return NULL;
}
