// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#ifndef PARTSYSTEM_H
#define PARTSYSTEM_H

#include <math.h>
#include <vector>

class ParticleSystem;
#include "particle.hpp"
#include "gridcell.hpp"

class Container;

#define PI 3.1415926535f

//-----------------------------------------------------------------------------
// partsystem class
//=============================================================================
class ParticleSystem {
//=============================================================================
protected:
	
	int id;
	
	// particle physics parameters	
	float particle_radius,particle_mass,h,h2,k;
	
	int num_of_total_particles;
	int num_of_particles;	
	
	float mass;
	float density;
	float viscosity;
	
	int nextfree(int begin); 
	void addParticle(int id,float m,float x,float y,float z, float velx, float vely, float velz, bool rcv, bool mnull, bool mthrow);
	void addDensityPressure(int id,float density,float pressure);

public:

	ParticleSystem() { firstfree = 0; }
	ParticleSystem(int i,float dens,float visc,float kk);
	
	int firstfree;
	
	// particles
	Particle* particles;
	
	// parameters
	int getID() const;
	float getH() const;
	int getPartNum() const;
	int getTPartNum() const;
	float getDensity() const;
	float getViscosity() const;
	float getK() const;
	float getM() const;
	
	void sceneReset();														// reset the scene					
	void createParticles(Container* container);								// create the particles of the system
	void resetStatus();														// reset the particles
	void resetParticleNeighbours();
	void setParticlesToCells(GridCell*** gridcells,Container* container);	// set the particles to cells
	void setNeighbours(GridCell*** gridcells,Container* container);			// set all particles neighbours
	
	// communication
	void communicateParticles(Container* container,GridCell*** gridcells);
	void communicateDensityPressure(Container* container,GridCell*** gridcells);
	void sendParticles(std::vector<Particle> *p, int dest);
	void sendParticleDensityPressure(std::vector<Particle> *p, int dest);
	
	void deleteOutside(Container* container,GridCell*** gridcells);
	
	int receiveParticles(int source);
	int receiveParticleDensityPressure(int source);
	void removeParticle(int pos, GridCell*** gridcells);	 	
		
	void draw();															// draw all particle to the screen
};

// the fluid class
//=============================================================================
class Fluid : public ParticleSystem {
//=============================================================================
public:
	inline Fluid() {}
	inline Fluid(const ParticleSystem& f);
};

//-----------------------------------------------------------------------------
// duplicate a fluid system for this container
//-----------------------------------------------------------------------------
Fluid::Fluid(const ParticleSystem& f) {
//-----------------------------------------------------------------------------
	id = f.getID();
	density = f.getDensity();
	viscosity = f.getViscosity();
	k = f.getK();
}

#endif /*PARTSYSTEM_H*/
