//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================

//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file EventTransmission.h
/// Classes for event transmission throught the network 

#ifndef __MINGLE_PARALLEL_EVENT_TRANSMISSION_H__
#define __MINGLE_PARALLEL_EVENT_TRANSMISSION_H__

// Include class definitions
#include "parallel-classes.h"

#include <string>
#include <vector>
#include <queue>

namespace MinGLE {
	class EventSerializer {
		protected:
			static const size_t BUFFER_SIZE = 256;
			
		protected:
			char *mBuffer;

			// Destination window proxy for event deserialization
			WindowProxy *mDstWindowProxy;

		public:
			EventSerializer();
			EventSerializer(WindowProxy *dstWindowProxy);
			~EventSerializer();
			
			inline void *getData() const { 
				return mBuffer;
			}
			inline size_t getSize() const {
				return sizeof(WindowListener::WindowEvent);
			}
			
			void serializeKeyEvent(int key, unsigned modifiers, InputListener::KeyEventState state); 
    		void serializeMouseButtonEvent(int x, int y, InputListener::MouseButton button, InputListener::MouseEventState state); 
    		void serializeMouseMoveEvent(int x, int y, InputListener::MouseMoveState state); 
    		void serializeMouseMoveEvent(InputListener::MouseMoveState state); 
    		void serializeMouseWheelEvent(int x, int y, int delta); 

			void serializeRegisterKeyboardEvent(int key, InputListener::KeyEventState state); 
			void serializeRegisterKeyboardModifierEvent(unsigned modifiers); 
			void serializeRegisterMouseMoveEvent(int x, int y); 
			void serializeRegisterMouseButtonEvent(InputListener::MouseButton button, InputListener::MouseEventState state); 
			void serializeRegisterMouseWheelEvent(int delta);

    		// void serializeRenderEvent();  
    		// void serializeIdleEvent();  
    		void serializeResizeEvent(unsigned width, unsigned height);  
    		void serializeRepositionEvent(int left, int top);  
    		// void serializeShowEvent();  
    		// void serializeHideEvent();  
			
			void deserializeEvent(const WindowListener::WindowEvent* event);
			
		private:
			void _allocateBuffer();
			void _serializeEvent(const WindowListener::WindowEvent* event);
	};

// =============================================================================

	// Class predefinition
	class SimpleCommSock;

	class EventTransmitter {
		protected:
			static const unsigned DEFAULT_PORT = 16127;
		
		protected:
			SimpleCommSock *mCommSock;
			EventSerializer *mEventSerializer;
			unsigned mPort;
		
		public:
			// Contructor for the collector
			EventTransmitter(unsigned port = DEFAULT_PORT);
			// Constructor for the injector
			EventTransmitter(WindowProxy *dstWindowProxy, unsigned port = DEFAULT_PORT);
			~EventTransmitter();
			
			inline EventSerializer *getEventSerializer() {
				return mEventSerializer;
			}
	};

// =============================================================================

	class EventCollector : public EventTransmitter {
		protected:
			std::vector<std::string> mHostList;
		
		public:
			EventCollector(const std::vector<std::string>& hostList);
			~EventCollector();

			void waitForInjectors();
			void sendEvent();
	};

// =============================================================================

	class LocalEventInjector : public EventTransmitter {
		protected:
			typedef std::queue < WindowListener::WindowEvent > WindowEventQueue;

		protected:
			GuardedObject<WindowEventQueue> mWindowEventQueue; 
	
		public:
			LocalEventInjector(WindowProxy *dstWindowProxy);

			void addEventToQueue(const WindowListener::WindowEvent *event);
			void injectQueuedEvents();
	};

// =============================================================================

	class EventInjector : public LocalEventInjector, public Runnable {
		
		protected:
			Thread *mThread;
			std::string mThisHost;
		
		public:
			EventInjector(WindowProxy *dstWindowProxy, const std::string& thisHost);
			~EventInjector();
			
			void connectToCollector(const std::string& collectorHost);
			//void disconnectFromCollector();

		protected:
			virtual void run();
	};

// =============================================================================

	class EventQueueWindowProxy : public WindowProxy {
		protected:
			LocalEventInjector *mEventInjector;
		
		public:
			EventQueueWindowProxy(WindowProxy *mWindowProxy);
			virtual ~EventQueueWindowProxy();

			inline LocalEventInjector *getEventInjector() {
				return mEventInjector;
			}
	};

}

#endif // __MINGLE_PARALLEL_EVENT_TRANSMISSION_H__
