//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================

//
// Original source of this class: HP XC ParaComp Kit.
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file ParallelHelpers.h
/// Helper classes for parallel rendering

#ifndef __MINGLE_PARALLEL_PARALLEL_HELPERS_H__
#define __MINGLE_PARALLEL_PARALLEL_HELPERS_H__

// Include class definitions
#include "parallel-classes.h"

namespace MinGLE {

template<typename T> class BoxPartitioner {
public:

	// Child type for binary tree
	enum ChildType {FIRSTCHILD, SECONDCHILD};

	// Plane normals for each axis aligned twosided plane
	enum PlaneNormal {MINUSX, MINUSY, MINUSZ, PLUSX, PLUSY, PLUSZ};

	// Axis aligned box partition
	// This is a node of a bidirectionally linked binary tree
	struct Partition {

		/// Offset of the partition
		T offset[3];

		/// Size of the partition
		T size[3];

		/// Amount of borders on the lower x, y, z sides
		T bordersLower[3];

		/// Amount of borders on the upper x, y, z sides
		T bordersUpper[3];

		/// Parent partition (null on the top)
		Partition *parent;

		/// Child partitions (binary tree, null on the leaves)
		Partition *firstChild, *secondChild;

		/// Number of leaf partitions
		size_t firstChildPartitions, secondChildPartitions;

		/// Plane normal pointing in the direction of the partition
		PlaneNormal planeNormal;

		/// The center of the side of the partition to which the 
		/// slice plane fits of which planeNormal is given.
		double planePoint[3];
		
		/// Flag indicated that the partition is detached
		bool detached;
		
		Partition(T width = 0, T height = 0, T depth = 0, T left = 0,
				T top = 0, T front = 0) :
			parent(0), firstChild(0), secondChild(0), firstChildPartitions(0),
					secondChildPartitions(0), planeNormal(MINUSX), detached(false) {
			setOffset(left, top, front);
			setSize(width, height, depth);
			setBorders(0, 0, 0, 0, 0, 0);
			
			planePoint[0] = planePoint[1] = planePoint[2] = 0.0;
		}

		/// Clone constructor. See BoxPartitioner::cloneDetached()
		Partition(const Partition& toClone) {
			setOffset(toClone.offset[0], toClone.offset[1], toClone.offset[2]);
			setSize(toClone.size[0], toClone.size[1], toClone.size[2]);
			setBorders(toClone.bordersLower[0], toClone.bordersLower[1], toClone.bordersLower[2], 
					toClone.bordersUpper[0], toClone.bordersUpper[1], toClone.bordersUpper[2]);
			
			// Clone parent if exists
			parent = toClone.parent ? new Partition(toClone.parent) : 0;
			
			// No child partitions
			firstChild = secondChild = 0;
			firstChildPartitions = secondChildPartitions = 0;
			
			// Copy plane
			planeNormal = toClone.planeNormal;
			planePoint[0] = toClone.planePoint[0];
			planePoint[1] = toClone.planePoint[1];
			planePoint[2] = toClone.planePoint[2];
			
			// Make detached
			detached = true;
		}
		
		~Partition() {
			if (firstChild)
				delete firstChild;
			if (secondChild)
				delete secondChild;
			
			// Detached partitions destroys their parents
			// (upto the root)
			if(detached)
				delete parent;
		}

		inline void setOffset(T left, T top, T front) {
			offset[0] = left, offset[1] = top, offset[2] = front;
		}

		inline void setSize(T width, T height, T depth) {
			size[0] = width, size[1] = height, size[2] = depth;
		}

		void setBorders(T xLower, T yLower, T zLower, T xUpper, T yUpper,
				T zUpper) {
			bordersLower[0] = xLower, bordersLower[1] = yLower, bordersLower[2]
					= zLower;
			bordersUpper[0] = xUpper, bordersUpper[1] = yUpper, bordersUpper[2]
					= zUpper;
		}
		
		void _debug(unsigned indent = 0) const {
			for(unsigned i=0; i<indent; i++) printf(" ");
			printf("(%g %g %g)-(%g %g %g) [F:%lu|S:%lu] %c\n", 
					(double) offset[0], (double) offset[1], (double) offset[2], 
					(double) offset[0]+size[0], (double) offset[1]+size[1], (double) offset[2]+size[2],
					firstChildPartitions, secondChildPartitions, !firstChild&&!secondChild?'*':' ');
			if(firstChild) firstChild->_debug(indent+2);
			if(secondChild) secondChild->_debug(indent+2);
		}
	};

protected:
	const static double PLANE_NORMAL_COORDS[6][3];

	/// A partition representing the whole box
	Partition *mRootBox;

	/// Number of partitions
	size_t mPartitions;

public:
	BoxPartitioner(T width = 0, T height = 0, T depth = 0, T left = 0,
			T top = 0, T front = 0) {
		mRootBox = new Partition(width, height, depth, left, top, front);
	}

	BoxPartitioner(const Partition *rootBox) {
		mRootBox = new Partition(
				rootBox->size[0], rootBox->size[1], rootBox->size[2],
				rootBox->offset[0], rootBox->offset[1], rootBox->offset[2] 
			);
	}
	
	~BoxPartitioner() {
		delete mRootBox;
	}

	inline Partition* getPartition(const size_t partition) const {
		Assert(partition < mPartitions, "BoxPartitioner<T>::getPartition");
		return _getPartition(mRootBox, partition);
	}

	inline size_t getPartitions() const {
		return mPartitions;
	}
	
	inline Partition* getRootBox() const {
		return mRootBox;
	}

	/// Chop the root box into partitions with equal volume and minimal diameter
	void chop(size_t partitions);

	/// Chop the root box into partitions with the specified (maximal) sizes
	void chop(const T width, const T height, const T depth);
	
	inline static Partition *getRootPartition(Partition *partition) {
		while(partition->parent)
			partition = partition->parent;
			
		return partition;
	}
	
	/// Return a detached version of the specified partition.
	/// All partitions which are not detached are destoryed
	/// when the partitioner object is destroyed.
	///
	/// Both the specified partition and its anchestors are
	/// detached, however its children are NOT.
	/// (Actually a route from the specified node is copied to 
	/// the root of the tree)
	inline static Partition *cloneAsDetached(const Partition *partition) {
		Assert(partition, "BoxPartitioner<T>::cloneAsDetached()");
		return new Partition(partition);
	}

	/// Transform the current partition.
	/// For non-detached partitions, the child partitions are also
	/// transformed.
	/// For detached partitions, the anchestor partitions are 
	/// transformed, too.
	static void transformPartition(Partition *partition, 
			const double translateX, const double translateY, const double translateZ,
			const double scaleX, const double scaleY, const double scaleZ);
	
	/// Cast a partition with another type to a partition 
	/// suitable for this box partitioner
	template<typename CAST_T>
	static Partition *castPartition(const typename BoxPartitioner<CAST_T>::Partition *partition);
	
	/// Get sorting order for the specified partition and view direction.
	/// Note that the depth type is templated
	template<typename DEPTH>
	static DEPTH getSortOrder(const Partition *partition, double eyeX, double eyeY,
			double eyeZ);
	
protected:
	static double _specDotProduct(PlaneNormal planeNormal, const double x, const double y, const double z);

	void _chop(Partition* toChop, size_t partitions);
	void _chopXFirstLevel(Partition *toChop, const T width, const T height, const T depth);
	void _chopYSecondLevel(Partition *toChop, const T height, const T depth);
	void _chopZThirdLevel(Partition *toChop, const T depth);
	void _cutX(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond);
	void _cutY(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond);
	void _cutZ(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond);
	void _addPartition(Partition* child, Partition* parent,
			ChildType childType, size_t childPartitions);
	inline Partition *_getPartition(const Partition *current, const size_t partition) const {

		if(!partition && !current->firstChildPartitions && !current->secondChildPartitions)
			return (Partition *) current;
		else if (partition < current->firstChildPartitions)
			return _getPartition(current->firstChild, partition);
		else if(partition-current->firstChildPartitions < current->secondChildPartitions)
			return _getPartition(current->secondChild, partition
					- current->firstChildPartitions);

		// This should never happened
		Except("BoxPartitioner<T>::_getPartition()",
				"Internal error while searching for a partition");
		return 0;
	}
	
	/// Calculate proper values of firstChildPartitions and secondChildPartitions 
	/// Also return the number of children 
	/// (scan the tree instead of returning firstChildPartitions+secondChildPartitions) 
	size_t _fixChildCounts(Partition* partition);
};

template<typename T> void BoxPartitioner<T>::chop(size_t partitions) {
	// Chop into partitions recursively
	_chop(mRootBox, mPartitions = partitions);
}

template<typename T> void BoxPartitioner<T>::chop(const T width, const T height, const T depth) {
	_chopXFirstLevel(mRootBox, width, height, depth);
	
	// Fix firstChildPartitions and secondChildPartitions values
	mPartitions = _fixChildCounts(mRootBox);
}

template<typename T> void BoxPartitioner<T>::transformPartition(Partition *partition, 
		const double translateX, const double translateY, const double translateZ,
		const double scaleX, const double scaleY, const double scaleZ) {
	Assert(partition, "BoxPartitioner<T>::transformPartition()");
	
	// a) Detached partition: transform the 
	// partition and the anchestors 
	if(partition->detached) {
		Except("BoxPartitioner<T>::transformPartition()", "TODO: Implement");
	}
	
	// b) non-detached partition: transform the
	// partition and the descendants
	else {
		partition->setOffset(
				(T) (partition->offset[0] + translateX),
				(T) (partition->offset[1] + translateY),
				(T) (partition->offset[2] + translateZ)
			);
		partition->setSize(
				(T) (partition->size[0] * scaleX),
				(T) (partition->size[1] * scaleY),
				(T) (partition->size[2] * scaleZ)
			);
		if(partition->firstChild)
			transformPartition(partition->firstChild, translateX, translateY, translateZ, scaleX, scaleY, scaleZ);
		if(partition->secondChild)
			transformPartition(partition->secondChild, translateX, translateY, translateZ, scaleX, scaleY, scaleZ);
	}
}

template<typename T> template<typename CAST_T>
typename BoxPartitioner<T>::Partition *BoxPartitioner<T>::castPartition(const typename BoxPartitioner<CAST_T>::Partition *partition) {
	Assert(partition, "BoxPartitioner<T>::castPartition()");

	Except("BoxPartitioner<T>::castPartition()", "TODO: Implement");
	
	/*
	// Get root partition for the partition to be cast
	const typename BoxPartitioner<CAST_T>::Partition *castRootPartition = 
		BoxPartitioner<CAST_T>::getRootPartition(partition); 

	// a) Detached partition: cast the 
	// partition and the anchestors 
	if(partition->detached) {
		Except("BoxPartitioner<T>::transformPartition()", "TODO: Implement");
	}
		
	// b) non-detached partition: cast the
	// partition and the descendants
	else {
		
	}
	*/
		
	return 0;
}

template<typename T> template<typename DEPTH> DEPTH BoxPartitioner<T>::getSortOrder(
		const Partition *partition, double eyeX, double eyeY, double eyeZ) {

	DEPTH order = 0;

	// Calculate plane-behind/plane-front flags for all partitions
	// until the top partition
	for(;;) {
		// If parent exists then there is valid plane normal
		if(partition->parent) {
			// Shift previous orderings to higher places
			order <<= 1;

			// Calculate corresponding bits
			double viewVector[3] = {
				partition->planePoint[0] - eyeX,
				partition->planePoint[1] - eyeY,
				partition->planePoint[2] - eyeZ
			};
			MathUtils::vectorNormalize(viewVector[0], viewVector[1], viewVector[2]);
			order |= (_specDotProduct(partition->planeNormal, viewVector[0], viewVector[1], viewVector[2])>=0.0 ? 1 : 0);

			// Step to the parent
			partition = partition->parent;
		} 
		
		// Else quit
		else
			break;
	}

	// Reverse the bits and return
	return MathUtils::reverseBits<DEPTH>(order);
}

template<typename T> double BoxPartitioner<T>::_specDotProduct(
		PlaneNormal planeNormal, const double x, const double y, const double z) {

	// Handle dot product between a plane normal and 
	// the specified vector
	
	switch(planeNormal) {
	case MINUSX: 	return -x; break;
	case PLUSX: 	return  x; break;
	case MINUSY: 	return -y; break;
	case PLUSY: 	return  y; break;
	case MINUSZ: 	return -z; break;
	case PLUSZ: 	return  z; break;
	}

	Except("BoxPartitioner<T>::_specDotProduct", "Invalid plane normal. This should have not happened.");
	return 0;
}

template<typename T> void BoxPartitioner<T>::_chop(Partition* toChop,
		size_t partitions) {

	// Handle trivial case
	if(partitions == 1) return;
	
	// Number of partitions in the two groups
	size_t partitionsFirst = partitions / 2;
	size_t partitionsSecond = partitions - partitionsFirst;

	// Partitions for the two groups
	Partition *partitionFirst, *partitionSecond;

	// X is the longest
	if (toChop->size[0] >= toChop->size[1]&& toChop->size[0] >= toChop->size[2]) {
		T cutPoint = toChop->size[0] * partitionsFirst / partitions;
		_cutX(cutPoint, toChop, &partitionFirst, &partitionSecond);
	}

	// Y is the longest
	else if (toChop->size[1] >= toChop->size[0]&& toChop->size[1]
			>= toChop->size[2]) {
		T cutPoint = toChop->size[1] * partitionsFirst / partitions;
		_cutY(cutPoint, toChop, &partitionFirst, &partitionSecond);
	}

	// Z is the longest
	else {
		T cutPoint = toChop->size[2] * partitionsFirst / partitions;
		_cutZ(cutPoint, toChop, &partitionFirst, &partitionSecond);
	}

	// Add partitions
	if(partitionsFirst) _addPartition(partitionFirst, toChop, FIRSTCHILD, partitionsFirst);
	if(partitionsSecond) _addPartition(partitionSecond, toChop, SECONDCHILD, partitionsSecond);

	// Call recursively if needed
	if (partitionsFirst > 1)
		_chop(partitionFirst, partitionsFirst);
	if (partitionsSecond > 1)
		_chop(partitionSecond, partitionsSecond);
}

template<typename T> void BoxPartitioner<T>::_chopXFirstLevel(Partition *toChop, const T width, const T height, const T depth) {
	// Need chopping along axis X
	if(toChop->size[0] > width) {
		// Number of partitions in the two groups
		size_t partitions = (size_t) ::ceil((double) toChop->size[0] / width);
		size_t partitionsFirst = partitions / 2;
		size_t partitionsSecond = partitions - partitionsFirst;
		
		// Partitions for the two groups
		Partition *partitionFirst, *partitionSecond;
		
		// Cut along axis X
		T cutPoint = width * partitionsFirst;
		_cutX(cutPoint, toChop, &partitionFirst, &partitionSecond);

		// Add partitions
		// (with invalid child partition count)
		if(partitionsFirst) _addPartition(partitionFirst, toChop, FIRSTCHILD, 0);
		if(partitionsSecond) _addPartition(partitionSecond, toChop, SECONDCHILD, 0);

		// Call recursively
		_chopXFirstLevel(partitionFirst, width, height, depth);
		_chopXFirstLevel(partitionSecond, width, height, depth);
	}

	// Chop along axis Y
	else {
		_chopYSecondLevel(toChop, height, depth);
	}
}

template<typename T> void BoxPartitioner<T>::_chopYSecondLevel(Partition *toChop, const T height, const T depth) {
	// Need chopping along axis Y
	if(toChop->size[1] > height) {
		// Number of partitions in the two groups
		size_t partitions = (size_t) ::ceil((double) toChop->size[1] / height);
		size_t partitionsFirst = partitions / 2;
		size_t partitionsSecond = partitions - partitionsFirst;
		
		// Partitions for the two groups
		Partition *partitionFirst, *partitionSecond;
		
		// Cut along axis X
		T cutPoint = height * partitionsFirst;
		_cutY(cutPoint, toChop, &partitionFirst, &partitionSecond);

		// Add partitions
		// (with invalid child partition count)
		if(partitionsFirst) _addPartition(partitionFirst, toChop, FIRSTCHILD, 0);
		if(partitionsSecond) _addPartition(partitionSecond, toChop, SECONDCHILD, 0);

		// Call recursively
		_chopYSecondLevel(partitionFirst, height, depth);
		_chopYSecondLevel(partitionSecond, height, depth);
	}

	// Chop along axis Z
	else {
		_chopZThirdLevel(toChop, depth);
	}
}

template<typename T> void BoxPartitioner<T>::_chopZThirdLevel(Partition *toChop, const T depth) {
	// Need chopping along axis Z
	if(toChop->size[2] > depth) {
		// Number of partitions in the two groups
		size_t partitions = (size_t) ::ceil((double) toChop->size[2] / depth);
		size_t partitionsFirst = partitions / 2;
		size_t partitionsSecond = partitions - partitionsFirst;
		
		// Partitions for the two groups
		Partition *partitionFirst, *partitionSecond;
		
		// Cut along axis X
		T cutPoint = depth * partitionsFirst;
		_cutZ(cutPoint, toChop, &partitionFirst, &partitionSecond);

		// Add partitions
		// (with invalid child partition count)
		if(partitionsFirst) _addPartition(partitionFirst, toChop, FIRSTCHILD, 0);
		if(partitionsSecond) _addPartition(partitionSecond, toChop, SECONDCHILD, 0);

		// Call recursively if needed
		if(partitionsFirst > 1)
			_chopZThirdLevel(partitionFirst, depth);
		if(partitionsSecond > 1)
			_chopZThirdLevel(partitionSecond, depth);
	}
}

template<typename T> void BoxPartitioner<T>::_cutX(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond) {
	// First child partition
	(*partitionFirst) = new Partition(cutPoint, toChop->size[1], toChop->size[2],
			toChop->offset[0], toChop->offset[1], toChop->offset[2]);
	(*partitionFirst)->planeNormal = MINUSX;
	(*partitionFirst)->planePoint[0] = cutPoint;
	(*partitionFirst)->planePoint[1] = toChop->offset[1] + 0.5*toChop->size[1];
	(*partitionFirst)->planePoint[2] = toChop->offset[2] + 0.5*toChop->size[2];
	
	// Second child partition
	*partitionSecond
			= new Partition(toChop->size[0] - cutPoint, toChop->size[1],
					toChop->size[2], toChop->offset[0] + cutPoint, toChop->offset[1],
					toChop->offset[2]);
	(*partitionSecond)->planeNormal = PLUSX;
	(*partitionSecond)->planePoint[0] = cutPoint;
	(*partitionSecond)->planePoint[1] = toChop->offset[1] + 0.5*toChop->size[1];
	(*partitionSecond)->planePoint[2] = toChop->offset[2] + 0.5*toChop->size[2];
}

template<typename T> void BoxPartitioner<T>::_cutY(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond) {
	// First child partition
	(*partitionFirst) = new Partition(toChop->size[0], cutPoint, toChop->size[2],
			toChop->offset[0], toChop->offset[1], toChop->offset[2]);
	(*partitionFirst)->planeNormal = MINUSY;
	(*partitionFirst)->planePoint[0] = toChop->offset[0] + 0.5*toChop->size[0];
	(*partitionFirst)->planePoint[1] = cutPoint;
	(*partitionFirst)->planePoint[2] = toChop->offset[2] + 0.5*toChop->size[2];
	
	// Second child partition
	*partitionSecond
			= new Partition(toChop->size[0], toChop->size[1] - cutPoint,
					toChop->size[2], toChop->offset[0], toChop->offset[1] + cutPoint,
					toChop->offset[2]);
	(*partitionSecond)->planeNormal = PLUSY;
	(*partitionSecond)->planePoint[0] = toChop->offset[0] + 0.5*toChop->size[0];
	(*partitionSecond)->planePoint[1] = cutPoint;
	(*partitionSecond)->planePoint[2] = toChop->offset[2] + 0.5*toChop->size[2];
}

template<typename T> void BoxPartitioner<T>::_cutZ(const T cutPoint, const Partition* toChop, Partition **partitionFirst, Partition **partitionSecond) {
	// First child partition
	(*partitionFirst) = new Partition(toChop->size[0], toChop->size[1], cutPoint,
			toChop->offset[0], toChop->offset[1], toChop->offset[2]);
	(*partitionFirst)->planeNormal = MINUSZ;
	(*partitionFirst)->planePoint[0] = toChop->offset[0] + 0.5*toChop->size[0];
	(*partitionFirst)->planePoint[1] = toChop->offset[1] + 0.5*toChop->size[1];
	(*partitionFirst)->planePoint[2] = cutPoint;
	
	// Second child partition
	*partitionSecond
			= new Partition(toChop->size[0], toChop->size[1],
					toChop->size[2]- cutPoint, toChop->offset[0], toChop->offset[1],
					toChop->offset[2] + cutPoint);
	(*partitionSecond)->planeNormal = PLUSZ;
	(*partitionSecond)->planePoint[0] = toChop->offset[0] + 0.5*toChop->size[0];
	(*partitionSecond)->planePoint[1] = toChop->offset[1] + 0.5*toChop->size[1];
	(*partitionSecond)->planePoint[2] = cutPoint;	
}

template<typename T> void BoxPartitioner<T>::_addPartition(Partition* child,
		Partition* parent, ChildType childType, size_t childPartitions) {

	switch (childType) {
	case FIRSTCHILD:
		parent->firstChild = child;
		parent->firstChildPartitions = childPartitions;
		break;
	case SECONDCHILD:
		parent->secondChild = child;
		parent->secondChildPartitions = childPartitions;
	}

	child->parent = parent;
}

template<typename T> size_t BoxPartitioner<T>::_fixChildCounts(Partition* partition) {
	if(partition->firstChild)
		partition->firstChildPartitions = _fixChildCounts(partition->firstChild);
	if(partition->secondChild)
		partition->secondChildPartitions = _fixChildCounts(partition->secondChild);
	
	return 
		// For internal nodes
		partition->firstChildPartitions + partition->secondChildPartitions + 
		// For leaf nodes
		(!partition->firstChildPartitions && !partition->secondChildPartitions ? 1 : 0);
}

} // namespace MinGLE

#endif // __MINGLE_PARALLEL_PARALLEL_HELPERS_H__
