//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================

//
// Original source of this class: HP XC ParaComp Kit.
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file ParallelRenderingWindowProxy.h
/// Window proxy class for that supplies parallel rendering support for a window

#ifndef __MINGLE_PARALLEL_PARALLEL_RENDERING_PROVIDER_H__
#define __MINGLE_PARALLEL_PARALLEL_RENDERING_PROVIDER_H__

// Include class definitions
#include "parallel-classes.h"

#include "ParallelRenderingSupport.h"
#include "ParallelHelpers.h"

namespace MinGLE {

class ParallelRenderingWindowProxy : public WindowProxy {

protected:
	static const int INVALID_RENDERER_INDEX = -1;
	static const bool DEFAULT_AUTOCALCULATE_FRAMELET = true;
	static const bool DEFAULT_AUTOCALCULATE_SORTORDER = true;
	static const bool DEFAULT_CONTRIBUTE_PIXELS = true;
	static const bool DEFAULT_DISPLAY_OUTPUT = false;
	static const unsigned DEFAULT_DISPLAY_DEBUG = 0x0;

	/// BOX_VERTICES[vertex][coord]
	static const float BOX_VERTICES[8][3];
	
	/// BOX_FACES[face][vertex] in ccw order
	static const unsigned BOX_FACES[6][4];
	
protected:
	int mRendererCount;
	int mThisRenderer;
	
	ParallelRenderingSupport::CompositingOperator mCompositingOperator;
	
	/// Indicates that the framework automatically 
	/// calculates the framelet offset and sizes
	bool mAutoCalculateFramelet;

	/// Indicates that the framework automatically
	/// calculates the depth sorting order when it is needed
	bool mAutoCalculateSortOrder;

	/// Indicates that this paracomp part is contributing pixels
	bool mContributePixels;
	
	/// Indicates that this paracomp part is 
	/// blitting the final output to its window
	bool mDisplayOutput;
	
	/// Display debug info
	unsigned mDisplayDebug;
	
	/// Box partition for framelet size and sort order calculation
	BoxPartitioner<double>::Partition *mPartition;
	
	/// Framelet offset and size
	unsigned mBoundingRectangleOffset[2], mBoundingRectangleSize[2];
	
	/// Indicates that the framelet offset and size values are valid
	bool mValidFramelet;
	
public:
	ParallelRenderingWindowProxy(Window *window, ParallelRenderingSupport::CompositingOperator compositingOperator);

	inline virtual int getRendererCount() const {
		return mRendererCount;
	}
	
	inline virtual int getThisRenderer() const {
		return mThisRenderer;
	}
	
	inline virtual const ParallelRenderingSupport::CompositingOperator getCompositingOperator() const {
		return mCompositingOperator;
	}
	
	inline virtual void setAutoCalculateFramelet(const bool flag) {
		mAutoCalculateFramelet = flag;
	}
	
	inline virtual void setAutoCalculateSortOrder(const bool flag) {
		mAutoCalculateSortOrder = flag;
	}
	
	inline virtual void setPartition(BoxPartitioner<double>::Partition *partition) {
		mPartition = partition;
	}
	
	inline virtual unsigned getDisplayDebug() {
		return mDisplayDebug;
	}
	
	inline virtual void setDisplayDebug(const unsigned displayDebug) {
		mDisplayDebug = displayDebug;
	}

	inline virtual void toggleDisplayDebug(const ParallelRenderingSupport::DisplayDebug flag) {
		mDisplayDebug ^= flag;
	}
	
	virtual void setSortOrder(unsigned order) = 0;
	
	/// Returns a color which is unique for all renderers 
	/// (and which looks nice for example for debugging)
	void getUniqueColor(double &r, double &g, double &b) const;

	/// Setup proper blending for the current compositing operator based on the
	/// parallel rendering extension.
	/// For instance, ParaComp needs glBlendFuncSeparateEXT(GL_SRC_ALPHA,  
	/// GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE_MINUS_SRC_ALPHA) for proper blending.
	virtual void setupBlending() const;
	
protected:
	virtual void _calculateFramelet();
	virtual void _calculateSortOrder();
	virtual void _displayBoundingRectangle(const bool filled) const;	
	virtual void _displayBoundingBox(const bool filled) const;	
	virtual void _renderBoundingRectangle(const bool filled) const;	
};

} // namespace MinGLE

#endif // __MINGLE_PARALLEL_PARALLEL_RENDERING_PROVIDER_H__
