//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file EventTransmission.cpp
/// Classes for event transmission throught the network 

#include <EventTransmission.h>
#include <SimpleCommSock.h>

#include <string.h>

namespace MinGLE {

EventSerializer::EventSerializer() :
	mDstWindowProxy(0) 
{
	_allocateBuffer();
}

EventSerializer::EventSerializer(WindowProxy *dstWindowProxy) :
	mDstWindowProxy(dstWindowProxy) 
{
	_allocateBuffer();
}

EventSerializer::~EventSerializer() {
	delete []mBuffer;
}
			
void EventSerializer::serializeKeyEvent(int key, unsigned modifiers, InputListener::KeyEventState state) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::KeyEvent;
	event.key = key;
	event.modifiers = modifiers;
	event.keyEventState = state;
	_serializeEvent(&event);
}
 
void EventSerializer::serializeMouseButtonEvent(int x, int y, InputListener::MouseButton button, InputListener::MouseEventState state) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::MouseButtonEvent;
	event.x = x, event.y = y;
	event.button = button;
	event.mouseEventState = state;
	_serializeEvent(&event);
}
 
void EventSerializer::serializeMouseMoveEvent(int x, int y, InputListener::MouseMoveState state) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::MouseMoveEvent;
	event.x = x, event.y = y;
	event.mouseMoveState = state;
	_serializeEvent(&event);
} 

void EventSerializer::serializeMouseMoveEvent(InputListener::MouseMoveState state) {
	// TODO is it required?
} 

void EventSerializer::serializeMouseWheelEvent(int x, int y, int delta) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::MouseWheelEvent;
	event.x = x, event.y = y;
	event.delta = delta;
	_serializeEvent(&event);
} 

void EventSerializer::serializeRegisterKeyboardEvent(int key, InputListener::KeyEventState state) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::RegisterKeyboardEvent;
	event.key = key;
	event.keyEventState = state;
	_serializeEvent(&event);
} 

void EventSerializer::serializeRegisterKeyboardModifierEvent(unsigned modifiers) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::RegisterKeyboardModifierEvent;
	event.modifiers = modifiers;
	_serializeEvent(&event);
} 

void EventSerializer::serializeRegisterMouseMoveEvent(int x, int y) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::RegisterMouseMoveEvent;
	event.x = x, event.y = y;
	_serializeEvent(&event);
} 

void EventSerializer::serializeRegisterMouseButtonEvent(InputListener::MouseButton button, InputListener::MouseEventState state) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::RegisterMouseButtonEvent;
	event.button = button;
	event.mouseEventState = state;
	_serializeEvent(&event);
} 

void EventSerializer::serializeRegisterMouseWheelEvent(int delta) {
	WindowListener::WindowEvent event;
	event.eventType = InputListener::RegisterMouseWheelEvent;
	event.delta = delta;
	_serializeEvent(&event);
} 

void EventSerializer::serializeResizeEvent(unsigned width, unsigned height) {
	WindowListener::WindowEvent event;
	event.eventType = WindowListener::ResizeEvent;
	event.width = width;
	event.height = height;
	_serializeEvent(&event);
}
  
void EventSerializer::serializeRepositionEvent(int left, int top) {
	WindowListener::WindowEvent event;
	event.eventType = WindowListener::RepositionEvent;
	event.left = left;
	event.top = top;
	_serializeEvent(&event);
}  

void EventSerializer::deserializeEvent(const WindowListener::WindowEvent* event) {
	switch(event->eventType) {
		case InputListener::KeyEvent:
			mDstWindowProxy->onKeyEvent(event->key, event->modifiers, event->keyEventState);
		break;
		case InputListener::MouseButtonEvent:
			mDstWindowProxy->onMouseButtonEvent(event->x, event->y, event->button, event->mouseEventState);
		break;
		case InputListener::MouseMoveEvent:
			mDstWindowProxy->onMouseMoveEvent(event->x, event->y, event->mouseMoveState);
		break;
		case InputListener::MouseWheelEvent:
			mDstWindowProxy->onMouseWheelEvent(event->x, event->y, event->delta);
		break;
		case InputListener::RegisterKeyboardEvent:
			mDstWindowProxy->_registerKeyboardEvent(event->key, event->keyEventState);
		break;
		case InputListener::RegisterKeyboardModifierEvent:
			mDstWindowProxy->_registerKeyboardModifierEvent(event->modifiers);
		break;
		case InputListener::RegisterMouseMoveEvent:
			mDstWindowProxy->_registerMouseMoveEvent(event->x, event->y);
		break;
		case InputListener::RegisterMouseButtonEvent:
			mDstWindowProxy->_registerMouseButtonEvent(event->button, event->mouseEventState);
		break;
		case InputListener::RegisterMouseWheelEvent:
			mDstWindowProxy->_registerMouseWheelEvent(event->delta);
		break;
		case WindowListener::ResizeEvent:
			mDstWindowProxy->setSize(event->width, event->height);
			// TODO: why do we need this manual even generation?
			// setSize should generate an event too
			mDstWindowProxy->onResize(event->width, event->height);
		break;
		case WindowListener::RepositionEvent:
			mDstWindowProxy->setPosition(event->left, event->top);
			// TODO: why do we need this manual even generation?
			// setSize should generate an event too
			mDstWindowProxy->onReposition(event->left, event->top);
		break;
		
		default:
			//printf("deserializedEvent %p\n", event);
		break;
	}
}

void EventSerializer::_allocateBuffer() {
	mBuffer = new char[BUFFER_SIZE];
}

void EventSerializer::_serializeEvent(const WindowListener::WindowEvent* event) {
	memcpy(mBuffer, event, sizeof(WindowListener::WindowEvent));
}

// =============================================================================

EventTransmitter::EventTransmitter(unsigned port) : 
	mCommSock(new SimpleCommSock()), mEventSerializer(new EventSerializer()), 
	mPort(port)
{
}

EventTransmitter::EventTransmitter(WindowProxy *dstWindowProxy, unsigned port) : 
	mCommSock(new SimpleCommSock()), mEventSerializer(new EventSerializer(dstWindowProxy)), 
	mPort(port)
{
}

EventTransmitter::~EventTransmitter() {
	delete mEventSerializer;
	delete mCommSock;
}

// =============================================================================

EventCollector::EventCollector(const std::vector<std::string>& hostList) :
	mHostList(hostList) 
{
}

EventCollector::~EventCollector() {
}

void EventCollector::waitForInjectors() {
	//printf("waiting for injectors\n");
	
	// Start server
	mCommSock->waitHosts(mHostList, mHostList[0], mPort);

	//printf("all injectors found\n");
}

void EventCollector::sendEvent() {
	mCommSock->broadcastMessage(mEventSerializer->getData(), mEventSerializer->getSize());
}

// =============================================================================

LocalEventInjector::LocalEventInjector(WindowProxy *dstWindowProxy) :
	EventTransmitter(dstWindowProxy) 
{
}

void LocalEventInjector::addEventToQueue(const WindowListener::WindowEvent *event) {
	// Lock the queue
	WindowEventQueue *eventQueue = mWindowEventQueue.lock();

	// Push the received event
	eventQueue->push(*event);

	// Release the queue
	mWindowEventQueue.unlock();
}

void LocalEventInjector::injectQueuedEvents() {

	// Lock the queue
	WindowEventQueue *eventQueue = mWindowEventQueue.lock();
	
	// Process all events in the queue
	while(eventQueue->size()) {
		
		// Get event from the queue
		WindowListener::WindowEvent event = eventQueue->front();
		
		// Deserialize the event
		mEventSerializer->deserializeEvent(&event);
		
		// Remove element
		eventQueue->pop();
	}
	
	// Release the queue
	mWindowEventQueue.unlock();
}

// =============================================================================

EventInjector::EventInjector(WindowProxy *dstWindowProxy, const std::string& thisHost) :
	LocalEventInjector(dstWindowProxy), mThisHost(thisHost) 
{
	//printf("starting injector: %s\n", thisHost.c_str());
}

EventInjector::~EventInjector() {
}

void EventInjector::connectToCollector(const std::string& collectorHost) {
	//printf("connect injector to %s\n", collectorHost.c_str());

	// Connect to the collector
	mCommSock->connect(mThisHost, collectorHost, mPort);
	
	// Create and start a joinable thread this object as runnable.
	// This thread will receive the events coming from the collector
	mThread = Threading::createThread(this, true);
	mThread->start();
	
	//printf("injector is up\n");
}

void EventInjector::run() {
	while(mThread->isRunning()) {
		// Read serialized event from the socket 
		WindowListener::WindowEvent* event = (WindowListener::WindowEvent *) mCommSock->recvMessage(sizeof(WindowListener::WindowEvent));

		// Add the event to the queue
		addEventToQueue(event);
	}
}

// =============================================================================

EventQueueWindowProxy::EventQueueWindowProxy(WindowProxy *mWindowProxy) :
	WindowProxy((Window *) mWindowProxy)
{
	// Create local event injector
	mEventInjector = new LocalEventInjector(this);
}

EventQueueWindowProxy::~EventQueueWindowProxy() {
	delete mEventInjector;
}

} // namespace MinGLE

