//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file ParallelRenderingSupport.cpp
/// General parallel rendering support manager class

#include <ParallelRenderingSupport.h>
#include <paracomp/include/Support.h>

#include <unistd.h>

#include <vector>
#include <string>

namespace MinGLE {

// Singleton instance
ParallelRenderingSupport *ParallelRenderingSupport::msInstance = 0;

// System type
ParallelRenderingSupport::SupportType ParallelRenderingSupport::msSupportType = ParallelRenderingSupport::UNKNOWN;

ParallelRenderingSupport::ParallelRenderingSupport() {
	// Register this object as a system listener
	// This is necessary, because the system class will 
	// delete this object before quitting
	System::registerSystemListener(this);
}

ParallelRenderingSupport::~ParallelRenderingSupport() {
}

void ParallelRenderingSupport::initialize(int *argc, char **argv, ParallelRenderingSupport::SupportType supportType) {
	
	// Parse command line
	unsigned sessionId;
	std::vector<std::string> hostList;
	_parseCommandLine(argc, argv, sessionId, hostList);
	
	// Call the other initialization method
	initialize(sessionId, hostList, supportType);
	
	/*
	switch(msSupportType = supportType) {
		case PARACOMP:
#if MINGLE_PARACOMP_SUPPORT == 1
			_initializeParaComp(argc, argv);
#else
	    	Except("ParallelRenderingSupport::initialize", "Unable to find parallel rendering support. Please recompile MinGLE-Parallel properly to enable at least one parallel rendering support.");
#endif
		break;

		case UNKNOWN:
#if MINGLE_PARACOMP_SUPPORT == 1
			_initializeParaComp(argc, argv);
#else
	    	Except("ParallelRenderingSupport::initialize", "Unable to find parallel rendering support. Please recompile MinGLE-Parallel properly to enable at least one parallel rendering support.");
#endif
	}
	*/
}

void ParallelRenderingSupport::initialize(unsigned sessionId, std::vector<std::string> hostList, SupportType supportType) {
	switch(msSupportType = supportType) {
		case PARACOMP:
#if MINGLE_PARACOMP_SUPPORT == 1
			_initializeParaComp(sessionId, hostList);
#else
	    	Except("ParallelRenderingSupport::initialize", "Unable to find parallel rendering support. Please recompile MinGLE-Parallel properly to enable at least one parallel rendering support.");
#endif
		break;

		case UNKNOWN:
#if MINGLE_PARACOMP_SUPPORT == 1
			_initializeParaComp(sessionId, hostList);
#else
	    	Except("ParallelRenderingSupport::initialize", "Unable to find parallel rendering support. Please recompile MinGLE-Parallel properly to enable at least one parallel rendering support.");
#endif
	}
}

void ParallelRenderingSupport::addParallelSupport(Window *window, CompositingOperator compositingOperator, SystemPart part) {
	// Call the corresponding method of the implementation 
	msInstance->_addParallelSupport(window, compositingOperator, part);

	//
	// Now some settings are done on the window
	//
	
	// Set screenshot and video frame filename pattern
	// containing the host name in order to avoid file occlusion
	std::string hostname = getHostName(); 
	window->setScreenShotPattern("screenshot_"+hostname+"_%03d.png");
	window->setVideoPattern(hostname + "_%02d_frame_%06d.tga");
}

std::string ParallelRenderingSupport::getHostName() {
	char hostname[256];
	Assert(!gethostname(hostname, sizeof(hostname)/sizeof(hostname[0])), "ParallelRenderingSupport::getHostName");
	
	return hostname;
}

void ParallelRenderingSupport::getUniqueColor(Window *window, double &r, double &g, double &b) {
	ParallelRenderingWindowProxy *prwProxy = _findParallelRenderingWindowProxy(window);
	Assert(prwProxy, "ParallelRenderingSupport::getUniqueColor");
	
	prwProxy->getUniqueColor(r, g, b);
}

/*
#if MINGLE_PARACOMP_SUPPORT == 1
void ParallelRenderingSupport::_initializeParaComp(int *argc, char **argv) {

	// TODO future work ---->
	const unsigned INVALID_SESSION_ID = 0;
	const std::string SESSION_ID_PARAMETER = "--session-id";
	const std::string HOSTLIST_PARAMETER = "--host-list";

	const std::string SESSION_ID_ERROR = "No session id is provided. Plase specify the session id with \'"+SESSION_ID_PARAMETER+" <sid>\'.";
	const std::string HOSTLIST_ERROR = "No hosts are provided. Plase specify the host list with \'"+HOSTLIST_PARAMETER+" <N> <host1> <host2> ... <hostN>\'.";

	// ParaComp session id
	unsigned sessionId = INVALID_SESSION_ID;
	
	// List of hosts related to this session
	// for the master: [master, slave1, slave2, ..., slaveN]
	// for the slaves: [slave]
	std::vector<std::string> hostList;

	// The remaining of the argument list
	std::vector<char *> remainingArgs; 

	// Process argument list
	for(int i=1; i<*argc; i++) {
		std::string arg = argv[i];
		
		if(SESSION_ID_PARAMETER == arg) {
			if(i+1<*argc) {
				
				i++;
			} 
			else
				Except("ParallelRenderingSupport::_initializeParaComp", SESSION_ID_ERROR);
		}
	}

	// Report error
	if(sessionId = INVALID_SESSION_ID) 
		Except("ParallelRenderingSupport::_initializeParaComp", SESSION_ID_ERROR);

	if(hostList.size())
		Except("ParallelRenderingSupport::_initializeParaComp", HOSTLIST_ERROR);
	// <---- TODO future work
	
	// Check command line args - we need atleast a session ID
    if(*argc < 4)
		Except("ParallelRenderingSupport::_initializeParaComp", "At least a session id, a host count, and a host name are required. Please specify them.");

	unsigned sessionId = StringUtils::parse<unsigned>(argv[1]);

	// Get host count
	int hostCount = StringUtils::parse<unsigned>(argv[2]);
	Assert(hostCount <= *argc-3, "ParallelRenderingSupport::_initializeParaComp");

	// Get host list
	std::vector<std::string> hostList;
	for(int i=0; i<hostCount; i++)
		hostList.push_back(argv[3+i]);

	// Initialize parallel rendering support
	msSupportType = PARACOMP;
	msInstance = new ParaCompSupport(sessionId, hostList);
	
	// Shift cmd line arguments
	for(int i=2+hostCount+1; i<*argc; i++) argv[i-2-hostCount] = argv[i];
	*argc-=2+hostCount;
}
#endif // MINGLE_PARACOMP_SUPPORT == 1
*/

void ParallelRenderingSupport::_parseCommandLine(int *argc, char **argv, unsigned &sessionId, std::vector<std::string> &hostList) {
	// Check command line args - we need atleast a session ID
    if(*argc < 4)
		Except("ParallelRenderingSupport::_parseCommandLine", "At least a session id, a host count, and a host name are required. Please specify them.");

	sessionId = StringUtils::parse<unsigned>(argv[1]);

	// Get host count
	int hostCount = StringUtils::parse<unsigned>(argv[2]);
	Assert(hostCount <= *argc-3, "ParallelRenderingSupport::_parseCommandLine");

	// Get host list
	for(int i=0; i<hostCount; i++)
		hostList.push_back(argv[3+i]);

	// Shift cmd line arguments
	for(int i=2+hostCount+1; i<*argc; i++) argv[i-2-hostCount] = argv[i];
	*argc-=2+hostCount;
}

#if MINGLE_PARACOMP_SUPPORT == 1
void ParallelRenderingSupport::_initializeParaComp(unsigned sessionId, std::vector<std::string> hostList) {
	// Initialize parallel rendering support
	msSupportType = PARACOMP;
	msInstance = new ParaCompSupport(sessionId, hostList);
}
#endif // MINGLE_PARACOMP_SUPPORT == 1

ParallelRenderingWindowProxy *ParallelRenderingSupport::_findParallelRenderingWindowProxy(Window *window) {
	return WindowProxy::findProxiedWindow<ParallelRenderingWindowProxy>(window);
}

} // namespace MinGLE

