//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================
//
// Original source of this class: HP XC ParaComp Kit.
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file ParallelRenderingWindowProxy.cpp
/// Window proxy class for that supplies parallel rendering support for a window

#include <ParallelRenderingWindowProxy.h>

namespace MinGLE {

// VERTICES

//     6,------,7                              
//     /|     /|                               
//    / |    / |                       
//  4+------+5 |         z|  /y                  
//   | 2|___|__|3         | /             
//   | /    |  /          |/      x        
//   |/     | /           +-------       
//   +------+/                          
//  0        1               

const float ParallelRenderingWindowProxy::BOX_VERTICES[8][3] = {
/* x,    y,	  z      vertex*/
{ 0.0, 0.0, 0.0 } /* 0 */,

{ 1.0, 0.0, 0.0 } /* 1 */,

{ 0.0, 1.0, 0.0 } /* 2 */,

{ 1.0, 1.0, 0.0 } /* 3 */,

{ 0.0, 0.0, 1.0 } /* 4 */,

{ 1.0, 0.0, 1.0 } /* 5 */,

{ 0.0, 1.0, 1.0 } /* 6 */,

{ 1.0, 1.0, 1.0 } /* 7 */
};

const unsigned ParallelRenderingWindowProxy::BOX_FACES[6][4] = {

{ 0, 1, 5, 4 },

{ 1, 3, 7, 5 },

{ 3, 2, 6, 7 },

{ 2, 0, 4, 6 },

{ 0, 1, 3, 2 },

{ 0, 2, 3, 1 }

};

ParallelRenderingWindowProxy::ParallelRenderingWindowProxy(Window *window,
		ParallelRenderingSupport::CompositingOperator compositingOperator) :
	WindowProxy(window), mRendererCount(INVALID_RENDERER_INDEX),
			mThisRenderer(INVALID_RENDERER_INDEX),
			mCompositingOperator(compositingOperator),
			mAutoCalculateFramelet(DEFAULT_AUTOCALCULATE_FRAMELET),
			mAutoCalculateSortOrder(DEFAULT_AUTOCALCULATE_SORTORDER),
			mContributePixels(DEFAULT_CONTRIBUTE_PIXELS),
			mDisplayOutput(DEFAULT_DISPLAY_OUTPUT),
			mDisplayDebug(DEFAULT_DISPLAY_DEBUG), mPartition(0),
			mValidFramelet(false) {
}

void ParallelRenderingWindowProxy::getUniqueColor(double &r, double &g,
		double &b) const {
	Assert(mRendererCount > 0, "ParallelRenderingWindowProxy::getUniqueColor");

	double h = (double) mThisRenderer / mRendererCount;
	double s = 1.0;
	double i = 1.0;

	MathUtils::hsi2rgb(h, s, i, r, g, b);
}

void ParallelRenderingWindowProxy::setupBlending() const {
	// Get blending factors
	unsigned sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha;
	ParallelRenderingSupport::getBlendingFactors(mCompositingOperator,
			sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);

	// Setup blending
	glEnable(GL_BLEND);
	glBlendFuncSeparateEXT(sfactorRGB, dfactorRGB, sfactorAlpha, dfactorAlpha);
}

void ParallelRenderingWindowProxy::_calculateFramelet() {
	// Get OpenGL viewport
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	// If the no partition is defined we specify whole screen framelet 
	if (!mPartition) {
		mBoundingRectangleOffset[0] = mBoundingRectangleOffset[1] = 0;
		mBoundingRectangleSize[0] = viewport[2];
		mBoundingRectangleSize[1] = viewport[3];

		mValidFramelet = true;
		return;
	}

	// Get OpenGL matrices
	GLdouble winx[8], winy[8], _winz;
	GLdouble modelMatrix[16], projMatrix[16];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

	// Project vertices
	for (int i=0; i<8; i++) {
		gluProject(mPartition->offset[0] + BOX_VERTICES[i][0]
				*mPartition->size[0], mPartition->offset[1]
				+ BOX_VERTICES[i][1]*mPartition->size[1], mPartition->offset[2]
				+ BOX_VERTICES[i][2]*mPartition->size[2], modelMatrix,
				projMatrix, viewport, winx+i, winy+i, &_winz);
	}

	// Calculate bounds
	GLdouble minWinX = winx[0], maxWinX = winx[0], minWinY = winy[0], maxWinY =
			winy[0];
	for (int i=0; i<8; i++) {
		if (winx[i] < minWinX)
			minWinX = winx[i];
		if (winx[i] > maxWinX)
			maxWinX = winx[i];
		if (winy[i] < minWinY)
			minWinY = winy[i];
		if (winy[i] > maxWinY)
			maxWinY = winy[i];
	}

	// Truncate coorinates on viewport borders
	if (minWinX < 0)
		minWinX = 0;
	if (minWinY < 0)
		minWinY = 0;
	if (maxWinX < 0)
		maxWinX = 0;
	if (maxWinY < 0)
		maxWinY = 0;
	if (minWinX > viewport[2]-1)
		minWinX = viewport[2]-1;
	if (minWinY > viewport[2]-1)
		minWinY = viewport[3]-1;
	if (maxWinX > viewport[2]-1)
		maxWinX = viewport[2]-1;
	if (maxWinY > viewport[2]-1)
		maxWinY = viewport[3]-1;

	// Calculate offsets and sizes
	mBoundingRectangleOffset[0] = (unsigned) minWinX,
			mBoundingRectangleOffset[1] = (unsigned) minWinY;
	mBoundingRectangleSize[0] = (unsigned) (maxWinX - minWinX),
			mBoundingRectangleSize[1] = (unsigned) (maxWinY - minWinY);

	// Check zero framelet dimensions
	if (mBoundingRectangleSize[0] && mBoundingRectangleSize[1])
		mValidFramelet = true;
	else
		mValidFramelet = false;
}

void ParallelRenderingWindowProxy::_calculateSortOrder() {

	// If the no partition is defined we specify the
	// id of this renderer as sort order
	if (!mPartition) {
		setSortOrder(mThisRenderer);
		return;
	}

	// Get current eye position
	double eyeX, eyeY, eyeZ;
	GLUtils::getViewPoint(eyeX, eyeY, eyeZ);

	// Calculate and set sort order
	setSortOrder(BoxPartitioner<double>::getSortOrder<unsigned>(mPartition, eyeX, eyeY, eyeZ)
	);
}

void ParallelRenderingWindowProxy::_displayBoundingRectangle(const bool filled) const {
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0.0, viewport[2], 0.0, viewport[3]);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	GLboolean isLighting = glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);
	GLboolean isBlending = glIsEnabled(GL_BLEND);

	// Set unique color
	double r, g, b;
	getUniqueColor(r, g, b);

	//
	// Wireframe
	//
	// Display an empty bounding rectangle with full alpha
	glColor3d(r, g, b);
	glDisable(GL_BLEND);
	_renderBoundingRectangle(false);

	//
	// Filled
	//
	if(filled) {
		// Using the default blending function, the graphics card will not hold the correct
		// alpha values (what is called destination alpha).  We need the correct values when
		// we composite the images later.  The blending function below will produce correct
		// destination alpha values.
		setupBlending();

		// Display a filled bounding rectangle with low alpha
		glColor4d(r, g, b, 0.66);
		_renderBoundingRectangle(true);
	}

	// Restore settings
	if(isLighting) glEnable(GL_LIGHTING);
	if(isBlending) glEnable(GL_BLEND);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void ParallelRenderingWindowProxy::_displayBoundingBox(const bool filled) const {

	GLboolean isLighting = glIsEnabled(GL_LIGHTING); glDisable(GL_LIGHTING);

	// Set unique color
	double r, g, b;
	getUniqueColor(r, g, b);

	glColor3d(0.75*r, 0.75*g, 0.75*b);

	for(unsigned face=0; face<6; face++) {
		glBegin(GL_LINE_LOOP);
		for(unsigned vertex=0; vertex<4; vertex++)
		glVertex3f(
		mPartition->offset[0] + BOX_VERTICES[BOX_FACES[face][vertex]][0]*mPartition->size[0],
		mPartition->offset[1] + BOX_VERTICES[BOX_FACES[face][vertex]][1]*mPartition->size[1],
		mPartition->offset[2] + BOX_VERTICES[BOX_FACES[face][vertex]][2]*mPartition->size[2]
		);
		glEnd();
	}

	// Restore settings
	if(isLighting) glEnable(GL_LIGHTING);

}

void ParallelRenderingWindowProxy::_renderBoundingRectangle(const bool filled) const {
	glBegin(filled ? GL_QUADS : GL_LINE_LOOP);
	glVertex2i(mBoundingRectangleOffset[0]+1, mBoundingRectangleOffset[1]+1);
	glVertex2i(mBoundingRectangleOffset[0]+mBoundingRectangleSize[0]-1, mBoundingRectangleOffset[1]+1);
	glVertex2i(mBoundingRectangleOffset[0]+mBoundingRectangleSize[0]-1, mBoundingRectangleOffset[1]+mBoundingRectangleSize[1]-1);
	glVertex2i(mBoundingRectangleOffset[0]+1, mBoundingRectangleOffset[1]+mBoundingRectangleSize[1]-1);
	glEnd();
}

} // namespace MinGLE

