//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.   
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.  
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \. 
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \.
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\. 
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.   
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
//                       Parallel Rendering Extension
// =============================================================================
//
// Original source of this class: HP XC ParaComp Kit.
//
// Copyright 2007 Balazs Domonkos
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/// @file SimpleCommSock.cpp
/// Simple socket based communication

#include <SimpleCommSock.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

namespace MinGLE {

SimpleCommSock::SimpleCommSock() {
    mMySock = -1;
}

SimpleCommSock::~SimpleCommSock() {
	// TODO why not??
	cleanup();
}
			
void SimpleCommSock::waitHosts(const std::vector<std::string>& hosts, const std::string& me, unsigned port) {
	
	try {
		// Allready connected to hosts, close first
		Assert(mMySock==-1, "SimpleCommSock::waitHosts");

		// Create socket
		mMySock = socket(AF_INET, SOCK_STREAM, 0);
    	Assert(mMySock>=0, "SimpleCommSock::waitHosts");

	// Set sock options for high speed data transfer
	setOpts();
    
		struct sockaddr_in serv_addr;
    	memset(&serv_addr, 0, sizeof(serv_addr));
    	serv_addr.sin_family=AF_INET;
    	serv_addr.sin_addr.s_addr=INADDR_ANY;
    	serv_addr.sin_port=htons(port);
    
    	// Bind to socket
    	//printf("bind elott\n");
    	Assert(bind(mMySock, reinterpret_cast<sockaddr*>(&serv_addr), sizeof(serv_addr)) >= 0, "SimpleCommSock::waitHosts");
		//printf("Server waiting on port %d for %lu slaves to connect\n", port, hosts.size()-1);    	

        // Allow connections to be in progress
        Assert(listen(mMySock, MAX_CONNECTIONS)>=0, "SimpleCommSock::waitHosts");

        // Master cannot send to itself
        mHostToIndex[me]=0;
        mHostSock.push_back(-1);

        // Wait till everybody connects
        while(mHostSock.size() < hosts.size())
        {
        	sockaddr_in from;
            socklen_t len = sizeof(from);

			// Accept connection
            int slaveSock = accept(mMySock, reinterpret_cast<sockaddr*>(&from),&len);
            Assert(slaveSock>=0, "SimpleCommSock::waitHosts");
            
            // First thing any slave sends us is the name
            readSock(slaveSock, mMsgBuffer, sizeof(mMsgBuffer));
            mHostToIndex[mMsgBuffer]=mHostSock.size();
            mHostSock.push_back(slaveSock);
        }


        //printf("Server is up\n");
        
	} catch(Exception e) {

		// Clean up on error
		cleanup();
		throw e;
	}
}
			
void SimpleCommSock::connect(const std::string& me, const std::string& master, unsigned port) {

	try {
		// Allready connected to hosts, close first
		Assert(mMySock==-1, "SimpleCommSock::connect");

        struct hostent* hp;
        struct hostent ret;
        char buffer[1000];
        int h_errnop;
        int errval = gethostbyname_r(master.c_str(), &ret, buffer, sizeof(buffer), &hp, &h_errnop);

		// Resolve master host
		Assert(!errval, "SimpleCommSock::connect");

		// Create socket
        mMySock = socket(AF_INET, SOCK_STREAM, 0);
        Assert(mMySock>=0, "SimpleCommSock::connect");

	// Set sock options for high speed data transfer
	setOpts();
        
        sockaddr_in si;
        memset(&si,0,sizeof(si));
        si.sin_family = AF_INET;
        memcpy((void*)&si.sin_addr.s_addr, (void*)hp->h_addr, 4); // IPv4...
        si.sin_port = htons(port);

        bool connected = false;
        int nMaxRetries = 10;
        int nRetries = nMaxRetries;

		// Connect to master
        while((!connected) && (nRetries--)) {
			if(::connect(mMySock, (sockaddr*)&si, sizeof(si))<0)
				sleep(1);
			else {
				connected = true;
				break;
			}
        }
		Assert(connected, "SimpleCommSock::connect");

        // Copy our name and send it to master
        strcpy(mMsgBuffer, me.c_str());
		writeSock(mMySock, mMsgBuffer, sizeof(mMsgBuffer));


		//printf("slave %s connected to master", me.c_str());
	} catch(Exception e) {

		// Clean up on error
		cleanup();
		throw e;
	}
}
			
void SimpleCommSock::broadcastMessage(void *data, size_t size) {
	try {
    	// master=index0
    	for(unsigned int i=1;i<mHostSock.size();i++) 
			writeSock(mHostSock[i], data, size);
	} catch(Exception e) {
		// Clean up on error
		cleanup();
		throw e;
	}
}
			
void* SimpleCommSock::recvMessage(unsigned size) {
	try {
		readSock(mMySock, mMsgBuffer, size);
		return mMsgBuffer;
	} catch(Exception e) {
		// Clean up on error
		cleanup();
		throw e;
	}
	
	return 0;
}
			
void SimpleCommSock::close() {
	cleanup();
}

void SimpleCommSock::readSock(int sock, void *buffer, unsigned int nBytes) {
	// Note no cleanup catch block is needed since it is a private method.
	// The generated exeptions will be handles in the public caller methods
	
	int nRemaining = nBytes;
	char* ptr= (char*)buffer;

    while(nRemaining>0) {
		int nRead = ::recv(sock, ptr, nRemaining, MSG_NOSIGNAL);
		if(nRead<0) {
			// Tolerable error
			// TODO???
			//Assert(errno == EINTR, "SimpleCommSock::readSock");
			continue;
		}
        nRemaining -= nRead;
        ptr += nRead;
	}
}

void SimpleCommSock::writeSock(int sock, void *buffer, unsigned int nBytes) {
	// Note no cleanup catch block is needed since it is a private method.
	// The generated exeptions will be handles in the public caller methods

	int nRemaining = nBytes;
    const char* ptr= (const char*)buffer;

    while(nRemaining>0) {
    	int nWrite = ::send(sock, ptr, nRemaining, MSG_NOSIGNAL);
		if(nWrite<0) {
			// Tolerable error
			// TODO???
			//Assert(errno == EINTR, "SimpleCommSock::writeSock");
			continue;
		}
        
        nRemaining -= nWrite;
        ptr += nWrite;
	}

}

void SimpleCommSock::cleanup() {
	// Wait for a packet from the slaves
    for(unsigned int i=0;i<mHostSock.size();i++) {
    	if(mHostSock[i]!=-1)
			// slave will write a packet before closing
			// and then close it's end, so we close after it to avoid the TIME_WAIT state!
            readSock(mHostSock[i], mMsgBuffer, sizeof(mMsgBuffer));
	}
    
    if(mHostSock.size()>0)
    	for(unsigned int i=0;i<mHostSock.size();i++)
			if(mHostSock[i]!=-1)
				::close(mHostSock[i]);

	// Then close the socket
    if(mMySock!=-1) {
		// Slave - write a full packet to the master to indicate exit..
		if(mHostSock.size()==0)
            writeSock(mMySock, mMsgBuffer, sizeof(mMsgBuffer));
    	
    	// and then close
        ::close(mMySock); mMySock=-1;
	}

	//if(mHostSock.size()>0) printf("Master closing sockets...\n");

	mHostSock.clear();
    mHostToIndex.clear();
}

void SimpleCommSock::setOpts() {
    // Set no delay TCP mode
    int flag=1;
    Assert(::setsockopt(mMySock, IPPROTO_TCP, TCP_NODELAY, (char*) &flag, sizeof(int)) >= 0, "SimpleCommSock::setOpts");
    
    // Reusing address and port
    int nAllowReuse = 1;
    Assert(::setsockopt(mMySock, SOL_SOCKET, SO_REUSEADDR, (char*) &nAllowReuse, sizeof(nAllowReuse)) >=0, "SimpleCommSock::setOpts");
    //Assert(::setsockopt(mMySock, SOL_SOCKET, SO_REUSEPORT, (char*) &nAllowReuse, sizeof(nAllowReuse)) >=0, "SimpleCommSock::setOpts");

    // The system shall block the calling thread during close() until it can transmit the data or until the time expires
    linger lingerSet;
    lingerSet.l_onoff = 0;
    lingerSet.l_linger = 0;
    Assert(::setsockopt(mMySock, SOL_SOCKET, SO_LINGER, (char*)&lingerSet, sizeof(lingerSet)) >=0, "SimpleCommSock::setOpts");
}

} // namespace MinGLE

