//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Atlas.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Atlas.h
/// Generic, image, and texture atlas classes

#ifndef __MINGLE_ATLAS_H__
#define __MINGLE_ATLAS_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Imaging.h"
#include "StringUtils.h"
#include "mingle-gl.h"

// System headers
#include <stdlib.h>
#include <map>

namespace MinGLE {

template<typename INDEX, typename COORDT> class Atlas {

public:
	typedef struct {
		COORDT u1, v1;
		COORDT u2, v2;
	} CoordSet2D;

protected:
	std::map<INDEX, CoordSet2D> mRegions;

public:
	Atlas() {
	}
	virtual ~Atlas() {
	}

	virtual CoordSet2D *getCoordSet(const INDEX& key) {
		if (mRegions.find(key) != mRegions.end())
			return &mRegions[key];

		return 0;
	}
	virtual void addCoordSet(const INDEX& key, const CoordSet2D *coordSet) {
		CoordSet2D *c = &mRegions[key];
		memcpy(c, coordSet, sizeof(CoordSet2D));
	}
};

// =============================================================================

template < typename INDEX >
class ImageAtlas : public Atlas<INDEX, unsigned> {

protected:
	typedef typename std::map<INDEX, typename Atlas< INDEX, unsigned >::CoordSet2D>::iterator CoordSet2DIterator;
	static const std::string XML_ROOT_ELEMENT;
	static const std::string XML_IMAGEFILE_ELEMENT;
	static const std::string XML_REGION_ELEMENT;
	static const std::string XML_REGION_KEY_ATTRIBUTE;
	static const std::string XML_REGION_U1_ATTRIBUTE;
	static const std::string XML_REGION_V1_ATTRIBUTE;
	static const std::string XML_REGION_U2_ATTRIBUTE;
	static const std::string XML_REGION_V2_ATTRIBUTE;
	static const std::string XML_IMAGEFILE_EXTENSION;

protected:
	Image *mImage;
	
public:
	ImageAtlas(unsigned width, unsigned height, GLenum format = Image::DEFAULT_FORMAT,
			GLenum type = Image::DEFAULT_TYPE, void *data = 0)
	{
		// Create new image atlas
		mImage = Imaging::createImage(width, height, format, type, data);
	}

	ImageAtlas(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) {
		// Load image atlas from file
		_loadFromFile(filename, vfs);
	}

	virtual ~ImageAtlas() {
		delete mImage;
	}
	
	inline Image *getImage() {
		return mImage;
	}
	
	virtual TextureAtlas<INDEX> *convertToTextureAtlas() {
		// TODO
		return 0;
	}

	// Note only a few implementations exist for these methods
	// in order to reduce the refrenced headers (tinyxml is needed for
	// saving and loading atlas files)
	virtual void saveToFile(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
protected:
	virtual void _loadFromFile(const std::string& filename, VirtualFileSystem *vfs);
};

// =============================================================================

template<typename INDEX> class TextureAtlas : public Atlas<INDEX, double> {

	virtual ImageAtlas<INDEX> *convertToImageAtlas() {
		// TODO
		return 0;
	}
};

} // namespace MinGLE

#endif // __MINGLE_ATLAS_H__
