//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Cameras.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Cameras.h
/// Camera classes

#ifndef __MINGLE_CAMERAS_H__
#define __MINGLE_CAMERAS_H__

// Include class definitions
#include "classes.h"

namespace MinGLE {

class Camera {
protected:
	static const double DEFAULT_CAMERA_POSITION[3];
	static const double DEFAULT_CAMERA_TARGET[3];
	static const double DEFAULT_UP_VECTOR[3];
	static const double DEFAULT_NEAR_CLIP = 0.01;
	static const double DEFAULT_FAR_CLIP = 100.0;
	static const double DEFAULT_IMAGE_WIDTH = 1.0;
	static const double DEFAULT_IMAGE_HEIGHT = 1.0;
	static const bool DEFAULT_AUTORECALCULATE = true;

protected:
	double mPosition[3];
	double mTarget[3];
	double mUpVector[3];
	double mNearClip, mFarClip;
	double mImageWidth, mImageHeight;

	bool mAutoRecalculate;

public:
	Camera();
	virtual ~Camera();

	// Getters
	inline virtual const double *getPosition() const {
		return mPosition;
	}
	inline virtual const double *getTarget() const {
		return mTarget;
	}
	inline virtual const double *getUpVector() const {
		return mUpVector;
	}
	inline virtual const double getNearClip() const {
		return mNearClip;
	}
	inline virtual const double getFarClip() const {
		return mFarClip;
	}
	inline virtual const double getImageWidth() const {
		return mImageWidth;
	}
	inline virtual const double getImageHeight() const {
		return mImageHeight;
	}

	// Setters
	inline virtual void setPosition(double x, double y, double z) {
		mPosition[0]=x, mPosition[1]=y, mPosition[2]=z;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setTarget(double x, double y, double z) {
		mTarget[0]=x, mTarget[1]=y, mTarget[2]=z;
		if (mAutoRecalculate)
			recalculateMatrices();
	}

	/// Since the camera is target oriented the setDirection() methods
	/// should be called after setting the position of the camera with
	/// setPosition()
	inline virtual void setDirection(double dirX, double dirY, double dirZ) {
		mTarget[0] = mPosition[0] + dirX, mTarget[1] = mPosition[1] + dirY,
				mTarget[2] = mPosition[2] + dirZ;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setUpVector(double x, double y, double z) {
		mUpVector[0]=x, mUpVector[1]=y, mUpVector[2]=z;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setNearClip(double near) {
		mNearClip = near;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setFarClip(double far) {
		mFarClip = far;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setImageWidth(double width) {
		mImageWidth = width;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setImageHeight(double height) {
		mImageHeight = height;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setImageSize(double width, double height) {
		mImageWidth = width, mImageHeight = height;
		if (mAutoRecalculate)
			recalculateMatrices();
	}
	inline virtual void setAutoRecalculate(bool recalculate) {
		mAutoRecalculate = recalculate;
	}

	// Calculation
	inline virtual void recalculateMatrices() {
		recalculateProjectionMatrix();
		recalculateModelViewMatrix();
	}
	virtual void recalculateProjectionMatrix() = 0;
	virtual void recalculateModelViewMatrix() = 0;
	virtual void getProjectionMatrix(double *matrix) = 0;
	virtual void getModelViewMatrix(double *matrix) = 0;
};

// =============================================================================

class GLManagedCamera : public Camera {

public:
	virtual ~GLManagedCamera();

	virtual void getProjectionMatrix(double *matrix);
	virtual void getModelViewMatrix(double *matrix);
};

// =============================================================================

class SymmetricFrustumCamera : public GLManagedCamera {
protected:
	static const double DEFAULT_FOVY = 45.0;

protected:
	double mFovY;

public:
	SymmetricFrustumCamera(double fovy = DEFAULT_FOVY);
	virtual ~SymmetricFrustumCamera();

	// Getters
	inline virtual const double getFovy() const {
		return mFovY;
	}

	// Setters
	inline virtual void setFovy(double fovy) {
		mFovY = fovy;
	}

	virtual void recalculateProjectionMatrix();
	virtual void recalculateModelViewMatrix();
};

} // namespace MinGLE

#endif // __MINGLE_CAMERAS_H__
