//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Configuration.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Configuration.h
/// Configuration file support

#ifndef __MINGLE_CONFIGURATION_H__
#define __MINGLE_CONFIGURATION_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Exception.h"
#include "VirtualFileSystem.h"
#include "StringUtils.h"

// System headers
#include <vector>
#include <string>

namespace MinGLE {

class Configuration {

public:
	enum EntryType {
		SIMPLE_ENTRY,
		COMPOSITE_ENTRY
	};

	struct Entry;
	typedef std::vector<Entry *> EntryList;

	union EntryValue {
		std::string *simpleValue;
		EntryList *compositeValue;
	};

	struct Entry {
		std::string name;
		EntryType type;
		EntryValue value;

		Entry(const std::string& _name, const std::string& _value);
		Entry(const std::string& _name);
		Entry(const std::string& _name, EntryList *_value);
		~Entry();

		const std::string& getStringValue();
		template<typename T> inline T getValue() {
			return StringUtils::parse<T>(getStringValue());
		}

		Entry *getChild(const unsigned index);
		Entry *getFirstChild(const std::string& name);
		EntryList *getChildren();
		EntryList findChildren(const std::string& name);

		EntryList getEntries(const std::string& searchPath);
	};

	static const char ENTRY_SEPARATOR;

protected:
	Entry *mRootEntry;

public:
	Configuration();
	~Configuration();

	inline Entry *getRootEntry() { return mRootEntry; }
	Entry *getFirstChild(const std::string& name);
	EntryList getEntries(const std::string& searchPath, bool onlySimpleEntries = true);

	void dumpForDebug();

protected:
	Entry * _addChild(Entry *parent, const std::string& name, const std::string& value);
	Entry * _addChild(Entry *parent, const std::string& name);

	void _dumpForDebug(const Entry *entry, const unsigned level);

}; // class Configuration

// =============================================================================

class ConfigurationManager :
public ProviderManager<ConfigurationManager, FileFormatFunctionality> {
public:
	virtual ~ConfigurationManager();

	static Configuration *createConfiguration(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());

	static Configuration *createConfiguration(const std::string& blockName, const std::string& filename);

protected:
	virtual Configuration *_createConfiguration(const std::string& filename, VirtualFileSystem *vfs) = 0;
}; // class ConfigurationManager

} // namespace MinGLE

#endif // __MINGLE_CONFIGURATION_H__
