//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/FontRendering.h
// Creation Date: December 12th 2007
// Revision Date: January 25th 2008
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file FontRendering.h
/// Generic font rendering classes

#ifndef __MINGLE_FONT_RENDERING_H__
#define __MINGLE_FONT_RENDERING_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Manager.h"
#include "VirtualFileSystem.h"
#include "mingle-gl.h"

// System headers
#include <string>

namespace MinGLE {

class Font {

public:
	static const float DEFAULT_FONT_COLOR[4];

protected:
	float mColor[4];

public:
	Font();
	virtual ~Font();

	inline void setColor(float r, float g, float b, float a = 1.0) {
		mColor[0] = r, mColor[1] = g, mColor[2] = b, mColor[3] = a;
	}
	inline void setColor(float color[4]) {
		mColor[0] = color[0], mColor[1] = color[1], mColor[2] = color[2],
				mColor[3] = color[3];
	}

	virtual size_t getHeight() const = 0;
	virtual void
			renderToFrameBuffer(const std::string& str, int left, int top) const = 0;
	//virtual void saveToFile(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) = 0;

}; // class Font

// =============================================================================

class RasterFont : public Font {
public:
	static const unsigned DEFAULT_FONT_WIDTH = 0;
	static const unsigned DEFAULT_FONT_HEIGHT = 32;
	static const unsigned DEFAULT_FIRST_CHARACTER = 33;
	static const unsigned DEFAULT_LAST_CHARACTER = 167;
	static const GLenum DEFAULT_FONT_TYPE = GL_UNSIGNED_BYTE;
	// static const GLenum DEFAULT_FONT_FORMAT = GL_LUMINANCE_ALPHA;
	static const GLenum DEFAULT_FONT_FORMAT = GL_RGBA; // IL_LUMINANCE_ALPHA bug is fixed only in DevIL 1.6.8

	enum ImageScaleMode {
		/// Use no scale mode if possible
		IMAGE_SCALE_TRY_NOSCALE,
		/// Force no scaling
		IMAGE_CREATION_NOSCALE,
		/// Always scale image to the nearset power of 2 size (compatibility mode)
		IMAGE_CREATION_SCALE
	};
	static const ImageScaleMode DEFAULT_IMAGE_SCALE_MODE = IMAGE_CREATION_SCALE;

protected:
	size_t mWidth, mHeight;
	long mFaceIndex;
	size_t mFirstCharacter, mLastCharacter;
	GLenum mFormat;
	GLenum mType;

	unsigned mBytesPerPixel;

	// Image atlas
	ImageAtlas<int> *mImageAtlas;

	ImageScaleMode mImageScaleMode;

public:
	RasterFont(unsigned height = DEFAULT_FONT_HEIGHT,
			unsigned width = DEFAULT_FONT_WIDTH, long faceIndex = 0,
			unsigned firstCharacter = DEFAULT_FIRST_CHARACTER,
			unsigned lastCharacter = DEFAULT_LAST_CHARACTER,
			GLenum format = DEFAULT_FONT_FORMAT,
			GLenum type = DEFAULT_FONT_TYPE,
			ImageScaleMode imageScaleMode = DEFAULT_IMAGE_SCALE_MODE);
	virtual ~RasterFont();

	inline size_t getWidth() const {
		return mWidth;
	}
	inline size_t getHeight() const {
		return mHeight;
	}
	inline ImageAtlas<int> *getImageAtlas() {
		return mImageAtlas;
	}
	inline void setImageScaleMode(ImageScaleMode mode) {
		mImageScaleMode = mode;
	}

	virtual void
			renderToFrameBuffer(const std::string& str, int left, int top) const;
};

// =============================================================================

class FontRendering :
	public ProviderManager<FontRendering, FileFormatFunctionality> {
public:
	virtual ~FontRendering();

	static Font
			*createFont(
					const std::string& filename,
					VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem(),
					size_t height = RasterFont::DEFAULT_FONT_HEIGHT,
					size_t width = RasterFont::DEFAULT_FONT_WIDTH,
					long faceIndex = 0,
					size_t firstCharacter = RasterFont::DEFAULT_FIRST_CHARACTER,
					size_t lastCharacter = RasterFont::DEFAULT_LAST_CHARACTER,
					GLenum format = RasterFont::DEFAULT_FONT_FORMAT,
					GLenum type = RasterFont::DEFAULT_FONT_TYPE,
					RasterFont::ImageScaleMode imageScaleMode = RasterFont::DEFAULT_IMAGE_SCALE_MODE);
	static Font
			*createFont(
					const std::string& blockName,
					const std::string& filename,
					size_t height = RasterFont::DEFAULT_FONT_HEIGHT,
					size_t width = RasterFont::DEFAULT_FONT_WIDTH,
					long faceIndex = 0,
					size_t firstCharacter = RasterFont::DEFAULT_FIRST_CHARACTER,
					size_t lastCharacter = RasterFont::DEFAULT_LAST_CHARACTER,
					GLenum format = RasterFont::DEFAULT_FONT_FORMAT,
					GLenum type = RasterFont::DEFAULT_FONT_TYPE,
					RasterFont::ImageScaleMode imageScaleMode = RasterFont::DEFAULT_IMAGE_SCALE_MODE);

	static void
			saveFont(
					Font *font,
					const std::string& filename,
					VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());

protected:
	virtual Font *_createFont(const std::string& filename,
			VirtualFileSystem *vfs, size_t height, size_t width,
			long faceIndex, size_t firstCharacter, size_t lastCharacter,
			GLenum format, GLenum type,
			RasterFont::ImageScaleMode imageScaleMode) = 0;
	virtual void _saveFont(Font *font, const std::string& filename,
			VirtualFileSystem *vfs) = 0;

}; // class FontRendering

} // namespace MinGLE

#endif // __MINGLE_FONT_RENDERING_H__
