//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Gui.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Gui.h
/// Very simple gui support

#ifndef __MINGLE_GUI_H__
#define __MINGLE_GUI_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "VirtualFileSystem.h"
#include "Atlas.h"
#include "FontRendering.h"
#include "WindowListener.h"
#include "Timer.h"
#include "TimerListener.h"

// System headers
#include <string>
#include <list>
#include <map>

namespace MinGLE {

class Widget : public InputListener {
	friend class Panel;

protected:
	enum ScaleMode {ABSOLUTE, RELATIVE};
	struct Bounds {
		ScaleMode leftMode;
		union {
			int absLeft;
			float relLeft;
		};
		ScaleMode topMode;
		union {
			int absTop;
			float relTop;
		};
		ScaleMode widthMode;
		union {
			int absWidth;
			float relWidth;
		};
		ScaleMode heightMode;
		union {
			int absHeight;
			float relHeight;
		};
	};
	struct EffectiveSizes {
		int width, height;
	};

protected:
	static const float DEFAULT_COLOR[4];
	static const unsigned char QUAD_INDICES[4];
	static float mDefaultColor[4];

	float mColor[4];

protected:
	Bounds mBounds;
	EffectiveSizes mEffectiveSizes;
	int mEffectiveCoords[8];
	std::string mId;
	Panel *mParent;

	/* Absolute bound are needed to recalculate */
	bool mDirty;

	/* The widget is in focus */
	bool mFocus;

public:
	Widget();
	Widget(const std::string& id);
	virtual ~Widget();

	inline static void setDefaultColor(const float r, const float g,
			const float b, const float a = 1.0) {
		mDefaultColor[0] = r, mDefaultColor[1] = b, mDefaultColor[2] = g,
				mDefaultColor[3] = a;
	}
	inline static void setDefaultColor(const float color[4]) {
		mDefaultColor[0] = color[0], mDefaultColor[1] = color[1],
				mDefaultColor[2] = color[2], mDefaultColor[3] = color[3];
	}

	inline void setLeftAbs(const int left) {
		mBounds.leftMode = ABSOLUTE;
		mBounds.absLeft = left;
		mDirty=true;
	}
	inline void setLeftRel(const float left) {
		mBounds.leftMode = RELATIVE;
		mBounds.relLeft = left;
		mDirty=true;
	}
	inline void setTopAbs(const int top) {
		mBounds.topMode = ABSOLUTE;
		mBounds.absTop = top;
		mDirty=true;
	}
	inline void setTopRel(const float top) {
		mBounds.topMode = RELATIVE;
		mBounds.relTop = top;
		mDirty=true;
	}
	inline void setWidthAbs(const int width) {
		mBounds.widthMode = ABSOLUTE;
		mBounds.absWidth = width;
		mDirty=true;
	}
	inline void setWidthRel(const float width) {
		mBounds.widthMode = RELATIVE;
		mBounds.relWidth = width;
		mDirty=true;
	}
	inline void setHeightAbs(const int height) {
		mBounds.heightMode = ABSOLUTE;
		mBounds.absHeight = height;
		mDirty=true;
	}
	inline void setHeightRel(const float height) {
		mBounds.heightMode = RELATIVE;
		mBounds.relHeight = height;
		mDirty=true;
	}

	inline void setSizeAbs(const int width, const int height) {
		setWidthAbs(width);
		setHeightAbs(height);
	}
	inline void setSizeRel(const float width, const float height) {
		setWidthRel(width);
		setHeightRel(height);
	}

	inline void setPositionAbs(const int left, const int top) {
		setLeftAbs(left);
		setTopAbs(top);
	}
	inline void setPositionRel(const float left, const float top) {
		setLeftRel(left);
		setTopRel(top);
	}

	inline void setBoundsAbs(const int left, const int top, const int width,
			const int height) {
		setLeftAbs(left);
		setTopAbs(top);
		setWidthAbs(width);
		setHeightAbs(height);
	}
	inline void setBoundsRel(const float left, const float top,
			const float width, const float height) {
		setLeftRel(left);
		setTopRel(top);
		setWidthRel(width);
		setHeightRel(height);
	}

	inline bool isVisible() const {
		return mColor[3] > 0.0;
	}
	inline bool isTransparent() const {
		return mColor[3] < 1.0;
	}

	inline void setColor(const float r, const float g, const float b,
			const float a = 1.0) {
		mColor[0] = r, mColor[1] = g, mColor[2] = b, mColor[3] = a;
	}

	inline void setColor(const float color[4]) {
		mColor[0] = color[0], mColor[1] = color[1], mColor[2] = color[2],
				mColor[3] = color[3];
	}

	// The second parameter is for compatibility
	inline virtual void setAlpha(const float alpha, const bool recursive = false) {
		mColor[3] = alpha;
	}

	inline const std::string& getId() const {
		return mId;
	}

	inline void setId(const std::string& id) {
		mId = id;
	}

	inline void getEffectiveBounds(int& left, int& top, int& width, int& height) {
		// It is assumed that the effective sizes are up to date
		// for speed. This query should be performed on an up to date
		// widget during the rendering. The rendering should be done
		// in parent->content order.
		left = mEffectiveCoords[0];
		top = mEffectiveCoords[1];
		width = mEffectiveSizes.width;
		height = mEffectiveSizes.height;
	}

	inline bool contains(const int x, const int y) {
		return x>=mEffectiveCoords[0] && x<=mEffectiveCoords[2] && y
				>=mEffectiveCoords[1] && y<=mEffectiveCoords[5];
	}

	virtual void getParentEffectiveBounds(int& left, int& top, int& width,
			int& height);

	virtual void render();

protected:
	void _setDefaultProperties();
	virtual void _recalculateEffectiveBounds();
	virtual void _render() = 0;
}; // class Widget

// =============================================================================

class Panel : public Widget {
	friend class GuiSheet;

public:
	static const char WIDGET_SEPARATOR;
	typedef std::map<std::string, Widget *> WidgetMap;

protected:
	WidgetMap mContents;

	bool mEventTransparent;

public:
	inline Panel() :
		Widget(), mEventTransparent(false) {
	}
	inline Panel(const std::string& id) :
		Widget(id), mEventTransparent(false) {
	}
	virtual ~Panel();

	void add(Widget *widget);
	void remove(Widget *widget);
	void remove(const std::string& id);
	Widget *get(const std::string& id);
	Widget *getWidgetByName(const std::string& searchPath);
	template<class T> std::list<T*> getWidgetsByType();

	virtual void render();

	virtual void setAlpha(const float alpha, const bool recursive = false);
	virtual void recalculateEffectiveBounds();

	virtual void setEventTransparent(const bool flag) {
		mEventTransparent = flag;
	}

protected:
	virtual bool onMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state);
	virtual bool onMouseMoveEvent(int x, int y,
			InputListener::MouseMoveState state);
	virtual bool onMouseWheelEvent(int x, int y, int delta);

	virtual void _recalculateEffectiveBounds();
	virtual void _render();
	virtual void _setContentAlpha(const float alpha);
}; // class Panel

template<class T>
std::list<T*> Panel::getWidgetsByType() {
	std::list<T*> widgets;
	
	for(WidgetMap::iterator i = mContents.begin(); i!=mContents.end(); i++) {
		T *widget = dynamic_cast<T*>(i->second);
		if(widget) widgets.push_back(widget);
	}
	
	return widgets;
}


// =============================================================================

class GuiSheet : public WindowListener {
protected:
	static const std::string DESKTOP_ID;
	static const int DEFAULT_GUI_TOGGLE_KEY = KEY_TAB;

	class GuiToggler : public TimerListener {
		friend class GuiSheet;

protected:
		static const double DEFAULT_GUI_TOGGLE_TIMESTEP = 0.04;
		static const double DEFAULT_GUI_TOGGLE_ALPHASTEP = 0.04;
		static const double DEFAULT_GUI_TOGGLE_MAXALPHA = 0.9;

		enum FadeMode {NOFADE, FADEIN, FADEOUT};

		Timer *mGuiToggleTimer;
		GuiSheet *mSheet;
		FadeMode mFadeMode;
		double mMaxAlpha, mAlpha;

protected:
		GuiToggler(GuiSheet *sheet,
				double maxAlpha = DEFAULT_GUI_TOGGLE_MAXALPHA);
		~GuiToggler();

		virtual void onTick();

		void show();
		void hide();
		void toggle();

		void _createTimer();
	};

	Panel *mDesktop;

	int mGuiToggleKey;
	GuiToggler *mGuiToggler;

	bool mBlind;

public:
	GuiSheet();
	GuiSheet(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	GuiSheet(const std::string& blockName, const std::string& filename);
	virtual ~GuiSheet();

	inline void add(Widget *widget) {
		mDesktop->add(widget);
	}
	inline void remove(Widget *widget) {
		mDesktop->remove(widget);
	}
	inline void remove(const std::string& id) {
		mDesktop->remove(id);
	}
	inline Widget *get(const std::string& id) {
		return mDesktop->get(id);
	}
	template<class T> inline T* getWidgetByName(const std::string& searchPath) {
		T *widget = dynamic_cast<T*>(mDesktop->getWidgetByName(searchPath));
		Assert(widget, "GuiSheet::getWidgetByName<T>");
		return widget;
	}

	virtual bool onRender();
	virtual bool onResize(unsigned width, unsigned height);

	virtual bool onKeyEvent(int key, unsigned modifiers, KeyEventState state);

	inline virtual bool onMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state) {
		// Redisplays the window when the desktop handles the event
		if (mDesktop->onMouseButtonEvent(x, y, button, state)) {
			redisplayRequest();
			return true;
		}

		return false;
	}
	inline virtual bool onMouseMoveEvent(int x, int y,
			InputListener::MouseMoveState state) {
		// Redisplays the window when the desktop handles the event
		if (mDesktop->onMouseMoveEvent(x, y, state)) {
			redisplayRequest();
			return true;
		}

		return false;
	}
	inline virtual bool onMouseWheelEvent(int x, int y, int delta) {
		// Redisplays the window when the desktop handles the event
		if (mDesktop->onMouseWheelEvent(x, y, delta)) {
			redisplayRequest();
			return true;
		}

		return false;
	}

	// This method is called when the gui is to be rendered again
	inline virtual void onUpdate() {
	}

	virtual void setAlpha(const float alpha);
	inline virtual void setBlind(const bool blind) {
		mBlind = blind;
	}

	virtual void setEventTransparent(const bool flag) {
		mDesktop->setEventTransparent(flag);
	}
	
	void loadFromFile(const std::string& filename, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	void loadFromFile(const std::string& blockName, const std::string& filename);
	
protected:
	void _loadFromFile(const std::string& filename, VirtualFileSystem *vfs);

}; // class GuiSheet

// =============================================================================

class TextField : public Widget {
public:
	enum VerticalAlign {TOP, MIDDLE, BOTTOM};

	class Data {
protected:
		void *mData;

public:
		inline Data(const void *data) :
			mData(const_cast<void *>(data)) {
		}
		inline virtual ~Data() {
		}

		virtual std::string toString() = 0;
	};

	template<class T> class DataTpl : public Data {
public:
		inline DataTpl(const T *data) :
			Data(data) {
		}
		inline T *getData() const {
			return (T *) mData;
		}
	};

	class DoubleData : public DataTpl<double> {
protected:
		int mPrecision;

public:
		inline DoubleData(
				const double *data,
				const int precision = StringUtils::DEFAULT_FLOATING_POINT_PRECISION) :
			DataTpl<double>(data), mPrecision(precision) {
		}

		inline virtual std::string toString() {
			return StringUtils::toString(*getData(), mPrecision);
		}
	};

	class StringData : public DataTpl<std::string> {
	public:
		inline StringData(
				std::string *data) :
			DataTpl<std::string>(data) {
		}

		inline virtual std::string toString() {
			return *getData();
		}
	};
	
protected:
	static Font *msDefaultFont;
	static const VerticalAlign DEFAULT_VERTICAL_ALIGN;

	std::string mText;
	Font *mFont;
	VerticalAlign mVerticalAlign;

	Data *mUserData;
	std::string mPrefixText;
	std::string mPostfixText;

public:
	TextField();
	TextField(const std::string& id);
	TextField(const std::string& id, const std::string& text);
	TextField(const std::string& id, Data* userData);
	TextField(const std::string& id, const std::string& text, Font* font);
	virtual ~TextField();

	inline const std::string& getText() const {
		return mText;
	}
	inline void setText(const std::string& text) {
		mText = text;
	}

	inline const std::string& getPrefixText() const {
		return mPrefixText;
	}
	inline void setPrefixText(const std::string& prefixText) {
		mPrefixText = prefixText;
	}

	inline const std::string& getPostfixText() const {
		return mPostfixText;
	}
	inline void setPostfixText(const std::string& postfixText) {
		mPostfixText = postfixText;
	}
	
	inline const Font *getFont() const {
		return mFont;
	}
	inline void setFont(Font *font) {
		mFont = font;
	}

	inline const VerticalAlign getVerticalAlign() {
		return mVerticalAlign;
	}
	inline void setVerticalAlign(const VerticalAlign& align) {
		mVerticalAlign = align;
	}

	inline const Data *getUserData() const {
		return mUserData;
	}
	inline void setUserData(Data *userData) {
		mUserData = userData;
	}
	
	inline static const Font *getDefaultFont() {
		return msDefaultFont;
	}
	inline static void setDefaultFont(Font *font) {
		msDefaultFont = font;
	}
	
protected:
	virtual void _render();
	void _setDefaultFont();
}; // class TextField

// =============================================================================

class CheckBox : public Widget {
protected:
	static const unsigned char UNCHECKED_INDICES[4];
	static const unsigned char CHECKED_INDICES[8];

	bool mChecked;
	bool mIsRadio;

	bool *mUserData;

public:
	CheckBox();
	CheckBox(const std::string& id);
	CheckBox(const std::string& id, const bool checked, const bool isRadio = false);
	CheckBox(const std::string& id, bool* userData, const bool isRadio = false);
	virtual ~CheckBox();

	inline const bool isChecked() const {
		return mChecked;
	}
	void setChecked(const bool checked);

	inline const bool isRadio() const {
		return mIsRadio;
	}
	inline void setRadio(const bool isRadio) {
		mIsRadio = isRadio;
	}
	
	inline const bool *getUserData() const {
		return mUserData;
	}
	inline void setUserData(bool *userData) {
		mChecked = *(mUserData = userData);
	}

protected:
	virtual bool onMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state);

	virtual void _render();
}; // class CheckBox

// =============================================================================

class Slider : public Widget {
public:
	enum FillType { BAR, FILL_TO, FILL_FROM };

protected:
	static const double DEFAULT_MINIMUM = 0.0;
	static const double DEFAULT_MAXIMUM = 1.0;
	static const double DEFAULT_VALUE = 0.0;
	static const bool DEFAULT_READONLY = false;
	static const FillType DEFAULT_FILLTYPE = FILL_TO;

protected:
	double mMinimum, mMaximum, mValue;
	bool mReadOnly;
	FillType mFillType;

	bool mButtonDown;

	double *mUserData;

public:
	Slider();
	Slider(const std::string& id, const double value = DEFAULT_VALUE);
	Slider(const std::string& id, double* userData);
	Slider(const std::string& id, const double value, const double minimum,
			const double maximum, const bool readonly = DEFAULT_READONLY, const FillType fillType = DEFAULT_FILLTYPE);
	Slider(const std::string& id, double* userData, const double minimum,
			const double maximum, const bool readonly = DEFAULT_READONLY, const FillType fillType = DEFAULT_FILLTYPE);
	virtual ~Slider();

	inline const double getMinimum() const {
		return mMinimum;
	}
	inline void setMinimum(const double minimum) {
		mMinimum = minimum;
	}
	inline const double getMaximum() const {
		return mMaximum;
	}
	inline void setMaximum(const double maximum) {
		mMaximum = maximum;
	}

	inline const double getValue() const {
		return mValue;
	}
	inline void setValue(const double value) {
		mValue = value;
		if (mUserData)
			*mUserData = value;
	}

	inline const bool getReadOnly() const {
		return mReadOnly;
	}
	inline void setReadOnly(const bool readOnly) {
		mReadOnly = readOnly;
	}
	
	inline const FillType getFillType() const {
		return mFillType;
	}
	inline void setFillType(const FillType fillType) {
		mFillType = fillType;
	}
	
	inline const double *getUserData() const {
		return mUserData;
	}
	inline void setUserData(double *userData) {
		mValue = *(mUserData = userData);
	}

protected:
	virtual bool onMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state);
	virtual bool onMouseMoveEvent(int x, int y,
			InputListener::MouseMoveState state);

	virtual void _render();
	virtual bool _handleSeek(const int x);
}; // class Slider

} // namespace MinGLE

#endif // __MINGLE_GUI_H__
