//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Imaging.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Imaging.h
/// Raster image class

#ifndef __MINGLE_IMAGING_H__
#define __MINGLE_IMAGING_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Manager.h"
#include "MathUtils.h"
#include "VirtualFileSystem.h"
#include "mingle-gl.h"

// System headers
#include <string>

namespace MinGLE {
class Image {

public:
	static const unsigned UNKOWN_SIZE = 0;
	static const GLenum DEFAULT_FORMAT = GL_RGB;
	static const GLenum DEFAULT_TYPE = GL_UNSIGNED_BYTE;

public:
	static MathUtils::ScaleMode mScaleMode;
	
public:
	// TODO
	unsigned mWidth, mHeight;
	unsigned mBytesPerPixel;
	unsigned mTotalSize;
	GLenum mFormat;
	GLenum mType;
	void * mData;

public:
	Image();
	virtual ~Image();

	// Getters
	inline static const MathUtils::ScaleMode& getScaleMode() {
		return mScaleMode;
	}

	inline static void setScaleMode(const MathUtils::ScaleMode& scaleMode) {
		mScaleMode = scaleMode;
	}
	
	inline unsigned getWidth() const {
		return mWidth;
	}
	inline unsigned getHeight() const {
		return mHeight;
	}
	inline unsigned getTotalSize() const {
		return mTotalSize;
	}
	inline GLenum getFormat() const {
		return mFormat;
	}
	inline GLenum getType() const {
		return mType;
	}
	inline void * getData() {
		return mData;
	}
	inline unsigned getPitch() {
		return mBytesPerPixel*mWidth;
	}
	
	// Input
	virtual void loadFromFile(const std::string& filename,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) = 0;
	virtual void createFromMemory(unsigned width, unsigned height,
			GLenum format = DEFAULT_FORMAT, GLenum type = DEFAULT_TYPE,
			void *data = 0) = 0;
	virtual void grabFromScreen();
	virtual void grabFromTexture2D(GLint level = 0);

	// Output
	virtual void saveToFile(const std::string& filename,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) = 0;
	virtual void saveToFile(const std::string& pattern, unsigned index,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) = 0;
	virtual void saveToFile(const std::string& pattern, unsigned index1,
			unsigned index2, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem()) = 0;
	virtual Image *clone();
	virtual void texImage2D(GLint level = 0) = 0;

	// Convert
	virtual void convert(GLenum format, GLenum type) = 0;
	virtual void mirror();
	virtual void flip();

	// Draw
	virtual void clear(float r = 0.0, float g = 0.0, float b = 0.0,
			float a = 0.0);

protected:
	void *_allocateMemory(unsigned width, unsigned height, GLenum format,
			GLenum type);

}; // class Image

// =============================================================================

class Imaging : public Manager<Imaging> {
public:
	virtual ~Imaging();

	static Image *createImage();
	static Image *createImage(unsigned width, unsigned height,
			GLenum format = Image::DEFAULT_FORMAT, GLenum type = Image::DEFAULT_TYPE,
			void *data = 0);
	static Image *createImage(const std::string& filename,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	static Image *createImage(const std::string& blockName, const std::string& filename);
	static Image *createImage(const Image * image);

protected:
	virtual Image *_createImage() = 0;
	virtual Image *_createImage(unsigned width, unsigned height, GLenum format,
			GLenum type, void *data = 0) = 0;
	virtual Image *_createImage(const std::string& filename,
			VirtualFileSystem *vfs) = 0;
	virtual Image *_createImage(const Image * image) = 0;

}; // class Imaging

} // namespace MinGLE

#endif // __MINGLE_IMAGING_H__
