//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/InputListener.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file InputListener.h
/// Input device listener class

#ifndef __MINGLE_INPUTLISTENER_H__
#define __MINGLE_INPUTLISTENER_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Listener.h"

namespace MinGLE {

class InputListener : public Listener {
	friend class Window;
	friend class GLUTWindow;

public:
	static const int MIN_KEY_CODE = -1024;
	static const int MAX_KEY_CODE = 1024;
	static const unsigned KEY_CODE_COUNT = (unsigned) (MAX_KEY_CODE
			- MIN_KEY_CODE + 1);
	static const unsigned MAX_MOUSE_BUTTONS = 5;

	// Keyboard code constants
	static const int KEY_ESCAPE = 27;
	static const int KEY_BACKSPACE = 8;
	static const int KEY_TAB = 9;
	static const int KEY_ENTER = 13;
	static const int KEY_SPACE = 32;
	static const int KEY_DELETE = 127;

	// Keyboard key constants (negative IBM PC codes)
	static const int KEY_F1 = -59;
	static const int KEY_F2 = -60;
	static const int KEY_F3 = -61;
	static const int KEY_F4 = -62;
	static const int KEY_F5 = -63;
	static const int KEY_F6 = -64;
	static const int KEY_F7 = -65;
	static const int KEY_F8 = -66;
	static const int KEY_F9 = -67;
	static const int KEY_F10 = -68;
	static const int KEY_F11 = -133;
	static const int KEY_F12 = -134;
	static const int KEY_LEFT_ARROW = -75;
	static const int KEY_UP_ARROW = -72;
	static const int KEY_RIGHT_ARROW = -77;
	static const int KEY_DOWN_ARROW = -80;
	static const int KEY_PAGE_UP = -73;
	static const int KEY_PAGE_DOWN = -81;
	static const int KEY_HOME = -71;
	static const int KEY_END = -79;
	static const int KEY_INSERT = -82;

	// Modifier constants
	static const unsigned KEY_LEFT_CTRL = 0x00000001;
	static const unsigned KEY_RIGHT_CTRL = 0x00000002;
	static const unsigned KEY_LEFT_SHIFT = 0x00000004;
	static const unsigned KEY_RIGHT_SHIFT = 0x00000008;
	static const unsigned KEY_LEFT_ALT = 0x00000010;
	static const unsigned KEY_RIGHT_ALT = 0x00000020;
	static const unsigned KEY_LEFT_META = 0x00000040;
	static const unsigned KEY_RIGHT_META = 0x00000080;

	static const unsigned KEY_CTRL = KEY_LEFT_CTRL | KEY_RIGHT_CTRL;
	static const unsigned KEY_SHIFT = KEY_LEFT_SHIFT | KEY_RIGHT_SHIFT;
	static const unsigned KEY_ALT = KEY_LEFT_ALT;
	static const unsigned KEY_ALT_GR = KEY_RIGHT_ALT;
	static const unsigned KEY_META = KEY_LEFT_META | KEY_RIGHT_META;

	enum KeyEventState {KEY_DOWN, KEY_UP, KEY_PRESS};
	enum MouseEventState {BUTTON_DOWN, BUTTON_UP, BUTTON_CLICK, BUTTON_DBLCLICK};
	enum MouseMoveState {ENTER, HOVER, LEAVE};
	enum MouseButton {LEFT_BUTTON = 0, RIGHT_BUTTON, MIDDLE_BUTTON, FOURTH_BUTTON, FIFTH_BUTTON};

	enum InputEventType {
		KeyEvent, MouseButtonEvent , MouseMoveEvent, MouseWheelEvent,
		RegisterKeyboardEvent, RegisterKeyboardModifierEvent, RegisterMouseMoveEvent,
		RegisterMouseButtonEvent, RegisterMouseWheelEvent
	};

	struct InputEvent {
		unsigned eventType;
		InputListener *source;
		int key;
		unsigned modifiers;
		KeyEventState keyEventState;
		int x, y;
		MouseButton button;
		MouseEventState mouseEventState;
		MouseMoveState mouseMoveState;
		int delta;
	};

public:
	InputListener();
	virtual ~InputListener();

protected:
	virtual bool onKeyEvent(int key, unsigned modifiers, KeyEventState state);

	virtual bool onMouseButtonEvent(int x, int y, MouseButton button,
			MouseEventState state);
	virtual bool onMouseMoveEvent(int x, int y, MouseMoveState state);
	virtual bool onMouseWheelEvent(int x, int y, int delta);
}; // class InputListener

// =============================================================================

class InputAdapter : public InputListener {
	friend class Window;
	friend class GLUTWindow;

protected:
	bool mKeyboardStatus[KEY_CODE_COUNT];
	unsigned mKeyboardModifiers;

	int mMouseX, mMouseY;
	int mMouseLastX, mMouseLastY;
	int mMouseDownX, mMouseDownY;
	bool mMouseButtonStatus[MAX_MOUSE_BUTTONS];
	int mMouseWheelPos;

public:
	InputAdapter();
	virtual ~InputAdapter();

	inline const int getMouseX() const {
		return mMouseX;
	}
	inline const int getMouseY() const {
		return mMouseY;
	}

	inline virtual bool isKeyDown(int key) const {
		if (key >= MIN_KEY_CODE && key <= MAX_KEY_CODE)
			return mKeyboardStatus[key-MIN_KEY_CODE];
		else
			return false;
	}

	inline virtual bool isMouseButtonDown(MouseButton button) {
		return mMouseButtonStatus[button];
	}

	inline virtual void _registerKeyboardEvent(int key, KeyEventState state) {
		if (key >= MIN_KEY_CODE && key <= MAX_KEY_CODE) {
			if (state == KEY_DOWN)
				mKeyboardStatus[key-MIN_KEY_CODE] = true;
			else
				mKeyboardStatus[key-MIN_KEY_CODE] = false;
		}
	}

	inline virtual void _registerKeyboardModifierEvent(unsigned modifiers) {
		mKeyboardModifiers = modifiers;
	}

	inline virtual void _registerMouseMoveEvent(int x, int y) {
		mMouseLastX = mMouseX;
		mMouseX = x;
		mMouseLastY = mMouseY;
		mMouseY = y;
	}

	inline virtual void _registerMouseButtonEvent(MouseButton button,
			MouseEventState state) {
		mMouseButtonStatus[button] = (state == BUTTON_DOWN);
		if (state == BUTTON_DOWN) {
			mMouseDownX = mMouseX;
			mMouseDownY = mMouseY;
		}
	}

	inline virtual void _registerMouseWheelEvent(int delta) {
		mMouseWheelPos += delta;
	}
}; // class InputAdapter

} // namespace MinGLE

#endif // __MINGLE_INPUTLISTENER_H__
