//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Manager.h
// Creation Date: January 11th 2008
// Revision Date: January 11th 2008
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Manager.h
/// Generic implementation manager class

#ifndef __MINGLE_MANAGER_H__
#define __MINGLE_MANAGER_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Exception.h"

// System headers
#include <string>
#include <list>
#include <map>
#include <algorithm>

namespace MinGLE {

template<typename T> class Manager {

protected:
	typedef std::map<std::string, T *> ImplementationMap;
	typedef ImplementationMap *ImplementationMapPtr;

protected:
	/// Map of implementations (string, T object)
	static ImplementationMapPtr mImplementations;

	/// Selected implementation
	static T *msInstance;

public:
	~Manager();

	static void shutdown();

	static void registerImplementation(const std::string& name,
			T *implementation, bool selectIfNoneSelected = true);

	static void unregisterImplementation(const std::string& name);

	static T *getActiveImplementation();
	static T *getImplementation(const std::string& name);

	static void selectImplementation(const std::string& name);
}; // class Manager

template<typename T> Manager<T>::~Manager() {
	// Destructing the main implementation
	// Note that several implementations can be loaded, however 
	// only one is used. That one destroys the other implementations
	if (msInstance == this) {
		// Unregister and delete all other implementations
		for (typename ImplementationMap::iterator i = mImplementations->begin(); i
				!= mImplementations->end(); i++) {
			// Delete any other implementations
			if (i->second != this)
				delete i->second;
		}

		// Erase implementations map
		delete mImplementations;
	}
}

template<typename T> void Manager<T>::shutdown() {
	T *manager = getActiveImplementation();
	if (manager)
		delete manager;
}

template<typename T> void Manager<T>::registerImplementation(
		const std::string& name, T *implementation, bool selectIfNoneSelected) {
	Assert(!getImplementation(name), "Manager::registerImplementation");
	(*mImplementations)[name] = implementation;

	// If there are no selected implementations, select this one
	if (!msInstance && selectIfNoneSelected)
		selectImplementation(name);
}

template<typename T> void Manager<T>::unregisterImplementation(
		const std::string& name) {
	Assert(getImplementation(name), "Manager::unregisterImplementation");
	mImplementations->erase(name);
}

template<typename T> T *Manager<T>::getActiveImplementation() {
	return msInstance;
}

template<typename T> T *Manager<T>::getImplementation(const std::string& name) {
	Assert(mImplementations, "Manager::getImplementation");
	if (mImplementations->find(name) != mImplementations->end())
		return (*mImplementations)[name];

	return 0;
}

template<typename T> void Manager<T>::selectImplementation(
		const std::string& name) {
	msInstance = getImplementation(name);
}

// =============================================================================

template<typename T, typename Functionality> class ProviderManager :
	public Manager<T> {
protected:
	typedef std::list<T*> ProviderList;
	typedef std::map<Functionality, ProviderList> FunctionalityMap;
	typedef FunctionalityMap *FunctionalityMapPtr;

public:
	enum RegisterPosition {REGISTER_AS_FIRST, REGISTER_AS_LAST};
	static const RegisterPosition DEFAULT_REGISTER_POSITION = REGISTER_AS_LAST;

protected:
	/// Map from functionality to provider list
	static FunctionalityMapPtr mFunctionalities;

public:
	~ProviderManager();

	static void registerFunctionality(const Functionality& functionality,
			const std::string& name,
			RegisterPosition position = DEFAULT_REGISTER_POSITION);
	static void unregisterFunctionality(const Functionality& functionality,
			const std::string& name);
	static void unregisterFunctionalities(const std::string& name);

	static T *getProvider(const Functionality& functionality);

	static void unregisterImplementation(const std::string& name);
}; // class ProviderManager

template<typename T, typename Functionality> ProviderManager<T, Functionality>::~ProviderManager() {
	// Destructing the main implementation
	if (Manager<T>::msInstance == this) {
		delete mFunctionalities;
	}
}

template<typename T, typename Functionality> void ProviderManager<T,
		Functionality>::registerFunctionality(
		const Functionality& functionality, const std::string& name,
		RegisterPosition position) {
	
	// Get or create provider list for the specified functionality
	ProviderList& providers = (*mFunctionalities)[functionality];

	// Verify that the provider list does not have the implementation already
	T *implementation = Manager<T>::getImplementation(name);
	Assert(std::find(providers.begin(), providers.end(), implementation) == providers.end(), "ProviderManager::registerFunctionality");
	
	// Register implementation
	switch (position) {
	case REGISTER_AS_FIRST:
		providers.push_front(implementation);
		break;
	case REGISTER_AS_LAST:
		providers.push_back(implementation);
		break;
	}
}

template<typename T, typename Functionality> void ProviderManager<T,
		Functionality>::unregisterFunctionality(
		const Functionality& functionality, const std::string& name) {

	// Get provider list for the specified functionality
	typename FunctionalityMap::iterator iProviders = mFunctionalities->find();
	Assert(iProviders != mFunctionalities->end(), "ProviderManager::unregisterFunctionality");
	
	// Verify that the provider list has the implementation
	typename ProviderList::iterator iImplementation = iProviders->find(Manager<T>::getImplementation(name)); 
	Assert(iImplementation != iProviders->end(), "ProviderManager::unregisterFunctionality");

	// Remove the implementation from the provider list
	iProviders->erase(iImplementation);
	
	// If the provider list has become empty, remove it from the functionality map
	if(!iProviders->size()) mFunctionalities->erase(iProviders);
}

template<typename T, typename Functionality> void ProviderManager<T,
		Functionality>::unregisterFunctionalities(const std::string& name) {

	T *implementation = Manager<T>::getImplementation(name);
	
	// Iterate on the functionality map
	for(typename FunctionalityMap::iterator iProviders=mFunctionalities->begin(); iProviders != mFunctionalities->end(); iProviders++) {
		// Remove the implementation from the map (if have)
		iProviders->remove(implementation);
	}
}

template<typename T, typename Functionality> T *ProviderManager<T,
		Functionality>::getProvider(const Functionality& functionality) {
	
	// Get provider list for the specified functionality
	// and check that the list exists and that it is not empty
	typename FunctionalityMap::iterator iProviders = mFunctionalities->find(functionality);
	Assert(iProviders != mFunctionalities->end() && iProviders->second.size(), "ProviderManager::getProvider");

	// Return the first element of the provider list
	return iProviders->second.front();
}

template<typename T, typename Functionality> void ProviderManager<T,
		Functionality>::unregisterImplementation(const std::string& name) {
	// Unregister functionalities 
	//unregisterFunctionalities(name);
	
	// Call base implementation
	Manager<T>::unregisterImplementation(name);
}

} // namespace MinGLE

#endif // __MINGLE_MANAGER_H__
