//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/MathUtils.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file MathUtils.h
/// Mathematical utils

#ifndef __MINGLE_MATHUTILS_H__
#define __MINGLE_MATHUTILS_H__

// Include class definitions
#include "classes.h"

#include <math.h>

namespace MinGLE {

class MathUtils {

public:
	enum ScaleMode {DOWNSCALE, UPSCALE, NEARESTSCALE};
	static const ScaleMode DEFAULT_SCALE_MODE = NEARESTSCALE;

protected:
	static const double DEFAULT_EPSILON = 1e-6;
	static double msEpsilon;
	static const double DEFAULT_PI= M_PI;
	static double msPi;

public:
	inline static double getEpsilon() {
		return msEpsilon;
	}
	inline static void setEpsilon(const double epsilon) {
		msEpsilon=epsilon;
	}
	inline static double getInfinity() {
		return (double) INFINITY;
	}
	inline static double getPi() {
		return msPi;
	}
	inline static void setPi(const double pi) {
		msPi=pi;
	}

	inline static int signum(double val) {
		return (val>0.0) ? 1 : ( (val<0.0) ? -1 : 0 );
	}

	inline static void vector4to3(double &x, double &y, double &z, double w) {
		// Homogenous divider is non-zero
		if (::fabs(w) > getEpsilon())
		x /= w, y /= w, z /= w;

		// Homogenous divider is zero
		else {
			x = y = z = getInfinity();
		}
	}

	static double vectorLength(const double x, const double y);
	static double vectorLength(const double x, const double y, const double z);
	static double vectorLength(const double x, const double y, const double z,
	const double w);

	static double dotProduct(const double x1, const double y1, const double x2,
	const double y2);
	static double dotProduct(const double x1, const double y1, const double z1,
	const double x2, const double y2, const double z2);

	static void crossProduct(const double x1, const double y1, const double z1,
	const double x2, const double y2, const double z2, double &x,
	double &y, double &z);

	static void vectorNormalize(double &x, double &y);
	static void vectorNormalize(double &x, double &y, double &z);
	static void vectorNormalize(double &x, double &y, double &z, double &w);

	static void extractMatrix3x3(double *m4x4, double *m3x3);

	static void matrix3x3VectorProduct(const double *m, const double *d,
	double *p);
	static void matrix4x4VectorProduct(const double *m, const double *d,
	double *p);
	static void matrix4x4Product(const double *m1, const double *m2,
	double *result);
	static void transposeMatrix4x4(const double *m, double *mT);
	static void invertMatrix4x4(double *m, double *invM);

	static bool intersectsRayPlane(const double *rayOrigin,
	const double *rayUnitDirection, const double *planeNormal,
	const double planeD, double *t);

	static void rodriguesMatrix3x3(double *m, const double *dir, const double theta);
	
	static bool isPowerOf2(unsigned n);
	static void getPowerOf2Bounds(unsigned n, unsigned &l, unsigned &u);
	static unsigned convertToPowerOf2(unsigned value,
	ScaleMode scaleMode = DEFAULT_SCALE_MODE);

	template<typename T> static T reverseBits(T value);

	static void hsi2rgb(double h, double s, double i, double &r, double &g, double &b);
}; // class MathUtils

//
// Inline / templated implementations 
//

// Generic reverse bits implementation
// Source: http://graphics.stanford.edu/~seander/bithacks.html
template<typename T> inline T MathUtils::reverseBits(T value) {
	unsigned s = sizeof(value) * 8; // bit size; must be power of 2 
	unsigned mask = ~0;
	while ((s >>= 1) > 0) {
		mask ^= (mask << s);
		value = ((value >> s) & mask) | ((value << s) & ~mask);
	}

	return value;
}

} // namespace MinGLE

#endif // __MINGLE_MATHUTILS_H__
