//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Navigators.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Navigators.h
/// Scene navigator classes

#ifndef __MINGLE_NAVIGATORS_H__
#define __MINGLE_NAVIGATORS_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "WindowListener.h"
#include "TimerListener.h"

namespace MinGLE {

class ApplicationKeyHandler : public WindowListener {

public:
	ApplicationKeyHandler();
	virtual ~ApplicationKeyHandler();

protected:
	virtual bool onKeyEvent(int key, unsigned modifiers, KeyEventState state);
};

// =============================================================================

class Navigator : public WindowListener {
protected:
	Camera *mCamera;

public:
	Navigator();
	Navigator(Camera *camera);
	virtual ~Navigator();

	// Getters
	inline virtual Camera *getCamera() {
		return mCamera;
	}

protected:
	virtual bool onResize(unsigned width, unsigned height);
}; // class Navigator

// =============================================================================

class Animator : public TimerListener, public WindowListener {

protected:
	Timer *mTimer;

public:
	Animator(Window *window, const double timeStep);
	virtual ~Animator();

protected:
	virtual void onTick();
	virtual bool onAnimation() = 0;
}; // class Animator

// =============================================================================

/*class SimpleAnimator : public Animator {
 public:
 SimpleAnimator();
 virtual ~SimpleAnimator();

 protected:
 virtual bool onAnimation();    		
 };*/

// =============================================================================

class ExaminerNavigator : public Navigator {
protected:
	static const bool DEFAULT_ZOOM_ENABLED = true;
	static const bool DEFAULT_MOVE_ENABLED = true;
	static const bool DEFAULT_WHEEL_ENABLED = true;
	static const double DEFAULT_DEFAULT_SENSITIVITY = 1.0;

	static const InputListener::MouseButton DEFAULT_TRACKBALL_BUTTON =
			InputListener::LEFT_BUTTON;
	static const InputListener::MouseButton DEFAULT_ZOOM_BUTTON =
			InputListener::RIGHT_BUTTON;
	static const InputListener::MouseButton DEFAULT_MOVE_BUTTON =
			InputListener::MIDDLE_BUTTON;

	static const double NO_DRAG_COORDINATE = -1.0;
	static const double WHEEL_ZOOM_IN_FACTOR = 1.1;
	static const double WHEEL_ZOOM_OUT_FACTOR = 0.9;

	static const double DEFAULT_ZOOM = 1.0;

protected:
	bool mZoomEnabled, mMoveEnabled, mWheelEnabled;
	double mSensitivity;

	// Button settings
	InputListener::MouseButton mTrackBallButton;
	InputListener::MouseButton mZoomButton;
	InputListener::MouseButton mMoveButton;

	// Dragged point in world coordinates
	//double mDragX, mDragY, mDragZ;

	//double mZoom;

public:
	ExaminerNavigator(bool zoomEnabled = DEFAULT_ZOOM_ENABLED,
			bool moveEnabled = DEFAULT_MOVE_ENABLED,
			bool wheelEnabled = DEFAULT_WHEEL_ENABLED,
			double sensitivity = DEFAULT_DEFAULT_SENSITIVITY);
	virtual ~ExaminerNavigator();

	//void setZoom(const double value);

protected:
	virtual bool onResize(unsigned width, unsigned height);
	virtual bool onMouseButtonEvent(int x, int y, MouseButton button,
			MouseEventState state);
	virtual bool onMouseMoveEvent(int x, int y, MouseMoveState state);
	virtual bool onMouseWheelEvent(int x, int y, int delta);
}; // class ExaminerNavigator

// =============================================================================

class FlyNavigator : public Navigator {
protected:
	/// Linear or angular move mode
	enum MoveMode {POSITION_CONTROLLED, VELOCITY_CONTROLLED, ACCELERATION_CONTROLLED};

	/// Mouse control mode
	enum MouseControlMode {NOT_CONTROLLED, MOUSE_MOTION_CONTROLLED, MOUSE_BUTTON_CONTROLLED};

	/// Overall (mouse and keyboard) control mode, activation button,
	/// and key
	struct MoveControl {
		MouseControlMode mouseControlMode;
		InputListener::MouseButton mouseButton;
		bool keyboardControlled;
		int key;
		unsigned modifiers;
		double moveOnKey;
	};

	static const double DEFAULT_ZERO_DIRECTION[3];
	static const double DEFAULT_ZERO_UP_VECTOR[3];
	static const double DEFAULT_ZERO_RIGHT_VECTOR[3];
	static const MoveMode DEFAULT_LINEAR_MOVE_MODE;
	static const MoveMode DEFAULT_ANGULAR_MOVE_MODE;
	static const double DEFAULT_LINEAR_CHANGE_RATE;
	static const double DEFAULT_ANGULAR_CHANGE_RATE;
	static const double DEFAULT_LINEAR_DECELERATION;
	static const double DEFAULT_ANGULAR_DECELERATION;
	static const double DEFAULT_MOVE_ON_KEY;
	static const bool DEFAULT_LINEAR_HANDBREAK;
	static const bool DEFAULT_ANGULAR_HANDBREAK;
	static const bool DEFAULT_LINEAR_STEERING;
	static const MoveControl DEFAULT_FORWARD_CONTROL;
	static const MoveControl DEFAULT_BACKWARD_CONTROL;
	static const MoveControl DEFAULT_LEFT_CONTROL;
	static const MoveControl DEFAULT_RIGHT_CONTROL;
	static const MoveControl DEFAULT_ROLL_CONTROL;
	static const MoveControl DEFAULT_PITCH_CONTROL;
	static const MoveControl DEFAULT_YAW_CONTROL;
	static const MoveControl DEFAULT_HANDBREAK_CONTROL;
	static const int INVALID_MOUSE_DOWN_COORD;

	class RedisplayAnimator : public Animator {
protected:
		FlyNavigator* mFlyNavigator;

public:
		RedisplayAnimator(Window *window, const double timeStep,
				FlyNavigator* flyNavigator) :
			Animator(window, timeStep), mFlyNavigator(flyNavigator) {
		}

protected:
		virtual bool onAnimation() {
			mFlyNavigator->_refreshCamera();

			return true;
		}
	};

protected:
	//
	// Geometric properties
	//

	/// Position of the navigator in object space
	double mPositionX, mPositionY, mPositionZ;

	/// Orientation of the navigator in eye space
	double mYaw, mPitch, mRoll;

	/// Direction (forward unit vector) of the navigator in object space
	double mDirection[3];

	/// Up unit vector of the navigator in object space
	double mUpVector[3];

	/// Right unit vector of the navigator in object space
	double mRightVector[3];

	//
	// Kinematic properties
	//

	/// (Linear) velocity of the navigator in object space in 1/sec
	double mVelocityX, mVelocityY, mVelocityZ;

	/// Angular velocity of the navigator in eye space in radian/sec.
	/// The rate of angle change is specified in Roll-Pitch-Yaw system.
	double mAngularVelocityYaw, mAngularVelocityPitch, mAngularVelocityRoll;

	/// (Linear) acceleration of the navigator in object space in 1/sec^2
	double mAccelerationX, mAccelerationY, mAccelerationZ;

	/// Angular velocity of the navigator in eye space in radian/sec.
	/// The rate of angular speed change is specified in Roll-Pitch-Yaw system.
	double mAngularAccelerationYaw, mAngularAccelerationPitch,
			mAngularAccelerationRoll;

	/// Linear move mode
	MoveMode mLinearMoveMode;

	/// Angular move mode
	MoveMode mAngularMoveMode;

	/// Amount of linear change (position, velocity, or acceleration) 
	/// depending on mLinearMoveMode
	double mLinearChangeRate;

	/// Amount of angular change (angle, angular velocity, or angular acceleration) 
	/// depending on mAngularMoveMode
	double mAngularChangeRate;

	/// (Linear) deceleration that slows the motion when the user does not
	/// affects the navigator
	double mLinearDeceleration;

	/// Angular deceleration that slows the motion when the user does not
	/// affects the navigator
	double mAngularDeceleration;

	/// (Linear) handbreak function: reverse control stops the motion
	bool mLinearHandBreak;

	/// Angular handbreak function: reverse control stops the motion
	bool mAngularHandBreak;

	/// (Linear) steering function: only that component of the velocity 
	/// is kept that parallel to the direction of the acceleration 
	bool mLinearSteering;

	/// Last time when move was performed 
	double mLastMoveTime;

	//
	// Control properties
	//

	/// Forward (linear)
	MoveControl mForwardControl;
	/// Backward (linear)
	MoveControl mBackwardControl;
	/// Left (linear)
	MoveControl mLeftControl;
	/// Right (linear)
	MoveControl mRightControl;

	/// Roll control (angular)
	MoveControl mRollControl;
	/// Pitch control (angular)
	MoveControl mPitchControl;
	/// Yaw control (angular)
	MoveControl mYawControl;

	/// Handbreak control (linear and angular)
	MoveControl mHandBreakControl;

	/// Amount of linear movement
	double mMoveForwardBackward, mMoveLeftRight;

	/// Amount of angular movement
	double mRotateYaw, mRotatePitch, mRotateRoll;

	/// Saved screen coordinates on mouse button down event
	int mMouseDownX, mMouseDownY;

	/// Saved angles on mouse button down event
	double mMouseDownYaw, mMouseDownPitch, mMouseDownRoll;

	//
	// Animation properties
	//

	/// Animator that requests redisplay
	RedisplayAnimator *mRedisplayAnimator;

public:
	FlyNavigator();
	FlyNavigator(Camera *camera);

protected:
	virtual void onRegistered();
	virtual void onUnRegistered();

	virtual bool onResize(unsigned width, unsigned height);
	virtual bool onKeyEvent(int key, unsigned modifiers, KeyEventState state);
	virtual bool onMouseButtonEvent(int x, int y, MouseButton button,
			MouseEventState state);
	virtual bool onMouseMoveEvent(int x, int y, MouseMoveState state);

	void _refreshCamera();
	bool _updateAccelerations(const double timeStep);
	bool _updateVelocities(const double timeStep);
	bool _updatePositionAndOrientation(const double timeStep);

	void _linearHandBreak();

	/// Handbreak stops motion (acceleration and velocity) when the sign of the given new 
	/// move direction does not equal the existing move direction
	void _linearHandBreak(const double moveDirX, const double moveDirY,
			const double moveDirZ);
	void _angularHandBreak();

	void _linearSteering();

	void _calculateRPYRotationMatrix(const double phi, const double theta,
			const double psi, double rotationMatrix[9]);
}; // class FlyNavigator

} // namespace MinGLE

#endif // __MINGLE_NAVIGATORS_H__
