//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/System.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file System.h
/// Singleton manager class

#ifndef __MINGLE_SYSTEM_H__
#define __MINGLE_SYSTEM_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Exception.h"
#include "DynLoad.h"
#include "Configuration.h"
#include "VirtualFileSystem.h"
#include "SystemListener.h"
#include "Window.h"

// System headers
#include <string>
#include <vector>
#include <map>

namespace MinGLE {

class System {

public:
	typedef struct {
		unsigned major, minor;
		std::string vendorSpecific;
	} GLVersion;

	static const unsigned UNKNOWN_RESOLUTION = 0;

public:
	typedef std::vector<std::string> DirectoryVector;

protected:
	typedef std::map<std::string, DynLib *> DynLibMap;
	typedef std::map<std::string, System *> SystemImplementationMap;
	typedef std::vector<SystemListener *> SystemListenerVector;
	typedef std::map<std::string, VirtualFileSystem *> VirtualFileSystemMap;

protected:
	static const std::string DEFAULT_CONFIG_FILE;
	
	static const std::string RENDER_SYSTEM_DIRECTORY_IN_CONFIG;
	static const std::string THREADING_DIRECTORY_IN_CONFIG;
	static const std::string FILE_SYSTEM_DIRECTORY_IN_CONFIG;
	static const std::string CONFIGURATION_DIRECTORY_IN_CONFIG;
	static const std::string IMAGING_DIRECTORY_IN_CONFIG;
	static const std::string FONTRENDERING_DIRECTORY_IN_CONFIG;

	static const std::string FILE_SYSTEMS_ENTRY_IN_CONFIG;
	static const std::string PROTOCOL_ENTRY_NAME_IN_CONFIG;
	static const std::string CONNECTIONSTRING_ENTRY_NAME_IN_CONFIG;

	static const std::string AUTOMATIC_SYSTEM_TYPE_STRING;

	static System *msInstance;

	static GLVersion mGLVersion;
	static std::string mGLEWVersionString;

	/// Name of the high level config file
	static std::string mConfigFile;

	/// High level configuration structure
	static Configuration *mConfiguration;

	/// List of loaded dynamic libraries
	static DynLibMap mDynLibs;

	/// Map of render system implementations (string, render system object)
	static SystemImplementationMap mSystemImplementations;

	/// Map of used virtual file systems (returned by getVirtualFileSystemForFile())
	static VirtualFileSystemMap mUsedVFSes;
	
	/// List of attached system listeners
	SystemListenerVector mSystemListeners;

	unsigned mScreenWidth, mScreenHeight;

protected:
	System();
	virtual ~System();

	virtual bool _initialize(int *argc, char **argv) = 0;

	virtual Window
			* _createWindow(
					unsigned width = Window::DEFAULT_WIDTH,
					unsigned height = Window::DEFAULT_HEIGHT,
					WindowAttributes::ColorMode colorMode = WindowAttributes::DIRECT_COLORS,
					bool doubleBuffer = true, bool fullScreen = false,
					unsigned videoRate = 0, unsigned char *colorBits = 0,
					unsigned char depthBits = 16,
					unsigned char stencilBits = 0,
					unsigned char *accumColorBits = 0, unsigned auxBuffers = 0,
					bool stereo = false) = 0;

	virtual void _enterMainLoop() = 0;

	virtual void _quit() = 0;

	virtual double _getTime() = 0;

public:
	static void setConfigFile(const std::string& configFile);
	/// Get virtual file system with which the specified file can be opened.
	/// The first file system is returned that can be open the file.
	/// The possible file systems are queried from the specified blocks in 
	/// the configuration.
	static VirtualFileSystem *getVirtualFileSystemForFile(const std::string& blockName, const std::string& fileName);

	static DynLib *loadDynamicLibrary(const std::string& name);
	static void unloadDynamicLibrary(const std::string& name);

	static void loadDynamicLibraries(const std::string& directory);

	static void registerSystemImplementation(const std::string& name,
			System *systemImplementation);

	static void unregisterSystemImplementation(const std::string& name);

	static System *getSystemImplementation(const std::string& name);
	
	static void initialize(int *argc, char **argv,
			const std::string& systemType = AUTOMATIC_SYSTEM_TYPE_STRING);

	static Window* createWindow(
			unsigned width = Window::DEFAULT_WIDTH,
			unsigned height = Window::DEFAULT_HEIGHT,
			WindowAttributes::ColorMode colorMode = WindowAttributes::DIRECT_COLORS,
			bool doubleBuffer = true, bool fullScreen = false,
			unsigned videoRate = 0, unsigned char *colorBits = 0,
			unsigned char depthBits = 16, unsigned char stencilBits = 0,
			unsigned char *accumColorBits = 0, unsigned auxBuffers = 0,
			bool stereo = false) {
		Window *win = msInstance->_createWindow(width, height, colorMode,
				doubleBuffer, fullScreen, videoRate, colorBits, depthBits,
				stencilBits, accumColorBits, auxBuffers, stereo);

		// Call system initializier method
		_onWindowCreated();
		return win;
	}

	inline static Window* createWindowWithDefaultEnv(
			WindowListener *sceneRenderer, GuiSheet *guiSheet = 0);

	inline static void enterMainLoop() {
		msInstance->_enterMainLoop();
	}

	inline static void quit() {
		msInstance->_quit();
	}

	// Return elapsed time since the initialization in seconds 
	inline static double getTime() {
		return msInstance->_getTime();
	}

	inline static unsigned getScreenWidth() {
		Assert(msInstance, "System::getScreenWidth");
		return msInstance->mScreenWidth;
	}

	inline static unsigned getScreenHeight() {
		Assert(msInstance, "System::getScreenHeight");
		return msInstance->mScreenHeight;
	}

	inline static const GLVersion& getGLVersion() {
		Assert(msInstance, "System::getGLVersion");
		return mGLVersion;
	}

	inline static const std::string& getGLEWVersionString() {
		Assert(msInstance, "System::getGLEWVersionString");
		return mGLEWVersionString;
	}

	// System listener support		
	static void registerSystemListener(SystemListener *systemListener);
	static void removeSystemListener(SystemListener *systemListener);

protected:
	static void _shutdownSubSystems();

	static void _loadDynamicLibrariesFromConfig(const std::string& searchPath);
	static void _unloadAllDynamicLibraries();

	static void _onWindowCreated();
	static void _retrieveGLVersion();
	static void _initGlew();

};
// class System

} // namespace MinGLE

#endif // __MINGLE_SYSTEM_H__
