//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Threading.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Threading.h
/// Threading classes for internal usage

#ifndef __MINGLE_THREADING_H__
#define __MINGLE_THREADING_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Manager.h"

// System headers
#include <map>
#include <string>

namespace MinGLE {

class Mutex {
public:
	inline virtual ~Mutex() {
	}

	virtual void lock() = 0;
	virtual void release() = 0;
}; // class Mutex

// =============================================================================

class Runnable {
public:
	virtual ~Runnable();

	virtual void run() = 0;
};

// =============================================================================

class Thread {

public:
	static const bool DEFAULT_JOINABLE_FLAG;
	static const bool DEFAULT_JOIN_AT_STOP_FLAG;

	static const int ERROR_RETURN_VALUE;
	static const int SUCCESS_RETURN_VALUE;

protected:
	Runnable *mRunnable;

	bool mRunning;
	bool mJoinable;
	bool mStopRequest;

public:
	Thread(Runnable *runnable, bool joinable = DEFAULT_JOINABLE_FLAG);
	virtual ~Thread();

	inline const bool isRunning() const {
		return mRunning;
	}
	
	static void sleep(double seconds);

	void start();
	void stop(bool joinAtStop = DEFAULT_JOIN_AT_STOP_FLAG);
	void join();

protected:
	virtual bool exceptionHandler(const Exception& exception);
	static void unexpected();

	virtual void _startThread() = 0;
	virtual void _joinThread() = 0;
	virtual void _killThread() = 0;

}; // class Thread

// =============================================================================

class Threading : public Manager<Threading> {

public:
	virtual ~Threading();
	
	static Mutex* createMutex();
	static Thread * createThread(Runnable *runnable,
			bool joinable = Thread::DEFAULT_JOINABLE_FLAG);

protected:
	virtual Mutex* _createMutex() = 0;
	virtual Thread * _createThread(Runnable *runnable, bool joinable) = 0;
};

// =============================================================================

template<typename T> class GuardedObject {
private:
	T* mObject;
	Mutex *mMutex;

public:
	inline GuardedObject() :
		mObject(new T()) {
		mMutex = Threading::createMutex();
	}
	inline GuardedObject(T* object) :
		mObject(object) {
		mMutex = Threading::createMutex();
	}
	inline ~GuardedObject() {
		delete mMutex;
		delete mObject;
	}

	inline T* lock() {
		mMutex->lock();
		return mObject;
	}

	inline void unlock() {
		mMutex->release();
	}
}; // class GuardedObject

// =============================================================================

template<typename T> class GuardedPointer {
private:
	T* mPointer;
	Mutex *mMutex;

public:
	inline GuardedPointer() :
		mPointer(0) {
		mMutex = 0;
	}
	inline GuardedPointer(T* pointer) :
		mPointer(pointer) {
		mMutex = Threading::createMutex();		
	}
	inline ~GuardedPointer() {
		delete mMutex;
	}

	inline void assign(T* pointer) {
		if(!mMutex) mMutex = Threading::createMutex();
		mMutex->lock();
		mPointer = pointer;
		mMutex->release();
	}
	
	inline T* lock() {
		// Note: mMutex can be null, it is not verified for better performance
		mMutex->lock();
		return mPointer;
	}

	inline void unlock() {
		// Note: mMutex can be null, it is not verified for better performance
		mMutex->release();
	}
}; // class GuardedObject

// =============================================================================

/*
 class Process : public Thread, public Runnable {
 protected:
 double mStatus;

 public:
 Process();

 inline const double getStatus() const {
 return mStatus;
 }

 inline const bool isFinished() const {
 return mStatus == 1.0;
 }

 virtual void run();
 virtual void process() = 0;
 }; // class Process
 */

} // namespace MinGLE

#endif // __MINGLE_THREADING_H__
