//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/VirtualFileSystem.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file VirtualFileSystem.h
/// Virtual file system class

#ifndef __MINGLE_VIRTUALFILESYSTEM_H__
#define __MINGLE_VIRTUALFILESYSTEM_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Manager.h"

// System headers
#include <string>
#include <vector>

namespace MinGLE {

class VirtualFileSystem {

public:
	static const char DIRECTORY_SEPARATOR;
	static const char END_OF_LINE;
	static const char END_OF_LINE_TO_DROP;
	static const int END_OF_FILE;
	static const int UNLIMITED_BYTES;

	typedef void * FileHandle;

	enum DirectoryEntryType {
		UNKNOWN = 0x01,
		DIRECTORY = 0x02,
		REGULAR = 0x04,
		FIFO = 0x08,
		SOCKET = 0x10,
		CHARACTER_DEVICE = 0x20,
		BLOCK_DEVICE = 0x40
	};

	typedef struct {
		std::string name;
		DirectoryEntryType type;
	} DirectoryEntry;

	typedef std::vector<DirectoryEntry> DirectoryEntryList;

	static const unsigned DIRECTORY_ENTRY_TYPE_MASK_ALL = UNKNOWN | DIRECTORY
			| REGULAR | FIFO | SOCKET | CHARACTER_DEVICE | BLOCK_DEVICE;

public:
	inline virtual ~VirtualFileSystem() {
	}

	virtual std::string getURL(const std::string& filename = "") const = 0;
	
	// Filename handling methods
	static std::string getDirectory(const std::string& filename);
	static std::string getFilename(const std::string& filename);
	static std::string getBaseName(const std::string& filename);
	static std::string getExtension(const std::string& filename);

	virtual FileHandle
			open(const std::string& filename, const std::string& mode) = 0;
	virtual void close(FileHandle f) = 0;
	virtual bool eof(FileHandle f) = 0;
	virtual int getchar(FileHandle f) = 0;
	virtual int putchar(int ch, FileHandle f) = 0;
	virtual int read(void *ptr, int size, int count, FileHandle f) = 0;
	virtual int write(const void *ptr, int size, int count, FileHandle f) = 0;
	virtual int seek(FileHandle f, int numbytes, int where) = 0;
	virtual int tell(FileHandle f) = 0;

	virtual bool exists(const std::string& path) = 0;
	
	virtual DirectoryEntryList listDirectory(const std::string& directoryName,
			const unsigned mask = DIRECTORY_ENTRY_TYPE_MASK_ALL) = 0;

	inline unsigned read16Lsb(FileHandle f) {
		// TODO: LSB/MSB
		return getchar(f) | getchar(f)<<8;
	}
	inline unsigned read32Lsb(FileHandle f) {
		// TODO: LSB/MSB
		return read16Lsb(f) | read16Lsb(f)<<16;
	}
	inline unsigned long long read64Lsb(FileHandle f) {
		// TODO: LSB/MSB
		return read32Lsb(f) | (unsigned long long) read32Lsb(f)<<32;
	}

	inline void write16Lsb(unsigned val, FileHandle f) {
		// TODO: LSB/MSB
		putchar(val&0x00FF, f);
		putchar((val&0xFF00)>>8, f);
	}
	inline void write32Lsb(unsigned val, FileHandle f) {
		// TODO: LSB/MSB
		write16Lsb(val&0x0000FFFF, f);
		write16Lsb((val&0xFFFF0000)>>16, f);
	}
	inline void write64Lsb(unsigned long long val, FileHandle f) {
		// TODO: LSB/MSB
		write32Lsb(val&0x00000000FFFFFFFF, f);
		write32Lsb((val&0xFFFFFFFF00000000LL)>>32, f);
	}

	std::string readLine(FileHandle f);
	void writeLine(const std::string& line, FileHandle f);
	
	unsigned char *readFile(const std::string& filename, int &length, unsigned char *buf = 0, int maxLength = UNLIMITED_BYTES);
	char *readFileAsString(const std::string& filename, unsigned char *buf = 0, int maxLength = UNLIMITED_BYTES);

protected:
	void _sortDirectoryEntryList(DirectoryEntryList& deList);
	unsigned char *_readData(const std::string& filename, int &length, unsigned char *buf, int maxLength, bool nullTerminated);

private:
	static bool _directoryEntrySortPredicate(const DirectoryEntry& lhs,
			const DirectoryEntry& rhs);

}; // class VirtualFileSystem

// =============================================================================

class VirtualFileSystemManager : public Manager<VirtualFileSystemManager> {

public:
	static const std::string PROTOCOL_SEPARATOR;
	static const std::string DEFAULT_FILE_SYSTEM_URL;
	
protected:
	static VirtualFileSystem *msDefaultFileSystem;
	
public:
	virtual ~VirtualFileSystemManager();

	static VirtualFileSystem *createVirtualFileSystem(const std::string& url);
	static VirtualFileSystem *createVirtualFileSystem(const std::string& protocol, const std::string& connectionString);
	static VirtualFileSystem *getDefaultVirtualFileSystem();

protected:
	virtual VirtualFileSystem *_createVirtualFileSystem(const std::string& connectionString) = 0;
}; // class VirtualFileSystemManager

// ============================================================================

class FileFormatFunctionality {
public:
	enum Mode { READ = 0x01, WRITE = 0x02, READ_WRITE = READ | WRITE };
	
protected:
	std::string mExtension;
	Mode mMode;
	
public:
	FileFormatFunctionality(const std::string& extension, Mode mode);
	inline bool operator<(const FileFormatFunctionality& right) const {
		return mExtension < right.mExtension || mExtension == right.mExtension && !(mMode & right.mMode);
	}
	
}; // class FileFormatFunctionality

} // namespace MinGLE

#endif // __MINGLE_VIRTUALFILESYSTEM_H__
