//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: include/Window.h
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Window.h
/// Rendering window class

#ifndef __MINGLE_WINDOW_H__
#define __MINGLE_WINDOW_H__

// Include class definitions
#include "classes.h"

// MinGLE headers
#include "Exception.h"
#include "Timer.h"
#include "WindowListener.h"
#include "TimerListener.h"

// System headers
#include <string>
#include <vector>

namespace MinGLE {

struct WindowAttributes {
	enum ColorMode {DIRECT_COLORS, INDEXED_COLORS};
	enum RenderMode {
		RENDER_WHEN_NEEDED,
		RENDER_WHEN_IDLE,
		RENDER_ON_TIMER
	};

	int mLeft, mTop;
	unsigned mWidth, mHeight;

	// Recovery from full screen mode
	int mRecoveryLeft, mRecoveryTop;
	unsigned mRecoveryWidth, mRecoveryHeight;

	ColorMode mColorMode;
	bool mDoubleBuffer;
	bool mFullScreen;
	unsigned char mColorBits[4];
	unsigned char mAccumColorBits[4];
	unsigned char mDepthBits, mStencilBits;
	unsigned mVideoRate;
	unsigned mAuxBuffers;
	bool mStereo;

	bool mVisible, mIconified;
	std::string mTitle;

	// Automatic features
	bool mAutoFinishFrame;
	bool mAutoClearFrame;
	bool mAutoUpdateViewport;

	RenderMode mRenderMode;
	bool mListenOnIdle;

	bool mRedisplayRequest;

	// Recording
	Image *mRecordingImage;
	std::string mScreenShotPattern;
	std::string mVideoPattern;

	unsigned mScreenShotIndex;
	bool mScreenShotCapturing;

	unsigned mVideoSceneIndex;
	double mVideoStartTime;
	bool mVideoRecording;

	// Window closing
	bool mCloseRequest;

}; // struct WindowAttributes

class Window {
	friend class System;
	friend class WindowProxy;

public:
	static const unsigned DEFAULT_WIDTH = 640, DEFAULT_HEIGHT = 480;
	static const WindowAttributes::RenderMode DEFAULT_RENDER_MODE =
			WindowAttributes::RENDER_WHEN_NEEDED;

protected:

	/** Frame rate calculator inner class */
	class FrameRateCalculator {
private:
		size_t mFrameCount;
		double mStartTime;

		double mFrameRate;

public:
		inline FrameRateCalculator() :
			mFrameCount(0), mStartTime(Timer::getSystemTime()), mFrameRate(0.0) {
		}

		inline void frameFinished() {
			mFrameCount++;
			double timeNow = Timer::getSystemTime(), diffTime = timeNow
					- mStartTime;
			if (diffTime > 1.0) {
				mFrameRate = mFrameCount/diffTime;
				mFrameCount = 0;
				mStartTime = timeNow;
			}
		}

		inline const double *getFrameRatePtr() const {
			return &mFrameRate;
		}

		inline double getFrameStartTime() const {
			return mStartTime;
		}
	};

	/** Inner class of window that controls rendering when RENDER_ON_TIMER render mode is specified */
	class TimeTriggeredRenderController : public TimerListener {
protected:
		static const double DEFAULT_FRAME_TIME = 0.0334;

protected:
		double mDesiredFrameTime;
		double mLastFrameStartTime;
		Window *mWindow;
		Timer *mTimer;

public:
		inline TimeTriggeredRenderController(Window *window) :
			mDesiredFrameTime(DEFAULT_FRAME_TIME), mLastFrameStartTime(0.0),
					mWindow(window), mTimer(0) {
		}

		inline ~TimeTriggeredRenderController() {
			if (mTimer) {
				// The listener maintains the timer, therefore
				// the listener have to be remove before the timer is killed.
				// See Timer::~Timer().
				mTimer->removeTimerListener();
				delete mTimer;
			}
		}

		inline void setDesiredFrameTime(double frameTime) {
			mDesiredFrameTime = frameTime;
		}

		inline void startOperation() {
			if (!mTimer) {
				mTimer = mWindow->createTimer(mDesiredFrameTime);
				mTimer->registerTimerListener(this);
			}

			else {
				Assert(!mTimer->isRunning(),
						"TimeTriggeredRenderController::startOperation");
				mTimer->start();
			}
		}

		inline void stopOperation() {
			Assert(mTimer, "TimeTriggeredRenderController::stopOperation");
			mTimer->stop();
		}

		inline virtual void onTick() {
			if (Timer::getSystemTime() - mLastFrameStartTime
					>= mDesiredFrameTime)
				mWindow->mWindowAttributes->mRedisplayRequest = true;
		}

		inline void frameStarted() {
			mLastFrameStartTime = Timer::getSystemTime();
		}
	};

protected:
	typedef std::vector<WindowListener *> WindowListenerVector;
	static const double VIDEO_IS_NOT_STARTED = -1.0;

protected:
	bool mOwnWindowAttributes;
	WindowAttributes *mWindowAttributes;

	WindowListenerVector mWindowListeners;

	// Window proxy
	WindowProxy *mWindowProxy;

	// Navigator
	Navigator *mNavigator;

	// FPS calculation
	FrameRateCalculator mFrameRateCalculator;

	// Internal controller for time triggered controlling
	TimeTriggeredRenderController mTimeTriggeredRenderController;

protected:
			Window(
					unsigned width = DEFAULT_WIDTH,
					unsigned height = DEFAULT_HEIGHT,
					WindowAttributes::ColorMode colorMode = WindowAttributes::DIRECT_COLORS,
					bool doubleBuffer = true, bool fullScreen = false,
					unsigned videoRate = 0, unsigned char *colorBits = 0,
					unsigned char depthBits = 16,
					unsigned char stencilBits = 0,
					unsigned char *accumColorBits = 0, unsigned auxBuffers = 0,
					bool stereo = false);
	virtual ~Window();

public:
	// Getters
	inline virtual int getLeft() const {
		return mWindowAttributes->mLeft;
	}
	inline virtual int getTop() const {
		return mWindowAttributes->mTop;
	}
	inline virtual int getWidth() const {
		return mWindowAttributes->mWidth;
	}
	inline virtual int getHeight() const {
		return mWindowAttributes->mHeight;
	}
	inline virtual const std::string& getTitle() const {
		return mWindowAttributes->mTitle;
	}
	inline virtual bool isFullScreen() {
		return mWindowAttributes->mFullScreen;
	}
	inline virtual bool isClosing() {
		return mWindowAttributes->mCloseRequest;
	}

	// Setters
	virtual void setPosition(int left, int top) = 0;
	virtual void setSize(unsigned width, unsigned height) = 0;
	virtual void setFullScreen(bool fullScreen) = 0;
	virtual void setVisible(bool visible) = 0;
	virtual void iconify() = 0;
	virtual void setTitle(const std::string& title) = 0;

	inline virtual void setAutoFinishFrame(bool flag) {
		mWindowAttributes->mAutoFinishFrame = flag;
	}
	inline virtual void setAutoUpdateViewport(bool flag) {
		mWindowAttributes->mAutoUpdateViewport = flag;
	}

	virtual void startFrame();
	virtual void finishFrame();
	virtual void redisplay() = 0;
	virtual void clearFrame() = 0;
	inline virtual void close() {
		mWindowAttributes->mCloseRequest = true;

		// Render one more frame.
		// Note that window closing is performed only at the end of a frame.
		redisplay();
	}

	// Window listeners and window proxy handling
	virtual void registerWindowListener(WindowListener *windowListener,
			const bool topLevel = false);
	virtual void removeWindowListener(WindowListener *windowListener);
	virtual void insertWindowProxy(WindowProxy *proxy);
	virtual WindowProxy *removeWindowProxy();
	inline virtual WindowProxy *getWindowProxy() {
		return mWindowProxy;
	}
	inline virtual bool hasWindowProxy() {
		return mWindowProxy != 0;
	}
	inline virtual Window* getToplevelWindow() {
		// For windows it returns this pointer.
		// For window proxies it calls the same method of the proxied window.
		return this;
	}
	inline virtual void registerNavigator(Navigator *navigator) {
		mNavigator = navigator;
		registerWindowListener((WindowListener *) navigator);
	}
	inline Navigator *getNavigator() const {
		return mNavigator;
	}

	virtual void setRenderMode(WindowAttributes::RenderMode renderMode);

	inline virtual void setDesiredFPS(unsigned fps) {
		mTimeTriggeredRenderController.setDesiredFrameTime(1.0/fps);
	}
	virtual void setListenOnIdle(bool listen);

	// Timer
	virtual Timer *createTimer(double period,
			Timer::TimerType type = Timer::TIMER_PERIODIC) = 0;

	// Recording
	inline virtual void setScreenShotPattern(std::string pattern) {
		mWindowAttributes->mScreenShotPattern = pattern;
		mWindowAttributes->mScreenShotIndex = 0;
	}

	inline virtual void setVideoPattern(std::string pattern) {
		mWindowAttributes->mVideoPattern = pattern;
		mWindowAttributes->mVideoSceneIndex = 0;
		mWindowAttributes->mVideoStartTime = VIDEO_IS_NOT_STARTED;
		mWindowAttributes->mVideoRecording = false;
	}

	virtual void createScreenShot();
	virtual void startVideo();
	virtual void pauseVideo();
	inline virtual bool isRecording() {
		return mWindowAttributes->mVideoRecording;
	}

	virtual void checkRedisplay();

	virtual void unProject(int x, int y, double *objX, double *objY,
			double *objZ);
	virtual void unProject(int x, int y, double z, double *objX, double *objY,
			double *objZ);

	// Get FPS
	inline virtual double getFrameRate() const {
		return *(mFrameRateCalculator.getFrameRatePtr());
	}
	inline virtual const double *getFrameRatePtr() const {
		return mFrameRateCalculator.getFrameRatePtr();
	}
	inline virtual double getFrameStartTime() const {
		return mFrameRateCalculator.getFrameStartTime();
	}

	void debugWindow(unsigned indent = 0);

protected:

#define OPERATE_ON_ALLWINDOWLISTENERS(__op__)		for(WindowListenerVector::iterator __i__=mWindowListeners.begin(); __i__ != mWindowListeners.end(); __i__++) (*__i__)->__op__
#define OPERATE_ON_ALLWINDOWLISTENERS_BOOL(__op__)	for(WindowListenerVector::iterator __i__=mWindowListeners.begin(); __i__ != mWindowListeners.end(); __i__++) if((*__i__)->__op__) break
#define OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(__op__)	for(WindowListenerVector::reverse_iterator __i__=mWindowListeners.rbegin(); __i__ != mWindowListeners.rend(); __i__++) if((*__i__)->__op__) break
	inline void _doKeyEvent(int key, unsigned modifiers,
			InputListener::KeyEventState state) {
		// Input listeners are called in reverse order
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(onKeyEvent(key, modifiers, state));
		checkRedisplay();
	}
	inline void _doMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state) {
		// Input listeners are called in reverse order
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(onMouseButtonEvent(x, y, button, state));
		checkRedisplay();
	}
	inline void _doMouseMoveEvent(int x, int y,
			InputListener::MouseMoveState state) {
		// Input listeners are called in reverse order
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(onMouseMoveEvent(x, y, state));
		checkRedisplay();
	}
	inline void _doMouseMoveEvent(InputListener::MouseMoveState state) {
		// Input listeners are called in reverse order
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(onMouseMoveEvent((*__i__)->mMouseX, (*__i__)->mMouseY, state));
		checkRedisplay();
	}
	inline void _doMouseWheelEvent(int x, int y, int delta) {
		// Input listeners are called in reverse order
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE(onMouseWheelEvent(x, y, delta));
		checkRedisplay();
	}
	inline void _doRender() {
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL(onRender());
		// No redisplay check is performed in this case,
		// because the frame has to be finished before
		// using the finishFrame method
	}
	inline void _doIdle() {
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL(onIdle());
		checkRedisplay();
	}
	inline void _doResize(unsigned width, unsigned height) {
		OPERATE_ON_ALLWINDOWLISTENERS(onResize(width, height));
		checkRedisplay();
	}
	inline void _doReposition(int left, int top) {
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL(onReposition(left, top));
		checkRedisplay();
	}
	inline void _doShow() {
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL(onShow());
		checkRedisplay();
	}
	inline void _doHide() {
		OPERATE_ON_ALLWINDOWLISTENERS_BOOL(onHide());
		checkRedisplay();
	}

	inline void _doRegisterKeyboardEvent(int key,
			InputListener::KeyEventState state) {
		OPERATE_ON_ALLWINDOWLISTENERS(_registerKeyboardEvent(key, state));
	}
	inline void _doRegisterKeyboardModifierEvent(unsigned modifiers) {
		OPERATE_ON_ALLWINDOWLISTENERS(_registerKeyboardModifierEvent(modifiers));
	}
	inline void _doRegisterMouseMoveEvent(int x, int y) {
		OPERATE_ON_ALLWINDOWLISTENERS(_registerMouseMoveEvent(x, y));
	}
	inline void _doRegisterMouseButtonEvent(InputListener::MouseButton button,
			InputListener::MouseEventState state) {
		OPERATE_ON_ALLWINDOWLISTENERS(_registerMouseButtonEvent(button, state));
	}
	inline void _doRegisterMouseWheelEvent(int delta) {
		OPERATE_ON_ALLWINDOWLISTENERS(_registerMouseWheelEvent(delta));
	}
#undef OPERATE_ON_ALLWINDOWLISTENERS
#undef OPERATE_ON_ALLWINDOWLISTENERS_BOOL
#undef OPERATE_ON_ALLWINDOWLISTENERS_BOOL_REVERSE

	void _grabVideoFrame();

	virtual void _passWindowListeners(Window *to);

	virtual void _setupWindowListener() = 0;
	virtual void _setupWindowListenerIdle() = 0;
	virtual void _windowIsReadyToBeShown() = 0;

	virtual void _close() = 0;
}; // class Window

// =============================================================================

class WindowProxy : public Window, public WindowListener {
protected:
	Window *mWindow;

public:
	WindowProxy(Window *window);
	virtual ~WindowProxy();

	inline virtual Window* getProxiedWindow() {
		return mWindow;
	}

	inline virtual void setPosition(int left, int top) {
		mWindow->setPosition(left, top);
	}
	inline virtual void setSize(unsigned width, unsigned height) {
		mWindow->setSize(width, height);
	}
	inline virtual void setFullScreen(bool fullScreen) {
		mWindow->setFullScreen(fullScreen);
	}
	inline virtual void setVisible(bool visible) {
		mWindow->setVisible(visible);
	}
	inline virtual void iconify() {
		mWindow->iconify();
	}
	inline virtual void setTitle(const std::string& title) {
		mWindow->setTitle(title);
	}
	inline virtual void redisplay() {
		mWindow->redisplay();
	}
	inline virtual void clearFrame() {
		mWindow->clearFrame();
	}
	inline virtual void close() {
		mWindow->close();
	}
	inline virtual Window* getToplevelWindow() {
		// For windows it returns this pointer.
		// For window proxies it calls the same method of the proxied window.
		return mWindow->getToplevelWindow();
	}
	inline virtual void registerNavigator(Navigator *navigator) {
		mWindow->registerNavigator(navigator);
	}
	inline Navigator *getNavigator() const {
		return mWindow->getNavigator();
	}
	inline virtual void setRenderMode(WindowAttributes::RenderMode renderMode) {
		mWindow->setRenderMode(renderMode);
	}
	inline virtual void setDesiredFPS(unsigned fps) {
		mWindow->setDesiredFPS(fps);
	}
	virtual void setListenOnIdle(bool listen) {
		mWindow->setListenOnIdle(listen);
	}

	inline virtual Timer *createTimer(double period,
			Timer::TimerType type = Timer::TIMER_PERIODIC) {
		return mWindow->createTimer(period, type);
	}

	/*inline virtual void setScreenShotPattern(std::string pattern) {
	 printf("WindowProxy::setScreenShotPattern\n");
	 return mWindow->setScreenShotPattern(pattern);
	 }
	 inline virtual void setVideoPattern(std::string pattern) {
	 return mWindow->setVideoPattern(pattern);
	 }*/

	inline virtual double getFrameRate() const {
		return mWindow->getFrameRate();
	}
	inline virtual const double *getFrameRatePtr() const {
		return mWindow->getFrameRatePtr();
	}

	inline virtual bool onKeyEvent(int key, unsigned modifiers,
			InputListener::KeyEventState state) {
		_doKeyEvent(key, modifiers, state);
		return false;
	}
	inline virtual bool onMouseButtonEvent(int x, int y,
			InputListener::MouseButton button,
			InputListener::MouseEventState state) {
		_doMouseButtonEvent(x, y, button, state);
		return false;
	}
	inline virtual bool onMouseMoveEvent(int x, int y,
			InputListener::MouseMoveState state) {
		_doMouseMoveEvent(x, y, state);
		return false;
	}
	inline virtual bool onMouseMoveEvent(InputListener::MouseMoveState state) {
		_doMouseMoveEvent(state);
		return false;
	}
	inline virtual bool onMouseWheelEvent(int x, int y, int delta) {
		_doMouseWheelEvent(x, y, delta);
		return false;
	}
	inline virtual bool onRender() {
		_doRender();
		return false;
	}
	inline virtual bool onIdle() {
		_doIdle();
		return false;
	}
	inline virtual bool onResize(unsigned width, unsigned height) {
		_doResize(width, height);
		return false;
	}
	inline virtual bool onReposition(int left, int top) {
		_doReposition(left, top);
		return false;
	}
	inline virtual bool onShow() {
		_doShow();
		return false;
	}
	inline virtual bool onHide() {
		_doHide();
		return false;
	}

	inline virtual void _registerKeyboardEvent(int key,
			InputListener::KeyEventState state) {
		_doRegisterKeyboardEvent(key, state);
	}

	inline virtual void _registerKeyboardModifierEvent(unsigned modifiers) {
		_doRegisterKeyboardModifierEvent(modifiers);
	}

	inline virtual void _registerMouseMoveEvent(int x, int y) {
		_doRegisterMouseMoveEvent(x, y);
	}

	inline virtual void _registerMouseButtonEvent(
			InputListener::MouseButton button,
			InputListener::MouseEventState state) {
		_doRegisterMouseButtonEvent(button, state);
	}

	inline virtual void _registerMouseWheelEvent(int delta) {
		_doRegisterMouseWheelEvent(delta);
	}

	/// Find the window which is proxied by the specified proxy window class.
	/// The search is performed up along the proxy chain 
	template<class PROXYWINDOWCLASS> static PROXYWINDOWCLASS
			* findProxiedWindow(Window *window);

protected:
	inline virtual void _setupWindowListener() {
		mWindow->_setupWindowListener();
	}
	inline virtual void _setupWindowListenerIdle() {
		mWindow->_setupWindowListenerIdle();
	}
	inline virtual void _windowIsReadyToBeShown() {
		mWindow->_windowIsReadyToBeShown();
	}
	inline virtual void _close() {
		mWindow->close();
	}
}; // class WindowProxy

//
// Template implementation
//
template<class PROXYWINDOWCLASS> PROXYWINDOWCLASS* WindowProxy::findProxiedWindow(
		Window *window) {
	for (;;) {
		// Check the proxy of the window.
		// If it is a ParaCompPart window proxy return it.
		PROXYWINDOWCLASS *part =
				dynamic_cast<PROXYWINDOWCLASS *>(window->getWindowProxy());
		if (part)
			return part;

		// Otherwise, check whether the window is a window proxy itself
		// and shedule invertigation its proxied window in the next iteration.
		// (Ie. step up along the proxy chain.)
		WindowProxy* proxy = dynamic_cast<WindowProxy *>(window);
		if (proxy)
			window = proxy->getProxiedWindow();

		// Otherwise, return null pointer
		// (This should be checked when using.)
		else
			return 0;
	}

	return 0;
}

} // namespace MinGLE

#endif // __MINGLE_WINDOW_H__
