//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Atlas.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Atlas.cpp
/// Generic, image, and texture atlas classes

// Header that contains the declaration
#include <Atlas.h>

// MinGLE headers
#include <Exception.h>
#include <tinyxml/include/MingleTinyXml.h>

// System headers
#include <string.h>

namespace MinGLE {

// Static string constants are defined for this cpp module only
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_ROOT_ELEMENT = "image-atlas";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_IMAGEFILE_ELEMENT = "file";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_ELEMENT = "region";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_KEY_ATTRIBUTE = "key";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_U1_ATTRIBUTE = "u1";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_V1_ATTRIBUTE = "v1";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_U2_ATTRIBUTE = "u2";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_REGION_V2_ATTRIBUTE = "v2";
template <typename INDEX> const std::string ImageAtlas<INDEX>::XML_IMAGEFILE_EXTENSION = "png";

// Image atlas save implementation
#define ImageAtlas__saveToFile_IMPLEMENTATION(INDEX) \
template <> \
void ImageAtlas<INDEX>::saveToFile(const std::string& filename, VirtualFileSystem *vfs) { \
	/* Create document */ \
	TiXmlDocument *doc = TinyXML::createXML(); \
	/* Create root element */ \
	TiXmlElement *root = new TiXmlElement(XML_ROOT_ELEMENT.c_str()); \
	doc->LinkEndChild(root); \
	/* Add image file name */ \
	std::string imageFilename = VirtualFileSystem::getBaseName(filename) + '.' + XML_IMAGEFILE_EXTENSION.c_str(); \
	TiXmlElement *imageFile = new TiXmlElement(XML_IMAGEFILE_ELEMENT.c_str()); \
	imageFile->LinkEndChild(new TiXmlText(imageFilename.c_str())); \
	root->LinkEndChild(imageFile); \
	/* Add regions */ \
	std::string regionKey, u1, v1, u2, v2;  \
	for(CoordSet2DIterator i=Atlas<INDEX, unsigned>::mRegions.begin(); i!=Atlas<INDEX, unsigned>::mRegions.end(); i++) { \
    	/* Convert to string */ \
    	StringUtils::toString<INDEX>(regionKey, i->first); \
    	StringUtils::toString<unsigned>(u1, i->second.u1); \
    	StringUtils::toString<unsigned>(v1, i->second.v1); \
    	StringUtils::toString<unsigned>(u2, i->second.u2); \
    	StringUtils::toString<unsigned>(v2, i->second.v2); \
		 \
		/* Create element */ \
		TiXmlElement *regionElement = new TiXmlElement(XML_REGION_ELEMENT.c_str()); \
		regionElement->SetAttribute(XML_REGION_KEY_ATTRIBUTE.c_str(), regionKey.c_str()); \
		regionElement->SetAttribute(XML_REGION_U1_ATTRIBUTE.c_str(), u1.c_str()); \
		regionElement->SetAttribute(XML_REGION_V1_ATTRIBUTE.c_str(), v1.c_str()); \
		regionElement->SetAttribute(XML_REGION_U2_ATTRIBUTE.c_str(), u2.c_str()); \
		regionElement->SetAttribute(XML_REGION_V2_ATTRIBUTE.c_str(), v2.c_str()); \
		root->LinkEndChild(regionElement); \
	} \
	/* Save XML document: TODO serialize and save using vfs */ \
	TinyXML::saveXML(doc, filename, vfs); \
	/* Save image file */ \
    mImage->flip(); /* TODO DevIL need flip */ \
    mImage->saveToFile(imageFilename, vfs); \
    mImage->flip(); \
    /* Delete XML document */ \
	delete doc;\
}

// Image atlas load implementation
#define ImageAtlas__loadFromFile_IMPLEMENTATION(INDEX) \
template <> \
void ImageAtlas<INDEX>::_loadFromFile(const std::string& filename, VirtualFileSystem *vfs) { \
	/* Load XML file */ \
	char *xmlSource = vfs->readFileAsString(filename); \
	TiXmlDocument doc; \
	doc.Parse(xmlSource); \
	delete [] xmlSource; \
	const TiXmlElement *root = doc.RootElement(); \
	Assert(XML_ROOT_ELEMENT == root->Value(), "ImageAtlas<INDEX>::_loadFromFile"); \
	/* Get image file */ \
	const TiXmlNode *imageFile = root->FirstChild("file"); \
	Assert(imageFile, "ImageAtlas<INDEX>::_loadFromFile"); \
	const TiXmlNode *imageFileText = imageFile->FirstChild(); \
	Assert(imageFileText && imageFileText->Type() == TiXmlNode::TEXT, "ImageAtlas<INDEX>::_loadFromFile"); \
	/* Calculate image directory and load image file */ \
	std::string directory = VirtualFileSystem::getDirectory(filename); \
	if(!directory.size()) directory = "./"; \
	std::string imageFilename = directory + imageFileText->ToText()->Value(); \
	mImage = Imaging::createImage(imageFilename, vfs); \
	/* Read regions */ \
	for(TiXmlNode *regionNode = (TiXmlNode *) root->FirstChild(XML_REGION_ELEMENT.c_str()); regionNode; \
		regionNode = (TiXmlNode *) root->IterateChildren(XML_REGION_ELEMENT.c_str(), regionNode)) { \
			/* Cast to element */ \
			const TiXmlElement *regionElement = regionNode->ToElement(); \
			Assert(regionElement, "ImageAtlas<INDEX>::_loadFromFile"); \
			/* Read key */ \
			const char *attrib = regionElement->Attribute(XML_REGION_KEY_ATTRIBUTE.c_str()); Assert(attrib, "ImageAtlas<INDEX>::_loadFromFile"); \
			INDEX key = StringUtils::parse<INDEX>(attrib); \
			/* Read region borders */ \
			/*typename*/ Atlas< INDEX, unsigned >::CoordSet2D coordSet2D; \
			attrib = regionElement->Attribute(XML_REGION_U1_ATTRIBUTE.c_str()); Assert(attrib, "ImageAtlas<INDEX>::_loadFromFile"); \
			coordSet2D.u1 = StringUtils::parse<unsigned>(attrib); \
			attrib = regionElement->Attribute(XML_REGION_V1_ATTRIBUTE.c_str()); Assert(attrib, "ImageAtlas<INDEX>::_loadFromFile"); \
			coordSet2D.v1 = StringUtils::parse<unsigned>(attrib); \
			attrib = regionElement->Attribute(XML_REGION_U2_ATTRIBUTE.c_str()); Assert(attrib, "ImageAtlas<INDEX>::_loadFromFile"); \
			coordSet2D.u2 = StringUtils::parse<unsigned>(attrib); \
			attrib = regionElement->Attribute(XML_REGION_V2_ATTRIBUTE.c_str()); Assert(attrib, "ImageAtlas<INDEX>::_loadFromFile"); \
			coordSet2D.v2 = StringUtils::parse<unsigned>(attrib); \
			/* Define region */ \
			addCoordSet(key, &coordSet2D); \
	} \
}

// Template implementations
ImageAtlas__saveToFile_IMPLEMENTATION(int)
ImageAtlas__loadFromFile_IMPLEMENTATION(int)

#undef ImageAtlas__saveToFile_IMPLEMENTATION
#undef ImageAtlas__loadFromFile_IMPLEMENTATION

} // namespace MinGLE
