//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Configuration.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Configuration.cpp
/// Configuration file support

// Header that contains the declaration
#include <Configuration.h>

// MinGLE headers
#include <System.h>

// System headers
#include <string.h>

namespace MinGLE {

const char Configuration::ENTRY_SEPARATOR =
		VirtualFileSystem::DIRECTORY_SEPARATOR;

Configuration::Entry::Entry(const std::string& _name, const std::string& _value) {
	name = _name;
	type = SIMPLE_ENTRY;
	value.simpleValue = new std::string(_value);
}

Configuration::Entry::Entry(const std::string& _name) {
	name = _name;
	type = COMPOSITE_ENTRY;
	value.compositeValue = new EntryList();
}

Configuration::Entry::Entry(const std::string& _name, EntryList *_value) {
	name = _name;
	type = COMPOSITE_ENTRY;
	value.compositeValue = _value;
}

Configuration::Entry::~Entry() {
	switch (type) {
	case SIMPLE_ENTRY:
		delete value.simpleValue;
		break;
	case COMPOSITE_ENTRY:

		// Delete list entries
		EntryList *eList = value.compositeValue;
		for (EntryList::iterator i=eList->begin(); i!=eList->end(); i++) {
			delete *i;
		}

		// Delete the list
		delete value.compositeValue;
		break;
	}
}

const std::string& Configuration::Entry::getStringValue() {
	Assert(type == SIMPLE_ENTRY, "Configuration::Entry::getStringValue");
	return *value.simpleValue;
}

Configuration::Entry *Configuration::Entry::getChild(const unsigned index) {
	Assert(type == COMPOSITE_ENTRY && index < (*(value.compositeValue)).size(),
			"Configuration::Entry::getChild");
	return (*(value.compositeValue))[index];
}

Configuration::Entry *Configuration::Entry::getFirstChild(
		const std::string& name) {
	Assert(type == COMPOSITE_ENTRY, "Configuration::Entry::getFirstChild");
	EntryList *eList = value.compositeValue;
	for (EntryList::iterator i=eList->begin(); i!=eList->end(); i++)
		if ((*i)->name == name)
			return *i;

	return 0;
}

Configuration::EntryList *Configuration::Entry::getChildren() {
	Assert(type == COMPOSITE_ENTRY, "Configuration::Entry::getChildren");
	return value.compositeValue;
}

Configuration::EntryList Configuration::Entry::findChildren(
		const std::string& name) {
	Assert(type == COMPOSITE_ENTRY, "Configuration::Entry::findChildren");

	EntryList *eList = value.compositeValue;
	EntryList result;

	for (EntryList::iterator i=eList->begin(); i!=eList->end(); i++)
		if ((*i)->name == name)
			result.push_back(*i);

	return result;
}

Configuration::EntryList Configuration::Entry::getEntries(
		const std::string& searchPath) {

	// Search for the separator char
	std::string::size_type delimPos = searchPath.find(ENTRY_SEPARATOR);

	// Not the last level
	if (delimPos != std::string::npos) {
		// Chop search path into two pieces
		std::string head = searchPath.substr(0, delimPos);
		std::string tail = searchPath.substr(delimPos+1);

		// Call recursively 
		EntryList result;
		EntryList children = findChildren(head);
		for (EntryList::iterator i=children.begin(); i!=children.end(); i++) {
			Configuration::EntryList childResult = (*i)->getEntries(tail);
			result.insert(result.end(), childResult.begin(), childResult.end());
		}

		return result;
	}

	// The last level
	return findChildren(searchPath);
}

Configuration::Configuration() :
	// Create composite root node with blank name
			mRootEntry(new Entry("")) {
}

Configuration::~Configuration() {
	delete mRootEntry;
}

Configuration::Entry *Configuration::getFirstChild(const std::string& name) {
	return mRootEntry->getFirstChild(name);
}

Configuration::EntryList Configuration::getEntries(
		const std::string& searchPath, bool onlySimpleEntries) {

	// Return only simple entries
	if (onlySimpleEntries) {
		// Retrieve all
		Configuration::EntryList result = mRootEntry->getEntries(searchPath);

		// Remove non-simple entries
		for (EntryList::iterator i=result.begin(); i!=result.end(); i++) {
			if ((*i)->type != SIMPLE_ENTRY)
				result.erase(i);
		}

		return result;
	}

	// Return anything
	return mRootEntry->getEntries(searchPath);
}

void Configuration::dumpForDebug() {
	// Print children of the root element
	EntryList *eList = mRootEntry->value.compositeValue;
	for (EntryList::iterator i=eList->begin(); i!=eList->end(); i++)
		_dumpForDebug(*i, 0);

}

Configuration::Entry *Configuration::_addChild(Entry *parent,
		const std::string& name, const std::string& value) {
	Assert(parent->type == COMPOSITE_ENTRY, "Configuration::Entry::_addChild");

	Entry *entry = new Entry(name, value);
	parent->value.compositeValue->push_back(entry);
	return entry;
}

Configuration::Entry *Configuration::_addChild(Entry *parent,
		const std::string& name) {
	Assert(parent->type == COMPOSITE_ENTRY, "Configuration::Entry::_addChild");

	Entry *entry = new Entry(name);
	parent->value.compositeValue->push_back(entry);
	return entry;
}

void Configuration::_dumpForDebug(const Entry *entry, const unsigned level) {
	switch (entry->type) {
	case SIMPLE_ENTRY:
		// Indenting
		for (unsigned i=0; i<level; i++)
			printf(" ");

		// Print (name, value) pair
		printf("%s = %s\n", entry->name.c_str(),
				entry->value.simpleValue->c_str());

		break;
	case COMPOSITE_ENTRY:
		// Indenting
		for (unsigned i=0; i<level; i++)
			printf(" ");

		// Print begin sign
		printf("%s {\n", entry->name.c_str());

		// Print children
		EntryList *eList = entry->value.compositeValue;
		for (EntryList::iterator i=eList->begin(); i!=eList->end(); i++)
			_dumpForDebug(*i, level+1);

		// Indenting
		for (unsigned i=0; i<level; i++)
			printf(" ");

		// Print end sign
		printf("}\n");

		break;
	}
}

// =============================================================================

template<> Manager<ConfigurationManager>::ImplementationMapPtr Manager<
		ConfigurationManager>::mImplementations =
		new Manager<ConfigurationManager>::ImplementationMap();
template<> ProviderManager<ConfigurationManager, FileFormatFunctionality>::FunctionalityMapPtr
		ProviderManager<ConfigurationManager, FileFormatFunctionality>::mFunctionalities =
				new ProviderManager<ConfigurationManager, FileFormatFunctionality>::FunctionalityMap();

// Singleton instance
template<> ConfigurationManager *Manager<ConfigurationManager>::msInstance = 0;

ConfigurationManager::~ConfigurationManager() {
	// Nothing to do
}

Configuration *ConfigurationManager::createConfiguration(
		const std::string& filename, VirtualFileSystem *vfs) {
	return getProvider(FileFormatFunctionality(VirtualFileSystem::getExtension(filename), FileFormatFunctionality::READ))->_createConfiguration(
			filename, vfs);
}

Configuration *ConfigurationManager::createConfiguration(
		const std::string& blockName, const std::string& filename) {
	return createConfiguration(filename, System::getVirtualFileSystemForFile(
			blockName, filename));
}

} // namespace MinGLE
