//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/DynLoad.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file DynLoad.cpp
/// Dynamic load library support

// Header that contains the declaration
#include <DynLoad.h>

// MinGLE headers
#include <VirtualFileSystem.h>

// System headers
#include <dlfcn.h>

namespace MinGLE {

// TODO platform specific
const std::string DynLoad::DYNAMIC_LIBRARY_EXTENSION = "so";

// TODO platform specific
const unsigned DynLoad::DYNAMIC_LIBRARY_TYPE_MASK = VirtualFileSystem::UNKNOWN
		| VirtualFileSystem::REGULAR;

DynLoad::DynLoad(const std::string& libraryName, LoadMode loadMode) :
	mHandle(0) {

	// Load dynamic library
	_load(libraryName, loadMode);
}

DynLoad::~DynLoad() {
	if (mHandle)
		_unload();
}

bool DynLoad::hasSymbol(const std::string& symbol) {
	Assert(mHandle, "DynLoad::hasSymbol");
	return dlsym(mHandle, symbol.c_str()) != 0;
}

void DynLoad::_load(const std::string& libraryName, LoadMode loadMode) {
	int mode= RTLD_LAZY;
	switch (loadMode) {
	case LAZY:
		mode = RTLD_LAZY;
		break;
	case IMMEDIATE:
		mode = RTLD_NOW;
		break;
	}

	// Open dynamic library
	if (!(mHandle = dlopen(libraryName.c_str(), mode)))
		Except("DynLoad::_load", dlerror());
}

void DynLoad::_unload() {
	// Close dynamic library
	if (dlclose(mHandle))
		Except("DynLoad::_unload()", dlerror());
}

void *DynLoad::_getSymbol(const std::string& symbol) {
	void *funcPtr = dlsym(mHandle, symbol.c_str());

	if (!funcPtr)
		Except("DynLoad::_getSymbol()", dlerror());

	return funcPtr;
}

// =============================================================================

const std::string DynLib::DYNAMIC_LIBRARY_LOADER_FUNCTION = "loadDynLib";
const std::string DynLib::DYNAMIC_LIBRARY_UNLOADER_FUNCTION = "unloadDynLib";

DynLib::DynLib(const std::string& libraryName, LoadMode loadMode) :
	DynLoad(libraryName, loadMode) {

	// Get loader function
	DynLibLoaderFunction loaderFunction =
			getSymbol<DynLibLoaderFunction>(DYNAMIC_LIBRARY_LOADER_FUNCTION);

	// Call loader function
	loaderFunction();
}

DynLib::~DynLib() {
	// Get unloader function
	DynLibUnLoaderFunction
			unloaderFunction =
					getSymbol<DynLibUnLoaderFunction>(DYNAMIC_LIBRARY_UNLOADER_FUNCTION);

	// Call unloader function
	unloaderFunction();
}

} // namespace MinGLE
