//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/FontRendering.cpp
// Creation Date: December 12th 2007
// Revision Date: January 25th 2008
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file FontRendering.cpp
/// Generic font rendering classes

// Header that contains the declaration
#include <FontRendering.h>

// MinGLE headers
#include <Exception.h>
#include <System.h>
#include <GLUtils.h>
#include <Atlas.h>

namespace MinGLE {

const float Font::DEFAULT_FONT_COLOR[4] = { 1.0, 1.0, 1.0, 1.0 };

Font::Font() {
	// Copy the default color
	mColor[0] = DEFAULT_FONT_COLOR[0], mColor[1] = DEFAULT_FONT_COLOR[1],
			mColor[2] = DEFAULT_FONT_COLOR[2], mColor[3]
					= DEFAULT_FONT_COLOR[3];
}

Font::~Font() {
}

// =============================================================================

RasterFont::RasterFont(unsigned height, unsigned width, long faceIndex,
		unsigned firstCharacter, unsigned lastCharacter, GLenum format,
		GLenum type, ImageScaleMode imageScaleMode) :
	mWidth(width), mHeight(height), mFaceIndex(faceIndex),
			mFirstCharacter(firstCharacter), mLastCharacter(lastCharacter),
			mFormat(format), mType(type),
			mBytesPerPixel(GLUtils::calculateBytesPerPixel(mFormat, mType)),
			mImageAtlas(0), mImageScaleMode(imageScaleMode) {
}

RasterFont::~RasterFont() {
	delete mImageAtlas;
}

void RasterFont::renderToFrameBuffer(const std::string& str, int left, int top) const {
	// Clear depth mask
	GLboolean isDepthMask;
	glGetBooleanv(GL_DEPTH_WRITEMASK, &isDepthMask);
	glDepthMask(GL_FALSE);

	// Disable texturing
	GLboolean isTex1D = glIsEnabled(GL_TEXTURE_1D), isTex2D =
			glIsEnabled(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	// Turn on blending
	// TODO blending function is not saved
	GLboolean isBlending = glIsEnabled(GL_BLEND);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Set pixel storage mode
	GLint packAlignment;
	glGetIntegerv(GL_PACK_ALIGNMENT, &packAlignment);
	GLint unpackAlignment;
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &unpackAlignment);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	// Set color
	float saveColor[4];
	glGetFloatv(GL_RED_SCALE, saveColor);
	glGetFloatv(GL_GREEN_SCALE, saveColor+1);
	glGetFloatv(GL_BLUE_SCALE, saveColor+2);
	glGetFloatv(GL_ALPHA_SCALE, saveColor+3);
	glPixelTransferf(GL_RED_SCALE, mColor[0]);
	glPixelTransferf(GL_GREEN_SCALE, mColor[1]);
	glPixelTransferf(GL_BLUE_SCALE, mColor[2]);
	glPixelTransferf(GL_ALPHA_SCALE, mColor[3]);

	unsigned char *imageData = (unsigned char *) mImageAtlas->getImage()->getData();
	unsigned imagePitch = mImageAtlas->getImage()->getPitch();

	for (char *ch = (char *) str.c_str(), *end = ch + str.size(); ch != end; ch++) {
		ImageAtlas<int>::CoordSet2D *coordSet = mImageAtlas->getCoordSet(*ch);

		// Character not found
		if (!coordSet) {
			// Do a space
			left += 10;
			continue;
		}

		// Pointer on the start of the row (initialized at u1,v1)
		unsigned char *pu1_v = imageData+coordSet->v1*imagePitch+mBytesPerPixel
				*coordSet->u1;

		// Pointer to u1,v2
		unsigned char *pu1_v2 = imageData+coordSet->v2*imagePitch
				+mBytesPerPixel*coordSet->u1;

		// Sizes of the character
		unsigned charWidth = coordSet->u2 - coordSet->u1;
		unsigned charHeight = coordSet->v2 - coordSet->v1;

		// Scan each line in the character
		for (unsigned row = 0; pu1_v != pu1_v2; row++, pu1_v+=imagePitch) {
			// Position on the screen
			glRasterPos2i(left, top-charHeight+row);

			// Draw a scanline of the character
			glDrawPixels(charWidth, 1, mFormat, mType, pu1_v);
		}

		// increment pen position
		left += charWidth;
	}

	// Set saved state
	glDepthMask(isDepthMask ? GL_TRUE : GL_FALSE);
	if (isTex1D)
		glEnable(GL_TEXTURE_1D);
	if (isTex2D)
		glEnable(GL_TEXTURE_2D);
	if (!isBlending)
		glDisable(GL_BLEND);
	glPixelStorei(GL_PACK_ALIGNMENT, packAlignment);
	glPixelStorei(GL_UNPACK_ALIGNMENT, unpackAlignment);
	glPixelTransferf(GL_RED_SCALE, saveColor[0]);
	glPixelTransferf(GL_GREEN_SCALE, saveColor[1]);
	glPixelTransferf(GL_BLUE_SCALE, saveColor[2]);
	glPixelTransferf(GL_ALPHA_SCALE, saveColor[3]);
}

// =============================================================================

template<> Manager<FontRendering>::ImplementationMapPtr
		Manager<FontRendering>::mImplementations =
				new Manager<FontRendering>::ImplementationMap();
template<> ProviderManager<FontRendering, FileFormatFunctionality>::FunctionalityMapPtr
		ProviderManager<FontRendering, FileFormatFunctionality>::mFunctionalities =
				new ProviderManager<FontRendering, FileFormatFunctionality>::FunctionalityMap();

// Singleton instance
template<> FontRendering *Manager<FontRendering>::msInstance = 0;

FontRendering::~FontRendering() {
	// Nothing to do
}

Font *FontRendering::createFont(const std::string& filename,
		VirtualFileSystem *vfs, size_t height, size_t width, long faceIndex,
		size_t firstCharacter, size_t lastCharacter, GLenum format,
		GLenum type, RasterFont::ImageScaleMode imageScaleMode) {
	Assert(msInstance, "FontRendering::createFont");

	return getProvider(FileFormatFunctionality(VirtualFileSystem::getExtension(filename), FileFormatFunctionality::READ))->_createFont(
			filename, vfs, height, width, faceIndex, firstCharacter,
			lastCharacter, format, type, imageScaleMode);
}

Font *FontRendering::createFont(const std::string& blockName,
		const std::string& filename, size_t height, size_t width,
		long faceIndex, size_t firstCharacter, size_t lastCharacter,
		GLenum format, GLenum type, RasterFont::ImageScaleMode imageScaleMode) {
	return createFont(filename, System::getVirtualFileSystemForFile(blockName,
			filename), height, width, faceIndex, firstCharacter, lastCharacter,
			format, type, imageScaleMode);
}

void FontRendering::saveFont(Font *font, const std::string& filename,
		VirtualFileSystem *vfs) {
	getProvider(FileFormatFunctionality(VirtualFileSystem::getExtension(filename), FileFormatFunctionality::WRITE))->_saveFont(
			font, filename, vfs);
}

} // namespace MinGLE
