//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/GLUtils.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file GLUtils.cpp
/// OpenGL utils

// Header that contains the declaration
#include <GLUtils.h>

// MinGLE headers
#include <Exception.h>
#include <MathUtils.h>

// System headers
#include <math.h>

namespace MinGLE {

unsigned GLUtils::calculateBytesPerPixel(GLenum format, GLenum type) {

	// Pixel format
	unsigned size = calculateNumberOfChannels(format);

	// Pixel type
	switch (type) {
	case GL_BYTE:
		size *= 1;
		break;
	case GL_UNSIGNED_BYTE:
		size *= 1;
		break;
	case GL_SHORT:
		size *= 2;
		break;
	case GL_UNSIGNED_SHORT:
		size *= 2;
		break;
	case GL_INT:
		size *= 4;
		break;
	case GL_UNSIGNED_INT:
		size *= 4;
		break;
	case GL_FLOAT:
		size *= 4;
		break;
	case GL_DOUBLE:
		size *= 8;
		break;
	default:
		Except("GLUtils::calculateBytesPerPixel", "Invalid pixel type");
	}

	return size;
}

unsigned GLUtils::calculateNumberOfChannels(GLenum format) {
	switch (format) {
	case GL_COLOR_INDEX:
		return 1;
		break;
	case GL_RGB:
		return 3;
		break;
	case GL_RGBA:
		return 4;
		break;
#if defined(GL_BGR) && defined(GL_BGRA) 
	case GL_BGR:
		return 3;
		break;
	case GL_BGRA:
		return 4;
		break;
#endif
	case GL_LUMINANCE:
		return 1;
		break;
	case GL_LUMINANCE_ALPHA:
		return 2;
		break;
	default:
		Except("GLUtils::calculateNumberOfChannels", "Invalid format");
	}

	return 0;
}

GLenum GLUtils::getTextureType(GLenum target, GLint level, bool forceStandard) {
	// (There are no explicit getter for data type, therefore 
	// we try to find out from its components)
	GLint format = 0;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_INTERNAL_FORMAT,
	&format);

	GLenum type = 0;

	switch (format) {
	case GL_RGB:
#if defined(GL_BGR)
	case GL_BGR:
#endif
	{
		GLint r, g, b;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_RED_TYPE_ARB,
		&r);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_GREEN_TYPE_ARB, &g);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_BLUE_TYPE_ARB, &b);

		Assert(r == g && r == b, "GLUtils::getTextureType");
		type = (GLenum) r;
	}
		break;

	case GL_RGBA:
#if defined(GL_BGRA)
	case GL_BGRA:
#endif
	{
		GLint r, g, b, a;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_RED_TYPE_ARB,
		&r);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_GREEN_TYPE_ARB, &g);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_BLUE_TYPE_ARB, &b);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_ALPHA_TYPE_ARB, &a);

		Assert(r == g && r == b && r == a, "GLUtils::getTextureType");
		type = (GLenum) r;
	}
		break;

	case GL_LUMINANCE: {
		GLint l;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_LUMINANCE_TYPE_ARB, &l);
		type = (GLenum) l;
	}
		break;

	case GL_LUMINANCE_ALPHA: {
		GLint l, a;
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_LUMINANCE_TYPE_ARB, &l);
		glGetTexLevelParameteriv(GL_TEXTURE_2D, level,
		GL_TEXTURE_ALPHA_TYPE_ARB, &a);

		Assert(l == a, "GLUtils::getTextureType");
		type = (GLenum) l;
	}
		break;

	default:
		Except("GLUtils::getTextureType", "Invalid format");
	}
	return forceStandard ? convertToStandardType(type) : type;
}

GLenum GLUtils::convertToStandardType(GLenum type) {

	switch (type) {
	case GL_BYTE:
	case GL_UNSIGNED_BYTE:
	case GL_SHORT:
	case GL_UNSIGNED_SHORT:
	case GL_INT:
	case GL_UNSIGNED_INT:
	case GL_FLOAT:
	case GL_DOUBLE:
		return type;

#ifdef GL_ARB_texture_float

	case GL_UNSIGNED_NORMALIZED_ARB:
	case GL_RGBA32F_ARB:
	case GL_RGB32F_ARB:
	case GL_ALPHA32F_ARB:
	case GL_INTENSITY32F_ARB:
	case GL_LUMINANCE32F_ARB:
	case GL_LUMINANCE_ALPHA32F_ARB:
	case GL_RGBA16F_ARB:
	case GL_RGB16F_ARB:
	case GL_ALPHA16F_ARB:
	case GL_INTENSITY16F_ARB:
	case GL_LUMINANCE16F_ARB:
	case GL_LUMINANCE_ALPHA16F_ARB:
		return GL_FLOAT;
#endif // GL_ARB_texture_float	
	default:
		Except("GLUtils::convertToStandardType", "Invalid pixel type");
	}

	return 0;
}

double GLUtils::calculateDepth(const double *projectionModelViewMatrix,
		const double *position) {
	double v[4];

	MathUtils::matrix4x4VectorProduct(projectionModelViewMatrix, position, v);
	return:: fabs(v[3]) > MathUtils::getEpsilon() ? v[2]/v[3] : /*v[2]*/(v[3]>0.0?1.0:-1.0);
}

double GLUtils::calculateDepth(const double *position) {
	double projectionModelViewMatrix[16];
	GLUtils::getProjectionModelViewMatrix(projectionModelViewMatrix);
	return calculateDepth(projectionModelViewMatrix, position);
}

void GLUtils::getModelViewMatrix(double *modelViewMatrix, bool transpose) {
	if(transpose) {
		double matrixMV[16];
		glGetDoublev(GL_MODELVIEW_MATRIX, matrixMV);
		MathUtils::transposeMatrix4x4(matrixMV, modelViewMatrix);
	}
	else
	glGetDoublev(GL_MODELVIEW_MATRIX, modelViewMatrix);
}

void GLUtils::setTranslation(double *modelViewMatrix, const double x, const double y, const double z, bool transpose) {
	if(transpose) {
		modelViewMatrix[4*0+3] = x,
		modelViewMatrix[4*1+3] = y,
		modelViewMatrix[4*2+3] = z;
	}
	else {
		modelViewMatrix[4*3+0] = x,
		modelViewMatrix[4*3+1] = y,
		modelViewMatrix[4*3+2] = z;
	}
}

void GLUtils::getProjectionModelViewMatrix(double *projectionModelViewMatrix, bool transpose) {
	if(transpose) {
		double matrixMV[16], matrixP[16], matrixMVT[16], matrixPT[16];

		glGetDoublev(GL_MODELVIEW_MATRIX, matrixMV);
		glGetDoublev(GL_PROJECTION_MATRIX, matrixP);

		MathUtils::transposeMatrix4x4(matrixMV, matrixMVT);
		MathUtils::transposeMatrix4x4(matrixP, matrixPT);

		MathUtils::matrix4x4Product(matrixPT, matrixMVT, projectionModelViewMatrix);
	}
	else {
		double matrixMV[16], matrixP[16];

		glGetDoublev(GL_MODELVIEW_MATRIX, matrixMV);
		glGetDoublev(GL_PROJECTION_MATRIX, matrixP);

		MathUtils::matrix4x4Product(matrixMV, matrixP, projectionModelViewMatrix);
	}
}

void GLUtils::getViewDirection(double &x, double &y, double &z) {
	// Get projection-modelview matrix for the calculations
	double modelViewMatrix[16], invModelViewMatrix[16];
	GLUtils::getModelViewMatrix(modelViewMatrix);
	MathUtils::invertMatrix4x4(modelViewMatrix, invModelViewMatrix);

	// Convert minus Z vector to object space
	x = 0.0, y = 0.0, z = -1.0;
	modelTransformation(invModelViewMatrix, x, y, z);

	// And normalize
	MathUtils::vectorNormalize(x, y, z);
}

void GLUtils::getViewPoint(double &x, double &y, double &z) {
	// Get projection-modelview matrix for the calculations
	double modelViewMatrix[16], invModelViewMatrix[16];
	GLUtils::getModelViewMatrix(modelViewMatrix);
	MathUtils::invertMatrix4x4(modelViewMatrix, invModelViewMatrix);

	// Convert eye origin to object space
	x = 0.0, y = 0.0, z = 0.0;
	modelTransformation(invModelViewMatrix, x, y, z);
}

void GLUtils::modelTransformation(const double *matrix, double &x, double &y, double &z) {
	const double pointSource[4] = {x, y, z, 1.0};
	double pointDestination[4];
	MathUtils::matrix4x4VectorProduct(matrix, pointSource, pointDestination);

	// Convert homogenous vector to Descatian one
	MathUtils::vector4to3(pointDestination[0], pointDestination[1], pointDestination[2], pointDestination[3]);
	x = pointDestination[0];
	y = pointDestination[1];
	z = pointDestination[2];
}

unsigned GLUtils::getMax3DTextureSize() {
	GLint maxsize;
	glGetIntegerv(GL_MAX_3D_TEXTURE_SIZE, &maxsize);

	return (unsigned) maxsize;
}

// ORIGINAL SOURCE:
//
// Copyright (C) 1999-2001  Brian Paul   All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
void GLUtils::drawGears(GLfloat angle) {
	static const GLfloat red[4] = {0.8, 0.1, 0.0, 1.0};
	static const GLfloat green[4] = {0.0, 0.8, 0.2, 1.0};
	static const GLfloat blue[4] = {0.2, 0.2, 1.0, 1.0};

	// Gear 1
	glPushMatrix();
	glTranslatef(-3.0, -2.0, 0.0);
	glRotatef(angle, 0.0, 0.0, 1.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);
	_drawGear(1.0, 4.0, 1.0, 20, 0.7);
	glPopMatrix();

	// Gear 2
	glPushMatrix();
	glTranslatef(3.1, -2.0, 0.0);
	glRotatef(-2.0 * angle - 9.0, 0.0, 0.0, 1.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green);
	_drawGear(0.5, 2.0, 2.0, 10, 0.7);
	glPopMatrix();

	// Gear 3
	glPushMatrix();
	glTranslatef(-3.1, 4.2, 0.0);
	glRotatef(-2.0 * angle - 25.0, 0.0, 0.0, 1.0);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue);
	_drawGear(1.3, 2.0, 0.5, 10, 0.7);
	glPopMatrix();
}

// ORIGINAL SOURCE:
//
// Copyright (C) 1999-2001  Brian Paul   All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
void GLUtils::_drawGear(const GLfloat innerRadius, const GLfloat outerRadius,
		const GLfloat width, const GLint teeth, const GLfloat toothDepth) {
	GLint i;
	GLfloat angle;
	GLfloat u, v, len;

	const GLfloat r0 = innerRadius;
	const GLfloat r1 = outerRadius - 0.5 * toothDepth;
	const GLfloat r2 = outerRadius + 0.5 * toothDepth;
	const GLfloat da = 0.5 * M_PI / teeth;

	glShadeModel(GL_FLAT);

	glNormal3f(0.0, 0.0, 1.0);

	/* draw front face */
	glBegin(GL_QUAD_STRIP);
	for (i = 0; i <= teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;
		glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		if (i < teeth) {
			glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
			glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
					width * 0.5);
		}
	}
	glEnd();

	/* draw front sides of teeth */
	glBegin(GL_QUADS);
	for (i = 0; i < teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;

		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
				width * 0.5);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
				width * 0.5);
	}
	glEnd();

	glNormal3f(0.0, 0.0, -1.0);

	/* draw back face */
	glBegin(GL_QUAD_STRIP);
	for (i = 0; i <= teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		if (i < teeth) {
			glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
					-width * 0.5);
			glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		}
	}
	glEnd();

	/* draw back sides of teeth */
	glBegin(GL_QUADS);
	for (i = 0; i < teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;

		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
				-width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
				-width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
	}
	glEnd();

	/* draw outward faces of teeth */
	glBegin(GL_QUAD_STRIP);
	for (i = 0; i < teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;

		glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
		glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
		u = r2 * cos(angle + da) - r1 * cos(angle);
		v = r2 * sin(angle + da) - r1 * sin(angle);
		len = sqrt(u * u + v * v);
		u /= len;
		v /= len;
		glNormal3f(v, -u, 0.0);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
		glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
		glNormal3f(cos(angle), sin(angle), 0.0);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
				width * 0.5);
		glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da),
				-width * 0.5);
		u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
		v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
		glNormal3f(v, -u, 0.0);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
				width * 0.5);
		glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da),
				-width * 0.5);
		glNormal3f(cos(angle), sin(angle), 0.0);
	}

	glVertex3f(r1 * cos(0), r1 * sin(0), width * 0.5);
	glVertex3f(r1 * cos(0), r1 * sin(0), -width * 0.5);

	glEnd();

	glShadeModel(GL_SMOOTH);

	/* draw inside radius cylinder */
	glBegin(GL_QUAD_STRIP);
	for (i = 0; i <= teeth; i++) {
		angle = i * 2.0 * M_PI / teeth;
		glNormal3f(-cos(angle), -sin(angle), 0.0);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
		glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
	}
	glEnd();
}

} // namespace MinGLE
