//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Gui.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Gui.cpp
/// Very simple gui support

// Header that contains the declaration
#include <Gui.h>

// MinGLE headers
#include <Exception.h>
#include <System.h>
#include <Window.h>
#include <Configuration.h>
#include <Serialization.h>
#include <StringUtils.h>
#include <mingle-gl.h>

namespace MinGLE {

// Declarations for serialization
EntrySerialization_Declaration(Widget);
EntrySerialization_Declaration(Panel);
EntrySerialization_Declaration(GuiSheet);
EntrySerialization_Declaration(TextField);
EntrySerialization_Declaration(CheckBox);
EntrySerialization_Declaration(Slider);

const float Widget::DEFAULT_COLOR[4] = { 1.0, 1.0, 1.0, 1.0 };
float Widget::mDefaultColor[4] = { DEFAULT_COLOR[0], DEFAULT_COLOR[1],
		DEFAULT_COLOR[2], DEFAULT_COLOR[3] };

// Vertex indices of the triangulized quad using triangle strip 
const unsigned char Widget::QUAD_INDICES[4] = { 0, 1, 2, 3 };

Widget::Widget() : 
	mId(""), mParent(0), mDirty(true), mFocus(false)
{
	_setDefaultProperties();
}

Widget::Widget(const std::string& id) :
	mId(id), mParent(0), mDirty(true), mFocus(false) {
	_setDefaultProperties();
}

Widget::~Widget() {
}

void Widget::getParentEffectiveBounds(int& parentLeft, int& parentTop,
		int& parentWidth, int& parentHeight) {
	Assert(mParent, "Widget::getParentEffectiveBounds");
	mParent->getEffectiveBounds(parentLeft, parentTop, parentWidth,
			parentHeight);
}

void Widget::render() {
	if (mDirty)
		_recalculateEffectiveBounds();
	_render();
}

void Widget::_setDefaultProperties() {
	setLeftRel(0.0);
	setTopRel(0.0);
	setWidthRel(1.0);
	setHeightRel(1.0);

	mColor[0] = mDefaultColor[0], mColor[1] = mDefaultColor[1], mColor[2]
			= mDefaultColor[2], mColor[3] = mDefaultColor[3];
}

void Widget::_recalculateEffectiveBounds() {

	// Get effective bounds for the parent component
	int parentLeft, parentTop, parentWidth, parentHeight;
	getParentEffectiveBounds(parentLeft, parentTop, parentWidth, parentHeight);

	//printf("parent: %d %d %d %d\n", parentLeft, parentTop, parentWidth, parentHeight);

	// Effective bounds
	int left, top, right, bottom;

	// Effective vertex coords
	// vertex #0 : (left, top)
	// vertex #1 : (right, top)
	// vertex #2 : (right, bottom)
	// vertex #3 : (left, bottom)

	// Effective bounds are window pixel coordinates
	left = mBounds.leftMode == ABSOLUTE ? mBounds.absLeft : (int)(parentWidth
			*mBounds.relLeft);
	if (left < 0)
		left = parentWidth - left;
	left += parentLeft;
	mEffectiveCoords[0] = mEffectiveCoords[6] = left;

	top = mBounds.topMode == ABSOLUTE ? mBounds.absTop : (int)(parentHeight
			*mBounds.relTop);
	if (top < 0)
		top = parentHeight - top;
	top += parentTop;
	mEffectiveCoords[1] = mEffectiveCoords[3] = top;

	right = mBounds.widthMode == ABSOLUTE ? mBounds.absWidth
			: (int)(parentWidth*mBounds.relWidth);
	if (right < 0)
		right = parentWidth - right;
	mEffectiveSizes.width = right;
	right += left;
	mEffectiveCoords[2] = mEffectiveCoords[4] = right;

	bottom = mBounds.heightMode == ABSOLUTE ? mBounds.absHeight
			: (int)(parentHeight*mBounds.relHeight);
	if (bottom < 0)
		bottom = parentHeight - bottom;
	mEffectiveSizes.height = bottom;
	bottom += top;
	mEffectiveCoords[5] = mEffectiveCoords[7] = bottom;

	//printf("%d %d %d %d\n", left, top, right, bottom);
	//printf("(%d,%d)-(%d,%d)-(%d,%d)-(%d,%d)\n", mEffectiveCoords[0], mEffectiveCoords[1], mEffectiveCoords[2], mEffectiveCoords[3], mEffectiveCoords[4], mEffectiveCoords[5], mEffectiveCoords[6], mEffectiveCoords[7]);

	mDirty = false;
}

Configuration::Entry* EntrySerialization<Widget>::serialize(Widget *widget, Configuration::Entry *serialData) {
	Except("EntrySerialization<Widget *>::serialize", "Implement"); 
	return serialData;
}

Widget* EntrySerialization<Widget>::deserialize(Configuration::Entry *serialData, Widget *widget) {
	Assert(widget, "EntrySerialization<Widget>::deserialize");
	
	// Attribute: id
	Configuration::Entry *idEntry = serialData->getFirstChild("id");
	Assert(idEntry, "EntrySerialization<Widget>::deserialize");
	widget->setId(idEntry->getStringValue());

	// Attribute: color
	// TODO: support both (r,g,b) and (r,g,b,a) formats
	Configuration::Entry *colorEntry = serialData->getFirstChild("color");
	if(colorEntry) {
		float r, g, b;
		sscanf(colorEntry->getStringValue().c_str(), "(%f,%f,%f)", &r, &g, &b);
		widget->setColor(r, g, b);
	}
	
	// Attribute: width
	Configuration::Entry *widthEntry = serialData->getFirstChild("width");
	if(widthEntry) {
		std::string widthStr = widthEntry->getStringValue();
		if(widthStr.find('.') != std::string::npos)
			// Relative width
			widget->setWidthRel(StringUtils::parse<float>(widthStr));
		else
			// Absolute width
			widget->setWidthAbs(StringUtils::parse<int>(widthStr));
	}

	// Attribute: height
	Configuration::Entry *heightEntry = serialData->getFirstChild("height");
	if(heightEntry) {
		std::string heightStr = heightEntry->getStringValue();
		if(heightStr.find('.') != std::string::npos)
			// Relative height
			widget->setHeightRel(StringUtils::parse<float>(heightStr));
		else
			// Absolute height
			widget->setHeightAbs(StringUtils::parse<int>(heightStr));
	}

	// Attribute: left
	Configuration::Entry *leftEntry = serialData->getFirstChild("left");
	if(leftEntry) {
		std::string leftStr = leftEntry->getStringValue();
		if(leftStr.find('.') != std::string::npos)
			// Relative left
			widget->setLeftRel(StringUtils::parse<float>(leftStr));
		else
			// Absolute left
			widget->setLeftAbs(StringUtils::parse<int>(leftStr));
	}

	// Attribute: top
	Configuration::Entry *topEntry = serialData->getFirstChild("top");
	if(topEntry) {
		std::string topStr = topEntry->getStringValue();
		if(topStr.find('.') != std::string::npos)
			// Relative top
			widget->setTopRel(StringUtils::parse<float>(topStr));
		else
			// Absolute top
			widget->setTopAbs(StringUtils::parse<int>(topStr));
	}
	
	return widget;
}

// =============================================================================

const char Panel::WIDGET_SEPARATOR = VirtualFileSystem::DIRECTORY_SEPARATOR;

Panel::~Panel() {
	// TODO remove contents
}

void Panel::add(Widget *widget) {
	Assert(widget->getId() != "" && mContents.find(widget->getId()) == mContents.end(), "Panel::add");
	mContents[widget->getId()] = widget;
	widget->mParent = this;
}

void Panel::remove(Widget *widget) {
	Assert(mContents.find(widget->getId()) != mContents.end(), "Panel::remove");
	mContents.erase(widget->getId());
	widget->mParent = 0;
}

void Panel::remove(const std::string& id) {
	std::map<std::string, Widget *>::iterator i = mContents.find(id);
	Assert(i != mContents.end(), "Panel::remove");
	i->second->mParent = 0;
	mContents.erase(id);
}

Widget *Panel::get(const std::string& id) {
	std::map<std::string, Widget *>::iterator i = mContents.find(id);
	return (i != mContents.end()) ? i->second : 0;
}

Widget *Panel::getWidgetByName(const std::string& searchPath) {
	
	// Search for the separator char
	std::string::size_type delimPos = searchPath.find(WIDGET_SEPARATOR);

	// Not the last level
	if (delimPos != std::string::npos) {
		// Chop search path into two pieces
		std::string head = searchPath.substr(0, delimPos);
		std::string tail = searchPath.substr(delimPos+1);

		// Get child panel
		Widget *childWidget = get(head);
		Assert(childWidget, "Panel::getWidgetByName");
		Panel *childPanel = dynamic_cast<Panel *>(childWidget);
		Assert(childPanel, "Panel::getWidgetByName");

		// Call recursively
		return childPanel->getWidgetByName(tail);
	}

	// The last level
	return get(searchPath);
}

void Panel::render() {
	// Recalculate effective bounds if needed
	recalculateEffectiveBounds();

	// Render the panel
	_render();

	// Render contents
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		i->second->render();
	}
}

void Panel::setAlpha(const float alpha, const bool recursive) {
	Widget::setAlpha(alpha);
	if (recursive)
		_setContentAlpha(alpha);
}

void Panel::recalculateEffectiveBounds() {
	if (mDirty) {
		this->_recalculateEffectiveBounds();
	}
}

bool Panel::onMouseButtonEvent(int x, int y, InputListener::MouseButton button,
		InputListener::MouseEventState state) {
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		Widget *w = i->second;

		if (w->contains(x, y) && w->onMouseButtonEvent(x, y, button, state))
			return !mEventTransparent;
	}

	// If the panel is visible it blocks the events 
	return isVisible() && !mEventTransparent;
}

bool Panel::onMouseMoveEvent(int x, int y, InputListener::MouseMoveState state) {
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		Widget *w = i->second;

		if (w->contains(x, y) && w->onMouseMoveEvent(x, y, state))
			return !mEventTransparent;
	}

	// If the panel is visible it blocks the events 
	return isVisible() && !mEventTransparent;
}

bool Panel::onMouseWheelEvent(int x, int y, int delta) {
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		Widget *w = i->second;

		if (w->contains(x, y) && w->onMouseWheelEvent(x, y, delta))
			return !mEventTransparent;
	}

	// If the panel is visible it blocks the events 
	return isVisible() && !mEventTransparent;
}

void Panel::_recalculateEffectiveBounds() {
	// Recalculate the panel
	Widget::_recalculateEffectiveBounds();

	// Recalculate the contents
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		i->second->_recalculateEffectiveBounds();
	}
}

void Panel::_render() {
	// Render only opaque panel
	if (isVisible()) {
		if (isTransparent())
			glEnable(GL_BLEND);
		else
			glDisable(GL_BLEND);
		glColor4fv(mColor);
		glVertexPointer(2, GL_INT, 0, mEffectiveCoords);
		glDrawElements(GL_TRIANGLE_FAN, 4, GL_UNSIGNED_BYTE, QUAD_INDICES);
	}
}

void Panel::_setContentAlpha(const float alpha) {
	for (std::map<std::string, Widget *>::iterator i = mContents.begin(); i
			!=mContents.end(); i++) {
		i->second->setAlpha(alpha, true);
	}
}

Configuration::Entry* EntrySerialization<Panel>::serialize(Panel *panel, Configuration::Entry *serialData) {
	Except("EntrySerialization<Panel *>::serialize", "Implement"); 
	return serialData;
}

Panel* EntrySerialization<Panel>::deserialize(Configuration::Entry *serialData, Panel *panel) {
	Assert(serialData->name == "panel", "EntrySerialization<Panel>::deserialize");
	
	// Create panel object if does not exist
	if(!panel) panel = new Panel();
	
	// Deserialize as a widget
	panel = dynamic_cast<Panel *>(EntrySerialization<Widget>::deserialize(serialData, dynamic_cast<Widget *>(panel)));
	
	// Deserialize the children of the panel
	Configuration::EntryList *childEntries = serialData->getChildren();
	for(Configuration::EntryList::iterator i=childEntries->begin(); i!=childEntries->end(); i++) {
		if((*i)->name == "panel")
			panel->add(EntrySerialization<Panel>::deserialize(*i));
		else if((*i)->name == "text")
			panel->add(EntrySerialization<TextField>::deserialize(*i));
		else if((*i)->name == "checkbox")
			panel->add(EntrySerialization<CheckBox>::deserialize(*i));
		else if((*i)->name == "slider")
			panel->add(EntrySerialization<Slider>::deserialize(*i));
	}
	
	return panel;
}

// =============================================================================

const std::string GuiSheet::DESKTOP_ID = "desktop";

GuiSheet::GuiSheet() :
	mGuiToggleKey(DEFAULT_GUI_TOGGLE_KEY), mBlind(false) {

	// Create "looped back" panel	
	mDesktop = new Panel(DESKTOP_ID);
	mDesktop->setColor(0.0, 0.0, 0.0, 0.0);
	mDesktop->mParent = mDesktop;

	// Create gui toggler
	mGuiToggler = new GuiToggler(this);
}

GuiSheet::GuiSheet(const std::string& filename, VirtualFileSystem *vfs) :
	mGuiToggleKey(DEFAULT_GUI_TOGGLE_KEY), mBlind(false) {
	_loadFromFile(filename, vfs);

	// Create gui toggler
	mGuiToggler = new GuiToggler(this);
}

GuiSheet::GuiSheet(const std::string& blockName, const std::string& filename) :
	mGuiToggleKey(DEFAULT_GUI_TOGGLE_KEY), mBlind(false) {
	_loadFromFile(filename, System::getVirtualFileSystemForFile(blockName,
			filename));

	// Create gui toggler
	mGuiToggler = new GuiToggler(this);
}

GuiSheet::~GuiSheet() {
	delete mDesktop;
}

bool GuiSheet::onRender() {

	// Call update event listener
	onUpdate();

	if (!mBlind) {
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(0.0, mWindow->getWidth(), mWindow->getHeight(), 0.0);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		glDisable(GL_DEPTH_TEST);
		GLboolean isLighting = glIsEnabled(GL_LIGHTING);
		glDisable(GL_LIGHTING);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glEnableClientState(GL_VERTEX_ARRAY);

		// Render as a panel
		mDesktop->render();

		glDisableClientState(GL_VERTEX_ARRAY);

		// Restore matrices
		glDisable(GL_BLEND);
		if (isLighting)
			glEnable(GL_LIGHTING);
		glEnable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	else
		mDesktop->recalculateEffectiveBounds();

	return false;
}

bool GuiSheet::onResize(unsigned width, unsigned height) {
	mDesktop->mDirty = true;

	mDesktop->mEffectiveCoords[0] = mDesktop->mEffectiveCoords[1] = 0;
	mDesktop->mEffectiveSizes.width = width;
	mDesktop->mEffectiveSizes.height = height;

	redisplayRequest();

	return false;
}

bool GuiSheet::onKeyEvent(int key, unsigned modifiers, KeyEventState state) {

	if (key == mGuiToggleKey && !modifiers && state != InputListener::KEY_DOWN) {
		mGuiToggler->toggle();
		return true;
	}

	return false;
}

void GuiSheet::setAlpha(const float alpha) {
	mDesktop->_setContentAlpha(alpha);
}

void GuiSheet::loadFromFile(const std::string& filename, VirtualFileSystem *vfs) {
	// Delete current desktop
	delete mDesktop;

	// Load GUI from file
	_loadFromFile(filename, vfs);

	// Recalculate and redisplay
	mDesktop->mEffectiveCoords[0] = mDesktop->mEffectiveCoords[1] = 0;
	mDesktop->mEffectiveSizes.width = mWindow->getWidth();
	mDesktop->mEffectiveSizes.height = mWindow->getHeight();
	redisplayRequest();
}

void GuiSheet::loadFromFile(const std::string& blockName, const std::string& filename) {
	// Delete current desktop
	delete mDesktop;
	
	// Load GUI from file
	_loadFromFile(filename, System::getVirtualFileSystemForFile(blockName,
			filename));

	// Recalculate and redisplay
	mDesktop->mEffectiveCoords[0] = mDesktop->mEffectiveCoords[1] = 0;
	mDesktop->mEffectiveSizes.width = mWindow->getWidth();
	mDesktop->mEffectiveSizes.height = mWindow->getHeight();
	redisplayRequest();
}

void GuiSheet::_loadFromFile(const std::string& filename,
		VirtualFileSystem *vfs) {

	// Create "looped back" panel	
	mDesktop = new Panel(DESKTOP_ID);
	mDesktop->setColor(0.0, 0.0, 0.0, 0.0);
	mDesktop->mParent = mDesktop;
	mDesktop->mDirty = true;

	// Load as a configuration
	Configuration *conf = ConfigurationManager::createConfiguration(filename, vfs);
	
	// Deserialize from the configuration
	(void) EntrySerialization<GuiSheet>::deserialize(conf->getRootEntry(), this);
	
	// Delete configuration
	delete conf;
}

Configuration::Entry* EntrySerialization<GuiSheet>::serialize(GuiSheet *guiSheet, Configuration::Entry *serialData) {
	Except("EntrySerialization<GuiSheet *>::serialize", "Implement"); 
	return serialData;
}

GuiSheet* EntrySerialization<GuiSheet>::deserialize(Configuration::Entry *serialData, GuiSheet *guiSheet) {
	Assert(serialData->name == "guisheet", "EntrySerialization<GuiSheet>::deserialize");
	Assert(guiSheet, "EntrySerialization<GuiSheet>::deserialize");
	
	// Deserialize the children of the panel
	Configuration::EntryList childPanelEntries = serialData->findChildren("panel");
	for(Configuration::EntryList::iterator i=childPanelEntries.begin(); i!=childPanelEntries.end(); i++) {
		guiSheet->add(EntrySerialization<Panel>::deserialize(*i));
	}
	
	return guiSheet;
}

// =============================================================================

GuiSheet::GuiToggler::GuiToggler(GuiSheet *sheet, double maxAlpha) :
	mGuiToggleTimer(0), mSheet(sheet), mFadeMode(NOFADE), mMaxAlpha(maxAlpha),
			mAlpha(maxAlpha) {
}

GuiSheet::GuiToggler::~GuiToggler() {
	if (mGuiToggleTimer) {
		mGuiToggleTimer->stop();
		delete mGuiToggleTimer;
	}
}

void GuiSheet::GuiToggler::onTick() {
	switch (mFadeMode) {
	case NOFADE:
		mGuiToggleTimer->stop();
		return;
	case FADEOUT:
		mAlpha -= DEFAULT_GUI_TOGGLE_ALPHASTEP;
		if (mAlpha <= 0.0) {
			mFadeMode = NOFADE;
			mAlpha = 0.0;
			mGuiToggleTimer->stop();
			return;
		}
		break;
	case FADEIN:
		mAlpha += DEFAULT_GUI_TOGGLE_ALPHASTEP;
		if (mAlpha >= mMaxAlpha) {
			mFadeMode = NOFADE;
			mAlpha = mMaxAlpha;
			mGuiToggleTimer->stop();
			return;
		}
		break;
	}

	mSheet->setAlpha(mAlpha);
	mSheet->redisplayRequest();
}

void GuiSheet::GuiToggler::show() {
	mFadeMode = FADEIN;

	if (!mGuiToggleTimer)
		_createTimer();
	else
		mGuiToggleTimer->start();
}

void GuiSheet::GuiToggler::hide() {
	mFadeMode = FADEOUT;

	if (!mGuiToggleTimer)
		_createTimer();
	else
		mGuiToggleTimer->start();
}

void GuiSheet::GuiToggler::toggle() {
	if (mAlpha < mMaxAlpha)
		show();
	else
		hide();
}

void GuiSheet::GuiToggler::_createTimer() {
	mGuiToggleTimer = mSheet->mWindow->createTimer(DEFAULT_GUI_TOGGLE_TIMESTEP);
	mGuiToggleTimer->registerTimerListener(this);
}

// =============================================================================

Font *TextField::msDefaultFont = 0;

const TextField::VerticalAlign TextField::DEFAULT_VERTICAL_ALIGN = MIDDLE;

TextField::TextField() :
	Widget(), mFont(0), mVerticalAlign(DEFAULT_VERTICAL_ALIGN), mUserData(0) {
	_setDefaultFont();
}

TextField::TextField(const std::string& id) :
	Widget(id), mFont(0), mVerticalAlign(DEFAULT_VERTICAL_ALIGN), mUserData(0) {
	_setDefaultFont();
}

TextField::TextField(const std::string& id, const std::string& text) :
	Widget(id), mText(text), mFont(0), mVerticalAlign(DEFAULT_VERTICAL_ALIGN),
			mUserData(0) {
	_setDefaultFont();
}

TextField::TextField(const std::string& id, TextField::Data* userData) :
	Widget(id), mFont(0), mVerticalAlign(DEFAULT_VERTICAL_ALIGN), mUserData(userData) {
	_setDefaultFont();
}

TextField::TextField(const std::string& id, const std::string& text, Font* font) :
	Widget(id), mText(text), mFont(font),
			mVerticalAlign(DEFAULT_VERTICAL_ALIGN), mUserData(0) {
}

TextField::~TextField() {
	delete mUserData;
}

void TextField::_render() {
	if (isVisible()) {

		// If have data object, update text
		if (mUserData)
			mText = mUserData->toString();

		// Concatenate prefix and postfix texts
		mText = mPrefixText + mText + mPostfixText;
		
		mFont->setColor(mColor);
		switch (mVerticalAlign) {
		case TOP:
			mFont->renderToFrameBuffer(mText, mEffectiveCoords[0],
					mEffectiveCoords[1] + mFont->getHeight());
			break;
		case MIDDLE:
			mFont->renderToFrameBuffer(mText, mEffectiveCoords[0],
					(mEffectiveCoords[1] + mEffectiveCoords[5])>>1);
			break;
		case BOTTOM:
			mFont->renderToFrameBuffer(mText, mEffectiveCoords[0],
					mEffectiveCoords[5]);
			break;
		}
	}
}

void TextField::_setDefaultFont() {
	Assert(msDefaultFont, "TextField::_setDefaultFont");
	mFont = msDefaultFont;
}

Configuration::Entry* EntrySerialization<TextField>::serialize(TextField *panel, Configuration::Entry *serialData) {
	Except("EntrySerialization<TextField *>::serialize", "Implement"); 
	return serialData;
}

TextField* EntrySerialization<TextField>::deserialize(Configuration::Entry *serialData, TextField *textField) {
	Assert(serialData->name == "text", "EntrySerialization<TextField>::deserialize");
	
	// Create panel object if does not exist
	if(!textField) textField = new TextField();
	
	// Deserialize as a widget
	textField = dynamic_cast<TextField *>(EntrySerialization<Widget>::deserialize(serialData, dynamic_cast<Widget *>(textField)));
	
	// Attribute: text
	Configuration::Entry *textEntry = serialData->getFirstChild("text");
	Assert(textEntry, "EntrySerialization<TextField>::deserialize");
	textField->setText(textEntry->getStringValue());

	// Attribute: valign
	// TODO: implement
	
	return textField;
}

// =============================================================================

const unsigned char CheckBox::UNCHECKED_INDICES[4] = { 0, 1, 2, 3 };
const unsigned char CheckBox::CHECKED_INDICES[8] = { 0, 1, 2, 3, 0, 2, 1, 3 };

CheckBox::CheckBox() :
	Widget(), mChecked(false), mIsRadio(false), mUserData(0) {
	setSizeAbs(16, 16); // TODO wired code
}

CheckBox::CheckBox(const std::string& id) :
	Widget(id), mChecked(false), mIsRadio(false), mUserData(0) {
	setSizeAbs(16, 16); // TODO wired code
}

CheckBox::CheckBox(const std::string& id, const bool checked, const bool isRadio) :
	Widget(id), mChecked(checked), mIsRadio(isRadio), mUserData(0) {
	setSizeAbs(16, 16); // TODO wired code
}

CheckBox::CheckBox(const std::string& id, bool* userData, const bool isRadio) :
	Widget(id), mChecked(*userData), mIsRadio(isRadio), mUserData(userData) {
	setSizeAbs(16, 16); // TODO wired code
}

CheckBox::~CheckBox() {
}

void CheckBox::setChecked(const bool checked) {
	// Radio buttons cannot be unchecked
	if(!checked && mIsRadio) return;
	
	mChecked = checked;
	if (mUserData)
		*mUserData = checked;

	// If it is a radio style checkbox set all other radio check boxes to false
	if(checked && mIsRadio) {
		std::list<CheckBox*> radioCheckBoxes = mParent->Panel::getWidgetsByType<CheckBox>();
		for(std::list<CheckBox*>::iterator i=radioCheckBoxes.begin(); i!=radioCheckBoxes.end(); i++)
			if(*i != this) {
				(*i)->mChecked = false;
				if ((*i)->mUserData)
						*((*i)->mUserData) = false;
			}
	}
}


bool CheckBox::onMouseButtonEvent(int x, int y,
		InputListener::MouseButton button, InputListener::MouseEventState state) {
	if (button == InputListener::LEFT_BUTTON
			&& (state == InputListener::BUTTON_CLICK || state
					== InputListener::BUTTON_UP)) {
		setChecked(!mChecked);
		return true;
	}

	return false;
}

void CheckBox::_render() {
	// Render only opaque panel
	if (isVisible()) {
		if (isTransparent())
			glEnable(GL_BLEND);
		else
			glDisable(GL_BLEND);
		glColor4fv(mColor);
		glVertexPointer(2, GL_INT, 0, mEffectiveCoords);
		if (mChecked)
			glDrawElements(GL_LINE_STRIP, 8, GL_UNSIGNED_BYTE, CHECKED_INDICES);
		else
			glDrawElements(GL_LINE_LOOP, 4, GL_UNSIGNED_BYTE, UNCHECKED_INDICES);
	}
}

Configuration::Entry* EntrySerialization<CheckBox>::serialize(CheckBox *panel, Configuration::Entry *serialData) {
	Except("EntrySerialization<CheckBox *>::serialize", "Implement"); 
	return serialData;
}

CheckBox* EntrySerialization<CheckBox>::deserialize(Configuration::Entry *serialData, CheckBox *checkBox) {
	Assert(serialData->name == "checkbox", "EntrySerialization<CheckBox>::deserialize");
	
	// Create panel object if does not exist
	if(!checkBox) checkBox = new CheckBox();
	
	// Deserialize as a widget
	checkBox = dynamic_cast<CheckBox *>(EntrySerialization<Widget>::deserialize(serialData, dynamic_cast<Widget *>(checkBox)));
	
	// Attribute: checked
	Configuration::Entry *checkedEntry = serialData->getFirstChild("checked");
	if(checkedEntry)
		checkBox->setChecked(StringUtils::parse<bool>(checkedEntry->getStringValue()));

	// Attribute: radio
	Configuration::Entry *isRadioEntry = serialData->getFirstChild("radio");
	if(isRadioEntry)
		checkBox->setRadio(StringUtils::parse<bool>(isRadioEntry->getStringValue()));
	
	return checkBox;
}

// =============================================================================

Slider::Slider() :
	Widget(), mMinimum(DEFAULT_MINIMUM), mMaximum(DEFAULT_MAXIMUM),
			mValue(DEFAULT_VALUE), mReadOnly(DEFAULT_READONLY), mFillType(DEFAULT_FILLTYPE), mButtonDown(false),
			mUserData(0) {
	setHeightAbs(16); // TODO wired code
}

Slider::Slider(const std::string& id, const double value) :
	Widget(id), mMinimum(DEFAULT_MINIMUM), mMaximum(DEFAULT_MAXIMUM),
			mValue(value), mReadOnly(DEFAULT_READONLY), mFillType(DEFAULT_FILLTYPE), mButtonDown(false),
			mUserData(0) {
	setHeightAbs(16); // TODO wired code
}

Slider::Slider(const std::string& id, double* userData) :
	Widget(id), mMinimum(DEFAULT_MINIMUM), mMaximum(DEFAULT_MAXIMUM),
			mValue(*userData), mReadOnly(DEFAULT_READONLY), mFillType(DEFAULT_FILLTYPE), mButtonDown(false),
			mUserData(userData) {
	setHeightAbs(16); // TODO wired code
}

Slider::Slider(const std::string& id, const double value, const double minimum,
		const double maximum, const bool readonly, const FillType fillType) :
	Widget(id), mMinimum(minimum), mMaximum(maximum), mValue(value),
			mReadOnly(readonly), mFillType(fillType), mButtonDown(false), mUserData(0) {
	setHeightAbs(16); // TODO wired code
}

Slider::Slider(const std::string& id, double* userData, const double minimum,
		const double maximum, const bool readonly, const FillType fillType) :
	Widget(id), mMinimum(minimum), mMaximum(maximum), mValue(*userData),
			mReadOnly(readonly), mFillType(fillType), mButtonDown(false), mUserData(userData) {
	setHeightAbs(16); // TODO wired code
}

Slider::~Slider() {
}

bool Slider::onMouseButtonEvent(int x, int y,
		InputListener::MouseButton button, InputListener::MouseEventState state) {
	if (button == InputListener::LEFT_BUTTON) {
		switch (state) {
		case BUTTON_DOWN:
			mButtonDown = true;
			return _handleSeek(x);
			break;
		case BUTTON_CLICK:
			mButtonDown = false;
			return _handleSeek(x);
		case BUTTON_UP:
		case BUTTON_DBLCLICK:
			mButtonDown = false;
		}
	}

	return false;
}

bool Slider::onMouseMoveEvent(int x, int y, InputListener::MouseMoveState state) {
	if (state == InputListener::ENTER)
		mButtonDown = false;

	else if (mButtonDown)
		return _handleSeek(x);

	return false;
}

void Slider::_render() {
	// Render only opaque panel
	if (isVisible()) {
		if (isTransparent())
			glEnable(GL_BLEND);
		else
			glDisable(GL_BLEND);
		glColor4fv(mColor);
		glVertexPointer(2, GL_INT, 0, mEffectiveCoords);
		glDrawElements(GL_LINE_LOOP, 4, GL_UNSIGNED_BYTE, QUAD_INDICES);

		// TODO not effective
		double pos = (mValue-mMinimum)/(mMaximum-mMinimum);
		int effValue = (int) ((1.0-pos)*mEffectiveCoords[0] + pos
				*mEffectiveCoords[2]);

		switch(mFillType) {
		case BAR:
			glBegin(GL_QUADS);
			glVertex2i(effValue-2, mEffectiveCoords[1]);
			glVertex2i(effValue+2, mEffectiveCoords[1]);
			glVertex2i(effValue+2, mEffectiveCoords[5]);
			glVertex2i(effValue-2, mEffectiveCoords[5]);
			glEnd();
			break;
		case FILL_TO:
			glBegin(GL_QUADS);
			glVertex2i(mEffectiveCoords[0], mEffectiveCoords[1]);
			glVertex2i(effValue, mEffectiveCoords[1]);
			glVertex2i(effValue, mEffectiveCoords[5]);
			glVertex2i(mEffectiveCoords[0], mEffectiveCoords[5]);
			glEnd();
			break;
		case FILL_FROM:
			if(mFillType) {
				glBegin(GL_QUADS);
				glVertex2i(effValue, mEffectiveCoords[1]);
				glVertex2i(mEffectiveCoords[2], mEffectiveCoords[3]);
				glVertex2i(mEffectiveCoords[4], mEffectiveCoords[5]);
				glVertex2i(effValue, mEffectiveCoords[5]);
				glEnd();
			}
			break;
		}
	}
}

bool Slider::_handleSeek(const int x) {

	// Translate to relative position
	double pos = (x-mEffectiveCoords[0]) / (double) mEffectiveSizes.width;

	setValue((1.0-pos)*mMinimum + pos*mMaximum);

	return true;
}

Configuration::Entry* EntrySerialization<Slider>::serialize(Slider *panel, Configuration::Entry *serialData) {
	Except("EntrySerialization<Slider *>::serialize", "Implement"); 
	return serialData;
}

Slider* EntrySerialization<Slider>::deserialize(Configuration::Entry *serialData, Slider *slider) {
	Assert(serialData->name == "slider", "EntrySerialization<Slider>::deserialize");
	
	// Create panel object if does not exist
	if(!slider) slider = new Slider();
	
	// Deserialize as a widget
	slider = dynamic_cast<Slider *>(EntrySerialization<Widget>::deserialize(serialData, dynamic_cast<Widget *>(slider)));
	
	// Attribute: min
	Configuration::Entry *minEntry = serialData->getFirstChild("min");
	if(minEntry)
		slider->setMinimum(StringUtils::parse<double>(minEntry->getStringValue()));

	// Attribute: max
	Configuration::Entry *maxEntry = serialData->getFirstChild("max");
	if(maxEntry)
		slider->setMaximum(StringUtils::parse<double>(maxEntry->getStringValue()));

	// Attribute: value
	Configuration::Entry *valueEntry = serialData->getFirstChild("value");
	if(valueEntry)
		slider->setValue(StringUtils::parse<double>(valueEntry->getStringValue()));

	// Attribute: readonly
	Configuration::Entry *readOnlyEntry = serialData->getFirstChild("readonly");
	if(readOnlyEntry)
		slider->setReadOnly(StringUtils::parse<bool>(readOnlyEntry->getStringValue()));
	
	// Attribute: invertfill
	Configuration::Entry *fillTypeEntry = serialData->getFirstChild("type");
	if(fillTypeEntry) {
		std::string strFillType = fillTypeEntry->getStringValue();
		if(strFillType == "bar")
			slider->setFillType(Slider::BAR);
		else if(strFillType == "fillto")
			slider->setFillType(Slider::FILL_TO);
		else if(strFillType == "fillfrom")
			slider->setFillType(Slider::FILL_FROM);
	}
	
	return slider;
}

} // namespace MinGLE
