//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Imaging.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Imaging.cpp
/// Raster imaging class

// Header that contains the declaration
#include <Imaging.h>

#include <Exception.h>
#include <MathUtils.h>
#include <GLUtils.h>
#include <System.h>
#include <Threading.h>

// System headers
#include <math.h>

// Add support to misc formats
#include <misc-formats/include/pfs.h>

namespace MinGLE {

MathUtils::ScaleMode Image::mScaleMode = MathUtils::DEFAULT_SCALE_MODE;

Image::Image() : 
	mWidth(UNKOWN_SIZE), mHeight(UNKOWN_SIZE), mBytesPerPixel(0),
	mTotalSize(0), mFormat(DEFAULT_FORMAT), mType(DEFAULT_TYPE),
	mData(0)
{
	
}

Image::~Image() {
	// Verify that the memory has been freed
	Assert(!mData, "Image::~Image");
}

void Image::grabFromScreen() {
	Except("Image::clone", "Method is not implemented.");
}

void Image::grabFromTexture2D(GLint level) {
	// Get dimensions	
	GLint width = 0, height = 0;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_WIDTH, &width);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_HEIGHT, &height);

	// Get format
	GLint format = DEFAULT_FORMAT;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, level, GL_TEXTURE_INTERNAL_FORMAT, &format);

	// Get type (forced to convert to standard OpenGL type)
	GLenum type = GLUtils::getTextureType(GL_TEXTURE_2D, level, true);

	// Get texture	
	void *data = _allocateMemory(width, height, format, type);
	glGetTexImage(GL_TEXTURE_2D, level, format, type, data);

	// Create image
	createFromMemory(width, height, format, type, data);
	free(data);
}

Image *Image::clone() {
	return Imaging::createImage(this);
}

void Image::mirror() {
	Except("Image::mirror", "Method is not implemented.");
}

void Image::flip() {
	Except("Image::flip", "Method is not implemented.");
}

void Image::clear(float r, float g, float b, float a) {
	
}

void *Image::_allocateMemory(unsigned width, unsigned height, GLenum format,
		GLenum type) {
	// TODO: use common temporary memory area to avoid continous allocation and free
	return calloc(width*height, GLUtils::calculateBytesPerPixel(format, type));
}

// =============================================================================

template<> Manager<Imaging>::ImplementationMapPtr Manager<Imaging>::mImplementations = new Manager<Imaging>::ImplementationMap();

// Singleton instance
template<> Imaging *Manager<Imaging>::msInstance = 0;
	
Imaging::~Imaging() {
	// Nothing to do
}
	
Image *Imaging::createImage() {
	Assert(msInstance, "Imaging::createImage");
	return msInstance->_createImage();	
}

Image *Imaging::createImage(unsigned width, unsigned height, GLenum format, GLenum type, void *data) {
	Assert(msInstance, "Imaging::createImage");
	return msInstance->_createImage(width, height, format, type, data);		
}

Image *Imaging::createImage(const std::string& filename, VirtualFileSystem *vfs) {
	Assert(msInstance, "Imaging::createImage");
	return msInstance->_createImage(filename, vfs);	
}

Image *Imaging::createImage(const std::string& blockName, const std::string& filename) {
	return createImage(filename, System::getVirtualFileSystemForFile(blockName, filename));
}

Image *Imaging::createImage(const Image * image) {
	Assert(msInstance, "Imaging::createImage");
	return msInstance->_createImage(image);	
}
	
} // namespace MinGLE
