//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/MathUtils.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file MathUtils.cpp
/// Mathematical utils

// Header that contains the declaration
#include <MathUtils.h>

// MinGLE headers
#include <Exception.h>

// System headers
#include <math.h>

namespace MinGLE {

double MathUtils::msEpsilon = MathUtils::DEFAULT_EPSILON;
double MathUtils::msPi = MathUtils::DEFAULT_PI;

double MathUtils::vectorLength(const double x, const double y) {
	return:: sqrt(x*x+y*y);
}

double MathUtils::vectorLength(const double x, const double y, const double z) {
	return ::sqrt(x*x+y*y+z*z);
}

double MathUtils::vectorLength(const double x, const double y, const double z, const double w) {
	Assert(w, "MathUtils::vectorLength");
	return ::sqrt(x*x+y*y+z*z)/w;
}

double MathUtils::dotProduct(const double x1, const double y1, const double x2, const double y2) {
	return x1*x2 + y1*y2;
}

double MathUtils::dotProduct(const double x1, const double y1, const double z1,
		const double x2, const double y2, const double z2) {
	return x1*x2 + y1*y2 + z1*z2;
}

void MathUtils::crossProduct(const double x1, const double y1, const double z1,
		const double x2, const double y2, const double z2, double &x,
		double &y, double &z) {
	x = y1*z2 - y2*z1;
	y = x2*z1 - x1*z2;
	z = x1*y2 - x2*y1;
}

void MathUtils::vectorNormalize(double &x, double &y) {
	const double invLength = 1.0/vectorLength(x, y);
	x *= invLength;
	y *= invLength;
}

void MathUtils::vectorNormalize(double &x, double &y, double &z) {
	const double invLength = 1.0/vectorLength(x, y, z);
	x *= invLength;
	y *= invLength;
	z *= invLength;
}

void MathUtils::vectorNormalize(double &x, double &y, double &z, double &w) {
	const double invLength = w/vectorLength(x, y, z);
	x *= invLength;
	y *= invLength;
	z *= invLength;
	w = 1.0;
}

void MathUtils::extractMatrix3x3(double *m4x4, double *m3x3) {
	m3x3[3*0+0] = m4x4[4*0+0]; m3x3[3*0+1] = m4x4[4*0+1]; m3x3[3*0+2] = m4x4[4*0+2];
	m3x3[3*1+0] = m4x4[4*1+0]; m3x3[3*1+1] = m4x4[4*1+1]; m3x3[3*1+2] = m4x4[4*1+2];
	m3x3[3*2+0] = m4x4[4*2+0]; m3x3[3*2+1] = m4x4[4*2+1]; m3x3[3*2+2] = m4x4[4*2+2];
}

void MathUtils::matrix3x3VectorProduct(const double *m, const double *d, double *p) {
	p[0] = m[3*0+0]*d[0] + m[3*0+1]*d[1] + m[3*0+2]*d[2];
	p[1] = m[3*1+0]*d[0] + m[3*1+1]*d[1] + m[3*1+2]*d[2];
	p[2] = m[3*2+0]*d[0] + m[3*2+1]*d[1] + m[3*2+2]*d[2];
}

void MathUtils::matrix4x4VectorProduct(const double *m, const double *d, double *p) {
	p[0] = m[4*0+0]*d[0] + m[4*0+1]*d[1] + m[4*0+2]*d[2] + m[4*0+3]*d[3];
	p[1] = m[4*1+0]*d[0] + m[4*1+1]*d[1] + m[4*1+2]*d[2] + m[4*1+3]*d[3];
	p[2] = m[4*2+0]*d[0] + m[4*2+1]*d[1] + m[4*2+2]*d[2] + m[4*2+3]*d[3];
	p[3] = m[4*3+0]*d[0] + m[4*3+1]*d[1] + m[4*3+2]*d[2] + m[4*3+3]*d[3];
}

void MathUtils::matrix4x4Product(const double *m1, const double *m2, double *result) {
	for(int x=0; x<4; x++)
	for(int y=0; y<4; y++) {
		result[4*y+x] = 0.0;
		for(int i=0; i<4; i++)
		result[4*y+x]+=m1[4*y+i]*m2[4*i+x];
	}
}

void MathUtils::transposeMatrix4x4(const double *m, double *mT) {
	mT[0] = m[0]; mT[1] = m[4]; mT[2] = m[8]; mT[3] = m[12];
	mT[4] = m[1]; mT[5] = m[5]; mT[6] = m[9]; mT[7] = m[13];
	mT[8] = m[2]; mT[9] = m[6]; mT[10] = m[10]; mT[11] = m[14];
	mT[12] = m[3]; mT[13] = m[7]; mT[14] = m[11]; mT[15] = m[15];
}

// From Mesa-2.2\src\glu\project.c
// Compute the inverse of a 4x4 matrix.  Contributed by scotter@lafn.org
void MathUtils::invertMatrix4x4(double *m, double *invM) {

	// NB. OpenGL Matrices are COLUMN major.
#define MAT(m,r,c) (m)[(c)*4+(r)]

	// Here's some shorthand converting standard (row,column) to index.
#define m11 MAT(m,0,0)
#define m12 MAT(m,0,1)
#define m13 MAT(m,0,2)
#define m14 MAT(m,0,3)
#define m21 MAT(m,1,0)
#define m22 MAT(m,1,1)
#define m23 MAT(m,1,2)
#define m24 MAT(m,1,3)
#define m31 MAT(m,2,0)
#define m32 MAT(m,2,1)
#define m33 MAT(m,2,2)
#define m34 MAT(m,2,3)
#define m41 MAT(m,3,0)
#define m42 MAT(m,3,1)
#define m43 MAT(m,3,2)
#define m44 MAT(m,3,3)

	// pre-compute 2x2 dets for last two rows when computing
	// cofactors of first two rows.
	double d12, d13, d23, d24, d34, d41;
	d12 = (m31*m42-m41*m32);
	d13 = (m31*m43-m41*m33);
	d23 = (m32*m43-m42*m33);
	d24 = (m32*m44-m42*m34);
	d34 = (m33*m44-m43*m34);
	d41 = (m34*m41-m44*m31);

	double tmp[16];
	tmp[0] = (m22 * d34 - m23 * d24 + m24 * d23);
	tmp[1] = -(m21 * d34 + m23 * d41 + m24 * d13);
	tmp[2] = (m21 * d24 + m22 * d41 + m24 * d12);
	tmp[3] = -(m21 * d23 - m22 * d13 + m23 * d12);

	// Compute determinant as early as possible using these cofactors.
	double det = m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2] + m14 * tmp[3];

	// Run singularity test.
	Assert(::fabs(det) > msEpsilon, "MathUtils::invertMatrix4x4");

	double invDet = 1.0 / det;
	// Compute rest of inverse.
	tmp[0] *= invDet;
	tmp[1] *= invDet;
	tmp[2] *= invDet;
	tmp[3] *= invDet;

	tmp[4] = -(m12 * d34 - m13 * d24 + m14 * d23) * invDet;
	tmp[5] = (m11 * d34 + m13 * d41 + m14 * d13) * invDet;
	tmp[6] = -(m11 * d24 + m12 * d41 + m14 * d12) * invDet;
	tmp[7] = (m11 * d23 - m12 * d13 + m13 * d12) * invDet;

	// Pre-compute 2x2 dets for first two rows when computing
	// cofactors of last two rows.
	d12 = m11*m22-m21*m12;
	d13 = m11*m23-m21*m13;
	d23 = m12*m23-m22*m13;
	d24 = m12*m24-m22*m14;
	d34 = m13*m24-m23*m14;
	d41 = m14*m21-m24*m11;

	tmp[8] = (m42 * d34 - m43 * d24 + m44 * d23) * invDet;
	tmp[9] = -(m41 * d34 + m43 * d41 + m44 * d13) * invDet;
	tmp[10] = (m41 * d24 + m42 * d41 + m44 * d12) * invDet;
	tmp[11] = -(m41 * d23 - m42 * d13 + m43 * d12) * invDet;
	tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23) * invDet;
	tmp[13] = (m31 * d34 + m33 * d41 + m34 * d13) * invDet;
	tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12) * invDet;
	tmp[15] = (m31 * d23 - m32 * d13 + m33 * d12) * invDet;

	memcpy(invM, tmp, 16 * sizeof(double));

#undef m11
#undef m12
#undef m13
#undef m14
#undef m21
#undef m22
#undef m23
#undef m24
#undef m31
#undef m32
#undef m33
#undef m34
#undef m41
#undef m42
#undef m43
#undef m44
#undef MAT
}

bool MathUtils::intersectsRayPlane(const double *rayOrigin, const double *rayUnitDirection,
		const double *planeNormal, const double planeD, double *t) {

	double denom = dotProduct(
			planeNormal[0], planeNormal[1], planeNormal[2],
			rayUnitDirection[0], rayUnitDirection[1], rayUnitDirection[2]
	);

	if(::fabs(denom) < msEpsilon)
	return false;

	double nom = dotProduct(
			planeNormal[0], planeNormal[1], planeNormal[2],
			rayOrigin[0], rayOrigin[1], rayOrigin[2]
	) + planeD;

	// Intersection point in ray parameter	
	*t = -(nom/denom);

	return *t >= 0.0;
}

void MathUtils::rodriguesMatrix3x3(double *m, const double *dir, const double theta) {

	const double cosTheta = ::cos(theta);
	const double oneMinusCosTheta = 1.0 - cosTheta;
	const double sinTheta = ::sin(theta);
	
	m[0] = cosTheta + dir[0]*dir[0]*oneMinusCosTheta; 
	m[1] = dir[0]*dir[1]*oneMinusCosTheta - dir[2]*sinTheta; 
	m[2] = dir[1]*sinTheta + dir[0]*dir[2]*oneMinusCosTheta;

	m[3] = dir[2]*sinTheta + dir[0]*dir[1]*oneMinusCosTheta; 
	m[4] = cosTheta + dir[1]*dir[1]*oneMinusCosTheta; 
	m[5] = -dir[0]*sinTheta + dir[1]*dir[2]*oneMinusCosTheta; 

	m[6] = -dir[1]*sinTheta + dir[0]*dir[2]*oneMinusCosTheta; 
	m[7] = dir[0]*sinTheta + dir[1]*dir[2]*oneMinusCosTheta; 
	m[8] = cosTheta + dir[2]*dir[2]*oneMinusCosTheta;	
}

bool MathUtils::isPowerOf2(unsigned n) {
	// Zero case
	if(!n) return false;

	while(!(n%2)) n >>= 1;
	return (n == 1) ? true : false;
}

void MathUtils::getPowerOf2Bounds(unsigned n, unsigned &l, unsigned &u) {

	// Zero case
	if(!n) {
		l = u = 0;
		return;
	}

	// Get exponent
	int e; unsigned _n = n; for(e = 0; _n; _n>>=1, e++);

	// Power of 2 case
	if(n==1) {
		l = u = n;
	}

	// General case
	else {
		l = 1<<(e - 1);
		u = l<<1;
	}
}

unsigned MathUtils::convertToPowerOf2(unsigned value, ScaleMode scaleMode) {

	if(isPowerOf2(value)) return value;

	unsigned lowerBound = 0, upperBound = 0;
	getPowerOf2Bounds(value, lowerBound, upperBound);

	switch(scaleMode) {
		case DOWNSCALE:
		return lowerBound;
		break;

		case UPSCALE:
		return upperBound;
		break;

		case NEARESTSCALE:
		return value - lowerBound < upperBound - value ? lowerBound : upperBound;
		break;
	}

	return 0;
}

void MathUtils::hsi2rgb(double h, double s, double i, double &r, double &g, double &b) {
	if(h < 1.0/3.0) {
		b = (1-s)/3;
		r = (1+s*::cos(2*getPi()*h)/::cos(getPi()/3-2*getPi()*h))/3.0;
		g = 1 - (b + r);
	} else if(h < 2.0/3.0) {
		h = h - 1.0/3.0;
		r = (1-s)/3;
		g = (1+s*::cos(2*getPi()*h)/::cos(getPi()/3-2*getPi()*h))/3.0;
		b = 1 - (r+g);
	} else {
		h = h - 2.0/3.0;
		g = (1-s)/3;
		b = (1+s*::cos(2*getPi()*h)/::cos(getPi()/3-2*getPi()*h))/3.0;
		r = 1 - (g+b);
	}

	r *= i, g *= i, b *= i;
}

} // namespace MinGLE
