//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Navigators.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Navigators.cpp
/// Scene navigator classes

// Header that contains the declaration
#include <Navigators.h>

// MinGLE headers
#include <Exception.h>
#include <Cameras.h>
#include <Window.h>
#include <MathUtils.h>

// System headers
#include <math.h>
#include <mingle-gl.h>

namespace MinGLE {

ApplicationKeyHandler::ApplicationKeyHandler() {
}

ApplicationKeyHandler::~ApplicationKeyHandler() {
}

bool ApplicationKeyHandler::onKeyEvent(int key, unsigned modifiers,
		KeyEventState state) {
	if (state != KEY_UP) {
		switch (key) {
		// Quit
		case KEY_ESCAPE:
			if (!modifiers) {
				mWindow->close();
				return true;
			}
			break;

			// Toggle full screen
		case KEY_ENTER:
			if (modifiers & KEY_ALT) {
				mWindow->setFullScreen(!mWindow->isFullScreen());
				return true;
			}
			break;

			// Create a screen shot
		case 16: // 'p'
			if (modifiers & KEY_CTRL) {
				mWindow->createScreenShot();
				return true;
			}
			break;

			// Toggle video recording
		case 18: // 'r'
			if (modifiers & KEY_CTRL) {
				if (mWindow->isRecording())
					mWindow->pauseVideo();
				else
					mWindow->startVideo();
				return true;
			}
			break;
		}
	}

	return false;
}

// =============================================================================

Animator::Animator(Window *window, const double timeStep) {
	// Register animator as window listener
	window->registerWindowListener(dynamic_cast<WindowListener *>(this));

	// Create time for animation
	mTimer = window->createTimer(timeStep);

	// Register animator as timer listener
	mTimer->registerTimerListener(dynamic_cast<TimerListener *>(this));
}

Animator::~Animator() {
	mTimer->removeTimerListener();
	delete mTimer;
}

void Animator::onTick() {
	glMatrixMode(GL_MODELVIEW);
	if (onAnimation())
		redisplayRequest();
}

// =============================================================================

/*SimpleAnimator::SimpleAnimator() {
 }

 SimpleAnimator::~SimpleAnimator() {
 }

 bool SimpleAnimator::onAnimation() {
 glRotatef(30.0 * mTimer->getElapsedTime(), 0.0, 1.0, 0.0);
 return true;
 }*/

// =============================================================================

Navigator::Navigator() :
	mCamera(new SymmetricFrustumCamera()) {
}

Navigator::Navigator(Camera *camera) :
	mCamera(camera) {
	Assert(mCamera, "Navigator::Navigator");
}

Navigator::~Navigator() {
	if (mCamera)
		delete mCamera;
}

bool Navigator::onResize(unsigned width, unsigned height) {
	mCamera->setImageSize(width, height);
	redisplayRequest();

	// Pass the event
	return false;
}

// =============================================================================

ExaminerNavigator::ExaminerNavigator(bool zoomEnabled, bool moveEnabled,
		bool wheelEnabled, double sensitivity) :
	mZoomEnabled(zoomEnabled), mMoveEnabled(moveEnabled),
			mWheelEnabled(wheelEnabled), mSensitivity(sensitivity),
			mTrackBallButton(DEFAULT_TRACKBALL_BUTTON),
			mZoomButton(DEFAULT_ZOOM_BUTTON), mMoveButton(DEFAULT_MOVE_BUTTON)
			//mDragX(NO_DRAG_COORDINATE), mDragY(NO_DRAG_COORDINATE), mDragZ(NO_DRAG_COORDINATE), mZoom(DEFAULT_ZOOM) 
{
}

bool ExaminerNavigator::onResize(unsigned width, unsigned height) {
	if (Navigator::onResize(width, height))
		return true;

	/*
	// TODO: this not should be here
	// Full class rewrite is needed
	glMatrixMode(GL_MODELVIEW);
	glScaled(mZoom, mZoom, mZoom);
	*/

	redisplayRequest();

	// Pass the control to the next event handler
	return false;
}

ExaminerNavigator::~ExaminerNavigator() {
}

/*
void ExaminerNavigator::setZoom(const double value) {
	glMatrixMode(GL_MODELVIEW);
	glScaled(mZoom = value, value, value);

	redisplayRequest();
}
*/

bool ExaminerNavigator::onMouseButtonEvent(int x, int y, MouseButton button,
		MouseEventState state) {
	
	/*
	if (state == BUTTON_DOWN) {
		if (button == mTrackBallButton || mZoomEnabled && button == mZoomButton
				|| mMoveEnabled && button == mMoveButton) {
			mWindow->unProject(x, y, &mDragX, &mDragY, &mDragZ);
			return true;
		}
	} else if (state == BUTTON_UP) {
		if (button == mTrackBallButton || mZoomEnabled && button == mZoomButton
				|| mMoveEnabled && button == mMoveButton) {
			mDragX = mDragY = mDragZ = NO_DRAG_COORDINATE;
			return true;
		}
	}
	*/

	// Pass the control to the next event handler
	return false;
}

bool ExaminerNavigator::onMouseMoveEvent(int x, int y, MouseMoveState state) {

	const double *position = mCamera->getPosition();
	const double *target = mCamera->getTarget();

	// Inverse view vector (from the target to the position)
	const double invView[] = {
		position[0] - target[0],	
		position[1] - target[1],	
		position[2] - target[2]	
	};
	
	// Up vector
	const double *up = mCamera->getUpVector();
	
	// Jump out on not hoovering
	if (state != HOVER)
		return false;

	// Trackball motion
	else if (mMouseButtonStatus[mTrackBallButton]) {
		
		// Matrix that rotates around the up vector
		double rodriguesUp[9];
		const double angleUp = mSensitivity * (mMouseLastX - x) * MathUtils::getPi() / mCamera->getImageWidth(); 
		MathUtils::rodriguesMatrix3x3(rodriguesUp, up, angleUp);
		
		// Matrix that rotates around the left vector
		double left[3];
		MathUtils::crossProduct(invView[0], invView[1], invView[2], up[0], up[1], up[2], left[0], left[1], left[2]);
		MathUtils::vectorNormalize(left[0], left[1], left[2]);
		double rodriguesLeft[9];
		const double angleLeft = mSensitivity * (y - mMouseLastY) * MathUtils::getPi() / mCamera->getImageHeight(); 
		MathUtils::rodriguesMatrix3x3(rodriguesLeft, left, angleLeft);

		// Rotate the inverse view vector around the up vector
		double rotatedInvViewUp[3];
		MathUtils::matrix3x3VectorProduct(rodriguesUp, invView, rotatedInvViewUp);
		double rotatedInvViewUpLeft[3];
		MathUtils::matrix3x3VectorProduct(rodriguesLeft, rotatedInvViewUp, rotatedInvViewUpLeft);
		
		// New position
		double newPosition[] = {
			target[0] + rotatedInvViewUpLeft[0],
			target[1] + rotatedInvViewUpLeft[1],
			target[2] + rotatedInvViewUpLeft[2]
		};
		mCamera->setPosition(newPosition[0], newPosition[1], newPosition[2]);
		
		// New up vector
		double newUp[3];
		MathUtils::matrix3x3VectorProduct(rodriguesLeft, up, newUp);
		mCamera->setUpVector(newUp[0], newUp[1], newUp[2]);

		redisplayRequest();
		return true;
	}

	// Zooming
	else if (mZoomEnabled && mMouseButtonStatus[mZoomButton]) {
		// Calculate exponential zoom factor
		const double zoomFactor = mSensitivity * ::exp((mMouseLastY - y) * 0.01);
		
		// New position
		double newPosition[] = {
			target[0] + zoomFactor*invView[0],
			target[1] + zoomFactor*invView[1],
			target[2] + zoomFactor*invView[2]
		};
		mCamera->setPosition(newPosition[0], newPosition[1], newPosition[2]);
		
		redisplayRequest();
		return true;
	}

	// Moving
	else if(mMoveEnabled && mMouseButtonStatus[mMoveButton]) {
		/*
		double _matrix[16];
					glGetDoublev(GL_MODELVIEW_MATRIX, _matrix);

					double px,py,pz;
					mWindow->unProject(x, y, &px, &py, &pz);

					glMatrixMode(GL_MODELVIEW);
					glLoadIdentity();
					glTranslatef(px-mDragX, py-mDragY, pz-mDragZ);
					glMultMatrixd(_matrix);

					mDragX = px;
					mDragY = py;
					mDragZ = pz;
		*/	
					
	}

	// Pass the control to the next event handler
	else
		return false;
	
	// Redisplay and cut event chain
	redisplayRequest();
	return true;
}

bool ExaminerNavigator::onMouseWheelEvent(int x, int y, int delta) {

	const double *position = mCamera->getPosition();
	const double *target = mCamera->getTarget();

	// Inverse view vector (from the target to the position)
	const double invView[] = {
		position[0] - target[0],	
		position[1] - target[1],	
		position[2] - target[2]	
	};

	// Linear zoom
	const double zoomFactor = mSensitivity * (delta > 0 ? WHEEL_ZOOM_OUT_FACTOR : WHEEL_ZOOM_IN_FACTOR);

	// New position
	double newPosition[] = {
		target[0] + zoomFactor*invView[0],
		target[1] + zoomFactor*invView[1],
		target[2] + zoomFactor*invView[2]
	};
	mCamera->setPosition(newPosition[0], newPosition[1], newPosition[2]);
	
	// Do not pass the control to the next event handler
	redisplayRequest();
	return true;
}

// =============================================================================

const double FlyNavigator::DEFAULT_ZERO_DIRECTION[3] = {0.0, 0.0, -1.0};
const double FlyNavigator::DEFAULT_ZERO_UP_VECTOR[3] = {0.0, 1.0, 0.0};
const double FlyNavigator::DEFAULT_ZERO_RIGHT_VECTOR[3] = {1.0, 0.0, 0.0};
const FlyNavigator::MoveMode FlyNavigator::DEFAULT_LINEAR_MOVE_MODE = FlyNavigator::ACCELERATION_CONTROLLED;
const FlyNavigator::MoveMode FlyNavigator::DEFAULT_ANGULAR_MOVE_MODE = FlyNavigator::POSITION_CONTROLLED;
const double FlyNavigator::DEFAULT_LINEAR_CHANGE_RATE = 1.0;
const double FlyNavigator::DEFAULT_ANGULAR_CHANGE_RATE = 1.0;
const double FlyNavigator::DEFAULT_LINEAR_DECELERATION = 0.5;
const double FlyNavigator::DEFAULT_ANGULAR_DECELERATION = 0.5;
const double FlyNavigator::DEFAULT_MOVE_ON_KEY = 1.0;
const bool FlyNavigator::DEFAULT_LINEAR_HANDBREAK = true;
const bool FlyNavigator::DEFAULT_ANGULAR_HANDBREAK = true;
const bool FlyNavigator::DEFAULT_LINEAR_STEERING = false;
const int FlyNavigator::INVALID_MOUSE_DOWN_COORD = -1;

const FlyNavigator::MoveControl FlyNavigator::DEFAULT_FORWARD_CONTROL = {
		NOT_CONTROLLED,
		InputListener::LEFT_BUTTON,
		true,
		InputListener::KEY_UP_ARROW,
		0x0,
		DEFAULT_MOVE_ON_KEY
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_BACKWARD_CONTROL = {
		NOT_CONTROLLED,
		InputListener::LEFT_BUTTON,
		true,
		InputListener::KEY_DOWN_ARROW,
		0x0,
		-DEFAULT_MOVE_ON_KEY
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_LEFT_CONTROL = {
		NOT_CONTROLLED,
		InputListener::LEFT_BUTTON,
		true,
		InputListener::KEY_LEFT_ARROW,
		0x0,
		-DEFAULT_MOVE_ON_KEY
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_RIGHT_CONTROL = {
		NOT_CONTROLLED,
		InputListener::LEFT_BUTTON,
		true,
		InputListener::KEY_RIGHT_ARROW,
		0x0,
		DEFAULT_MOVE_ON_KEY
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_ROLL_CONTROL = {
		MOUSE_BUTTON_CONTROLLED,
		InputListener::RIGHT_BUTTON,
		false,
		0x0,
		0x0
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_PITCH_CONTROL = {
		MOUSE_BUTTON_CONTROLLED,
		InputListener::LEFT_BUTTON,
		false,
		0x0,
		0x0
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_YAW_CONTROL = {
		MOUSE_BUTTON_CONTROLLED,
		InputListener::LEFT_BUTTON,
		false,
		0x0,
		0x0
};
const FlyNavigator::MoveControl FlyNavigator::DEFAULT_HANDBREAK_CONTROL = {
		NOT_CONTROLLED,
		InputListener::LEFT_BUTTON,
		true,
		InputListener::KEY_SPACE,
		0x0
};

#define INITIALIZE_FLY_NAVIGATOR 	mPositionX(0.0), mPositionY(0.0), mPositionZ(5.0), \
	mYaw(0.0), mPitch(0.0), mRoll(0.0), \
	mVelocityX(0.0), mVelocityY(0.0), mVelocityZ(0.0), \
	mAngularVelocityYaw(0.0), mAngularVelocityPitch(0.0), mAngularVelocityRoll(0.0), \
	mAccelerationX(0.0), mAccelerationY(0.0), mAccelerationZ(0.0), \
	mAngularAccelerationYaw(0.0), mAngularAccelerationPitch(0.0), mAngularAccelerationRoll(0.0), \
	mLinearMoveMode(DEFAULT_LINEAR_MOVE_MODE), mAngularMoveMode(DEFAULT_ANGULAR_MOVE_MODE), \
	mLinearChangeRate(DEFAULT_LINEAR_CHANGE_RATE), mAngularChangeRate(DEFAULT_ANGULAR_CHANGE_RATE), \
	mLinearDeceleration(DEFAULT_LINEAR_DECELERATION), mAngularDeceleration(DEFAULT_ANGULAR_DECELERATION), \
	mLinearHandBreak(DEFAULT_LINEAR_HANDBREAK), mAngularHandBreak(DEFAULT_ANGULAR_HANDBREAK), \
	mLinearSteering(DEFAULT_LINEAR_STEERING), \
	mLastMoveTime(Timer::getSystemTime()), \
	mForwardControl(DEFAULT_FORWARD_CONTROL), mBackwardControl(DEFAULT_BACKWARD_CONTROL), \
	mLeftControl(DEFAULT_LEFT_CONTROL), mRightControl(DEFAULT_RIGHT_CONTROL), \
	mRollControl(DEFAULT_ROLL_CONTROL), mPitchControl(DEFAULT_PITCH_CONTROL), mYawControl(DEFAULT_YAW_CONTROL), \
	mHandBreakControl(DEFAULT_HANDBREAK_CONTROL), \
	mMoveForwardBackward(0.0), mMoveLeftRight(0.0), \
	mRotateYaw(0.0), mRotatePitch(0.0), mRotateRoll(0.0), \
	mMouseDownX(INVALID_MOUSE_DOWN_COORD), mMouseDownY(INVALID_MOUSE_DOWN_COORD), \
	mMouseDownYaw(0.0), mMouseDownPitch(0.0), mMouseDownRoll(0.0), \
	mRedisplayAnimator(0)

FlyNavigator::FlyNavigator() :
	Navigator(), 
	INITIALIZE_FLY_NAVIGATOR
{
	// The automatric recalculation functionality of the camera is disabled
	mCamera->setAutoRecalculate(false);
}

FlyNavigator::FlyNavigator(Camera *camera) : 
	Navigator(camera), 
	INITIALIZE_FLY_NAVIGATOR
{
	
}

#undef INITIALIZE_FLY_NAVIGATOR

void FlyNavigator::onRegistered() {
	mRedisplayAnimator = new RedisplayAnimator(mWindow, 0.04, this);
}

void FlyNavigator::onUnRegistered() {
}

bool FlyNavigator::onResize(unsigned width, unsigned height) {
	// Set camera image size
	mCamera->setImageSize(1.0, (double) height / width);
	
	// Refresh the camera
	_refreshCamera();
	
	// Pass the control to the next event handler
	return false;
}

bool FlyNavigator::onKeyEvent(int key, unsigned modifiers, KeyEventState state) {
	// Forward key
	if(mForwardControl.keyboardControlled && mForwardControl.key == key && mForwardControl.modifiers == modifiers)
		switch(state) {
		case InputListener::KEY_DOWN: 
			// Handbreak
			if(mLinearHandBreak) 
				_linearHandBreak(mForwardControl.moveOnKey * mDirection[0], 
						mForwardControl.moveOnKey * mDirection[1], 
						mForwardControl.moveOnKey * mDirection[2]);

			// Set amount of movement
			mMoveForwardBackward = mForwardControl.moveOnKey; 
			return true;
		case InputListener::KEY_UP: 
			mMoveForwardBackward = 0.0; 
			return true;
		case InputListener::KEY_PRESS: 
			// Handbreak only
			if(mLinearHandBreak) 
				_linearHandBreak(mForwardControl.moveOnKey * mDirection[0], 
						mForwardControl.moveOnKey * mDirection[1], 
						mForwardControl.moveOnKey * mDirection[2]);
		break;
		}

	// Backward key
	else if(mBackwardControl.keyboardControlled && mBackwardControl.key == key && mBackwardControl.modifiers == modifiers)
		switch(state) {
		case InputListener::KEY_DOWN: 
			// Handbreak
			if(mLinearHandBreak) 
				_linearHandBreak(mBackwardControl.moveOnKey * mDirection[0], 
						mBackwardControl.moveOnKey * mDirection[1], 
						mBackwardControl.moveOnKey * mDirection[2]);

			// Set amount of movement
			mMoveForwardBackward = mBackwardControl.moveOnKey; 
			return true;
		case InputListener::KEY_UP: mMoveForwardBackward = 0.0; return true;
		case InputListener::KEY_PRESS: 
			// Handbreak only
			if(mLinearHandBreak) 
				_linearHandBreak(mBackwardControl.moveOnKey * mDirection[0], 
						mBackwardControl.moveOnKey * mDirection[1], 
						mBackwardControl.moveOnKey * mDirection[2]);
			break;
		}

	// Left key
	else if(mLeftControl.keyboardControlled && mLeftControl.key == key && mLeftControl.modifiers == modifiers)
		switch(state) {
		case InputListener::KEY_DOWN: 
			// Handbreak
			if(mLinearHandBreak)
				_linearHandBreak(mLeftControl.moveOnKey * mRightVector[0], 
						mLeftControl.moveOnKey * mRightVector[1], 
						mLeftControl.moveOnKey * mRightVector[2]);

			// Set amount of movement
			mMoveLeftRight = mLeftControl.moveOnKey; 
			return true;
		case InputListener::KEY_UP: mMoveLeftRight = 0.0; return true;
		case InputListener::KEY_PRESS: 
			// Handbreak only
			if(mLinearHandBreak) 
				_linearHandBreak(mLeftControl.moveOnKey * mRightVector[0], 
						mLeftControl.moveOnKey * mRightVector[1], 
						mLeftControl.moveOnKey * mRightVector[2]);
			break;
		}

	// Right key
	else if(mRightControl.keyboardControlled && mRightControl.key == key && mRightControl.modifiers == modifiers)
		switch(state) {
		case InputListener::KEY_DOWN: 
			// Handbreak
			if(mLinearHandBreak)
				_linearHandBreak(mRightControl.moveOnKey * mRightVector[0], 
						mRightControl.moveOnKey * mRightVector[1], 
						mRightControl.moveOnKey * mRightVector[2]);

			// Set amount of movement
			mMoveLeftRight = mRightControl.moveOnKey; 
			return true;
		case InputListener::KEY_UP: mMoveLeftRight = 0.0; return true;
		case InputListener::KEY_PRESS: 
			// Handbreak only
			if(mLinearHandBreak) 
				_linearHandBreak(mRightControl.moveOnKey * mRightVector[0], 
						mRightControl.moveOnKey * mRightVector[1], 
						mRightControl.moveOnKey * mRightVector[2]);
			break;
		}

	// Handbreak
	else if(mHandBreakControl.keyboardControlled && mHandBreakControl.key == key && mHandBreakControl.modifiers == modifiers
	    && state != InputListener::KEY_UP) {
	    if(mLinearHandBreak) _linearHandBreak();
	}
	
	// TODO roll pitch yaw with keyboard
	
	// Pass the control to the next event handler
	return false;
}

bool FlyNavigator::onMouseButtonEvent(int x, int y, MouseButton button, MouseEventState state) {

	switch(state) {
	case BUTTON_DOWN:
		mMouseDownX = x, mMouseDownY = y;
		mMouseDownRoll = mRoll, mMouseDownPitch = mPitch, mMouseDownYaw = mYaw; 
		break;
	case BUTTON_UP:
		mMouseDownX = INVALID_MOUSE_DOWN_COORD, mMouseDownY = INVALID_MOUSE_DOWN_COORD;
	default:
		return false;
	}
	
	// Stop event propagation
	return true;
}

bool FlyNavigator::onMouseMoveEvent(int x, int y, MouseMoveState state) {
	bool handled = false;
	
	// Yaw
	if(mYawControl.mouseControlMode == MOUSE_BUTTON_CONTROLLED && isMouseButtonDown(mYawControl.mouseButton)) {
		// TODO: fov and image size
		mRotateYaw = -(x - mMouseDownX) / 600.0;
		handled = true;
	}

	// Pitch
	if(mPitchControl.mouseControlMode == MOUSE_BUTTON_CONTROLLED && isMouseButtonDown(mPitchControl.mouseButton)) {
		// TODO: fov and image size
		mRotatePitch = -(y - mMouseDownY) / 600.0;
		handled = true;
	}
	
	// Pass the control to the next event handler
	return handled;
}

void FlyNavigator::_refreshCamera() {
	double timeNow = Timer::getSystemTime();
	double timeStep = timeNow - mLastMoveTime;
	mLastMoveTime = timeNow;
	
	// Calculate linear and angular accelerations
	_updateAccelerations(timeStep); 
	
	// Calculate linear and angular velocities
	_updateVelocities(timeStep);
	
	// Calculate position and direction
	_updatePositionAndOrientation(timeStep);
	
	// Set position and orientation of the camera
	mCamera->setPosition(mPositionX, mPositionY, mPositionZ);
	mCamera->setDirection(mDirection[0], mDirection[1], mDirection[2]);
	mCamera->setUpVector(mUpVector[0], mUpVector[1], mUpVector[2]);
	
	// And recalculate its matrices
	// (Note, that the automatric recalculation functionality of the camera is
	//  disabled in the default constructor of FlyNavigator)
	mCamera->recalculateMatrices();

	// Request redisplay
	redisplayRequest();
}

bool FlyNavigator::_updateAccelerations(const double timeStep) {
	//
	// Linear movement
	//
	
	// Set acceleration from control
	if(mLinearMoveMode == ACCELERATION_CONTROLLED) {
		// Forward-backward move
		mAccelerationX = mMoveForwardBackward * mDirection[0];
		mAccelerationY = mMoveForwardBackward * mDirection[1];
		mAccelerationZ = mMoveForwardBackward * mDirection[2];

		// Left-right mode
		mAccelerationX += mMoveLeftRight * mRightVector[0];
		mAccelerationY += mMoveLeftRight * mRightVector[1];
		mAccelerationZ += mMoveLeftRight * mRightVector[2];
	}
	
	return false;
}

bool FlyNavigator::_updateVelocities(const double timeStep) {
	//
	// Linear movement
	//

	// Set velocity from control
	if(mLinearMoveMode == VELOCITY_CONTROLLED) {
		mVelocityX = mMoveForwardBackward * mDirection[1];
		mVelocityY = mMoveForwardBackward * mDirection[2];
		mVelocityZ = mMoveForwardBackward * mDirection[3];
	}

	// Update velocity from acceleration
	else {
		// If steering is on, only that component of the velocity 
		// is kept that parallel to the direction of the acceleration
		if(mLinearSteering)
			_linearSteering();
		
		mVelocityX += mAccelerationX * timeStep;
		mVelocityY += mAccelerationY * timeStep;
		mVelocityZ += mAccelerationZ * timeStep;
		
		// Deceleration
		if(::fabs(mMoveForwardBackward + mMoveLeftRight) < MathUtils::getEpsilon()) {
			double decelVelocity = mLinearDeceleration * timeStep;
			mVelocityX = mVelocityX >= 0.0 ?
					mVelocityX > decelVelocity ? mVelocityX - decelVelocity : 0.0 :
					mVelocityX < decelVelocity ? mVelocityX + decelVelocity : 0.0;
			mVelocityY = mVelocityY >= 0.0 ?
					mVelocityY > decelVelocity ? mVelocityY - decelVelocity : 0.0 :
					mVelocityY < decelVelocity ? mVelocityY + decelVelocity : 0.0;
			mVelocityZ = mVelocityZ >= 0.0 ?
					mVelocityZ > decelVelocity ? mVelocityZ - decelVelocity : 0.0 :
					mVelocityZ < decelVelocity ? mVelocityZ + decelVelocity : 0.0;
		}
	}
	
	//
	// Angular movement
	//

	// Set velocity from control
	if(mAngularMoveMode == VELOCITY_CONTROLLED) {
		mAngularVelocityRoll = mRotateRoll;
		mAngularVelocityPitch = mRotatePitch;
		mAngularVelocityYaw = mRotateYaw;
	}
	
	return false;
}

bool FlyNavigator::_updatePositionAndOrientation(const double timeStep) {
	//
	// Linear movement
	//

	mPositionX += mVelocityX * timeStep;
	mPositionY += mVelocityY * timeStep;
	mPositionZ += mVelocityZ * timeStep;

	//
	// Angular movement (with RPY angles)
	//

	// Set angles from control
	if(mAngularMoveMode == POSITION_CONTROLLED) {
		mRoll  = mMouseDownRoll + mRotateRoll;
		mPitch = mMouseDownPitch + mRotatePitch;
		mYaw   = mMouseDownYaw + mRotateYaw;
	}
	else {
		mRoll  += mAngularVelocityRoll * timeStep;
		mPitch += mAngularVelocityPitch * timeStep;
		mYaw   += mAngularVelocityYaw * timeStep;
	}
	
	// Update direction, up vector, and right vector
	double rotationMatrix[9];
	_calculateRPYRotationMatrix(mRoll, mYaw, mPitch, rotationMatrix);
	
	MathUtils::matrix3x3VectorProduct(rotationMatrix, DEFAULT_ZERO_DIRECTION, mDirection);
	MathUtils::matrix3x3VectorProduct(rotationMatrix, DEFAULT_ZERO_UP_VECTOR, mUpVector);
	MathUtils::matrix3x3VectorProduct(rotationMatrix, DEFAULT_ZERO_RIGHT_VECTOR, mRightVector);
	
	return false;
}

void FlyNavigator::_linearHandBreak() {
    mAccelerationX = mAccelerationY = mAccelerationZ = 0.0;
    mVelocityX = mVelocityY = mVelocityZ = 0.0;
    mMoveForwardBackward = mMoveLeftRight = 0.0;
}

void FlyNavigator::_linearHandBreak(const double moveDirX, const double moveDirY, const double moveDirZ) {
	// Reset acceleration
	// Not needed it is rewritten
	/*
	if(::abs(MathUtils::signum(moveDirX) - MathUtils::signum(mAccelerationX)) > 1)
		mAccelerationX = 0.0;
	if(::abs(MathUtils::signum(moveDirY) - MathUtils::signum(mAccelerationY)) > 1)
		mAccelerationY = 0.0;
	if(::abs(MathUtils::signum(moveDirZ) - MathUtils::signum(mAccelerationZ)) > 1)
		mAccelerationZ = 0.0;
	*/

	// Reset velocity
	if(::abs(MathUtils::signum(moveDirX) - MathUtils::signum(mVelocityX)) > 1)
		mVelocityX = 0.0;
	if(::abs(MathUtils::signum(moveDirY) - MathUtils::signum(mVelocityY)) > 1)
		mVelocityY = 0.0;
	if(::abs(MathUtils::signum(moveDirZ) - MathUtils::signum(mVelocityZ)) > 1)
		mVelocityZ = 0.0;
}

void FlyNavigator::_angularHandBreak() {
}

// TODO does not operate
void FlyNavigator::_linearSteering() {
	// Get direction of the acceleration
	double dX = mAccelerationX, dY = mAccelerationY, dZ = mAccelerationZ;
	MathUtils::vectorNormalize(dX, dY, dZ);
	
	int sgnDX = MathUtils::signum(dX), sgnDY = MathUtils::signum(dY), sgnDZ = MathUtils::signum(dZ); 
	
	// Get velocity compontent that is parallel to the direction of the acceleration
	if(sgnDX) mVelocityX *= MathUtils::signum(mVelocityX) == sgnDX ? ::fabs(dX) : -::fabs(dX); 
	if(sgnDY) mVelocityY *= MathUtils::signum(mVelocityY) == sgnDY ? ::fabs(dY) : -::fabs(dY);
	if(sgnDZ) mVelocityZ *= MathUtils::signum(mVelocityZ) == sgnDZ ? ::fabs(dZ) : -::fabs(dZ); 
}

void FlyNavigator::_calculateRPYRotationMatrix(
		const double phi, const double theta, const double psi, double rotationMatrix[9]) {
	rotationMatrix[0] = ::cos(phi)*::cos(theta);
	rotationMatrix[1] = ::cos(phi)*::sin(theta)*::sin(psi) - ::sin(phi)*::cos(psi);
	rotationMatrix[2] = ::cos(phi)*::sin(theta)*::cos(psi) + ::sin(phi)*::sin(psi);

	rotationMatrix[3] = ::sin(phi)*::cos(theta);
	rotationMatrix[4] = ::sin(phi)*::sin(theta)*::sin(psi) + ::cos(phi)*::cos(psi);
	rotationMatrix[5] = ::sin(phi)*::sin(theta)*::cos(psi) - ::cos(phi)*::sin(psi);

	rotationMatrix[6] = -::sin(theta);
	rotationMatrix[7] = ::cos(theta)*::sin(psi);
	rotationMatrix[8] = ::cos(theta)*::cos(psi);
}

} // namespace MinGLE
