//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/StringUtils.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file StringUtils.cpp
/// String handling utils

// Header that contains the declaration
#include <StringUtils.h>

// MinGLE headers
#include <Exception.h>

namespace MinGLE {

const std::string StringUtils::DEFAULT_TRIM_DELIMITERS = " \t";
const int StringUtils::DEFAULT_FLOATING_POINT_PRECISION = -1;

void StringUtils::trim(std::string& str, const std::string& delimiters) {
	std::string::size_type pos = str.find_last_not_of(delimiters);
	if (pos != std::string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(delimiters);
		if (pos != std::string::npos)
			str.erase(0, pos);
	} else
		str.erase(str.begin(), str.end());
}

std::string StringUtils::toString(const double value, const int precision) {
	std::ostringstream osstr;
	if (precision == DEFAULT_FLOATING_POINT_PRECISION)
		osstr << value;

	else {
		osstr.precision(precision);
		osstr << std::fixed << value;
	}

	return osstr.str();
}

unsigned StringUtils::parseUnsigned(const std::string& str) {
	unsigned result;
	std::istringstream isstr(str);
	isstr >> result;

	return result;
}

double StringUtils::parseDouble(const std::string& str) {
	double result;
	std::istringstream isstr(str);
	isstr >> result;

	return result;
}

float StringUtils::parseFloat(const std::string& str) {
	float result;
	std::istringstream isstr(str);
	isstr >> result;

	return result;
}

void StringUtils::tokenize(const std::string& str,
		std::vector<std::string>& tokens, const std::string& tokenDelimiters,
		const bool needsTrim, const std::string& trimDelimiters) {

	// Skip delimiters at beginning.
	std::string::size_type lastPos = str.find_first_not_of(tokenDelimiters, 0);

	// Find first "non-delimiter".
	std::string::size_type pos = str.find_first_of(tokenDelimiters, lastPos);

	while (std::string::npos != pos || std::string::npos != lastPos) {

		// Found a token, add it to the vector
		tokens.push_back(str.substr(lastPos, pos - lastPos));

		// Trim the token if needed and drop if its length changes to zero
		if (needsTrim) {
			trim(tokens.back(), trimDelimiters);

			if (!tokens.back().size())
				tokens.pop_back();
		}

		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(tokenDelimiters, pos);

		// Find next "non-delimiter"
		pos = str.find_first_of(tokenDelimiters, lastPos);
	}
}

std::string StringUtils::subStringBefore(const std::string& str, const std::string& pattern) {
	std::string::size_type pos = str.find(pattern);
	return pos != std::string::npos ? str.substr(0, pos) : str;
}

std::string StringUtils::subStringAfter(const std::string& str, const std::string& pattern) {
	std::string::size_type pos = str.find(pattern);
	return pos != std::string::npos ? str.substr(pos+pattern.size()) : str;
}

} // namespace MinGLE
