//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/System.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//
/// @file System.cpp
/// Singleton manager class

// Header that contains the declaration
#include <System.h>

// Header that contains the configuration options
#include <config_wrapper.h>

// MinGLE headers
#include <Exception.h>
#include <VirtualFileSystem.h>
#include "filesystems/include/LocalFileSystem.h"
#include "configuration/include/TextConfiguration.h"
#include "configuration/include/XMLConfiguration.h"
#include <DynLoad.h>
#include <Threading.h>
#include <Imaging.h>
#include <FontRendering.h>
#include <Navigators.h>

// Proper inclusion of OpenGL functionality
#include <mingle-gl.h>

// System headers
#include <string.h>
#include <iostream>

namespace MinGLE {

// Map of loaded dynamic libraries
System::DynLibMap System::mDynLibs;

// Map of system implementations
System::SystemImplementationMap System::mSystemImplementations;

// Set of used virtual file systems
System::VirtualFileSystemMap System::mUsedVFSes;

// Singleton instance
System *System::msInstance = 0;

// Default configuration file name
const std::string System::DEFAULT_CONFIG_FILE = "mingle.cfg";

// Search path for render system implementation directories in the configuration
const std::string System::RENDER_SYSTEM_DIRECTORY_IN_CONFIG =
		"implementations/renderSystem/searchPath";

// Search path for threading implementation directories in the configuration
const std::string System::THREADING_DIRECTORY_IN_CONFIG =
		"implementations/threading/searchPath";

// Search path for file system implementation directories in the configuration
const std::string System::FILE_SYSTEM_DIRECTORY_IN_CONFIG =
		"implementations/fileSystem/searchPath";

// Search path for configuration implementation directories in the configuration
const std::string System::CONFIGURATION_DIRECTORY_IN_CONFIG =
		"implementations/configuration/searchPath";

// Search path for imaging implementation directories in the configuration
const std::string System::IMAGING_DIRECTORY_IN_CONFIG =
		"implementations/imaging/searchPath";

// Search path for imaging implementation directories in the configuration
const std::string System::FONTRENDERING_DIRECTORY_IN_CONFIG =
		"implementations/fontRendering/searchPath";

// Name of the composite entry for file systems in the configuration
const std::string System::FILE_SYSTEMS_ENTRY_IN_CONFIG = "fileSystems";

// Name of the simple entry for file system protocol in the configuration
const std::string System::PROTOCOL_ENTRY_NAME_IN_CONFIG = "protocol";

// Name of the simple entry for file system connection string in the configuration
const std::string System::CONNECTIONSTRING_ENTRY_NAME_IN_CONFIG = "connectionString";

// String constant for automatic system implementation detection
const std::string System::AUTOMATIC_SYSTEM_TYPE_STRING = "auto";

// OpenGL version
System::GLVersion
		System::mGLVersion = { 0 /* major */, 0 /* minor */, "" /* vendorSpecific */};

// GLEW version string
std::string System::mGLEWVersionString = "GLEW is not supported";

// Name of the high level config file
std::string System::mConfigFile = System::DEFAULT_CONFIG_FILE;

// High level configuration structure
Configuration *System::mConfiguration = 0;

// Protected constructor, cannot be called externally
System::System() :
	mScreenWidth(UNKNOWN_RESOLUTION), mScreenHeight(UNKNOWN_RESOLUTION) {
}

// Protected destructor, cannot be called externally
System::~System() {

	// Erase all system listeners
	for (SystemListenerVector::iterator i = mSystemListeners.begin(); i
			!= mSystemListeners.end(); i++) {
		// Unregister if we still have event listener
		(*i)->onUnRegistered();

		// Delete listener (it is necessary since the execution is in our
		// hands because of the System::enterMainLoop() call.)
		delete *i;
	}

	// Destructing the main system implementation
	// Note that several system implementations can be loaded, however 
	// only one is used. That one destroys the other implementations unload
	// the loaded dynamic libraries and exits from the application
	if (msInstance == this) {

		// Unregister and delete all system implementations
		for (SystemImplementationMap::iterator i =
				mSystemImplementations.begin(); i
				!= mSystemImplementations.end(); i++) {

			// Delete any other system implementations
			if (i->second != this)
				delete i->second;
		}

		// Shutdown subsystems
		_shutdownSubSystems();
		
		// Unload all dynamic libraries
		_unloadAllDynamicLibraries();

		// If the System is deleted nothing to do, but
		// exit from application
		exit(0);
	}
}

void System::setConfigFile(const std::string& configFile) {
	mConfigFile = configFile;
}

VirtualFileSystem *System::getVirtualFileSystemForFile(
		const std::string& blockName, const std::string& fileName) {
	
	// Get blocks
	Configuration::EntryList blocks = mConfiguration->getFirstChild(FILE_SYSTEMS_ENTRY_IN_CONFIG)->getEntries(blockName);
	
	// Test blocks
	for(Configuration::EntryList::iterator iBlock=blocks.begin(); iBlock!=blocks.end(); iBlock++) {
		
		// Get protocol
		std::string protocol = (*iBlock)->getFirstChild(PROTOCOL_ENTRY_NAME_IN_CONFIG)->getStringValue();
		
		// Test connection strings 
		Configuration::EntryList connStrings = (*iBlock)->getEntries(CONNECTIONSTRING_ENTRY_NAME_IN_CONFIG);
		for(Configuration::EntryList::iterator iConnStr=connStrings.begin(); iConnStr!=connStrings.end(); iConnStr++) {
			
			// Get connection string
			std::string connStr = (*iConnStr)->getStringValue();
			
			// Create file system
			VirtualFileSystem *vfs = VirtualFileSystemManager::createVirtualFileSystem(protocol, connStr);
			
			// Test file existence
			if(vfs->exists(fileName)) {
				std::string uniqueStringId = vfs->getURL();
				VirtualFileSystemMap::iterator iVfs = mUsedVFSes.find(uniqueStringId);
				
				// This is a new VFS, insert into the map
				if(iVfs == mUsedVFSes.end()) {
					mUsedVFSes[uniqueStringId] = vfs;
					return vfs;
				}

				// This is a vfs that is already in use
				delete vfs;
				return iVfs->second;
			}
			
			// Delete file system
			delete vfs;
		}
	}
	
	return 0;
}

DynLib *System::loadDynamicLibrary(const std::string& name) {
	Assert(mDynLibs.find(name) == mDynLibs.end(), "System::loadDynamicLibrary");

	// Create, load dynamic library and push it to the map 
	DynLib *dynLib = new DynLib(name);
	mDynLibs[name] = dynLib;

	return dynLib;
}

void System::unloadDynamicLibrary(const std::string& name) {
	Assert(mDynLibs.find(name) != mDynLibs.end(),
			"System::unloadDynamicLibrary");
	mDynLibs.erase(name);
}

void System::loadDynamicLibraries(const std::string& directory) {
	VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem();
	VirtualFileSystem::DirectoryEntryList dl = vfs->listDirectory(directory);

	for (VirtualFileSystem::DirectoryEntryList::iterator i=dl.begin(); i
			!=dl.end(); i++) {

		if (i->type & DynLoad::DYNAMIC_LIBRARY_TYPE_MASK
				&& VirtualFileSystem::getExtension(i->name)
						== DynLoad::DYNAMIC_LIBRARY_EXTENSION) {
			loadDynamicLibrary(directory + i->name);
		}
	}
}

void System::registerSystemImplementation(const std::string& name,
		System *systemImplementation) {
	Assert(!getSystemImplementation(name), "System::registerRenderSystem");

	mSystemImplementations[name] = systemImplementation;
}

void System::unregisterSystemImplementation(const std::string& name) {
	Assert(getSystemImplementation(name),
			"System::unregisterSystemImplementation");

	mSystemImplementations.erase(name);
}

System *System::getSystemImplementation(const std::string& name) {
	if (mSystemImplementations.find(name) != mSystemImplementations.end())
		return mSystemImplementations[name];

	return 0;
}

void System::initialize(int *argc, char **argv, const std::string& systemType) {

	// Register the in-built local file system implementation
	// to be able to load the configuration file
	LocalFileSystemManager::registerVirtualFileSystemManager();	
	
	// Register the in-built text and xml configuration implementations
	// to be able to parse the configuration file
	TextConfigurationManager::registerConfigurationManager();
	XMLConfigurationManager::registerConfigurationManager();
	
	// Load config file
	mConfiguration = ConfigurationManager::createConfiguration(mConfigFile);

	// Load implementations dynamic load libraries
	_loadDynamicLibrariesFromConfig(FILE_SYSTEM_DIRECTORY_IN_CONFIG);
	_loadDynamicLibrariesFromConfig(CONFIGURATION_DIRECTORY_IN_CONFIG);
	_loadDynamicLibrariesFromConfig(RENDER_SYSTEM_DIRECTORY_IN_CONFIG);
	_loadDynamicLibrariesFromConfig(THREADING_DIRECTORY_IN_CONFIG);
	_loadDynamicLibrariesFromConfig(IMAGING_DIRECTORY_IN_CONFIG);
	_loadDynamicLibrariesFromConfig(FONTRENDERING_DIRECTORY_IN_CONFIG);
	
	// The system implementation is explicitly set 
	if (systemType != AUTOMATIC_SYSTEM_TYPE_STRING) {
		System *systemImplementation = getSystemImplementation(systemType);
		Assert(systemImplementation, "System::initialize");

		// Assign the implementation to singleton instance
		// and initialize it with the command line arguments
		msInstance = systemImplementation;
		msInstance->_initialize(argc, argv);
	}

	// Automatic system implementation detection
	else {
		// Check that there are any system implementations
		Assert(mSystemImplementations.size(), "System::initialize");

		bool initializationSucceeded = false;

		for (SystemImplementationMap::iterator i=mSystemImplementations.begin(); i
				!=mSystemImplementations.end(); i++) {
			if (i->second->_initialize(argc, argv)) {
				initializationSucceeded = true;
				msInstance = i->second;
				break;
			}
		}

		// Check success
		Assert(initializationSucceeded, "System::initialize");
	}
}

Window* System::createWindowWithDefaultEnv(WindowListener *sceneRenderer,
		GuiSheet *guiSheet) {
	// Create window
	Window *win = System::createWindow();

	// Create and register scene renderer
	Assert(sceneRenderer, "System::createWindowWithDefaultEnv");
	win->registerWindowListener(sceneRenderer);

	// Add navigator
	win->registerNavigator(new ExaminerNavigator());

	// Add keyhandler
	ApplicationKeyHandler *keyHandler = new ApplicationKeyHandler();
	win->registerWindowListener(keyHandler);

	// Add gui sheet
	if (guiSheet)
		win->registerWindowListener((WindowListener *) guiSheet);

	return win;
}

void System::registerSystemListener(SystemListener *systemListener) {
	msInstance->mSystemListeners.push_back(systemListener);

	// Call registration event
	systemListener->onRegistered();
}

void System::removeSystemListener(SystemListener *systemListener) {
	SystemListenerVector::iterator i = std::find(
			msInstance->mSystemListeners.begin(),
			msInstance->mSystemListeners.end(), systemListener);
	Assert(i != msInstance->mSystemListeners.end(),
			"System::removeSystemListener");

	msInstance->mSystemListeners.erase(i);

	// Call unregistration event
	systemListener->onUnRegistered();
}

void System::_loadDynamicLibrariesFromConfig(const std::string& searchPath) {
	// Get implementation directories from the configuration
	Configuration::EntryList implementationDirectories =
			mConfiguration->getEntries(searchPath);

	// Scan for implementations in the directories set as a parameter
	for (Configuration::EntryList::iterator i=implementationDirectories.begin(); i
			!=implementationDirectories.end(); i++) {
		loadDynamicLibraries(*((*i)->value.simpleValue));
	}
}


void System::_shutdownSubSystems() {
	// Shutdown font rendering, imaging, and threading subsystems
	FontRendering::shutdown();
	Imaging::shutdown();
	Threading::shutdown();

	// Delete used virtual file systems and shutdown vfs subsystem
	for(VirtualFileSystemMap::iterator i=mUsedVFSes.begin(); i!=mUsedVFSes.end(); i++)
		delete i->second;
	VirtualFileSystemManager::shutdown();
}

void System::_unloadAllDynamicLibraries() {
	// Delete dynamic libraries
	for (DynLibMap::iterator i = mDynLibs.begin(); i != mDynLibs.end(); i++) {
		delete i->second;
	}

	// Empty map
	mDynLibs.clear();
}

// This method is called by the descendant of the Window class, when a window is created
void System::_onWindowCreated() {
	// Get and parse OpenGL version
	System::_retrieveGLVersion();

	// Initialize GLEW (if supported)
	System::_initGlew();
}

void System::_retrieveGLVersion() {
	// Get version string
	const char * versionString = (const char *) glGetString(GL_VERSION);
	Assert(versionString, "System::_retrieveGLVersion");

	// Parse version string
	sscanf(versionString, "%d.%d", &mGLVersion.major, &mGLVersion.minor);
	char *vendorSpecific = strchr(versionString, ' ');
	if (vendorSpecific)
		mGLVersion.vendorSpecific = vendorSpecific;
}

void System::_initGlew() {
#ifdef MINGLE_GLEW
	Assert(glewInit() == GLEW_OK, "System::_initGlew");
	mGLEWVersionString = (const char *) glewGetString(GLEW_VERSION);
#endif // MINGLE_GLEW
}

} // namespace MinGLE
