//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Threading.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Threading.cpp
/// Thread support classes for internal usage

// Header that contains the declaration
#include <Threading.h>

// MinGLE headers
#include <Exception.h>

// System headers
#include <iostream>
#include <sys/time.h>

namespace MinGLE {

// =============================================================================

Runnable::~Runnable() {
}

// =============================================================================

const bool Thread::DEFAULT_JOINABLE_FLAG = true;
const bool Thread::DEFAULT_JOIN_AT_STOP_FLAG = true
		&& Thread::DEFAULT_JOINABLE_FLAG;

const int Thread::ERROR_RETURN_VALUE = 0;
const int Thread::SUCCESS_RETURN_VALUE = 1;

Thread::Thread(Runnable *runnable, bool joinable) :
	mRunnable(runnable), mRunning(false), mJoinable(joinable),
			mStopRequest(true) {
}

Thread::~Thread() {
	// If the thread is joinable, the destructor will wait for it
	// FIXME indirect pure virtual method call from destructor (_joinThread)
	if (mJoinable)
		join();
}

void Thread::sleep(double seconds) {
	usleep((__useconds_t) (seconds * 1.0e6));
}

void Thread::start() {
	_startThread();
}

void Thread::stop(bool joinAtStop) {
	Assert(mJoinable || !joinAtStop, "Thread::stop");

	// Indicate stop request
	mStopRequest = true;

	// Join thread if needed
	if (joinAtStop)
		_joinThread();
}

void Thread::join() {
	Assert(mJoinable, "Thread::join");
	_joinThread();
}

bool Thread::exceptionHandler(const Exception& exception) {
	// Override this method.
	// This method is called when an unhandled MinGLE exception 
	// is thrown during the execution of the thread.

	// True return value means, the exception is handled,
	// false means that the unexpected() is needed to be called.
	std::cerr << "Error in a thread: " << exception.mDescription
			<< " (Thrown in " << exception.mFile << ':' << exception.mMethod
			<< " line " << exception.mLine << ')' << std::endl;

	return true;
}

void Thread::unexpected() {
	// Override this method.
	// This method is called when an unhandled unknown 
	// exception is thrown during the execution of the thread.
	std::cerr << "Unexpected exception caught in a MinGLE thread" << std::endl;
}

// =============================================================================

template<> Manager<Threading>::ImplementationMapPtr Manager<Threading>::mImplementations = new Manager<Threading>::ImplementationMap();

// Singleton instance
template<> Threading *Manager<Threading>::msInstance = 0;

Threading::~Threading() {
	// Nothing to do
}

Mutex* Threading::createMutex() {
	Assert(msInstance, "Threading::createMutex");
	return msInstance->_createMutex();
}

Thread * Threading::createThread(Runnable *runnable, bool joinable) {
	Assert(msInstance, "Threading::createThread");
	return msInstance->_createThread(runnable, joinable);
}

// =============================================================================

/*
 Process::Process() :
 Thread(this, true), mStatus(0.0) {
 }

 void Process::run() {
 mStatus = 0.0;
 this->process();
 mStatus = 1.0;
 }
 */

} // namespace MinGLE
