//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Timer.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Timer.cpp
/// Generic timer class

// Header that contains the declaration
#include <Timer.h>

// MinGLE headers
#include <Exception.h>
#include <TimerListener.h>
#include <System.h>

// System headers
#include <sys/time.h>

namespace MinGLE {

Timer::Timer(Window * window, double period, TimerType type) :
	mPeriod(period), mType(type), mLastTickTime(0), mElapsedTime(0.0),
			mRunning(false), mWindow(window), mTimerListener(0),

			mAutoStartOnRegistration(DEFAULT_START_ON_REGISTRATION),
			mAdaptToDesiredFPS(DEFAULT_ADAPT_TO_DESIRED_FPS) {
	Assert(period > getGranularity(), "Timer::Timer");
}

Timer::~Timer() {
	if (mTimerListener) {
		mTimerListener->onUnRegistered();
		delete mTimerListener;
	}
}

double Timer::getSystemTime() {
	struct timeval tv;
	struct timezone tz;

	gettimeofday(&tv, &tz);
	return tv.tv_sec * 1.0 + tv.tv_usec * 0.000001;
}

void Timer::start() {
	// Generic operations
	mLastTickTime = System::getTime();
	mRunning = true;
}

void Timer::pause() {
	// Generic operations
	mRunning = false;
}

void Timer::stop() {
	// Pause
	pause();

	// Reset timing info
	mLastTickTime = 0.0, mElapsedTime = 0.0;
}

void Timer::registerTimerListener(TimerListener *timerListener) {
	// Unregister previous listener
	if (mTimerListener)
		mTimerListener->onUnRegistered();

	// Register
	(mTimerListener = timerListener)->mTimer = this;
	mTimerListener->onRegistered();

	// Start timer automatically
	if (mAutoStartOnRegistration)
		start();
}

TimerListener *Timer::removeTimerListener() {
	Assert(mTimerListener, "Timer::removeTimerListener");
	mTimerListener->onUnRegistered();

	TimerListener *listener = mTimerListener;
	mTimerListener = 0;

	return listener;
}

} // namespace MinGLE
