//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/VirtualFileSystem.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file VirtualFileSystem.cpp
/// Virtual file system class

// Header that contains the declaration
#include <VirtualFileSystem.h>

// MinGLE headers
#include <StringUtils.h>
#include <System.h>
#include "filesystems/include/LocalFileSystem.h"

// System headers
#include <algorithm>
#include <sstream>

namespace MinGLE {

// TODO add OS support
const char VirtualFileSystem::DIRECTORY_SEPARATOR = '/';

const char VirtualFileSystem::END_OF_LINE = '\n';

const char VirtualFileSystem::END_OF_LINE_TO_DROP = '\r';

const int VirtualFileSystem::END_OF_FILE = -1;

const int VirtualFileSystem::UNLIMITED_BYTES = -1;

std::string VirtualFileSystem::getDirectory(const std::string& filename) {
	std::string::size_type pos = filename.rfind(DIRECTORY_SEPARATOR);
	if (std::string::npos != pos)
		return filename.substr(0, pos+1);

	return "";
}

std::string VirtualFileSystem::getFilename(const std::string& filename) {
	std::string::size_type pos = filename.rfind(DIRECTORY_SEPARATOR);
	if (std::string::npos != pos)
		return filename.substr(pos+1);

	return filename;
}

std::string VirtualFileSystem::getBaseName(const std::string& filename) {
	std::string::size_type dirPos = filename.rfind(DIRECTORY_SEPARATOR);
	std::string::size_type extPos = filename.rfind('.');

	if (std::string::npos == extPos)
		extPos = filename.size();

	if (std::string::npos != dirPos)
	    return filename.substr(dirPos+1, extPos-dirPos-1);
        return filename.substr(0, extPos-dirPos-1);
}

std::string VirtualFileSystem::getExtension(const std::string& filename) {
	std::string::size_type pos = filename.rfind('.');
	if (std::string::npos != pos)
		return filename.substr(pos+1);

	return "";
}

std::string VirtualFileSystem::readLine(FileHandle f) {
	std::ostringstream osstr;
	int ch;

	while ((ch = getchar(f)) != END_OF_LINE && ch != END_OF_FILE && ch != 0) {
		osstr << (char) ch;
	}

	return osstr.str();
}

void VirtualFileSystem::writeLine(const std::string& line, FileHandle f) {
	Except("VirtualFileSystem::writeLine", "TODO implement");
}

unsigned char *VirtualFileSystem::readFile(const std::string& filename, int &length, unsigned char *buf, int maxLength) {
	return _readData(filename, length, buf, maxLength, false);
}

char *VirtualFileSystem::readFileAsString(const std::string& filename, unsigned char *buf, int maxLength) {
	int length;
	return (char *) _readData(filename, length, buf, maxLength, true);
}

void VirtualFileSystem::_sortDirectoryEntryList(DirectoryEntryList& deList) {
	std::sort(deList.begin(), deList.end(),
			VirtualFileSystem::_directoryEntrySortPredicate);
}

unsigned char *VirtualFileSystem::_readData(const std::string& filename, int &length, unsigned char *buf, int maxLength, bool nullTerminated) {
	// Open file for binary read
	FileHandle f = open(filename, "rb");
	Assert(f, "VirtualFileSystem::readFile");
	
	// Get file length
	seek(f, 0, SEEK_END);
	length = tell(f) + (nullTerminated ? 1 : 0);
	seek(f, 0, SEEK_SET);
	
	// Calculate number of bytes to read
	if(maxLength != UNLIMITED_BYTES)
		length = length <= maxLength ? length : maxLength;
	
	// If null pointer is set as data the memory is allocated
	if(!buf) buf = new unsigned char[length];
	
	// Read data
	if(nullTerminated) {
		read(buf, sizeof(unsigned char), length-1, f);
		buf[length-1] = '\0';
	}
	else
		read(buf, sizeof(unsigned char), length, f);
	
	// Close file
	close(f);
	
	return buf;
}

bool VirtualFileSystem::_directoryEntrySortPredicate(const DirectoryEntry& lhs,
		const DirectoryEntry& rhs) {
	if (lhs.type == DIRECTORY && rhs.type != DIRECTORY)
		return true;
	else if (lhs.type != DIRECTORY && rhs.type == DIRECTORY)
		return false;

	return lhs.name < rhs.name;
}

// =============================================================================

const std::string VirtualFileSystemManager::PROTOCOL_SEPARATOR = "://";
const std::string VirtualFileSystemManager::DEFAULT_FILE_SYSTEM_URL = LocalFileSystemManager::PROTOCOL_STRING + PROTOCOL_SEPARATOR;

template<> Manager<VirtualFileSystemManager>::ImplementationMapPtr Manager<
		VirtualFileSystemManager>::mImplementations =
		new Manager<VirtualFileSystemManager>::ImplementationMap();

// Singleton instance
template<> VirtualFileSystemManager
		*Manager<VirtualFileSystemManager>::msInstance = 0;

VirtualFileSystem *VirtualFileSystemManager::msDefaultFileSystem = 0;

VirtualFileSystemManager::~VirtualFileSystemManager() {
	if (msInstance == this && msDefaultFileSystem)
		delete msDefaultFileSystem;
}

VirtualFileSystem *VirtualFileSystemManager::createVirtualFileSystem(
		const std::string& url) {
	// Split into protocol and connection string
	std::string protocol =
			StringUtils::subStringBefore(url, PROTOCOL_SEPARATOR);
	std::string connectionString = StringUtils::subStringAfter(url,
			PROTOCOL_SEPARATOR);

	return createVirtualFileSystem(protocol, connectionString);
}

VirtualFileSystem *VirtualFileSystemManager::createVirtualFileSystem(
		const std::string& protocol, const std::string& connectionString) {
	// Get corresponding file system implementation
	VirtualFileSystemManager *implementation = getImplementation(protocol);
	Assert(implementation, "VirtualFileSystemManager::createVirtualFileSystem");

	// Create file system
	return implementation->_createVirtualFileSystem(connectionString);
}

VirtualFileSystem *VirtualFileSystemManager::getDefaultVirtualFileSystem() {
	return msDefaultFileSystem ? msDefaultFileSystem : msDefaultFileSystem
			= createVirtualFileSystem(DEFAULT_FILE_SYSTEM_URL);
}

// ============================================================================

FileFormatFunctionality::FileFormatFunctionality(const std::string& extension, Mode mode) :
	mExtension(extension), mMode(mode)
{
}

} // namespace MinGLE
