//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/Window.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file Window.cpp
/// Rendering window class

// Header that contains the declaration
#include <Window.h>

// MinGLE headers
#include <Exception.h>
#include <Imaging.h>
#include <System.h>
#include <mingle-gl.h>

// System headers
#include <string.h>
#include <algorithm>

namespace MinGLE {

// Protected constructor, cannot be called externally (only by System)
Window::Window(unsigned width, unsigned height,
		WindowAttributes::ColorMode colorMode, bool doubleBuffer,
		bool fullScreen, unsigned videoRate, unsigned char *colorBits,
		unsigned char depthBits, unsigned char stencilBits,
		unsigned char *accumColorBits, unsigned auxBuffers, bool stereo) :
	mTimeTriggeredRenderController(this) {
	mOwnWindowAttributes = true;
	mWindowAttributes = new WindowAttributes();
	mWindowAttributes->mLeft = 0, mWindowAttributes->mTop = 0,
			mWindowAttributes->mWidth = width,
			mWindowAttributes->mHeight = height,
			mWindowAttributes->mRecoveryLeft = 0,
			mWindowAttributes->mRecoveryTop = 0,
			mWindowAttributes->mRecoveryWidth = width,
			mWindowAttributes->mRecoveryHeight = height,
			mWindowAttributes->mColorMode = colorMode,
			mWindowAttributes->mDoubleBuffer = doubleBuffer,
			mWindowAttributes->mFullScreen = fullScreen,
			mWindowAttributes->mDepthBits = depthBits,
			mWindowAttributes->mStencilBits = stencilBits,
			mWindowAttributes->mVideoRate = videoRate,
			mWindowAttributes->mAuxBuffers = auxBuffers,
			mWindowAttributes->mStereo = stereo,
			mWindowAttributes->mVisible = true,
			mWindowAttributes->mIconified = false,
			mWindowAttributes->mTitle = "OpenGL Window",
			mWindowAttributes->mAutoFinishFrame = true,
			mWindowAttributes->mAutoClearFrame = true,
			mWindowAttributes->mAutoUpdateViewport = true,
			mWindowAttributes->mRenderMode = DEFAULT_RENDER_MODE,
			mWindowAttributes->mListenOnIdle = false,
			mWindowAttributes->mRedisplayRequest = false,
			mWindowProxy = 0,

			// Recording
			mWindowAttributes->mScreenShotPattern = "screenshot_%03d.png",
			mWindowAttributes->mVideoPattern = "%02d_frame_%06d.tga",
			mWindowAttributes->mScreenShotCapturing = false,
			mWindowAttributes->mScreenShotIndex = 0,
			mWindowAttributes->mVideoSceneIndex = 0,
			mWindowAttributes->mVideoStartTime = VIDEO_IS_NOT_STARTED,
			mWindowAttributes->mVideoRecording = false,

			mWindowAttributes->mCloseRequest = false;

	// Copy color bits
	if (colorBits)
		memcpy(mWindowAttributes->mColorBits, colorBits, 4);
	else
		for (unsigned i=0; i<4; i++)
			mWindowAttributes->mColorBits[i] = 8;

	if (accumColorBits)
		memcpy(mWindowAttributes->mAccumColorBits, accumColorBits, 4);
	else
		for (unsigned i=0; i<4; i++)
			mWindowAttributes->mAccumColorBits[i] = 8;

	// Image for recording
	mWindowAttributes->mRecordingImage = Imaging::createImage();
}

// Protected destructor, cannot be called externally (only by System)
Window::~Window() {

	//printf("destr: "); debugWindow();

	// Remove and erase window proxy
	if (mWindowProxy)
		delete removeWindowProxy();

	// Erase all window listeners
	for (WindowListenerVector::iterator i = mWindowListeners.begin(); i
			!= mWindowListeners.end(); i++) {
		// Unregister if we still have event listener
		(*i)->onUnRegistered();

		// Delete listener (it is necessary since the execution is in our
		// hands because of the System::enterMainLoop() call.)
		delete *i;
	}

	if (mOwnWindowAttributes) {
		// Image for recording
		delete mWindowAttributes->mRecordingImage;
		delete mWindowAttributes;
	}
}

void Window::startFrame() {
	// Report a new frame to the time triggered render controller
	mTimeTriggeredRenderController.frameStarted();
}

void Window::finishFrame() {

	// Report a new frame to the frame rate calculator
	mFrameRateCalculator.frameFinished();

	// Screenshot capturing
	if (mWindowAttributes->mScreenShotCapturing) {
//#if MINGLE_DEVIL_SUPPORT == 1
		mWindowAttributes->mRecordingImage->grabFromScreen();
		mWindowAttributes->mRecordingImage->saveToFile(mWindowAttributes->mScreenShotPattern, mWindowAttributes->mScreenShotIndex++);
/*#else
		Except(
				"Window::createScreenShot",
				"Image handling is not compiled into your implementation. Please recompile MinGLE with DevIL support.");
#endif // MINGLE_DEVIL_SUPPORT == 1*/
		mWindowAttributes->mScreenShotCapturing = false;
	}

	// Video recording
	if(mWindowAttributes->mVideoRecording) _grabVideoFrame();

	// Closing the window is always performed at the end of a frame
	if (mWindowAttributes->mCloseRequest) {
		_close();
	}
}

void Window::registerWindowListener(WindowListener *windowListener,
		const bool topLevel) {

	// The window has window proxy
	if (mWindowProxy && !topLevel)
		mWindowProxy->registerWindowListener(windowListener);

	// The window has no proxy
	else {
		mWindowListeners.push_back(windowListener);
		windowListener->mWindow = this;

		// Call registration event
		windowListener->onRegistered();
	}

	// Do system dependent setup
	_setupWindowListener();
}

void Window::removeWindowListener(WindowListener *windowListener) {
	//printf("remove listener: %p from %p (proxy: %p)\n", windowListener, this, (WindowListener *)mWindowProxy);

	// The window has window proxy,
	// and this does not equal the listener to be removed
	if (mWindowProxy && mWindowProxy != windowListener)
		mWindowProxy->removeWindowListener(windowListener);

	// The window has no proxy,
	// or the proxy is to remove
	else {
		WindowListenerVector::iterator i = std::find(mWindowListeners.begin(),
				mWindowListeners.end(), windowListener);
		Assert(i != mWindowListeners.end(), "Window::removeWindowListener");
		mWindowListeners.erase(i);

		windowListener->mWindow = 0;

		// Call unregistration event
		windowListener->onUnRegistered();
	}
}

void Window::insertWindowProxy(WindowProxy *proxy) {
	Assert(proxy, "Window::insertWindowProxy");

	// Remove previous proxy
	if (mWindowProxy)
		removeWindowProxy();

	// Pass existing window listeners transparently.
	// Transparently means that no unregistration and
	// registration events will be generated during the
	// takeover.
	// Further window listeners will be assigned to 
	// the proxy. See Window::registerWindowListener.
	_passWindowListeners(proxy);

	// Register the proxy as window listener
	// Be careful, the order of these two calls is very important
	// See Window::registerWindowListener. 
	registerWindowListener(proxy);
	mWindowProxy = proxy;
}

WindowProxy *Window::removeWindowProxy() {
	//printf("remove proxy (%p): ", mWindowProxy); debugWindow();
	Assert(mWindowProxy, "Window::removeWindowProxy");

	// Recursively drop the proxy 
	if (mWindowProxy->hasWindowProxy())
		mWindowProxy->removeWindowProxy();

	// Unregister the proxy
	removeWindowListener(mWindowProxy);

	// Take back the event listeners
	mWindowProxy->_passWindowListeners(this);

	// Clear proxy pointer
	WindowProxy *proxy = mWindowProxy;
	mWindowProxy = 0;

	// Return removed proxy
	return proxy;
}

void Window::setRenderMode(WindowAttributes::RenderMode renderMode) {
	if (mWindowAttributes->mRenderMode == WindowAttributes::RENDER_ON_TIMER
			&& renderMode != WindowAttributes::RENDER_ON_TIMER)
		mTimeTriggeredRenderController.stopOperation();

	if (mWindowAttributes->mRenderMode != WindowAttributes::RENDER_WHEN_IDLE
			&& renderMode == WindowAttributes::RENDER_WHEN_IDLE) {
		// (Set render mode before calling idle listener setter) 
		mWindowAttributes->mRenderMode = renderMode;
		_setupWindowListenerIdle();
	}

	mWindowAttributes->mRenderMode = renderMode;
}

void Window::setListenOnIdle(bool listen) {
	mWindowAttributes->mListenOnIdle = listen;
	_setupWindowListenerIdle();
}

// Recording
void Window::createScreenShot() {
	mWindowAttributes->mScreenShotCapturing = true;
}

void Window::startVideo() {
//#if MINGLE_DEVIL_SUPPORT == 1
	mWindowAttributes->mVideoStartTime = System::getTime();
	mWindowAttributes->mVideoRecording = true;
/*#else
	Except(
			"Window::startVideo",
			"Image handling is not compiled into your implementation. Please recompile MinGLE with DevIL support.");
#endif*/ // MINGLE_DEVIL_SUPPORT == 1
}

void Window::pauseVideo() {
//#if MINGLE_DEVIL_SUPPORT == 1
	mWindowAttributes->mVideoSceneIndex++;
	mWindowAttributes->mVideoRecording = false;
/*#else
	Except(
			"Window::pauseVideo",
			"Image handling is not compiled into your implementation. Please recompile MinGLE with DevIL support.");
#endif */ // MINGLE_DEVIL_SUPPORT == 1
}

void Window::checkRedisplay() {
	// The request flag is initialized with the 
	// own redisplay request flag of the window
	bool request = mWindowAttributes->mRedisplayRequest;
	mWindowAttributes->mRedisplayRequest = false;

	for (WindowListenerVector::iterator i = mWindowListeners.begin(); i
			!= mWindowListeners.end(); i++) {
		if ((*i)->mNeedRedisplay)
			request = true;
		(*i)->mNeedRedisplay = false;
	}

	if (request)
		redisplay();
}

void Window::unProject(int x, int y, double *objX, double *objY, double *objZ) {
	/*
	 GLint viewport[4];
	 glGetIntegerv( GL_VIEWPORT, viewport );

	 GLfloat winY = (GLfloat) viewport[3] - (GLfloat) y, winZ;
	 glReadPixels(x, int(winY), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ);
	 unProject(x, y, winZ, objX, objY, objZ);
	 */

	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	double _top = 1.0;
	double _bottom = -1.0;
	double _left = -(double)viewport[2]/(double)viewport[3];
	double _right = -_left;
	double _zNear = 0.01;

	*objX = (double)(x-viewport[0])/(double)(viewport[2]);
	*objY = (double)(y-viewport[1])/(double)(viewport[3]);

	*objX = _left + (*objX)*(_right-_left);
	*objY = _top + (*objY)*(_bottom-_top);
	*objZ = _zNear;

	//unProject(x, y, 0.99, objX, objY, objZ);
}

void Window::unProject(int x, int y, double z, double *objX, double *objY,
		double *objZ) {
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	GLdouble modelview[16];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview);

	GLdouble projection[16];
	glGetDoublev(GL_PROJECTION_MATRIX, projection);

	GLdouble winX = (GLdouble) x, winY = (GLdouble) viewport[3] - (GLdouble) y,
			winZ = (GLdouble) z;
	gluUnProject(winX, winY, winZ, modelview, projection, viewport, objX, objY,
			objZ);
}

void Window::_grabVideoFrame() {
	mWindowAttributes->mRecordingImage->grabFromScreen();

	printf("%s\n", mWindowAttributes->mVideoPattern.c_str());

	// Use ms in frame file name
	mWindowAttributes->mRecordingImage->saveToFile(
			mWindowAttributes->mVideoPattern, mWindowAttributes->mVideoSceneIndex,
			(unsigned) (1000.0 * (System::getTime()-mWindowAttributes->mVideoStartTime))
	);
}

void Window::_passWindowListeners(Window *to) {
	for (WindowListenerVector::iterator i=mWindowListeners.begin(); i
			!= mWindowListeners.end(); i++) {
		(*i)->mWindow = to;
		to->mWindowListeners.push_back(*i);
	}
	mWindowListeners.clear();
}

#define INDENT(indent)	for(unsigned __i__=0;__i__<indent;__i__++) printf(" ")
void Window::debugWindow(unsigned indent) {
	INDENT(indent);
	printf("%p [%c]\n", this, mWindowProxy ? 'P' : ' ');
	for (WindowListenerVector::iterator i=mWindowListeners.begin(); i
			!= mWindowListeners.end(); i++) {
		if (dynamic_cast<WindowProxy *>(*i))
			dynamic_cast<WindowProxy *>(*i)->debugWindow(indent+1);
		else {
			INDENT(indent+1);
			printf("%p\n", *i);
		}
	}
	if (!indent)
		printf("=========\n");
}
#undef INDENT

// =============================================================================

WindowProxy::WindowProxy(Window *window) :
	mWindow(window) {
	// Take window attributes
	mWindowAttributes = window->mWindowAttributes;
	mOwnWindowAttributes = false;
}

WindowProxy::~WindowProxy() {
}

} // namespace MinGLE
