//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/imaging/include/DevilImage.h
// Creation Date: January 11th 2008
// Revision Date: January 11th 2008
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file DevilImaging.h
/// DevIL imaging implementation

#ifndef __MINGLE_DEVIL_IMAGING_H__
#define __MINGLE_DEVIL_IMAGING_H__

// Include class definitions
#include <classes.h>

// MinGLE headers
#include <Imaging.h>
#include <Threading.h>

// System headers
#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

namespace MinGLE {

class DevilImage : public Image {

protected:
	static GuardedPointer<VirtualFileSystem> msCurrentVfs;
	static VirtualFileSystem *msLockedVfs;
	ILuint mImageId;
	
public:
	DevilImage();
	DevilImage(unsigned width, unsigned height, GLenum format, GLenum type,
			void *data = 0);
	DevilImage(const std::string& filename, VirtualFileSystem *vfs);
	DevilImage(const Image * image);
	virtual ~DevilImage();

	// Input
	virtual void loadFromFile(const std::string& filename,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	virtual void createFromMemory(unsigned width, unsigned height,
			GLenum format = DEFAULT_FORMAT, GLenum type = DEFAULT_TYPE,
			void *data = 0);
	virtual void grabFromScreen();
	virtual void texImage2D(GLint level = 0);

	// Output
	virtual void saveToFile(const std::string& filename,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	virtual void saveToFile(const std::string& pattern, unsigned index,
			VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());
	virtual void saveToFile(const std::string& pattern, unsigned index1,
			unsigned index2, VirtualFileSystem *vfs = VirtualFileSystemManager::getDefaultVirtualFileSystem());

	// Convert
	virtual void convert(GLenum format, GLenum type);
	virtual void mirror();
	virtual void flip();

	// Draw
	virtual void clear(float r = 0.0, float g = 0.0, float b = 0.0,
			float a = 0.0);
	
protected:
	static void _setVFSFunctionPointers(VirtualFileSystem *vfs);
	
	void _loadFromFile(const std::string& filename, VirtualFileSystem *vfs);
	void _createFromMemory(unsigned width, unsigned height, GLenum format,
			GLenum type, void *data);

	void _genImage();
	void _deleteImage();
	void _bindImage();
	void _getAttributes();

private:
	// Callback functions for file reading
	static ILvoid _fCloseRProc(ILHANDLE handle);
	static ILboolean _fEofProc(ILHANDLE handle);
	static ILint _fGetcProc(ILHANDLE handle);
	static ILHANDLE _fOpenRProc(const ILstring filename);
	static ILint _fReadProc(void* ptr, ILuint size, ILuint count, ILHANDLE handle);
	static ILint _fSeekRProc(ILHANDLE handle, ILint offset, ILint whence);
	static ILint _fTellRProc(ILHANDLE handle);

	// Callback functions for file writing
	static ILvoid _fCloseWProc(ILHANDLE handle);
	static ILHANDLE _fOpenWProc(const ILstring filename);
	static ILint _fPutcProc(ILubyte ch, ILHANDLE handle);
	static ILint _fSeekWProc(ILHANDLE handle, ILint offset, ILint whence);
	static ILint _fTellWProc(ILHANDLE handle);
	static ILint _fWriteProc(const void* ptr, ILuint size, ILuint count, ILHANDLE handle);
	
};

// =============================================================================

class DevilImaging : public Imaging {

public:
	DevilImaging();
	virtual ~DevilImaging();

public:
	static GLenum _convertFormatIL2GL(unsigned format);
	static unsigned _convertFormatGL2IL(GLenum format);
	static GLenum _convertTypeIL2GL(unsigned type);
	static unsigned _convertTypeGL2IL(GLenum type);
	
protected:
	virtual Image *_createImage();
	virtual Image *_createImage(unsigned width, unsigned height, GLenum format,
			GLenum type, void *data = 0);
	virtual Image *_createImage(const std::string& filename,
			VirtualFileSystem *vfs);
	virtual Image *_createImage(const Image * image);
};

} // namespace MinGLE

#endif // __MINGLE_DEVIL_IMAGING_H__
