//      ___                       ___           ___           ___       ___     
//     /\__\          ___        /\__\         /\  \         /\__\     /\  \.    
//    /::|  |        /\  \      /::|  |       /::\  \       /:/  /    /::\  \.   
//   /:|:|  |        \:\  \    /:|:|  |      /:/\:\  \     /:/  /    /:/\:\  \.  
//  /:/|:|__|__      /::\__\  /:/|:|  |__   /:/  \:\  \   /:/  /    /::\~\:\  \. 
// /:/ |::::\__\  __/:/\/__/ /:/ |:| /\__\ /:/__/_\:\__\ /:/__/    /:/\:\ \:\__\.
// \/__/~~/:/  / /\/:/  /    \/__|:|/:/  / \:\  /\ \/__/ \:\  \    \:\~\:\ \/__/
//       /:/  /  \::/__/         |:/:/  /   \:\ \:\__\    \:\  \    \:\ \:\__\.  
//      /:/  /    \:\__\         |::/  /     \:\/:/  /     \:\  \    \:\ \/__/  
//     /:/  /      \/__/         /:/  /       \::/  /       \:\__\    \:\__\.    
//     \/__/                     \/__/         \/__/         \/__/     \/__/    
// 
// =============================================================================
//                       Minimalist OpenGL Environment
// =============================================================================
//
// This file is part of Minimalist OpenGL Environment (MinGLE)
//
// Version: 0.2.0
// Author: Balazs Domonkos
// Filename: src/GLUT/src/GLUTWindow.cpp
// Creation Date: December 12th 2007
// Revision Date: December 12th 2007
//
//
// The Minimalist OpenGL Environment is free software: you can
// redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// The Minimalist OpenGL Environment is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without
// even the implied warranty of MERCHANTABILITY or FITNESS FOR
// A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Minimalist OpenGL Environment.  If not, see 
// <http://www.gnu.org/licenses/>. See the COPYING file included 
// with this distribution file for license details.
//

/// @file GLUTWindow.cpp
/// GLUT rendering window class

// Header that contains the declaration
#include <GLUTWindow.h>

// MinGLE headers
#include <GLUTTimer.h>
#include <System.h>

// System headers
#include <GL/glut.h>

namespace MinGLE {

GLUTWindow *GLUTWindow::msInstance = 0;
bool GLUTWindow::mKeyRepeatOn = false;
double GLUTWindow::mLastKeyEventTime = 0.0;
unsigned char GLUTWindow::mLastNormalKey = 0;
int GLUTWindow::mLastSpecialKey = 0;

GLUTWindow::GLUTWindow(unsigned width, unsigned height,
		WindowAttributes::ColorMode colorMode, bool doubleBuffer,
		bool fullScreen, unsigned videoRate, unsigned char *colorBits,
		unsigned char depthBits, unsigned char stencilBits,
		unsigned char *accumColorBits, unsigned auxBuffers, bool stereo) :
	Window(width, height, colorMode, doubleBuffer, fullScreen, videoRate,
			colorBits, depthBits, stencilBits, accumColorBits, auxBuffers,
			stereo), mWindowId(-1), mGameModeSupported(false),
			mLastMouseActionTime(0.0) {
	// Save instance (only on GLUT window is possible)
	msInstance = this;

	unsigned int mode = 0;

	switch (mWindowAttributes->mColorMode) {
	case WindowAttributes::DIRECT_COLORS:
		mode |= GLUT_RGBA;
		break;
	case WindowAttributes::INDEXED_COLORS:
		mode |= GLUT_INDEX;
		break;
	}

	mode |= mWindowAttributes->mDoubleBuffer ? GLUT_DOUBLE : GLUT_SINGLE;
	if (mWindowAttributes->mColorBits[3])
		mode |= GLUT_ALPHA;
	if (mWindowAttributes->mDepthBits)
		mode |= GLUT_DEPTH;
	if (mWindowAttributes->mStencilBits)
		mode |= GLUT_STENCIL;
	if (mWindowAttributes->mAccumColorBits[0]
			|mWindowAttributes->mAccumColorBits[1]
			|mWindowAttributes->mAccumColorBits[2]
			|mWindowAttributes->mAccumColorBits[3])
		mode |= GLUT_ACCUM;
	// NOTE: Aux buffers are not supported
	if (mWindowAttributes->mStereo)
		mode |= GLUT_STEREO;

	// Init display mode	
	glutInitDisplayMode(mode);

	// Detect supports
	mGameModeSupported = glutGameModeGet(GLUT_GAME_MODE_POSSIBLE);
	mGameModeSupported = false;

	if (mWindowAttributes->mFullScreen) {
		setFullScreen(true);
	} else {
		glutInitWindowPosition(mWindowAttributes->mLeft,
				mWindowAttributes->mTop);
		glutInitWindowSize(mWindowAttributes->mWidth,
				mWindowAttributes->mHeight);
		mWindowId = glutCreateWindow(mWindowAttributes->mTitle.c_str());
	}

	// Try to disable key repeat
	_disableKeyRepeat();

	// TODO: check errors
}

GLUTWindow::~GLUTWindow() {
}

void GLUTWindow::setPosition(int left, int top) {
	// Return from full screen
	if (mWindowAttributes->mFullScreen)
		setFullScreen(false);

	_setPosition(mWindowAttributes->mRecoveryLeft = left,
			mWindowAttributes->mRecoveryTop = top);
}

void GLUTWindow::setSize(unsigned width, unsigned height) {
	// Return from full screen
	if (mWindowAttributes->mFullScreen)
		setFullScreen(false);

	_setSize(mWindowAttributes->mRecoveryWidth = width,
			mWindowAttributes->mRecoveryHeight = height);
}

void GLUTWindow::setFullScreen(bool fullScreen) {

	// Leave game mode
	if (mWindowAttributes->mFullScreen && !fullScreen && mGameModeSupported)
		glutLeaveGameMode();

	if (mWindowAttributes->mFullScreen = fullScreen) {
		// Enter game mode if supported
		if (mGameModeSupported) {

			// Just for sure
			_setPosition(0, 0);
			_setSize(System::getScreenWidth(), System::getScreenHeight());

			char gameModeString[64];
			sprintf(
					gameModeString,
					"%dx%d:%d@%d",
					mWindowAttributes->mWidth,
					mWindowAttributes->mHeight,
					mWindowAttributes->mColorBits[0]
							+mWindowAttributes->mColorBits[1]
							+mWindowAttributes->mColorBits[2]
							+mWindowAttributes->mColorBits[3],
					mWindowAttributes->mVideoRate ? mWindowAttributes->mVideoRate
							: 60);
			glutGameModeString(gameModeString);

			glutEnterGameMode();
		}

		// Just fullscreen the window
		else {
			glutFullScreen();
		}
	} else {
		setPosition(mWindowAttributes->mRecoveryLeft,
				mWindowAttributes->mRecoveryTop);
		setSize(mWindowAttributes->mRecoveryWidth,
				mWindowAttributes->mRecoveryHeight);
	}
}

void GLUTWindow::setVisible(bool visible) {
	if (mWindowAttributes->mVisible = visible)
		glutShowWindow();
	else
		glutHideWindow();

	mWindowAttributes->mIconified = false;
}

void GLUTWindow::iconify() {
	mWindowAttributes->mIconified = true;
	glutIconifyWindow();
}

void GLUTWindow::setTitle(const std::string& title) {
	mWindowAttributes->mTitle = title;
	glutSetWindowTitle(mWindowAttributes->mTitle.c_str());
	glutSetIconTitle(mWindowAttributes->mTitle.c_str());
}

void GLUTWindow::finishFrame() {
	if (mWindowAttributes->mDoubleBuffer)
		glutSwapBuffers();
	else
		glFlush();

	// Call base implementation
	Window::finishFrame();
}

void GLUTWindow::redisplay() {
	glutPostRedisplay();
}

void GLUTWindow::clearFrame() {
	GLbitfield mask= GL_COLOR_BUFFER_BIT;

	if (mWindowAttributes->mDepthBits)
		mask |= GL_DEPTH_BUFFER_BIT;
	if (mWindowAttributes->mStencilBits)
		mask |= GL_STENCIL_BUFFER_BIT;
	if (mWindowAttributes->mAccumColorBits[0]
			|mWindowAttributes->mAccumColorBits[1]
			|mWindowAttributes->mAccumColorBits[2]
			|mWindowAttributes->mAccumColorBits[3])
		mask |= GL_ACCUM_BUFFER_BIT;

	glClear(mask);
}

Timer *GLUTWindow::createTimer(double period, Timer::TimerType type) {
	return new GLUTTimer(this, period, type);
}

void GLUTWindow::_disableKeyRepeat() {
	// Try to disable key repeating in two ways
	glutIgnoreKeyRepeat(1);
	glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);

	// Check success
	if (glutDeviceGet(GLUT_DEVICE_KEY_REPEAT) != GLUT_KEY_REPEAT_OFF
			&& !glutDeviceGet(GLUT_DEVICE_IGNORE_KEY_REPEAT)) {
		// Setting software repeat filtering. But don't expect very good results
		mKeyRepeatOn = true;
		mLastKeyEventTime = Timer::getSystemTime();
	}
}

void GLUTWindow::_setupWindowListener() {
	// Always render when the window is invalid
	glutDisplayFunc(GLUTWindow::_onRender);

	glutReshapeFunc(GLUTWindow::_onResize);

	glutKeyboardFunc(_onNormalKeyboardEvent);
	glutSpecialFunc(_onSpecialKeyboardEvent);

	// FreeGLUT implementation supports not only keydown but also keyup events
#ifdef __FREEGLUT_STD_H__
	glutKeyboardUpFunc(_onNormalKeyboardUpEvent);
	glutSpecialUpFunc(_onSpecialKeyboardUpEvent);
#endif

	glutMouseFunc(_onMouseEvent);
	glutMotionFunc(_onMotionEvent);
	glutPassiveMotionFunc(_onPassiveMotionEvent);
	glutEntryFunc(_onEntryEvent);

	_setupWindowListenerIdle();
}

void GLUTWindow::_setupWindowListenerIdle() {

	// If render on idle	
	if (mWindowAttributes->mRenderMode == WindowAttributes::RENDER_WHEN_IDLE)
		glutIdleFunc(GLUTWindow::_onRender);

	// If we want to listen on idle event
	else if (mWindowAttributes->mListenOnIdle)
		glutIdleFunc(GLUTWindow::_onIdle);

	// Forget the idle event
	else
		glutIdleFunc(0);
}

void GLUTWindow::_windowIsReadyToBeShown() {
	// Start render triggering timer
	if (mWindowAttributes->mRenderMode == WindowAttributes::RENDER_ON_TIMER)
		mTimeTriggeredRenderController.startOperation();
}

void GLUTWindow::_close() {
	// We cannot quit from GLUT's event loop, kill this window
	delete this;

	// Closing a GLUT window means quitting from the application
	// since only one GLUT windows is allowed
	System::quit();
}

void GLUTWindow::_setPosition(int left, int top) {
	glutPositionWindow(mWindowAttributes->mLeft = left, mWindowAttributes->mTop
			= top);
}

void GLUTWindow::_setSize(unsigned width, unsigned height) {
	glutReshapeWindow(mWindowAttributes->mWidth = width,
			mWindowAttributes->mHeight = height);
}

void GLUTWindow::_onRender() {
	if (msInstance->mWindowAttributes->mAutoClearFrame)
		msInstance->clearFrame();
	msInstance->_doRender();
	if (msInstance->mWindowAttributes->mAutoFinishFrame)
		msInstance->finishFrame();

	msInstance->checkRedisplay();
}

void GLUTWindow::_onIdle() {
	msInstance->_doIdle();
}

void GLUTWindow::_onResize(int width, int height) {
	msInstance->mWindowAttributes->mWidth = (unsigned) width;
	msInstance->mWindowAttributes->mHeight = (unsigned) height;
	if (msInstance->mWindowAttributes->mAutoUpdateViewport)
		glViewport(0, 0, (GLint) width, (GLint) height);
	msInstance->_doResize((unsigned) width, (unsigned) height);
}

void GLUTWindow::_onNormalKeyboardEvent(unsigned char glutCode, int x, int y) {
	// Register the mouse information
	msInstance->_doRegisterMouseMoveEvent(x, y);

	// Keyrepeat cannot be disabled (software filtering)
	if (mKeyRepeatOn) {
		// Only raise event when the elapsed time is big enough
		double now = Timer::getSystemTime();
		if (glutCode != mLastNormalKey || now - mLastKeyEventTime >= KEY_EVENT_DELAY_THRESHOLD) {
			//printf("down: %.6f\n", now - mLastKeyEventTime);
			_handleNormalKeyboardEvent(glutCode, InputListener::KEY_DOWN);
		}

		mLastKeyEventTime = now;
		mLastNormalKey = glutCode;
	}

	else
		_handleNormalKeyboardEvent(glutCode, InputListener::KEY_DOWN);
}

void GLUTWindow::_onNormalKeyboardUpEvent(unsigned char glutCode, int x, int y) {
	// Register the mouse information
	msInstance->_doRegisterMouseMoveEvent(x, y);

	// Keyrepeat cannot be disabled (software filtering)
	if (mKeyRepeatOn) {
		// Only raise event when the elapsed time is big enough
		double now = Timer::getSystemTime();
		if (glutCode != mLastNormalKey || now - mLastKeyEventTime
				>= KEY_EVENT_DELAY_THRESHOLD) {
			//printf("up: %.6f\n", now - mLastKeyEventTime);
			_handleNormalKeyboardEvent(glutCode, InputListener::KEY_UP);
		}

		mLastKeyEventTime = now;
		mLastNormalKey = glutCode;
	}

	else
		_handleNormalKeyboardEvent(glutCode, InputListener::KEY_UP);
}

void GLUTWindow::_onSpecialKeyboardEvent(int glutKey, int x, int y) {
	msInstance->_doRegisterMouseMoveEvent(x, y);

	// Keyrepeat cannot be disabled (software filtering)
	if (mKeyRepeatOn) {
		// Only raise event when the elapsed time is big enough
		double now = Timer::getSystemTime();
		if (glutKey != mLastSpecialKey || now - mLastKeyEventTime
				>= KEY_EVENT_DELAY_THRESHOLD) {
			_handleSpecialKeyboardEvent(glutKey, InputListener::KEY_DOWN);
		}

		mLastKeyEventTime = now;
		mLastSpecialKey = glutKey;
	}

	else
		_handleSpecialKeyboardEvent(glutKey, InputListener::KEY_DOWN);
}

void GLUTWindow::_onSpecialKeyboardUpEvent(int glutKey, int x, int y) {
	msInstance->_doRegisterMouseMoveEvent(x, y);

	// Keyrepeat cannot be disabled (software filtering)
	if (mKeyRepeatOn) {
		// Only raise event when the elapsed time is big enough
		double now = Timer::getSystemTime();
		if (glutKey != mLastSpecialKey || now - mLastKeyEventTime
				>= KEY_EVENT_DELAY_THRESHOLD) {
			_handleSpecialKeyboardEvent(glutKey, InputListener::KEY_UP);
		}

		mLastKeyEventTime = now;
		mLastSpecialKey = glutKey;
	}

	else
		_handleSpecialKeyboardEvent(glutKey, InputListener::KEY_UP);
}

void GLUTWindow::_onKeyboardTimer(int key) {

	KeyEventMapIterator i = msInstance->mCachedKeyEvents.find(key);

	// The key down event is in the cache => generating simple key down event 
	if (i != msInstance->mCachedKeyEvents.end() && System::getTime()
			- i->second.downTime >= MAX_PRESS_TIME) {
		msInstance->_doRegisterKeyboardEvent(key, InputListener::KEY_DOWN);
		msInstance->_doRegisterKeyboardModifierEvent(i->second.modifiers);
		msInstance->_doKeyEvent(key, i->second.modifiers,
				InputListener::KEY_DOWN);
		msInstance->mCachedKeyEvents.erase(i);
	}
}

void GLUTWindow::_onMouseEvent(int button, int state, int x, int y) {
	msInstance->_doRegisterMouseMoveEvent(x, y);

	// Convert button
	InputListener::MouseButton _button = InputListener::LEFT_BUTTON;
	switch (button) {

	// Buttons
	case GLUT_MIDDLE_BUTTON:
		_button = InputListener::MIDDLE_BUTTON;
		break;
	case GLUT_RIGHT_BUTTON:
		_button = InputListener::RIGHT_BUTTON;
		break;
#ifdef GLUT_WHEEL_UP
		case GLUT_RIGHT_BUTTON+1: _button = InputListener::FOURTH_BUTTON; break;
		case GLUT_RIGHT_BUTTON+2: _button = InputListener::FIFTH_BUTTON; break;
#endif

		// Win32 and MacOSX GLUT wheel event
#ifdef GLUT_WHEEL_UP
		case GLUT_WHEEL_UP:
		if(state == GLUT_UP) {
			msInstance->_doRegisterMouseWheelEvent(-WHEEL_SCROLLING);
			msInstance->_doMouseWheelEvent(x, y, -WHEEL_SCROLLING);
		}
		return;
		break;
		case GLUT_WHEEL_DOWN:
		if(state == GLUT_UP) {
			msInstance->_doRegisterMouseWheelEvent(WHEEL_SCROLLING);
			msInstance->_doMouseWheelEvent(x, y, WHEEL_SCROLLING);
		}
		return;
		break;
#else

		// Linux GLUT wheel event (fourth and fifth button)
		// NOTE: FOURTH_BUTTON, FIFTH_BUTTON cannot be handled
	case GLUT_RIGHT_BUTTON+1:
		if (state == GLUT_UP) {
			msInstance->_doRegisterMouseWheelEvent(-WHEEL_SCROLLING);
			msInstance->_doMouseWheelEvent(x, y, -WHEEL_SCROLLING);
		}
		return;
		break;
	case GLUT_RIGHT_BUTTON+2:
		if (state == GLUT_UP) {
			msInstance->_doRegisterMouseWheelEvent(WHEEL_SCROLLING);
			msInstance->_doMouseWheelEvent(x, y, WHEEL_SCROLLING);
		}
		return;
		break;
#endif
	}

	// The button has been changed
	/*if(msInstance->mLastMouseActionTime != 0.0 && 
	 msInstance->mCachedMouseEvent.button != _button) _resetClickInfo();*/

	// Convert state
	double currentTime = System::getTime();
	switch (state) {
	case GLUT_DOWN:
		switch (msInstance->mCachedMouseEvent.state) {
		case BUTTON_STATE_REST:
			_startClickTimeOut(x, y, _button, BUTTON_STATE_DOWN1);
			break;
		case BUTTON_STATE_DOWN1:
			// NOTE: invalid
			break;
		case BUTTON_STATE_UP1:
			if (currentTime - msInstance->mLastMouseActionTime
					<= MAX_CLICK_TIME)
				_startClickTimeOut(x, y, _button, BUTTON_STATE_DOWN2);
			else
				_resetClickInfo();
			break;
		case BUTTON_STATE_DOWN2:
			// NOTE: invalid
			break;
		case BUTTON_STATE_UP2:
			_resetClickInfo();
			_startClickTimeOut(x, y, _button, BUTTON_STATE_DOWN1);
			break;
		}
		break;
	case GLUT_UP:
		switch (msInstance->mCachedMouseEvent.state) {
		case BUTTON_STATE_REST:
			msInstance->_doRegisterMouseButtonEvent(_button,
			//msInstance->mCachedMouseEvent.button,
					InputListener::BUTTON_UP);
			msInstance->_doMouseButtonEvent(x, y, _button,
					InputListener::BUTTON_UP);
			_resetClickInfo();
			break;
		case BUTTON_STATE_DOWN1:
			_startClickTimeOut(x, y, _button, BUTTON_STATE_UP1);
			break;
		case BUTTON_STATE_UP1:
			// NOTE: invalid
			break;
		case BUTTON_STATE_DOWN2:
			if (currentTime - msInstance->mLastMouseActionTime
					<= MAX_CLICK_TIME)
				_startClickTimeOut(x, y, _button, BUTTON_STATE_UP2);
			else
				_resetClickInfo();
			break;
		case BUTTON_STATE_UP2:
			// NOTE: invalid
			break;
		}
		break;
	}
}

void GLUTWindow::_onMotionEvent(int x, int y) {
	msInstance->_doRegisterMouseMoveEvent(x, y);
	_resetClickInfo();
	msInstance->_doMouseMoveEvent(x, y, InputListener::HOVER);
}

void GLUTWindow::_onPassiveMotionEvent(int x, int y) {
	msInstance->_doRegisterMouseMoveEvent(x, y);
	_resetClickInfo();
	msInstance->_doMouseMoveEvent(x, y, InputListener::HOVER);
}

void GLUTWindow::_onEntryEvent(int state) {
	_resetClickInfo();
	switch (state) {
	case GLUT_LEFT:
		msInstance->_doMouseMoveEvent(InputListener::LEAVE);
		break;
	case GLUT_ENTERED:
		msInstance->_doMouseMoveEvent(InputListener::ENTER);
		break;
	}
}

void GLUTWindow::_onMouseTimer(int value) {
	// Timer is not actual
	if (msInstance->mCachedMouseEvent.state != value)
		return;

	_resetClickInfo();
}

void GLUTWindow::_handleNormalKeyboardEvent(unsigned char glutCode,
		InputListener::KeyEventState state) {
	_handleKeyboardEvent(_convertKeyNormal(glutCode), _getModifiers(), state);
}

void GLUTWindow::_handleSpecialKeyboardEvent(int glutKey,
		InputListener::KeyEventState state) {
	_handleKeyboardEvent(_convertKeySpecial(glutKey), _getModifiers(), state);
}

void GLUTWindow::_handleKeyboardEvent(int key, unsigned modifiers,
		InputListener::KeyEventState state) {
	switch (state) {
	case InputListener::KEY_DOWN: {
		_startPressTimeOut(key, modifiers);
	}
		break;
	case InputListener::KEY_UP: {
		KeyEventMapIterator i = msInstance->mCachedKeyEvents.find(key);

		// The key down event is in the key press interval 
		if (i != msInstance->mCachedKeyEvents.end()) {
			msInstance->_doRegisterKeyboardEvent(key, InputListener::KEY_PRESS);
			msInstance->_doRegisterKeyboardModifierEvent(i->second.modifiers);
			msInstance->_doKeyEvent(key, i->second.modifiers,
					InputListener::KEY_PRESS);
			msInstance->mCachedKeyEvents.erase(i);
		}

		// The key belongs to a key down event
		else {
			msInstance->_doKeyEvent(key, i->second.modifiers,
					InputListener::KEY_UP);
		}
	}
		break;
	case InputListener::KEY_PRESS:
		// NOTE: not valid, this method is never called with this argument
		break;
	}

	msInstance->checkRedisplay();
}

int GLUTWindow::_convertKeyNormal(unsigned char glutCode) {
	return glutCode;
}

int GLUTWindow::_convertKeySpecial(int glutKey) {
	switch (glutKey) {
	case GLUT_KEY_F1:
		return InputListener::KEY_F1;
	case GLUT_KEY_F2:
		return InputListener::KEY_F2;
	case GLUT_KEY_F3:
		return InputListener::KEY_F3;
	case GLUT_KEY_F4:
		return InputListener::KEY_F4;
	case GLUT_KEY_F5:
		return InputListener::KEY_F5;
	case GLUT_KEY_F6:
		return InputListener::KEY_F6;
	case GLUT_KEY_F7:
		return InputListener::KEY_F7;
	case GLUT_KEY_F8:
		return InputListener::KEY_F8;
	case GLUT_KEY_F9:
		return InputListener::KEY_F9;
	case GLUT_KEY_F10:
		return InputListener::KEY_F10;
	case GLUT_KEY_F11:
		return InputListener::KEY_F11;
	case GLUT_KEY_F12:
		return InputListener::KEY_F12;

	case GLUT_KEY_LEFT:
		return InputListener::KEY_LEFT_ARROW;
	case GLUT_KEY_UP:
		return InputListener::KEY_UP_ARROW;
	case GLUT_KEY_RIGHT:
		return InputListener::KEY_RIGHT_ARROW;
	case GLUT_KEY_DOWN:
		return InputListener::KEY_DOWN_ARROW;

	case GLUT_KEY_PAGE_UP:
		return InputListener::KEY_PAGE_UP;
	case GLUT_KEY_PAGE_DOWN:
		return InputListener::KEY_PAGE_DOWN;
	case GLUT_KEY_HOME:
		return InputListener::KEY_HOME;
	case GLUT_KEY_END:
		return InputListener::KEY_END;
	case GLUT_KEY_INSERT:
		return InputListener::KEY_INSERT;
	default:
		return 0;
	}

	return 0;
}

void GLUTWindow::_startPressTimeOut(int key, unsigned modifiers) {
	KeyEvent ke = { key, modifiers, System::getTime() };
	msInstance->mCachedKeyEvents[key] = ke;
	glutTimerFunc((unsigned) (1000.0 * MAX_PRESS_TIME), _onKeyboardTimer, key);
}

void GLUTWindow::_startClickTimeOut(int x, int y,
		InputListener::MouseButton button, ClickState state) {
	msInstance->mLastMouseActionTime = System::getTime();
	msInstance->mCachedMouseEvent.x = x;
	msInstance->mCachedMouseEvent.y = y;
	msInstance->mCachedMouseEvent.button = button;
	msInstance->mCachedMouseEvent.state = state;
	glutTimerFunc((unsigned) (1000.0 * MAX_CLICK_TIME), _onMouseTimer, state);
}

unsigned GLUTWindow::_getModifiers() {
	int glutModifiers = glutGetModifiers();
	unsigned modifiers = 0;

	// Unable to distiguish left and right keys (only alt and alt gr)
	if (glutModifiers & GLUT_ACTIVE_SHIFT)
		modifiers |= InputListener::KEY_SHIFT;
	if (glutModifiers & GLUT_ACTIVE_ALT)
		modifiers |= InputListener::KEY_ALT;
	if (glutModifiers & GLUT_ACTIVE_CTRL)
		modifiers |= InputListener::KEY_CTRL;

	return modifiers;
}

void GLUTWindow::_fireMouseEvent() {
	switch (msInstance->mCachedMouseEvent.state) {
	case BUTTON_STATE_REST:
		// NOTE: do nothing
		break;
	case BUTTON_STATE_DOWN1:
		msInstance->_doRegisterMouseButtonEvent(
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_DOWN);
		msInstance->_doMouseButtonEvent(msInstance->mCachedMouseEvent.x,
				msInstance->mCachedMouseEvent.y,
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_DOWN);
		break;
	case BUTTON_STATE_UP1:
		msInstance->_doMouseButtonEvent(msInstance->mCachedMouseEvent.x,
				msInstance->mCachedMouseEvent.y,
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_CLICK);
		break;
	case BUTTON_STATE_DOWN2:
		msInstance->_doRegisterMouseButtonEvent(
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_DOWN);
		msInstance->_doMouseButtonEvent(msInstance->mCachedMouseEvent.x,
				msInstance->mCachedMouseEvent.y,
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_CLICK);
		msInstance->_doMouseButtonEvent(msInstance->mCachedMouseEvent.x,
				msInstance->mCachedMouseEvent.y,
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_DOWN);
		break;
	case BUTTON_STATE_UP2:
		msInstance->_doMouseButtonEvent(msInstance->mCachedMouseEvent.x,
				msInstance->mCachedMouseEvent.y,
				msInstance->mCachedMouseEvent.button,
				InputListener::BUTTON_DBLCLICK);
		break;
	}

	msInstance->checkRedisplay();
}

} // namespace MinGLE
