#!/bin/bash

. mingle_path
export LD_LIBRARY_PATH="${MINGLE_LD_PATH}:${MINGLE_PARALLEL_LD_PATH}:${LD_LIBRARY_PATH}"

#
# Startup script for Hewlett-Packard XC
# 

export theProgramName=${0##*/}
export theUpperProgramName=`echo $theProgramName|tr a-z A-Z`
export theProgramVersion="E1.0.002"

#------------------------------------------------------------------------------
# Module header and display function for help
DisplayHelp() {
    echo "This is the help"
}

# Display function for module usage
svaDisplayUsage() {
    echo "Usage: ${theProgramName} -r <n> [-p <name>] [-g <geometry>]"
}
    
# Initialize the SVA data. This sets the temporary data file name
# which is used until the job is allocated.
. svainit
if [[ $? -ne 0 ]]; then
    svaDisplayError "Unable to perform SVA initialization."
    exit
fi

if [[ -z "${SLURM_JOBID}" ]]; then

    # Setup job default from user configuration files if they exist
    svaGetJobDefaults

    export DMX_INPUT_SERVER=$DISPLAY
    SVA_Render_Count=0
    SVA_Display_Surface=""
fi

SVA_Local_Display=1 # Don't set this and SVA will try to allocate a display surface... 
SVA_Render_Count=1 # Minimum 1 node to show compositing, this makes it possible to run the script without args
_PC_Executable="./sph-mpi"

# Process Command-Line options
while [[ -n "$1" ]]; do
    case "$1" in
        "--help" | "-h" | "-?")    DisplayHelp
                                   exit
                                   ;;
        "--usage" | "-u")          svaDisplayUsage
                                   exit
                                   ;;
        "--version" | "-V")        svaDisplayVersion
                                   exit
                                   ;;
        "--verbose" | "-v")        SVA_Verbose_Flag=1
                                   shift
                                   continue
                                   ;;
       "--display_surface" | "-d") shift
                                   SVA_Display_Surface="$1"
                                   shift
                                   continue
                                   ;;
        "--render" | "-r")         shift
                                   SVA_Render_Count=$1
                                   shift
                                   continue
                                   ;;
        "--partition" | "-p")      shift
                                   SVA_Partition_Name=$1
                                   shift
                                   continue
                                   ;;
        "--tile_geometry" | "-g")  shift
                                   SVA_Tile_Geometry=$1
                                   shift
                                   continue
                                   ;;
    esac

    # The value in $1 isn't a local option, so break out of the processing
    break
done

# Validate command options. If it fails, display a usage message and exit
svaValidateOptions

if  [[ ${SVA_Command_Line} != "" ]] ; then
    echo "Extra options or arguments"
    svaDisplayUsage
    exit 1
fi

if [[ ${SVA_Usage_Error} -ne 0 ]]; then
    svaDisplayUsage
    exit
fi

if [[ -x ${_PC_Executable} ]] ; then
    echo "running ${_PC_Executable}"
else
    echo "program ${_PC_Executable} missing - run make to create it"
    exit
fi

####
# JOB PHASE 1: Allocate resources for the job.
# This will allocate Render resources for the job.
####

svaAllocateJob
svaDisplayVerbose "render count:  ${SVA_Render_Count}"
svaDisplayVerbose "display res:   ${SVA_Tile_Geometry}"

####
# JOB PHASE 2: Launch basic services.  This launches X servers.
# The application is not launched during this phase.
####

# Detect if we are running in a TurboVNC or VirtualGL session,
echo ${LD_PRELOAD}|grep librrfaker >/dev/null
if [ $? -eq 0 ] ; then
    SVA_TurboVNC_Session=1
fi

# The X servers should not load the fake gl library, so remove
# it from preload
if [ ${SVA_TurboVNC_Session:-0} -eq 1 ] ; then
    LD_PRELOAD_STATE=${LD_PRELOAD}
    export LD_PRELOAD=`echo ${LD_PRELOAD}| sed -e 's/librrfaker.so//g;s/libdlfaker.so//g'`
fi

# Start X Servers on display and render hosts,
export SVA_X_Options="-ac ${SVA_X_Options}"
export SVA_XServer_Delay=1
svaDisplayVerbose "Start X servers"
svaRunDisplayXServers -nowait
svaRunRenderXServers -nowait

# Wait for all of the X servers on display nodes to start up
svaRunDisplayHosts sva_wait_for_X.sh :0

# Wait for all of the X servers on render nodes to start up
svaRunRenderHosts sva_wait_for_X.sh :0

####
# JOB PHASE 3: Run the sample
####

echo
echo "================================================================================"
echo "SVA: Started the X server(s). Now running the application..."
echo "================================================================================"
MPI_RUN=/opt/hpmpi/bin/mpirun
SESSIONID=1
MASTER=`hostname`
SLAVES=`expandnodes ${SVA_Allocate_Hosts}`
echo "Allocated hosts = $SLAVES"
echo "Master = $MASTER"

# Save the DISPLAY variable
OLDDISPLAY=$DISPLAY

# Run slave apps
export DISPLAY=:0.0

NODELIST=`echo "$MASTER $SLAVES" | sed "s| |,|g" | sed "s|,$||"`
${MPI_RUN} -IBV -np $((SVA_Render_Count+1)) -hostlist "$NODELIST" -e DISPLAY=:0.0 -e LD_LIBRARY_PATH="$LD_LIBRARY_PATH" ${_PC_Executable} $SESSIONID $((SVA_Render_Count+1)) $MASTER $SLAVES

#NODELIST=`echo "$MASTER $SLAVES" | sed "s| |,|g" | sed "s|,$||"`
#${MPI_RUN} -srun -N$((SVA_Render_Count+1)) -w "$NODELIST" ${_PC_Executable} $SESSIONID

# No rendering on the master
#NODELIST=`echo "$SLAVES" | sed "s| |,|g" | sed "s|,$||"`
#${MPI_RUN} -srun -N$((SVA_Render_Count)) -w "$NODELIST" ${_PC_Executable} $SESSIONID

#${MPI_RUN} -e $DISPLAY -srun -N$((SVA_Render_Count+1)) -w "$NODELIST" ${_PC_Executable} $SESSIONID
#${MPI_RUN} -np $((SVA_Render_Count+1)) -hostlist "$NODELIST" -e $DISPLAY ${_PC_Executable} $SESSIONID

#for n in $SLAVES; do
#    echo "Running slave piece on $n"
#    #srun -N1 -w $n ${_PC_Executable} $SESSIONID &
#    ${MPI_RUN} -srun -N1 -w $n ${_PC_Executable} $SESSIONID &
#done
	
# Restore DISPLAY variable
export DISPLAY=$OLDDISPLAY

# Restore the preload state as we may need it for running the app itself...
if [ ${SVA_TurboVNC_Session:-0} -eq 1 ] ; then
    export LD_PRELOAD=${LD_PRELOAD_STATE}
fi
echo "Running master piece on $MASTER"
#${_PC_Executable} $SESSIONID
#${_PC_Executable} $SESSIONID $((SVA_Render_Count+1)) $MASTER $SLAVES

####
# JOB PHASE 4: Terminate the job.  This is where your job is terminated.  If you
#              have any special termination commands this is where you would
#              place them.
####

# Kill the Job
[ -n $SLURM_JOBID ] && svaTerminateJobStep $SLURM_JOBID

# Wait for everthing to terminate
wait
