// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#include "../include/fluidsim.hpp"
#include "../include/container.hpp"
#include <math.h>

//------------------------------------------------------------------------------
// build container
//------------------------------------------------------------------------------
Container::Container(int i,Vector p,float w,float h,float d,bool o,bool hl,int psid,int pn,float psh,std::vector<Container*> subs) {
//------------------------------------------------------------------------------
	id = i;
	rootid = i;
	subcontainers = subs;
	show = false;
	position = p;
	width = w; height = h; depth = d;
	radius = w>d ? 1.1*w/2 : 1.1*d/2;
	partsysid = psid;
	partsysheight = height * (psh/100);
	num_of_particles = pn;
	partsystem = NULL;
	max_x_cells = max_y_cells = max_z_cells = 0;
	cellsize = 0;
	framenum = 0;

}

//-----------------------------------------------------------------------------
// summarize the containers
//-----------------------------------------------------------------------------
void Container::getSubNumber(int* num) {
//-----------------------------------------------------------------------------
	for(unsigned int i=0;i<subcontainers.size();subcontainers[i++]->getSubNumber(num));
	(*num)++;
}

//-----------------------------------------------------------------------------
// fill the containers array
//-----------------------------------------------------------------------------
void Container::getSubs(Container** containers) {
//-----------------------------------------------------------------------------
	for(unsigned int i=0;i<subcontainers.size();subcontainers[i++]->getSubs(containers));
	containers[id] = this;
}

//-----------------------------------------------------------------------------
// create particles
//-----------------------------------------------------------------------------
void Container::initSystem(std::vector<ParticleSystem*> particlesystems) {
//-----------------------------------------------------------------------------
	
	// first the subcontainers
	for(unsigned int i=0;i<subcontainers.size();subcontainers[i++]->initSystem(particlesystems));
	// then this one
	if (partsysid >= 0) {
		ParticleSystem* ps = particlesystems[partsysid];
		partsystem = new Fluid(*ps);
		partsystem->createParticles(this);		
	}
}

//-----------------------------------------------------------------------------
// make grid partition
//-----------------------------------------------------------------------------
void Container::createGrid() {
//-----------------------------------------------------------------------------
	// calculate cellsize - max core
	setCellSize(&cellsize);
	// make gridcells objects in a 3D array
	if (cellsize) {
		max_x_cells = (int)(2*radius/cellsize)+1;
		max_y_cells = (int)(height/cellsize)+1;
		max_z_cells = (int)(2*radius/cellsize)+1;

		gridcells = new GridCell**[max_x_cells];
		for(int x=0;x<max_x_cells;x++) {
			gridcells[x] = new GridCell*[max_y_cells];
			for(int y=0;y<max_y_cells;y++) {
				gridcells[x][y] = new GridCell[max_z_cells];
			}
		}
	}
	// give to the subcontainers
	for(unsigned int i=0;i<subcontainers.size();i++)
		subcontainers[i]->setGridCell(gridcells,cellsize,id);		
}

//-----------------------------------------------------------------------------
// calculate cellsize
//-----------------------------------------------------------------------------
void Container::setCellSize(float* cs) {
//-----------------------------------------------------------------------------
	for(unsigned int i=0;i<subcontainers.size();subcontainers[i++]->setCellSize(cs));
	if (partsystem && partsystem->getH() > *cs) *cs = partsystem->getH();
}

//-----------------------------------------------------------------------------
// set the grid
//-----------------------------------------------------------------------------
void Container::setGridCell(GridCell*** gc,float cs,int rid) {
//---------------------------0.5---------------------------------------------------
	// first to the subcontainers
	for(unsigned int i=0;i<subcontainers.size();subcontainers[i++]->setGridCell(gc,cs,rid));
	// then to this one
	gridcells = gc;
	cellsize = cs;
	rootid = rid;	
}

//------------------------------------------------------------------------------
// move the container and the partsystem inside
//------------------------------------------------------------------------------
void Container::move(Container** containers,int num) {
//------------------------------------------------------------------------------
	gpuSPH.simulate(partsystem,containers,framenum);
	framenum++;	
}

//------------------------------------------------------------------------------
// distribute containers for parallel simulation
//------------------------------------------------------------------------------
void Container::distribute() {
//------------------------------------------------------------------------------
	int size = FluidSimApp::Instance()->size;
	int rank = FluidSimApp::Instance()->rank;

	// container width
	subcontainers[0]->width /= size;
		
	if(size % 2 == 1) {
		int step = size / 2;
		float begin = - subcontainers[0]->width * step; 
			
		for(int i=0;i<size;i++) {
			if(rank==i) {
			 subcontainers[0]->position.x = begin + i * subcontainers[0]->width; 
			 printf("Rank:%d %f\n",rank,subcontainers[0]->position.x);
			}
		}
	} else {
		int step = size / 2;
		float begin = - subcontainers[0]->width * step + subcontainers[0]->width/2.0; 
		
		for(int i=0;i<size;i++) {
			if(rank==i) 
			 subcontainers[0]->position.x = begin + i * subcontainers[0]->width; 
		}
	}
}

//------------------------------------------------------------------------------
// draw container
//------------------------------------------------------------------------------
void Container::draw() {
//------------------------------------------------------------------------------
	if (partsystem) partsystem->draw();	
}
