// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#include  <stdarg.h>

#include "../include/fluidsim.hpp"

// simulation data file
const char* scenefile = "./data/scene.dat";

FluidSimApp* FluidSimApp::mInstance = 0;

//----------------------------------------------------------------------
// application init
//----------------------------------------------------------------------
FluidSimApp::FluidSimApp() {
//----------------------------------------------------------------------
	grid_on = true; text_on = true;
	current_frame = 0; fps_frame = 0; fps=0;
	
	loadingtime = calctime = 0;
	initGL();	
}

//----------------------------------------------------------------------
// application init
//----------------------------------------------------------------------
void FluidSimApp::init() {
//----------------------------------------------------------------------
	start = clock();
	scene.initScene(scenefile);
		
	double end = clock();
	loadingtime = (double)(end-start) / CLOCKS_PER_SEC;
	printf("-- Loading time: %f sec --",loadingtime);
}

//----------------------------------------------------------------------
FluidSimApp* FluidSimApp::Instance ()  {
//----------------------------------------------------------------------
    // is it the first call?
    if(mInstance==0) {  
      mInstance = new FluidSimApp(); // create sole instance
    }
    return mInstance; // address of sole instance
 }

//----------------------------------------------------------------------
// opengl init
//----------------------------------------------------------------------
void FluidSimApp::initGL() {
//----------------------------------------------------------------------
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

//----------------------------------------------------------------------
// light init
//----------------------------------------------------------------------
void FluidSimApp::initLight() {
//----------------------------------------------------------------------
	GLfloat LightAmbient[] = {0.0f, 0.0f, 0.0f, 1.0f};
	GLfloat LightDiffuse[] = {0.8f, 0.8f, 0.8f, 1.0f};
	GLfloat LightPosition[] = {30.0f, 30.0f, 30.0f, 1.0f};	
	glLightfv(GL_LIGHT0, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
}

//----------------------------------------------------------------------
// camera init
//----------------------------------------------------------------------
void FluidSimApp::initCamera() {
//----------------------------------------------------------------------
	GLfloat aspect = (float)scene.active_camera->hres /(float)scene.active_camera->vres;
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();
	gluPerspective(scene.active_camera->vfov*2.0, aspect, scene.active_camera->nearClip, scene.active_camera->farClip);
	glMatrixMode(GL_MODELVIEW);		
	glLoadIdentity();
	gluLookAt(scene.active_camera->eyep.x,
			  scene.active_camera->eyep.y,
			  scene.active_camera->eyep.z,
			  scene.active_camera->lookp.x,
			  scene.active_camera->lookp.y,
			  scene.active_camera->lookp.z,
			  scene.active_camera->updir.x,
			  scene.active_camera->updir.y,
			  scene.active_camera->updir.z);
}

//----------------------------------------------------------------------
// generate a new frame
//----------------------------------------------------------------------
void FluidSimApp::newFrame() {
//----------------------------------------------------------------------
	start = clock();
	scene.newFrame();
	current_frame++;
	fps_frame++;
	double end = clock();
	
	calctime += (end-start) / CLOCKS_PER_SEC;
	
	// fps counter - 5 secs time
	if(calctime>5.0) {fps=fps_frame/5.0; fps_frame=0; calctime=0;}	 
}

//----------------------------------------------------------------------
// simulate frames
//----------------------------------------------------------------------
void FluidSimApp::simulate() {
//----------------------------------------------------------------------
		newFrame();
		display();
}

//----------------------------------------------------------------------
// display parameters
//----------------------------------------------------------------------
void FluidSimApp::displayDatas() {
//----------------------------------------------------------------------
	glDisable(GL_LIGHTING);
	glMatrixMode(GL_PROJECTION);
	glDepthFunc(GL_ALWAYS);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 12000, 0, 8000);


	glColor3f(0.5,0.5,0.5);
	
	// current frame number
	drawString(10800,7800,"Frame: %d",current_frame);
	
	// FPS	
	glColor3f(0.7,0.7,0.7);
	drawString(10800,7600,"FPS: %4.3f",fps);

	// camera names
	glColor3f(1.0,1.0,1.0);
	drawString(200, 7800, "Camera%d",scene.active_camera->getID());

	// particle numbers
	glColor3f(0.9,0.9,1.0);
	for(int i=0;i<scene.particlesystems.size();i++) {
		const char* act = i == scene.active_particlesystem->getID() ? "->" : "";
		drawString(200, 7600-i*200, "%s particlesystem%d: density: %4.2f, k: %4.2f, viscosity: %4.2f",act,i,scene.particlesystems[i]->getDensity(),scene.particlesystems[i]->getK(),scene.particlesystems[i]->getViscosity());
	}

	// container datas
	glColor3f(0.7,0.7,0.2);
	for(int i=0;i<scene.num_of_containers;i++) {
		if (scene.containers[i]->getFluid())
			drawString(200,7600-(scene.particlesystems.size()+i)*200, "Simulated particles: %d",scene.containers[i]->num_of_particles*FluidSimApp::Instance()->size);
	}
			drawString(200,7200, "Nodes: %d",FluidSimApp::Instance()->size);
	
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glPopAttrib();	
	glMatrixMode(GL_MODELVIEW);

}

//----------------------------------------------------------------------
// update the application
//----------------------------------------------------------------------   
void FluidSimApp::display() {  
//----------------------------------------------------------------------
	glClearColor(0.25, 0.25, 0.25, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	initLight();		// init default light
	initCamera();		// init scene camera	
	
	// master application displays grid and text
	if(FluidSimApp::Instance()->rank==0) {
		if(grid_on)
	 		createGrid(0.25,4);	// create viewport grid
	 	scene.display();		// draw the scene	 	
	 	
	 	if(text_on)
	 		displayDatas();		// display scene datas	 	
	} else {
		scene.display();		// draw the scene			
	} 	
}


//----------------------------------------------------------------------
// draw string
//----------------------------------------------------------------------   
void FluidSimApp::drawString(GLfloat x, GLfloat y, char *format,...) {
//----------------------------------------------------------------------
	va_list args;
	char buffer[200], *p;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);
	glPushMatrix();
	glTranslatef(x, y, 0);
	for (p = buffer; *p; p++)
		glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
	glPopMatrix();
}

//----------------------------------------------------------------------
// toggle grid
//----------------------------------------------------------------------
void FluidSimApp::toggleGrid() {
//----------------------------------------------------------------------
	if(grid_on) grid_on=false;
	else grid_on=true;
}

//----------------------------------------------------------------------
// toggle text
//----------------------------------------------------------------------
void FluidSimApp::toggleText() {
//----------------------------------------------------------------------
	if(text_on) text_on=false;
	else text_on=true;
}

//----------------------------------------------------------------------
// create viewport grid
//----------------------------------------------------------------------   
void FluidSimApp::createGrid(float step, float n) {  
//----------------------------------------------------------------------
	glDisable(GL_LIGHTING);
	for(float i = -n; i <= n; i += step) {
		if(!i) {
			glBegin(GL_LINES);
					glColor3f(0.5,0.5,0.5);			
					glVertex3f(-n, 0, i);
					glVertex3f(n, 0, i);
					glVertex3f(i, 0, -n);
					glVertex3f(i, 0, n);
			glEnd();
		} else {
			glBegin(GL_LINES);
					glColor3f(0.2,0.2,0.2);
			
					glVertex3f(-n, 0, i);
					glVertex3f(n, 0, i);
					glVertex3f(i, 0, -n);
					glVertex3f(i, 0, n);
			glEnd();
		}
    }
	glEnable(GL_LIGHTING);
}
