// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#include "../include/gpuSPH.hpp"

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <assert.h>
#include <GL/glut.h>

#include "../include/container.hpp"
#include "../include/fluidsim.hpp"

// context variable
CGcontext   cgContext;

//-----------------------------------------------------------------------------
// cg error callback
//-----------------------------------------------------------------------------
void cgErrorCallback() {
//-----------------------------------------------------------------------------
  CGerror lastError = cgGetError();
  
  if(lastError)
  {
    printf("%s\n\n", cgGetErrorString(lastError));
    printf("%s\n", cgGetLastListing(cgContext));
    printf("Cg error, exiting...\n");    
  }
} 

//-----------------------------------------------------------------------------
// neighbour map contruction on the CPU -> upload
//-----------------------------------------------------------------------------
void GpuSPH::constructNeighbourMap(ParticleSystem* ps, int num) {
//-----------------------------------------------------------------------------
	//int texHeight = pNum;
	int texHeight = pNum;
	int texWidth = nLength;
	
	// fill in the texture
	// R = particle index, G = neighbour index, B = dontcare
	
	for(int i=0;i<texHeight;i++) {
		int nid = 0;
		for(int j=0;j<3*texWidth;j=j+3) {
				
			if(i+num*4096 < ps->getTPartNum()) {			
				if(!ps->particles[i+num*4096].isNull && nid < ps->particles[i+num*4096].num_of_neighbours) {
					neighbourMapData[i*3*texWidth+j] = i+num*4096;
					neighbourMapData[i*3*texWidth+j+1] = ps->particles[i+num*4096].neighbours[nid].particle->lid;
					neighbourMapData[i*3*texWidth+j+2] = -1;				
				
				}			
		  		else {
			  		neighbourMapData[i*3*texWidth+j] = -1;
			  		neighbourMapData[i*3*texWidth+j+1] = -1;
		  			neighbourMapData[i*3*texWidth+j+2] = -1;		  			
		  		}			
			}	
			else {
				neighbourMapData[i*3*texWidth+j] = -1;
				neighbourMapData[i*3*texWidth+j+1] = -1;
				neighbourMapData[i*3*texWidth+j+2] = -1;
			}				
			nid++;	
		}		
	}
	
	// upload to the GPU
	neighbourMap[num]->BeginCapture();
   	 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    
     glDrawPixels(texWidth, texHeight, GL_RGB, GL_FLOAT, neighbourMapData);     
	neighbourMap[num]->EndCapture();
	
}

//-----------------------------------------------------------------------------
// construct position and velocity maps for the first iteration
//-----------------------------------------------------------------------------
void GpuSPH::constructPosAttribMap(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
	int texWidth = posAttribMap->GetWidth();
	int texHeight = posAttribMap->GetHeight();
	
	int length = ps->getTPartNum();
		
	// fill in the posMap
	// R = particle pos X, G = particle pos Y, B = particle pos Z
	for(int i=0;i<texWidth*texHeight;i++) {
 		if(i<length && !ps->particles[i].isNull) {
			posAttribMapData[3*i] = ps->particles[i].position.x;
			posAttribMapData[3*i+1] = ps->particles[i].position.y;
			posAttribMapData[3*i+2] = ps->particles[i].position.z;
		} else {
			posAttribMapData[3*i] = -1;
			posAttribMapData[3*i+1] = -1;
			posAttribMapData[3*i+2] = -1;
		}
	}
	
	// upload the posMap to the GPU
	posAttribMap->BeginCapture();
   	 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);         
   	 glDrawPixels(texWidth, texHeight, GL_RGB, GL_FLOAT, posAttribMapData);     
	posAttribMap->EndCapture();  
	
}

//-----------------------------------------------------------------------------
// construct position and velocity maps for the first iteration
//-----------------------------------------------------------------------------
void GpuSPH::constructVelAttribMap(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
	int texWidth = velAttribMap->GetWidth();
	int texHeight = velAttribMap->GetHeight();
	
	int length = ps->getTPartNum();
	
	// fill in the velMap
	// R = particle vel X, G = particle vel Y, B = particle vel Z
	for(int i=0;i<texWidth*texHeight;i++) {
 		if(i<length && !ps->particles[i].isNull) {
			velAttribMapData[3*i] = ps->particles[i].velocity.x;
			velAttribMapData[3*i+1] = ps->particles[i].velocity.y;
			velAttribMapData[3*i+2] = ps->particles[i].velocity.z;
		} else {
			velAttribMapData[3*i] = -1;
			velAttribMapData[3*i+1] = -1;
			velAttribMapData[3*i+2] = -1;
		}
	}
	
	// upload the velMap to the GPU
  	velAttribMap->BeginCapture();
  	 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  
  	 glDrawPixels(texWidth, texHeight, GL_RGB, GL_FLOAT, velAttribMapData);           
  	velAttribMap->EndCapture(); 
}

//-----------------------------------------------------------------------------
// construct position and velocity maps for the first iteration
//-----------------------------------------------------------------------------
void GpuSPH::constructDensityAndPressureMap(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
	int texWidth = densityAndPressureMap->GetWidth();
	int texHeight = densityAndPressureMap->GetHeight();
	
	int length = ps->getTPartNum();
	
	// fill in the velMap
	// R = density, G = pressure
	for(int i=0;i<texWidth*texHeight;i++) {
 		if(i<length /*&& !ps->particles[i].isNull*/) {
			densityAndPressureMapData[3*i] = ps->particles[i].density;
			densityAndPressureMapData[3*i+1] = ps->particles[i].pressure;
			densityAndPressureMapData[3*i+2] = 0;
		} /*else {
			densityAndPressureMapData[3*i] = -1;
			densityAndPressureMapData[3*i+1] = -1;
			densityAndPressureMapData[3*i+2] = -1;
		}*/
	}
	
	// upload the densityAndPressureMapData to the GPU
  	densityAndPressureMap->BeginCapture();
  	 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  
  	 glDrawPixels(texWidth, texHeight, GL_RGB, GL_FLOAT, densityAndPressureMapData);           
  	densityAndPressureMap->EndCapture();
}

//-----------------------------------------------------------------------------
// distance computation between neighbours
//-----------------------------------------------------------------------------
void GpuSPH::computeDistance(int num) {
//-----------------------------------------------------------------------------
   	distanceMap[num]->BeginCapture();
  	{ 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);      
    cgGLBindProgram(distanceFragProgram);
    cgGLEnableProfile(theProfile);   
    
    // texture length
    attrlenParam = cgGetNamedParameter(distanceFragProgram,"attrlen");
	cgGLSetParameter1f(attrlenParam,posAttribMap->GetWidth());    
    
    // attribMap texture
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posAttribMap->GetTextureID());

    // neighbourMap texture
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, neighbourMap[num]->GetTextureID());       
    
    
    // pixel-texel mapping    
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();  
	gluOrtho2D(-1, 1, -1, 1);  	
	
	glMatrixMode(GL_MODELVIEW);     
    glLoadIdentity();    
   	
    glBegin(GL_QUADS);
    
    	glTexCoord2f(0,0);
    	glVertex2f(-1, -1);
    	
    	glTexCoord2f(distanceMap[num]->GetWidth(),0);
    	glVertex2f(1,-1);
    	
    	glTexCoord2f(distanceMap[num]->GetWidth(),distanceMap[num]->GetHeight());
    	glVertex2f( 1,  1);
    	
    	glTexCoord2f(0,distanceMap[num]->GetHeight());
    	glVertex2f(-1,  1);
    	
   glEnd();
      
   cgGLDisableProfile(theProfile);
   glPopMatrix();
    
   }    
   distanceMap[num]->EndCapture();
    
}

//-----------------------------------------------------------------------------
// compute density and pressure values for the particles
//-----------------------------------------------------------------------------
void GpuSPH::computeDensityAndPressure(int num) {
//-----------------------------------------------------------------------------
   	densityAndPressureMap->BeginCapture();
  	{ 
  	//if(num == 0) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   
    cgGLBindProgram(densityFragProgram);
    cgGLEnableProfile(theProfile);   
    
    // texture length
    lenParam = cgGetNamedParameter(densityFragProgram,"len");
    
	// texture length
    attrlenParam = cgGetNamedParameter(densityFragProgram,"attrlen");
	cgGLSetParameter1f(attrlenParam,posAttribMap->GetWidth());
	
	// itNum
    itNumParam = cgGetNamedParameter(densityFragProgram,"itNum");    
    
    // smmothing radius
    hParam = cgGetNamedParameter(densityFragProgram,"h");
    
    // particle mass
    massParam = cgGetNamedParameter(densityFragProgram,"mass");
    
    // gas constant    
    gasconstParam = cgGetNamedParameter(densityFragProgram,"gas_const");
    
    // main density
    maindensParam = cgGetNamedParameter(densityFragProgram,"maindens");
    
    // poly6 coefficient
    poly6_coef = cgGetNamedParameter(densityFragProgram,"poly6_coef");	
	
	// set uniform parameters
	cgGLSetParameter1f(lenParam,neighbourMap[num]->GetWidth());
	cgGLSetParameter1f(hParam,neighbourMap[num]->GetHeight());
	cgGLSetParameter1f(itNumParam,num);
	cgGLSetParameter1f(hParam,h);
	cgGLSetParameter1f(massParam,mass);
	cgGLSetParameter1f(gasconstParam,gas_const);
	cgGLSetParameter1f(maindensParam,maindensity);
	cgGLSetParameter1f(poly6_coef,poly6coef());
    
    // distMap texture
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, distanceMap[num]->GetTextureID());
    
    // densityMap texture
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, densityAndPressureMap->GetTextureID());         
    
        
    // pixel-texel mapping        
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();  
	gluOrtho2D(-1, 1, -1, 1);  	
	
    glMatrixMode(GL_MODELVIEW);     
    glLoadIdentity();    
   	
	float stepY =  4096.0 / densityAndPressureMap->GetWidth();
	float divNum = (float)densityAndPressureMap->GetHeight() / stepY;    
   	
    glBegin(GL_QUADS);
    
    	glTexCoord2f(0,num*stepY);
    	glVertex2f(-1, -1 + 2.0*(float)num/divNum);
    	
    	glTexCoord2f(densityAndPressureMap->GetWidth(),num*stepY);
    	glVertex2f(1,-1 + 2.0*(float)num/divNum);
    	
    	glTexCoord2f(densityAndPressureMap->GetWidth(),(num+1)*stepY);
    	glVertex2f(1, -1 + 2.0*(float)(num+1)/divNum);
    	
    	glTexCoord2f(0,(num+1)*stepY);
    	glVertex2f(-1, -1 + 2.0*(float)(num+1)/divNum);
    	
    glEnd();
      
    cgGLDisableProfile(theProfile);
    glPopMatrix();
    
  	}
  	densityAndPressureMap->EndCapture();
}

//-----------------------------------------------------------------------------
// compute force, collision and acceleration of the particles
//-----------------------------------------------------------------------------
void GpuSPH::computeForceAndCollision(int num) {
//-----------------------------------------------------------------------------
  	forceMap->BeginCapture();
  	{ 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);      
    cgGLBindProgram(forceFragProgram);
    cgGLEnableProfile(theProfile);
        
    // texture length
    lenParam = cgGetNamedParameter(forceFragProgram,"len");
    
    // itNum
    itNumParam = cgGetNamedParameter(forceFragProgram,"itNum");  
    
    // texture length
    attrlenParam = cgGetNamedParameter(forceFragProgram,"attrlen");
	cgGLSetParameter1f(attrlenParam,posAttribMap->GetWidth());
    
    // mass
    massParam = cgGetNamedParameter(forceFragProgram,"mass");
    
    // smoothing radius
    hParam = cgGetNamedParameter(forceFragProgram,"h");
    
    // grad spiky coefficient
    grad_spiky_coef = cgGetNamedParameter(forceFragProgram,"grad_spiky_coef");
    
    // v_lap visc coefficient
    v_lap_visc_coef = cgGetNamedParameter(forceFragProgram,"v_lap_visc_coef");     	
	
	cgGLSetParameter1f(massParam,mass);
	cgGLSetParameter1f(itNumParam,num);
	cgGLSetParameter1f(hParam,h);
	cgGLSetParameter1f(lenParam,neighbourMap[num]->GetWidth());
	cgGLSetParameter1f(grad_spiky_coef,grad_spiky());
	cgGLSetParameter1f(v_lap_visc_coef,v_lap_visc());
	
	// neighbourMap texture
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, neighbourMap[num]->GetTextureID());  
    
    // distMap texture
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, distanceMap[num]->GetTextureID());
    
    // densityMap texture
    glActiveTextureARB(GL_TEXTURE2_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, densityAndPressureMap->GetTextureID());
    
    // velAttribMap texture
    glActiveTextureARB(GL_TEXTURE3_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, velAttribMap->GetTextureID());  
    
    // posAttribMap texture
    glActiveTextureARB(GL_TEXTURE4_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posAttribMap->GetTextureID());
    
    
    // pixel-texel mapping
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();  
	gluOrtho2D(-1, 1, -1, 1);  	
	
	glMatrixMode(GL_MODELVIEW);     
    glLoadIdentity();    
   	
    glBegin(GL_QUADS);
    
       	glTexCoord2f(0,0);
    	glVertex2f(-1, -1);
    	
    	glTexCoord2f(forceMap->GetWidth(),0);
    	glVertex2f(1,-1);
    	
    	glTexCoord2f(forceMap->GetWidth(),forceMap->GetHeight());
    	glVertex2f( 1,  1);
    	
    	glTexCoord2f(0,forceMap->GetHeight());
    	glVertex2f(-1,  1);
    	
    glEnd();
      
    cgGLDisableProfile(theProfile);
    glPopMatrix();
    
  	}    
  	forceMap->EndCapture();
}

//-----------------------------------------------------------------------------
// compute velocity of the particles
//-----------------------------------------------------------------------------
void GpuSPH::computeVelocity(int num) {
//-----------------------------------------------------------------------------
  	velAttribMap->BeginCapture();
  	{  	
    //if(num == 0) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);      
    cgGLBindProgram(velFragProgram);
    cgGLEnableProfile(theProfile);     
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();  
	gluOrtho2D(-1, 1, -1, 1);  	
	
	glMatrixMode(GL_MODELVIEW);     
    glLoadIdentity();
    
    // time step
    dtParam = cgGetNamedParameter(velFragProgram,"dt");	
	cgGLSetParameter1f(dtParam,dt);
	
	itNumParam = cgGetNamedParameter(velFragProgram,"itNum");
	cgGLSetParameter1f(itNumParam,num);    
		
	// texture length
    attrlenParam = cgGetNamedParameter(velFragProgram,"attrlen");
	cgGLSetParameter1f(attrlenParam,posAttribMap->GetWidth());
    
    // forceMap texture
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, forceMap->GetTextureID());  
    
    // velAttribMap texture
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, velAttribMap->GetTextureID());  
    
    float divNum = (float)velAttribMap->GetHeight() / (4096.0 / velAttribMap->GetWidth());
    float stepY =  velAttribMap->GetHeight() / divNum;
   	
    glBegin(GL_QUADS);
    
    	glTexCoord2f(0,num*stepY);
    	glVertex2f(-1, -1 + 2.0*(float)num/divNum);
    	
    	glTexCoord2f(velAttribMap->GetWidth(),num*stepY);
    	glVertex2f(1,-1 + 2.0*((float)num/divNum));
    	
    	glTexCoord2f(velAttribMap->GetWidth(),(num+1)*stepY);
    	glVertex2f(1, -1 + 2.0*(float)(num+1)/divNum);
    	
    	glTexCoord2f(0,(num+1)*stepY);
    	glVertex2f(-1, -1 + 2.0*(float)(num+1)/divNum);
    	
    glEnd();
      
    cgGLDisableProfile(theProfile);
    glPopMatrix();
    
  	}    
  	velAttribMap->EndCapture();
}

//-----------------------------------------------------------------------------
// compute position of the particles
//-----------------------------------------------------------------------------
void GpuSPH::computePosition() {
//-----------------------------------------------------------------------------
  	posAttribMap->BeginCapture();
  	{ 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);      
    cgGLBindProgram(posFragProgram);
    cgGLEnableProfile(theProfile);     
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();  
	gluOrtho2D(-1, 1, -1, 1);  	
	
	glMatrixMode(GL_MODELVIEW);     
    glLoadIdentity();    
    
    static int framenum = -1;
    framenum++;
    
	// time step					    
    dtParam = cgGetNamedParameter(posFragProgram,"dt");	
	cgGLSetParameter1f(dtParam,dt);	
	
	frameParam = cgGetNamedParameter(posFragProgram,"frame");	
	cgGLSetParameter1f(frameParam,framenum);   
	
	rankParam = cgGetNamedParameter(posFragProgram,"rank");	
	cgGLSetParameter1f(rankParam,FluidSimApp::Instance()->rank);   
    
    // posMap texture
    glActiveTextureARB(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posAttribMap->GetTextureID());  
    
    // velAttribMap texture
    glActiveTextureARB(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, velAttribMap->GetTextureID());
       	
    
    // pixel-texel mapping
    
    glBegin(GL_QUADS);
    
       glTexCoord2f(0,0);
    	glVertex2f(-1, -1);
    	
    	glTexCoord2f(posAttribMap->GetWidth(),0);
    	glVertex2f(1,-1);
    	
    	glTexCoord2f(posAttribMap->GetWidth(),posAttribMap->GetHeight());
    	glVertex2f( 1,  1);
    	
    	glTexCoord2f(0,posAttribMap->GetHeight());
    	glVertex2f(-1,  1);
    	
    glEnd();
      
    cgGLDisableProfile(theProfile);
    glPopMatrix();
    
  	}    
  	posAttribMap->EndCapture();
}

//-----------------------------------------------------------------------------
// download particle positions to the CPU
//-----------------------------------------------------------------------------
void GpuSPH::downloadDensityPressure(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
 	int texWidth = densityAndPressureMap->GetWidth();
 	int texHeight = densityAndPressureMap->GetHeight();
 
 	// bind position map
 	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, densityAndPressureMap->GetTextureID()); 	
 	glGetTexImage(GL_TEXTURE_RECTANGLE_ARB,0,GL_RGB,GL_FLOAT,downData);
    
 	// download positions
	for(int i=0;i<texWidth*texHeight;i++) {
		if(i< ps->getTPartNum()) {
  			if(!ps->particles[i].isNull) {
  					ps->particles[i].density = downData[3*i];
					ps->particles[i].pressure = downData[3*i+1];					
  			}
		}  	
	}
}

//-----------------------------------------------------------------------------
// download particle positions to the CPU
//-----------------------------------------------------------------------------
void GpuSPH::downloadPosition(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
 
 	// bind position map
 	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, posAttribMap->GetTextureID()); 	
 	glGetTexImage(GL_TEXTURE_RECTANGLE_ARB,0,GL_RGB,GL_FLOAT,downData);
    
 	// download positions
	for(int i=0;i<posAttribMap->GetWidth()*posAttribMap->GetHeight();i++) {
		if(i< ps->getTPartNum()) {
  			if(!ps->particles[i].isNull) {
  					ps->particles[i].position.x = downData[3*i];
					ps->particles[i].position.y = downData[3*i+1];
					ps->particles[i].position.z = downData[3*i+2];

					//fprintf(stderr, "%d ", ps->particles[i].num_of_neighbours);
  			}
		}  	
	}

	//fprintf(stderr,"\n\n");
}

//-----------------------------------------------------------------------------
// download particle velocities to the CPU
//-----------------------------------------------------------------------------
void GpuSPH::downloadVelocity(ParticleSystem* ps, GridCell*** gridcells) {
//-----------------------------------------------------------------------------
 
 	// bind position map
 	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, velAttribMap->GetTextureID()); 	
 	glGetTexImage(GL_TEXTURE_RECTANGLE_ARB,0,GL_RGB,GL_FLOAT,downData);
    
 	// download velocity
	for(int i=0;i<velAttribMap->GetWidth()*velAttribMap->GetHeight();i++) {
		if(i< ps->getTPartNum()) {
  			if(!ps->particles[i].isNull) {
  				if(ps->particles[i].toThrow) {  				
					ps->removeParticle(i,gridcells);
  				} else {
  					ps->particles[i].velocity.x = downData[3*i];
					ps->particles[i].velocity.y = downData[3*i+1];
					ps->particles[i].velocity.z = downData[3*i+2];
  				}
  			}
	}
		}	
}


//-----------------------------------------------------------------------------
// fluid solver init, variable init, texture load
//-----------------------------------------------------------------------------
void GpuSPH::init(ParticleSystem* ps) {
//-----------------------------------------------------------------------------
  
  // setup simulation parameters
  dt = 1.0/150;
  
  maindensity = ps->getDensity();
  gas_const = ps->getK();
  mass = ps->getM();
  h = ps->getH();
  visc = ps->getViscosity();
  
  // maximum neighbourlength
  nLength = 128;
	
  // particle number in textures
  while(pNum<ps->getTPartNum())
	pNum = pNum << 1;  
	
  if(pNum>4096) pNum = 4096;  
  
  // attrib texture width, length
  
  int attrWidth = 1;
  int attrLength = 1;  
  
  bool flag = false;
  while(attrWidth*attrLength < ps->getTPartNum()) {
  	if(flag) attrWidth = attrWidth << 1;
  	else attrLength = attrLength << 1;
  	flag = !flag; 
  } 
  
  printf("%d %d",attrWidth,attrLength);
  
  // alloc memory for neighbourmap
  neighbourMapData = (float*)malloc(3*nLength*pNum*sizeof(float));  
  
  itNum = (int)ceil(ps->getTPartNum() / 4096.0);  
  
  // init texmaps
  for(int i=0;i<itNum;i++) {
	// init neighbourMaps	
 	neighbourMap[i] = new RenderTexture(nLength,pNum);
 	neighbourMap[i]->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);
  }
  
  for(int i=0;i<itNum;i++) {
	// init neighbourMaps	
 	distanceMap[i] = new RenderTexture(nLength,pNum);
 	distanceMap[i]->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);
  }
 	
  // init posAttribMap  	
  posAttribMap = new RenderTexture(attrWidth,attrLength);
  posAttribMap->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);  
  
  // init velAttribMap
  velAttribMap = new RenderTexture(attrWidth,attrLength);
  velAttribMap->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);
   
  // init densityAndPressureMap
  densityAndPressureMap = new RenderTexture(attrWidth,attrLength);
  densityAndPressureMap->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);
  
  // init forceMapNum
  forceMap = new RenderTexture(pNum,1);
  forceMap->Initialize(true, false, false, false, false, 32, 32, 32, 32, RenderTexture::RT_COPY_TO_TEXTURE);
  
  // init data structures
  posAttribMapData = (float*)malloc(3*posAttribMap->GetWidth()*posAttribMap->GetHeight()*sizeof(float));
  velAttribMapData = (float*)malloc(3*velAttribMap->GetWidth()*velAttribMap->GetHeight()*sizeof(float));
  densityAndPressureMapData = (float*)malloc(3*densityAndPressureMap->GetWidth()*densityAndPressureMap->GetHeight()*sizeof(float));
    
  // alloc memory for position readback
  downData = (float*)malloc(3*posAttribMap->GetWidth()*posAttribMap->GetHeight()*sizeof(float));
    
  // setup Cg if needed
  cgSetErrorCallback(cgErrorCallback);
  
  // create cgContext
  cgContext = cgCreateContext();

  // get the best profile for this hardware
  theProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
  assert(theProfile != CG_PROFILE_UNKNOWN);
  cgGLSetOptimalOptions(theProfile);
  
  // create programs from files
  vertexProgram = cgCreateProgramFromFile(cgContext, 
                                           CG_SOURCE,
                                           distanceMap[0]->IsRectangleTexture() ? "shaders/textureRECT.cg" : "Shaders/texture2D.cg",
                                           theProfile,
                                           NULL,
                                           NULL);
                                           
  distanceFragProgram = cgCreateProgramFromFile(cgContext, 
                             			   CG_SOURCE,
                                          "shaders/compute_distance.cg", 
                                           theProfile,
                                           NULL,
                                           NULL);
                                           
  densityFragProgram = cgCreateProgramFromFile(cgContext, 
                                          CG_SOURCE,
                                          "shaders/compute_density.cg", 
                                           theProfile,
                                           NULL,
                                           NULL);  
                                           
  forceFragProgram = cgCreateProgramFromFile(cgContext, 
                                          CG_SOURCE,
                                          "shaders/compute_force.cg", 
                                           theProfile,
                                           NULL,
                                           NULL); 
                                           	
  velFragProgram = cgCreateProgramFromFile(cgContext, 
                                          CG_SOURCE,
                                          "shaders/compute_velocity.cg", 
                                           theProfile,
                                           NULL,
                                           NULL);
                                           
  posFragProgram = cgCreateProgramFromFile(cgContext, 
                                          CG_SOURCE,
                                          "shaders/compute_position.cg", 
                                           theProfile,
                                           NULL,
                                           NULL);
                                                                                      
  // load programs
  if(vertexProgram != NULL) cgGLLoadProgram(vertexProgram);
  if(distanceFragProgram != NULL) cgGLLoadProgram(distanceFragProgram);
  if(densityFragProgram != NULL) cgGLLoadProgram(densityFragProgram);
  if(forceFragProgram != NULL) cgGLLoadProgram(forceFragProgram);
  if(velFragProgram != NULL) cgGLLoadProgram(velFragProgram);
  if(posFragProgram != NULL) cgGLLoadProgram(posFragProgram);      
}

//-----------------------------------------------------------------------------
// do a simulation step
//-----------------------------------------------------------------------------
void GpuSPH::simulate(ParticleSystem* ps, Container** container, int framenum) {
//-----------------------------------------------------------------------------
	
	if(ps) {		
		
		// init gpu fluid solver
	 	if(framenum==0) {
	 		init(ps);	 		 		 		
	 	}	 	
	 	
	 	// iterations
	 	if(framenum % 1 == 0) {	 		
	 		// reset status
	 		ps->resetStatus();
	 		// first reset all the particles	
	 		ps->resetParticleNeighbours();	
	 		// communication	 		
	 		ps->communicateParticles(container[1],container[1]->gridcells); 		 		
	 		// see the cell changes
	 		ps->setParticlesToCells(container[1]->gridcells,container[1]);
	 		// set the neighbours
	 		ps->setNeighbours(container[1]->gridcells,container[1]);				
	 	}
	 	
	 	// construct attrib maps
	 	constructVelAttribMap(ps);	 			
	 	constructPosAttribMap(ps);	 	
	
	 	for(int i=0;i<itNum;i++) 	{
	 			 		
	 		if(framenum % 1 == 0) {
	 			// construct neighbourMap
    			constructNeighbourMap(ps,i);
	 		}
	 		
			// particle neighbour distance computation
			computeDistance(i);
			// particle density and pressure computation
			computeDensityAndPressure(i);
		}

		// download density and pressure
		downloadDensityPressure(ps);

		// send density and pressure 
		ps->communicateDensityPressure(container[1],container[1]->gridcells);
		
		// upload density and pressure
		constructDensityAndPressureMap(ps);

		// delete outside of the volume
		ps->deleteOutside(container[1],container[1]->gridcells); 

		for(int i=0;i<itNum;i++) 	{
			// particle force and collision computation
			computeForceAndCollision(i);
			// velocity computation
			computeVelocity(i);
		}
		
		// position computation
		computePosition();		

		// download result to the cpu
		downloadVelocity(ps,container[1]->gridcells);
		downloadPosition(ps);
	}
}
