// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 


#include <stdio.h>
#include <unistd.h>
#include <iostream>			
#include <assert.h>
#include <stdlib.h>

// MPI
#include <mpi.h>

// MinGLE
#include <mingle.h>
#include <mingle-parallel.h>
#include <GL/glut.h>

#include "../include/fluidsim.hpp"
#include "../RenderTexture/RenderTexture.h"

#define TRUE 1
#define FALSE 0

#define MASTER 0

void init(int argc, char** argv);
void display();

// hostname of the node
char g_hostName[256];

// MPI params
int		rank, size, len;
char	name[MPI_MAX_PROCESSOR_NAME];

// mouse coords
int prev_x;
int prev_y;

// keyboard events
int left_down;
int right_down;
int middle_down;
int alt_down;

// application pointer
FluidSimApp* app;

//------------------------------------------------------------------------------
// Custom window listener that does the rendering
//------------------------------------------------------------------------------
class MyListener : public MinGLE::WindowListener {
//------------------------------------------------------------------------------
    protected:
		// This method is called when to render
		virtual bool onRender() {
			// Render the scene
	    	display();

			// True return value means that the event 
			// has been fully handled, no further onRender
			// methods of other window listeners are called.
	    	return true;
		}
		
		// application hotkeys
		virtual bool onKeyEvent(int key, unsigned modifiers, KeyEventState state) {
			switch(key)	{
				// camera reset
	 			case 'r': app->scene.active_camera->reset(); break;
	 
 				// switch camera
	 			case 'c': app->scene.switchCamera(); break;
	 
				// toggle grid
	 			case 'g': app->toggleGrid(); break;
	 			
	 			// toggle text - TAB KEY
	 			case 9: app->toggleText(); break;
			}
			
			return false;
		}
		
		// mouse button events
		virtual bool onMouseButtonEvent(int x, int y, MouseButton button, MouseEventState state) {
			prev_x = x;
			prev_y = y;
			
			if (state == BUTTON_DOWN)	{
				if (button == LEFT_BUTTON) {
					left_down = TRUE;
				}			
				else if (button == RIGHT_BUTTON)
					right_down = TRUE;
				else if (button == MIDDLE_BUTTON)
					middle_down = TRUE;
			}
	
			else if (state == BUTTON_DOWN)	{
				if (button == LEFT_BUTTON)
					left_down = FALSE;
				else if (button == RIGHT_BUTTON)
					right_down = FALSE;
				else if (button == MIDDLE_BUTTON)
					middle_down = FALSE;
			}
			return false;
		}
		
		// mouse events
		virtual bool onMouseMoveEvent(int x, int y, MouseMoveState state) {
			int dx,dy;
			
			if(left_down) {		
				dx = x - prev_x;
				dy = y - prev_y;

				// ALT + LMB - camera orbit
				app->scene.active_camera->orbit(dx,dy);			

				prev_x = x;
				prev_y = y;
			}

			else if(right_down) {		
				dx = x - prev_x;
				dy = y - prev_y;

				// ALT + MMB - camera pan
				app->scene.active_camera->strafe(dx*0.001);
				app->scene.active_camera->moveUpDown(dy*0.001);

				prev_x = x;
				prev_y = y;
			}
			else if(middle_down) {
				dx = x - prev_x;
				dy = y - prev_y;

				// ALT + RMB - camera zoom
				if(dx<0) app->scene.active_camera->zoomIn();
				if(dx>0) app->scene.active_camera->zoomOut();
				if(dy<0) app->scene.active_camera->zoomIn();
				if(dy>0) app->scene.active_camera->zoomOut();

				prev_x = x;
				prev_y = y;
			}
			return false;
		}
};

//-----------------------------------------------------------------------------
// main function
//=============================================================================
int main(int argc,char* argv[]) {
//=============================================================================
    
	try {    
		
    	// Initialize Rendering system
   		MinGLE::System::initialize(&argc, argv);
   		
   		std::vector<std::string> hostList;   		
   		int arg_offset = 3;
   		
   		/* Get the hostname of this node - this will be this program's host  */
		gethostname(g_hostName, sizeof(g_hostName));
   		
   		// master
   		if(strcmp(argv[3],g_hostName)==0) {
   			// push back nodes
   			int sva_render_count = atoi(argv[2]);
   			for(int i=0;i<sva_render_count;i++)
   				hostList.push_back(argv[i+arg_offset]);
   		} else {
   			//push back current slave node;
   			hostList.push_back(g_hostName);
   		}
   		
   		for(unsigned int i=0;i<hostList.size();i++)
   			printf("%s ",hostList[i].c_str());
   		printf("\n");
   		
   		// Initialize parallel rendering support   		
		MinGLE::ParallelRenderingSupport::initialize(1, hostList, MinGLE::ParallelRenderingSupport::PARACOMP);   	
   		
    	// Create window
    	MinGLE::Window *win = MinGLE::System::createWindow(800,600);    
    	win->setRenderMode(MinGLE::WindowAttributes::RENDER_WHEN_IDLE);
		win->setTitle("Distributed SPH Fluid Simulator");

	   	// Add parallel rendering support to the window
		MinGLE::ParallelRenderingSupport::addParallelSupport(win, MinGLE::ParallelRenderingSupport::DEPTH_OPERATOR);
		
		// Register the window listener
    	win->registerWindowListener(new MyListener());
    	
		// Add keyhandler
		MinGLE::ApplicationKeyHandler *keyHandler = new MinGLE::ApplicationKeyHandler();
		win->registerWindowListener(keyHandler);
		
		// Initialize application
		init(argc,argv);	
	
		// Enter the event handing loop
    	MinGLE::System::enterMainLoop();
    
	    return 0;
	
	} catch(const MinGLE::Exception& e) {
		std::cerr << "Error: " << e.mDescription << " (Thrown in " << e.mFile << ':'
			<< e.mMethod << " line " << e.mLine << ')' << std::endl;
	}
	
	MPI_Finalize();
    return 0;
}   

//------------------------------------------------------------------------------
// application init
//------------------------------------------------------------------------------
void init(int argc, char** argv) {
//------------------------------------------------------------------------------
    // Create the application
	app = FluidSimApp::Instance();
	
	// Initialize MPI	
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	
	// MPI
	app->rank = rank;
	app->size = size;
	
	app->init();
}

//------------------------------------------------------------------------------
// display
//------------------------------------------------------------------------------
void display() {
//------------------------------------------------------------------------------
	app->simulate();	
	app->display();	
}
