// 
//    Distributed SPH Fluid Simulator (asterix@index.hu)
// =============================================================================
//
// Copyright 2007 David Illes
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

#include <mpi.h>

#include "../include/particlesystem.hpp"
#include "../include/container.hpp"
#include "../include/fluidsim.hpp"			// for the display modes and constants

// particle core
#define CORE 5.0

// slice size to send
#define BUFFER h

//-----------------------------------------------------------------------------
// create particlesystem
//-----------------------------------------------------------------------------
ParticleSystem::ParticleSystem(int i,float dens,float visc,float kk) {
//-----------------------------------------------------------------------------
	id = i;
	density = dens;
	viscosity = visc;
	k = kk;
	firstfree = 0;
}

//-----------------------------------------------------------------------------
// reset particles of the particlesystem
//-----------------------------------------------------------------------------
void ParticleSystem::sceneReset() {
//-----------------------------------------------------------------------------
	for(int i=0;i<getPartNum();i++)
		particles[i].resetPosition();
	
}

//-----------------------------------------------------------------------------
// properties
//-----------------------------------------------------------------------------
int ParticleSystem::getID() const {return id;}
float ParticleSystem::getH() const {return h;}
int ParticleSystem::getPartNum() const {return num_of_particles;}
int ParticleSystem::getTPartNum() const {return num_of_total_particles;}
float ParticleSystem::getDensity() const {return density;}
float ParticleSystem::getViscosity() const {return viscosity;}
float ParticleSystem::getK() const {return k;}
float ParticleSystem::getM() const {return particles[0].mass;}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// draw the particlesystem
//-----------------------------------------------------------------------------
void ParticleSystem::draw() {
//-----------------------------------------------------------------------------
	GLUquadricObj* sphere;
	sphere = gluNewQuadric();
		
	for(int i=0;i<getTPartNum();i++) {	
		if(!particles[i].isNull) {			 
				
			// set color based on density
			float intensity = particles[i].density / 2000;
								
			if(intensity>1) intensity = 1;
			 else if(intensity<0) intensity = 0;

			GLfloat Surface[] = {1-intensity,1-intensity,1,0.6f};
				 				
			glMaterialfv(GL_FRONT, GL_DIFFUSE, Surface);
			
			// draw particle as sphere
			glPushMatrix();

			glTranslatef(particles[i].position.x,particles[i].position.y,particles[i].position.z);
			gluSphere(sphere,0.01,8,7);

			glPopMatrix();				
		}	
	}
}

//-----------------------------------------------------------------------------
// create particles
//-----------------------------------------------------------------------------
void ParticleSystem::createParticles(Container* container) {
//-----------------------------------------------------------------------------

	// calculate the radius of a particle
	// first approximation by volume
	float volume = container->getFluidVolume();
	num_of_particles = container->num_of_particles;
	float pvolume = volume / num_of_particles;
	float diameter = exp(log(pvolume)/3);
	
	// new array for display
	particles = new Particle[num_of_particles*2];

	num_of_total_particles=num_of_particles*2;
	
	particles = new Particle[num_of_total_particles];
	 
	int rank = FluidSimApp::Instance()->rank;
	
	for(int i=0;i<num_of_total_particles;i++)
		particles[i].init(i+rank*num_of_total_particles,i,0,0,0,0,0,0,0,false,true,false);

	// number of particles in a direction with this diameter
	int defxnum = (int) (container->width / diameter);
	int defynum = (int) (container->partsysheight / diameter);
	int defznum = (int) (container->depth / diameter);
	
	int x0num = 0 ,y0num = 0,z0num = 0;
	int xnum = defxnum,ynum = defynum,znum = defznum;
	// difference between the set and the real particle number
	int diff = num_of_particles - defxnum*defynum*defznum;
	// we increase numbers in all dimensions for the best difference
	for(int i=0;i<7;i++) {
		switch(i) {
		 case 0: x0num = defxnum+1; y0num = defynum; z0num = defznum; break;
		 case 1: x0num = defxnum; y0num = defynum+1; z0num = defznum; break;
		 case 2: x0num = defxnum; y0num = defynum; z0num = defznum+1; break;
		 case 3: x0num = defxnum+1; y0num = defynum+1; z0num = defznum; break;
		 case 4: x0num = defxnum; y0num = defynum+1; z0num = defznum+1; break;
		 case 5: x0num = defxnum+1; y0num = defynum; z0num = defznum+1; break;
		 case 6: x0num = defxnum+1; y0num = defynum+1; z0num = defznum+1; break;
		}
		int pnum =  x0num*y0num*z0num;
		if (pnum <= num_of_particles && num_of_particles-pnum < diff) {
			diff = num_of_particles - pnum;
			xnum = x0num;
			ynum = y0num;
			znum = z0num;
		}
	}
	// after we have the best choise, we take the average of the radiuses
	diameter = ((container->width/xnum)+(container->partsysheight/ynum)+(container->depth/znum))/3-0.000005f;
	particle_radius = diameter/2;
	h = CORE*particle_radius;
	h2 = h*h;

	num_of_particles = xnum*ynum*znum;   

	// compute mass
	mass = density * container->getFluidVolume();
	particle_mass = 1.5 * mass / num_of_particles;
	//particles = new Particle[num_of_particles];
	
	// build the particles from bottom to up
	float posx0 = container->position.x-container->width/2+particle_radius;
	float posy0 = container->position.y-container->height/2+particle_radius;
	float posz0 = container->position.z-container->depth/2+particle_radius;
	
	for(int y=0;y<ynum;y++)
		for(int z=0;z<znum;z++)
			for(int x=0;x<xnum;x++)
				addParticle(-1,particle_mass,posx0+x*diameter,posy0+y*diameter,posz0+z*diameter,0.0,0.0,0.0,false,false,false);
			
}

//-----------------------------------------------------------------------------
// set particles to cells
//-----------------------------------------------------------------------------
void ParticleSystem::setParticlesToCells(GridCell*** gridcells,Container* container) {
//-----------------------------------------------------------------------------
	int ocx,ocy,ocz;	// old coordinates
	int cx,cy,cz;		// new coordinates
	for(int i=0;i<getTPartNum();i++) {		
		
		if(!particles[i].isNull) {
			ocx = particles[i].cellx;
			ocy = particles[i].celly;
			ocz = particles[i].cellz;

			cx = (int)((particles[i].position.x - container->position.x+container->radius) / container->cellsize);
			cy = (int)((particles[i].position.y - container->position.y+container->radius) / container->cellsize);
			cz = (int)((particles[i].position.z - container->position.z+container->radius) / container->cellsize);

			if (cx<0) cx=0;
			if (cy<0) cy=0;
			if (cz<0) cz=0;

			// change cell
			if (ocx!=cx || ocy!=cy || ocz!=cz) {
				particles[i].cellx = cx;
				particles[i].celly = cy;
				particles[i].cellz = cz;				
	
				if (ocx>=0 && ocx<container->max_x_cells && ocy>=0 && ocy<container->max_y_cells && ocz>=0 && ocz<container->max_z_cells) {
					gridcells[ocx][ocy][ocz].removeParticle(particles[i].cellid);
				}

				if (cx>=0 && cx<container->max_x_cells && cy>=0 && cy<container->max_y_cells && cz>=0 && cz<container->max_z_cells)
					gridcells[cx][cy][cz].addParticle(&particles[i]);
			}
		}
	}
}			

//-----------------------------------------------------------------------------
// set neighbours of each particle
//-----------------------------------------------------------------------------
void ParticleSystem::setNeighbours(GridCell*** gridcells,Container* container) {
//-----------------------------------------------------------------------------
	Vector dist_vector;
	float dist;
	Particle* pj;

	for(int n=0;n<getTPartNum();n++) {
		if(!particles[n].isNull) {
			particles[n].num_of_neighbours = 0;
			
			// calculate cell id
			int widthid = particles[n].cellx;
			int heightid = particles[n].celly;
			int depthid = particles[n].cellz;

			// see all the neighbour cells
			for(int i=widthid-1;i<=widthid+1;i++)
				for(int j=heightid-1;j<=heightid+1;j++)
					for(int k=depthid-1;k<=depthid+1;k++) {
						if (0<=i && i<container->max_x_cells && 0<=j && j<container->max_y_cells && 0<=k && k<container->max_z_cells) {
							// see each particle in the neighbour cell
							for(int m=0;m<gridcells[i][j][k].num_of_particles;m++) {
								if(particles[n].num_of_neighbours < 128) {
									pj = gridcells[i][j][k].particles[m];
								
									if(pj!=NULL) {

										if (pj->id != n) {
											dist_vector = (particles[n].position - pj->position);
											dist = dist_vector.Norm();								
								
											// inside radius
											if (dist < h) {
												particles[n].newNeighbour(pj);
											}
										}
									}
								} else break;							
							}
						}
					}
		}			
	}
}

//-----------------------------------------------------------------------------
// reset the particles
//-----------------------------------------------------------------------------
void ParticleSystem::resetParticleNeighbours() {
//-----------------------------------------------------------------------------
	for(int i=0;i<getTPartNum();i++)
	 particles[i].resetNeighbours();
	
}

//-----------------------------------------------------------------------------
// reset status
//-----------------------------------------------------------------------------
void ParticleSystem::resetStatus() {
//-----------------------------------------------------------------------------
	for(int i=0;i<getTPartNum();i++)
	 if(!particles[i].isNull) particles[i].resetStatus();
	
}

//------------------------------------------------------------------------------
// communicate particles
//------------------------------------------------------------------------------
void ParticleSystem::communicateParticles(Container* container,GridCell*** gridcells) {
//------------------------------------------------------------------------------
	
	int size = FluidSimApp::Instance()->size;
	int rank = FluidSimApp::Instance()->rank;

	std::vector<Particle> pSend1;
	std::vector<Particle> pSend2;

	// left end of the volume
	if(rank==0) {
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(conRight-BUFFER < particles[i].position.x)	
 				 	pSend2.push_back(particles[i]);
			}
		}
	} else		
	// right end of the volume	
	if(rank==size-1) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(particles[i].position.x < conLeft+BUFFER)
				  	pSend1.push_back(particles[i]);				
			}
		}
	// inside of the volume
	} else if (rank!=size-1 && rank!=0) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(particles[i].position.x < conLeft+BUFFER) {
			  	 pSend1.push_back(particles[i]);				
				} else
				if(conRight-BUFFER < particles[i].position.x)	
 				 pSend2.push_back(particles[i]);
			}
		}
	}
		
	// send - receive

	for(int id=0;id<size-1;id++) {
		if(rank==id)
		 sendParticles(&pSend2,id+1);
		else if(rank==id+1)
		 receiveParticles(id);	
	}

	for(int id=size-1;id>0;id--) {
		if(rank==id)
		 sendParticles(&pSend1,id-1);
		else if(rank==id-1)
		 receiveParticles(id);	
	}

}	

//------------------------------------------------------------------------------
// communicate density and pressure
//------------------------------------------------------------------------------
void ParticleSystem::communicateDensityPressure(Container* container,GridCell*** gridcells) {
//------------------------------------------------------------------------------
	
	int size = FluidSimApp::Instance()->size;
	int rank = FluidSimApp::Instance()->rank;

	std::vector<Particle> pSend1;
	std::vector<Particle> pSend2;

	// left end of the volume
	if(rank==0) {
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(conRight-BUFFER < particles[i].position.x)	
 				 	pSend2.push_back(particles[i]);
			}
		}
	} else		
	// right end of the volume	
	if(rank==size-1) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(particles[i].position.x < conLeft+BUFFER)
				  	pSend1.push_back(particles[i]);				
			}
		}
	// inside of the volume
	} else if (rank!=size-1 && rank!=0) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0 && particles[i].received==0) {			
				// step 1 - send
				if(particles[i].position.x < conLeft+BUFFER) {
			  	 pSend1.push_back(particles[i]);				
				} else
				if(conRight-BUFFER < particles[i].position.x)	
 				 pSend2.push_back(particles[i]);
			}
		}
	}
		
	// send - receive

	for(int id=0;id<size-1;id++) {
		if(rank==id)
		 sendParticleDensityPressure(&pSend2,id+1);
		else if(rank==id+1)
		 receiveParticleDensityPressure(id);	
	}

	for(int id=size-1;id>0;id--) {
		if(rank==id)
		 sendParticleDensityPressure(&pSend1,id-1);
		else if(rank==id-1)
		 receiveParticleDensityPressure(id);	
	}

}	
//------------------------------------------------------------------------------
// delete outside volume
//------------------------------------------------------------------------------
void ParticleSystem::deleteOutside(Container* container,GridCell*** gridcells) {
//------------------------------------------------------------------------------
	int size = FluidSimApp::Instance()->size;
	int rank = FluidSimApp::Instance()->rank;

	// left end of the volume
	if(rank==0) {
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0) {			
				if(conRight < particles[i].position.x)	
 				 	removeParticle(i,gridcells);

			}
		}
	} else		
	// right end of the volume	
	if(rank==size-1) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0) {			
				if(particles[i].position.x < conLeft)
				  	removeParticle(i,gridcells);				
			}
		}
	// inside of the volume
	} else if (rank!=size-1 && rank!=0) {
		float conLeft = container->subcontainers[0]->position.x - container->subcontainers[0]->width/2;
		float conRight = container->subcontainers[0]->position.x + container->subcontainers[0]->width/2;
		
		for(int i=0;i<getTPartNum();i++) {
			if(particles[i].isNull==0) {			
				if(conRight < particles[i].position.x || particles[i].position.x < conLeft)
			  	 removeParticle(i,gridcells);			
			}
		}
	}
}

//------------------------------------------------------------------------------
// finds the next free position in the particle system
//------------------------------------------------------------------------------
int ParticleSystem::nextfree(int begin) {
//------------------------------------------------------------------------------
	for(int i = begin;i<getTPartNum();i++) {
		if(particles[i].isNull) {
			return i;
		}
	}
	
	printf("%d error ",begin);
	return -1;	
}

//------------------------------------------------------------------------------
// add a particle to the particle system
//------------------------------------------------------------------------------
void ParticleSystem::addParticle(int id, float m,float x,float y,float z, float velx, float vely, float velz, bool rcv, bool mnull, bool mThrow) {
//------------------------------------------------------------------------------
	particles[firstfree].init(id,firstfree,m,x,y,z,velx,vely,velz,rcv,mnull,mThrow);	
	firstfree = nextfree(firstfree);	
}

//------------------------------------------------------------------------------
// add a particle to the particle system
//------------------------------------------------------------------------------
void ParticleSystem::addDensityPressure(int id, float dens, float pres) {
//------------------------------------------------------------------------------
	for(int i=0;i<getTPartNum();i++) {	
		if(!particles[i].isNull && particles[i].id == id) {
			particles[i].density = dens;
			particles[i].pressure = pres;
			return;
		}	
	}
}

//------------------------------------------------------------------------------
// remove a particle to the particle system
//------------------------------------------------------------------------------
void ParticleSystem::removeParticle(int pos, GridCell*** gridcells) {
//------------------------------------------------------------------------------

	// remove from cell
	int ocx = particles[pos].cellx;
	int ocy = particles[pos].celly;
	int ocz = particles[pos].cellz;

	if(particles[pos].cellid!=-1)
		gridcells[ocx][ocy][ocz].removeParticle(particles[pos].cellid);

	// reset position
	if(pos<firstfree) firstfree = pos;
	
	particles[pos].isNull = true;	
	particles[pos].cellid=-1;
}

//------------------------------------------------------------------------------
// send a particle
//------------------------------------------------------------------------------
void ParticleSystem::sendParticles(std::vector<Particle> *transParticles, int dest) {
//------------------------------------------------------------------------------
	
	int size = transParticles->size();
	
	// send size
	MPI_Send(&size, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
	
	// packet
	float* partArray = new float[7*size];
	
	// creating packet content
	for(int i=0;i<size;i++) {	
		partArray[7*i] = (float)(*transParticles)[i].id;
		
		partArray[7*i+1] = (*transParticles)[i].position.x;
		partArray[7*i+2] = (*transParticles)[i].position.y;
		partArray[7*i+3] = (*transParticles)[i].position.z;
		
		partArray[7*i+4] = (*transParticles)[i].velocity.x;
		partArray[7*i+5] = (*transParticles)[i].velocity.y;
		partArray[7*i+6] = (*transParticles)[i].velocity.z;
	}	
		
	// send coord packet
	MPI_Send(partArray, 7*size, MPI_FLOAT, dest, 1, MPI_COMM_WORLD);		
	
	delete[] partArray;
}

//------------------------------------------------------------------------------
// send particle density and pressure
//------------------------------------------------------------------------------
void ParticleSystem::sendParticleDensityPressure(std::vector<Particle> *transParticles, int dest) {
//------------------------------------------------------------------------------
	
	int size = transParticles->size();
	
	// send size
	MPI_Send(&size, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
	
	// packet
	float* partArray = new float[3*size];
	
	// creating packet content
	for(int i=0;i<size;i++) {	
		partArray[3*i] = (int)(*transParticles)[i].id;
		partArray[3*i+1] = (*transParticles)[i].density;
		partArray[3*i+2] = (*transParticles)[i].pressure;
	}	
		
	// send coord packet
	MPI_Send(partArray, 3*size, MPI_FLOAT, dest, 1, MPI_COMM_WORLD);		
	
	delete[] partArray;
}


//------------------------------------------------------------------------------
// receive particles
//------------------------------------------------------------------------------
int ParticleSystem::receiveParticles(int source) {
//------------------------------------------------------------------------------

	MPI_Status status;
	
	int id;
	float posX, posY, posZ;
	float velX, velY, velZ;
	int size;
	
	bool mThrow = false;
	
	
	float* partArray;
	
	// receive size of data block
	MPI_Recv(&size,1,MPI_INT, source, 1, MPI_COMM_WORLD, &status);
	
	partArray = new float[7*size];
	
	// receive coords
	MPI_Recv(partArray,7*size,MPI_FLOAT, source, 1, MPI_COMM_WORLD, &status);			
	
	for(int i=0;i<size;i++) {
		id = (int)partArray[7*i];
		
		posX = partArray[7*i+1];
		posY = partArray[7*i+2];
		posZ = partArray[7*i+3];
	
		velX = partArray[7*i+4];
		velY = partArray[7*i+5];
		velZ = partArray[7*i+6];

		if(source==1) { if(posX>0) mThrow = true; else mThrow = false; }
		else if(source==0) { if(posX<0) mThrow = true; else mThrow = false; } 		

		// add density and pressure		
		addParticle(id,particle_mass,posX, posY, posZ, velX, velY, velZ, true,false,mThrow);	
	}
	
	delete[] partArray;
	
	return size;	
}

//------------------------------------------------------------------------------
// receive density and pressure data
//------------------------------------------------------------------------------
int ParticleSystem::receiveParticleDensityPressure(int source) {
//------------------------------------------------------------------------------

	MPI_Status status;
	
	int id;
	float density, pressure;
	int size;
	
	float* partArray;
	
	// receive size of data block
	MPI_Recv(&size,1,MPI_INT, source, 1, MPI_COMM_WORLD, &status);
	
	partArray = new float[3*size];
	
	// receive coords
	MPI_Recv(partArray,3*size,MPI_FLOAT, source, 1, MPI_COMM_WORLD, &status);			
	
	for(int i=0;i<size;i++) {
		id = (int)partArray[3*i];
		density = partArray[3*i+1];
		pressure = partArray[3*i+2];

		// add density and pressure		
		addDensityPressure(id,density,pressure);		
	}
	
	delete[] partArray;
	
	return size;	
}

